//******************************************************************************
//* File       : FileDlgContext.cpp                                            *
//* Author     : Mahlon R. Smith                                               *
//*              Copyright (c) 2005-2020 Mahlon R. Smith, The Software Samurai *
//*                  GNU GPL copyright notice located in FileMangler.hpp       *
//* Date       : 21-Jul-2020                                                   *
//* Version    : (see FileDlgVersion string in FileDlg.cpp)                    *
//*                                                                            *
//* Description:                                                               *
//* This module of the FileDlg class is an extension of the FileDlgPrompt.cpp  *
//* module, and contains functionality to support the application 'ViewFile'   *
//* context menu.                                                              *
//*    1) View Contents:                                                       *
//*       View the contents of the selected file.                              *
//*    2) View File Stats:                                                     *
//*       View and optionally modify the file stats including dates,           *
//*       permission flags, ets.                                               *
//*    3) Execute (Run):                                                       *
//*       a) For executable files (binaries, script files), execute the        *
//*          selected file.                                                    *
//*       b) For media file and document files, invoke the system's default    *
//*          application for the selected file.                                *
//*    4) Find Link Target:                                                    *
//*       For symbolic link files, locate the link target and set the target's *
//*       directory as the current working directory (CWD).                    *
//*                                                                            *
//* Development Tools: see notes in FileMangler.cpp.                           *
//******************************************************************************
//* Version History (most recent first):                                       *
//*   See version history in FileDlg.cpp.                                      *
//******************************************************************************
//* Programmer's Notes:                                                        *
//*                                                                            *
//*                                                                            *
//******************************************************************************

#include "GlobalDef.hpp"
#include "FileDlg.hpp"

//******************************
//* Local definitions and data *
//******************************
extern const char** fdNotImplemented ;
extern const char** fmtypeName ;

//* Control indices for opeUserPrompt() *
enum upIndices : short
{
   upiLAUNCH = ZERO, upiCANCEL, upiGUI, upiCUR, upiNEW, upiARGS, upiGED, 
   upiCONTROLS,
   upiHEADER, upiEDIT, upiQHELP     // extra indices for various messages
} ;


//******************************
//*  Local non-member methods  *
//******************************

static bool vfcIsPerlSource ( const gString& trgPath, bool strict = false ) ;
static bool vfcIsPerlCommand ( const gString& gstoken ) ;
static bool vfcGetPerlToken ( gString& gssrc, gString& gstrg ) ;
static bool oepGooeyType ( const gString& fPath, bool guiAudio ) ;
static bool oepBinexeTest ( const gString& fPath ) ;
static bool htmlExtractTag ( gString& gsIn, gString& gsOut ) ;


//*************************
//*   ViewFileContents    *
//*************************
//******************************************************************************
//* View contents of the currently-highlighted file.                           *
//*                                                                            *
//* Output is first formated (if necessary) and then displayed using the       *
//* 'less' command-line utility.                                               *
//*                                                                            *
//* Input  : vfFmt : (optional, vfPrompt by default) -member of enum viewFormat*
//*                  format in which to view file contents                     *
//*                  [ CURRENTLY IGNORED ]                                     *
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*          ERR if file not found or file is not a 'regular' file             *
//******************************************************************************
//* Notes:                                                                     *
//* a) To view as text:                                                        *
//*    invoke the 'less' utility                                               *
//* b) To view as hex/decimal/octal:                                           *
//*    invoke the 'hexdump' utility to format the data and pipe it to 'less'.  *
//* c) To view as binary:                                                      *
//*    Call vfcDecodeBinary() to format the source data for binary output,     *
//*    then call the 'less' utility.                                           *
//*    Note: This was formerly handled by the 'xxd' utility, which is often    *
//*          not installed by default by modern Linux distributions.           *
//* d) To view the contents of a symlink file rather than it's target:         *
//*    invoke the 'readlink' C library function to format the data, then       *
//*    call the 'less' utility.                                                *
//* e) For archive files, note that the 'less' utility may invoke a            *
//*    preprocessor program to correctly display archive contents              *
//*    automagically. However, this is a bit TOO magical for our taste.        *
//*    On Fedora, this magic conversion done through a script specified by the *
//*    'LESSOPEN' environment variable. This is not guaranteed to be the same  *
//*    for all systems or for all environments on a given system, so we do not *
//*    rely upon it, and use our own decoding.                                 *
//* f) For OpenDocument files:                                                 *
//*    -- displayed as text, (.odt, .ods, .odp), the XML sub-file containing   *
//*       the document text is extracted from the document and scanned.        *
//*       The text data are formatted and copied to a temp file which is then  *
//*       displayed by the 'less' utility.                                     *
//*    -- displayed as a 'zip archive, (all document formats for which we have *
//*       samples), run the file through the archive filter (see item 'e')     *
//*       and then display the archive contents.                               *
//* g) For Office OpenXML files:                                               *
//*    -- displayed as text, (.docx, pptx, xlsx), the XML sub-file containing  *
//*       the document text is extracted from the document and scanned.        *
//*       The text data are formatted and copied to a temp file which is then  *
//*       displayed by the 'less' utility.                                     *
//*    -- displayed as a 'zip archive, (all document formats for which we have *
//*       samples), run the file through the archive filter (see item 'e')     *
//*       and then display the archive contents.                               *
//*                                                                            *
//* h) If the filename is too long to display in this dialog, it will result   *
//*    in a non-critical case of the uglies.                                   *
//*                                                                            *
//******************************************************************************

short FileDlg::ViewFileContents ( viewFormat vfFmt )
{
   short status = ERR ;

   //* If there is actually a file under the highlight *
   if ( this->fileCount > ZERO )
   {
      const short DLG_ROWS    = 13,       // dialog lines
                  dlgCols     = minfitCOLS ; // dialog columns
      const attr_t dColor = this->cs.sd,  // dialog background color
                   hColor = this->cs.em,  // bold text
                   rbnColor = this->cs.tn,// radio-button colors
                   rbfColor = this->cs.scheme == ncbcCOLORS ? 
                                                 this->cs.tf | ncgATTR : this->cs.pf ;
      short dlgRows = DLG_ROWS ;

      enum vfcControls 
      { vfcVIEW= ZERO, vfcCANCEL, vfcTEXT, vfcHEXB, vfcHEXW, 
        vfcDECI, vfcOCT, vfcBIN, vfcLINK, vfcARCH, vfcMETA, 
        controlsDEFINED } ;

      //* Retrieve the file stat informaton *
      tnFName  fn, fnt ;            // copy of basic file info
      this->GetStats ( fn ) ;       // stats for currently-highlighted file
      const char* trgName = fn.fName ;// pointer to target filename
      gString trgPath ;             // target path string
      this->fmPtr->CatPathFname ( trgPath, this->currDir, fn.fName ) ;
      gString origPath = trgPath ;  // save a copy of original target
      fmFType trgType = fn.fType ;  // target's file type
      short viewFormat = vfcTEXT ;  // viewing format, initially view-as-text

      //* 'true' if source is OpenDocument file with scanable text data *
      bool  isOD_File = this->odIsOpenDocFile ( trgPath ) ;
      //* 'true' if source is OpenXML file with scanable text data *
      bool  isOX_File = this->odIsOpenXmlFile ( trgPath ) ;
      bool  ansiColor  = false,     // 'true' if less should interpret ANSI color
            fmtSrcCode = false,     // 'true' formatting applied to source code (XML, etc.)
            regType    = true,      // 'true' if ultimate target file == fmREG_TYPE
            lnkType    = false,     // 'true' if original source == fmLINK_TYPE
            brokenLink = false,     // 'true' if sym-link target not found
            userCancel = false ;    // 'true' if user cancelled viewing of file

      //* If source file is a symbolic link *
      if ( fn.fType == fmLINK_TYPE )
      {  //* Get stats for sym-link target *
         lnkType = true ;
         gString fntPath ;
         if ( (this->fmPtr->GetLinkTargetStats ( trgPath, fnt, fntPath)) == OK )
         {  //* There is a valid link target. Use it as our source file *
            trgName = fnt.fName ;
            trgType = fnt.fType ;
            trgPath = fntPath ;
            regType = (bool)(fnt.fType == fmREG_TYPE ? true : false ) ;
         }
         else
         {  //* The stat failed, but traces of the broken link may exist.*
            //* Get its name && type if possible.                        *
            if ( fnt.readAcc )
            {
               trgName = fnt.fName ;
               trgType = fnt.fType ;
            }
            else
            { /* serious error */ }
            brokenLink = true ;
            regType = false ;
         }
      }
      //* Source file is neither a 'regular' file nor a link to a regular file.*
      else if ( fn.fType != fmREG_TYPE )
         regType = false ;

      //* We display contents ONLY for 'regular' files, (or as a special case, *
      //* the contents of a symlink), so if target is neither a regular file   *
      //* nor the special case, the radio-button group is not instantiated,    *
      //* and we can shrink the dialog window by an equivalent amount.         *
      if ( ! regType && ! lnkType )
         dlgRows -= 4 ;

      InitCtrl ic[controlsDEFINED] = 
      {
      {  //* 'VIEW' pushbutton - - - - - - - - - - - - - - - - - - - - vfcVIEW *
         dctPUSHBUTTON,                // type:      
         rbtTYPES,                     // rbSubtype: (n/a)
         false,                        // rbSelect:  (n/a)
         short(dlgRows - 5),           // ulY:       upper left corner in Y
         short(dlgCols / 3 - 4),       // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         8,                            // cols:      control columns
         "  ^VIEW  ",                  // dispText:  
         this->cs.pn,                  // nColor:    non-focus color
         this->cs.pf,                  // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "",                           // label:     (n/a)
         ZERO,                         // labY:      (n/a)
         ZERO,                         // labX       (n/a)
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcCANCEL],               // nextCtrl:  link in next structure
      },
      {  //* 'CANCEL" pushbutton - - - - - - - - - - - - - - - - - - vfcCANCEL *
         dctPUSHBUTTON,                // type:      
         rbtTYPES,                     // rbSubtype: (n/a)
         false,                        // rbSelect:  (n/a)
         short(ic[vfcVIEW].ulY),       // ulY:       upper left corner in Y
         short(dlgCols / 3 * 2 - 4),   // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         8,                            // cols:      control columns
         " ^CANCEL ",                  // dispText:  
         this->cs.pn,                  // nColor:    non-focus color
         this->cs.pf,                  // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "",                           // label:     (n/a)
         ZERO,                         // labY:      (n/a)
         ZERO,                         // labX       (n/a)
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         ((regType || lnkType) ? &ic[vfcTEXT] : NULL),// nextCtrl:  
      },
      {  //* 'Text' radio button   - - - - - - - - - - - - - - -  vfcTEXT   *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         true,                         // rbSelect:  initially set
         3,                            // ulY:       upper left corner in Y
         3,                            // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Text",                      // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcHEXB],                 // nextCtrl:  link in next structure
      },
      {  //* 'HexByte' radio button  - - - - - - - - - - - - - -  vfcHEXB   *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcTEXT].ulY + 1),   // ulY:       upper left corner in Y
         ic[vfcTEXT].ulX,              // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Hex (8-bit)",               // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcHEXW],                 // nextCtrl:  link in next structure
      },
      {  //* 'HexWord' radio button  - - - - - - - - - - - - - -  vfcHEXW   *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcHEXB].ulY + 1),   // ulY:       upper left corner in Y
         ic[vfcHEXB].ulX,              // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "He^x (16-bit)",              // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcDECI],                 // nextCtrl:  link in next structure
      },
      {  //* 'Decimal' radio button  - - - - - - - - - - - - - - -  vfcDECI *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         ic[vfcTEXT].ulY,              // ulY:       upper left corner in Y
         short(ic[vfcTEXT].ulX + 19),  // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Decimal",                   // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcOCT],                  // nextCtrl:  link in next structure
      },
      {  //* 'Octal' radio button    - - - - - - - - - - - - - -  vfcOCT    *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcDECI].ulY + 1),   // ulY:       upper left corner in Y
         ic[vfcDECI].ulX,              // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Octal",                     // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcBIN]                   // nextCtrl:  link in next structure
      },
      {  //* 'Binary' radio button   - - - - - - - - - - - - - - -   vfcBIN *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcOCT].ulY + 1),    // ulY:       upper left corner in Y
         ic[vfcOCT].ulX,               // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Binary",                    // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcLINK]                  // nextCtrl:  link in next structure
      },
      {  //* 'View Link' radio button  - - - - - - - - - - - - - -  vfcLINK *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         ic[vfcDECI].ulY,              // ulY:       upper left corner in Y
         short(ic[vfcDECI].ulX + 14),  // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^SymLink file",              // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcARCH]                  // nextCtrl:  link in next structure
      },
      {  //* 'List Archive' radio button  - - - - - - - - - - - - - -  vfcARCH *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcLINK].ulY + 1),   // ulY:       upper left corner in Y
         ic[vfcLINK].ulX,              // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "^Archive file",              // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[vfcMETA]                  // nextCtrl:  link in next structure
      },
      {  //* 'Media Metadata' radio button   - - - - - - - - - - - - - vfcMETA *
         dctRADIOBUTTON,               // type:      
         rbtS3s,                       // rbSubtype: standard, 3-wide
         false,                        // rbSelect:  initially reset
         short(ic[vfcARCH].ulY + 1),   // ulY:       upper left corner in Y
         ic[vfcARCH].ulX,              // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         0,                            // cols:      (n/a)
         NULL,                         // dispText:  (n/a)
         rbnColor,                     // nColor:    non-focus color
         rbfColor,                     // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "M^edia Metadata",            // label:     
         ZERO,                         // labY:      
         4,                            // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         NULL
      },
      } ;

      //* If file is not a symlink, disable the symlink Radiobutton *
      if ( ! lnkType )
      {
         ic[vfcLINK].active = false ;
         ic[vfcLINK].nColor = nc.gyR ;
      }
      //* If file is identified as a supported archive   *
      //* type, then set the 'Archive file' Radiobutton. *
      arcType aType ;
      if ( (aType = this->ArchiveTarget ( trgPath )) != atNONE )
      {
         //* If file is identified as an OpenDocument/OpenXML   *
         //* (zip) document, vfcARCH Radiobutton remains active *
         //* although selected viewFormat remains vfcTEXT.      *
         if ( (aType != atODOC) && (aType != atOXML) )
         {
            ic[vfcTEXT].rbSelect = false ;
            ic[vfcARCH].rbSelect = true ;
            viewFormat = vfcARCH ;
         }
      }
      //* Otherwise, target is not a recognized type *
      //* of archive, Disable archive Radiobutton.   *
      else
      {
         ic[vfcARCH].active = false ;
         ic[vfcARCH].nColor = nc.gyR ;
      }

      //* If file is identified as a supported audio file,*
      //* then set the 'Audio Metadata' Radiobutton.      *
      MediafileType mfType ;
      if ( (mfType = this->MediafileTarget ( trgPath )) != mftNONE )
      {
         ic[vfcTEXT].rbSelect = false ;
         ic[vfcMETA].rbSelect = true ;
         viewFormat = vfcMETA ;
      }
      //* If target is not a supported audio file,        *
      //* disable the audio metadata Radiobutton.         *
      else
      {
         ic[vfcMETA].active = false ;
         ic[vfcMETA].nColor = nc.gyR ;
      }

      short ctrY        = this->fulY + this->fMaxY / 2,
            ctrX        = this->fulX + this->fMaxX / 2,
            ulY         = ctrY - dlgRows / 2,
            ulX         = ctrX - dlgCols / 2,
            status = ERR ;
      if ( ulY < this->fulY )    ulY = this->fulY ;   // vertical dialog position

      //* Save the display for the parent dialog window *
      this->dPtr->SetDialogObscured () ;

      //* Initial parameters for dialog window *
      InitNcDialog dInit( dlgRows,        // number of display lines
                          dlgCols,        // number of display columns
                          ulY,            // Y offset from upper-left of terminal 
                          ulX,            // X offset from upper-left of terminal 
                          " View-File Format Options ", // dialog title
                          ncltSINGLE,     // border line-style
                          dColor,         // border color attribute
                          dColor,         // interior color attribute
                          ic              // pointer to list of control definitions
                        ) ;

      //* Instantiate the dialog window *
      NcDialog* dp = new NcDialog ( dInit ) ;

      //* Open the dialog window *
      if ( (dp->OpenWindow()) == OK )
      {
         gString msg ;        // message composition
         winPos wp( 2, 2 ) ;  // message position

         //* Group radio buttons *
         // Programmer's Note: Note special tests for inactive Radiobuttons.
         // This avoids a never-ending loop in the NcDialog library after item selection.
         if ( regType || lnkType )
         {
            short XorGroup1[9] = 
            { vfcTEXT, vfcHEXB, vfcHEXW, vfcDECI, vfcOCT, vfcBIN, 
              /*vcfLINK*/ -1, /*vcfARCH*/ -1, /*vcfMETA*/ -1 } ;
              if ( ic[vfcLINK].active )
                 XorGroup1[6] = vfcLINK ;
              if ( ic[vfcARCH].active )
                 XorGroup1[(XorGroup1[6] == (-1)) ? 6 : 7] = vfcARCH ;
              if ( ic[vfcMETA].active )
                 XorGroup1[(XorGroup1[6] == (-1)) ? 6 :
                           (XorGroup1[7] == (-1)) ? 7 : 8] = vfcMETA ;
            dp->GroupRadiobuttons ( XorGroup1 ) ;
         }

         if ( regType || lnkType )
         {
            msg.compose( L"View the file: '%s' as:", trgName ) ;
            msg.limitCols( dlgCols - 3 ) ;
            dp->WriteString ( --wp.ypos, wp.xpos, msg, hColor ) ;
            if ( fn.fType == fmLINK_TYPE )
               dp->WriteString ( ++wp.ypos, (wp.xpos + 16), "(sym-link target)", dColor ) ;
            wp = { short(dlgRows - 4), 2 } ;
            dp->WriteParagraph ( wp, 
               "  Output is first formated (if necessary) and then\n"
               "  displayed using the 'less' command-line utility.\n"
               "  (view with scroll keys, '/' to search, 'q' to quit)", dColor ) ;
         }
         else
         {
            msg.compose( L"The file: '%s'", trgName ) ;
            msg.limitCols( dlgCols - 3 ) ;
            dp->WriteString ( wp.ypos++, wp.xpos, msg, dColor ) ;
            msg.compose( L"is a %s", (const char*)(fmtypeName[trgType]) ) ;
            wp = dp->WriteString ( wp.ypos, wp.xpos, msg, dColor ) ;
            if ( brokenLink )
               dp->WriteString ( wp.ypos, wp.xpos, "  (broken link)", dColor ) ;
            else if ( fn.fType == fmLINK_TYPE )
               dp->WriteString ( wp.ypos, wp.xpos, "  (sym-link target)", dColor ) ;
            dp->WriteString ( ++wp.ypos, (wp.xpos = 2), 
                     "Only 'Regular-file' contents may be displayed.", dColor ) ;
            dp->NextControl () ; // place focus on 'Cancel' button
         }

         dp->RefreshWin () ;                 // make the text visible

         uiInfo   Info ;                     // user interface data returned here
         short    icIndex = ZERO ;           // index of control with input focus
         bool     done = false ;             // loop control
         while ( ! done )
         {
            //* If focus is currently on a Pushbutton   *
            if ( ic[icIndex].type == dctPUSHBUTTON )
            {
               if ( Info.viaHotkey )
                  Info.HotData2Primary () ;
               else
                  icIndex = dp->EditPushbutton ( Info ) ;
   
               //* If a Pushbutton was pressed *
               if ( Info.dataMod != false )
               {
                  //* If user has pressed 'VIEW' and we are viewing either *
                  //* a 'regular' file OR the contents of a 'symlink' file *
                  if ( (Info.ctrlIndex == vfcVIEW) && 
                       (regType || (lnkType && viewFormat == vfcLINK)) )
                     status = OK ;
                  else if ( Info.ctrlIndex == vfcCANCEL )
                     userCancel = true ;
                  done = true ;
               }
            }
            //* If focus is currently on a Radio Button *
            else if ( ic[icIndex].type == dctRADIOBUTTON )
            {
               if ( Info.viaHotkey )
                  Info.HotData2Primary () ;
               else
                  icIndex = dp->EditRadiobutton ( Info ) ;
               viewFormat = Info.selMember ; // 'selected' button of group
            }
            if ( ! done && ! Info.viaHotkey )
            {
               if ( Info.keyIn == nckSTAB )
                  icIndex = dp->PrevControl () ; 
               else
                  icIndex = dp->NextControl () ;
            }
         }  // while()
      }
      else              // dialog did not open (probably memory allocation error)
      {
         this->DisplayStatus ( ecMEMALLOC ) ;
      }

      if ( dp != NULL )                         // close the window
         delete ( dp ) ;

      //* Restore display of parent dialog window *
      this->dPtr->RefreshWin () ;

      //* If user has made a viewing choice *
      if ( status == OK )
      {
         //* Escape any "special characters" in the target pathspec.*
         this->fmPtr->EscapeSpecialChars ( trgPath, escMIN ) ;

         gString  cmdBuff ;
         switch ( viewFormat )
         {
            case vfcHEXB:           // view as hexidecimal (byte)
               cmdBuff.compose( L"hexdump -Cv \"%S\" | less -c", trgPath.gstr() ) ;
               break ;
            case vfcHEXW:           // view as hexidecimal (word)
               cmdBuff.compose( L"hexdump -xv \"%S\" | less -c", trgPath.gstr() ) ;
               break ;
            case vfcDECI:           // view as decimal
               cmdBuff.compose( L"hexdump -dv \"%S\" | less -c", trgPath.gstr() ) ;
               break ;
            case vfcOCT:            // view as octal
               cmdBuff.compose( L"hexdump -bv \"%S\" | less -c", trgPath.gstr() ) ;
               break ;
            case vfcBIN:            // view as binary
               //* If successful, path of decoded data is returned in trgPath.*
               if ( (status = this->vfcDecodeBinary ( origPath, trgPath )) == OK )
                  cmdBuff.compose( L"less -c \"%S\"", trgPath.gstr() ) ;
               break ;
            case vfcARCH:           // view archive contents
               //* If successful, path of decoded data is returned in trgPath.*
               if ( (status = this->vfcDecodeArchive ( origPath, trgPath )) == OK )
                  cmdBuff.compose( L"less -c \"%S\"", trgPath.gstr() ) ;
               break ;
            case vfcLINK:           // view symbolic link file
               //* If successful, path of decoded data is returned in trgPath.*
               if ( (status = this->vfcDecodeSymlink ( origPath, trgPath )) == OK )
                  cmdBuff.compose( L"less -c \"%S\"", trgPath.gstr() ) ;
               break ;
            case vfcMETA:           // view media metadata
               //* If successful, path of decoded data is returned in trgPath.*
               if ( (status = this->vfcDecodeMetadata ( origPath, trgPath, mfType )) == OK )
                  cmdBuff.compose( L"less -c \"%S\"", trgPath.gstr() ) ;
               break ;
            case vfcTEXT:           // view as text
            default:
               //* For HTML and XML source code, add ANSI colorization.*
               {
                  short scode = ZERO ;
                  if ( (trgPath.find( ".xml" )) == (trgPath.gschars() - 5) )
                     scode = 1 ;
                  else if ( ((trgPath.find( ".html" )) == (trgPath.gschars() - 6)) ||
                            ((trgPath.find( ".htm" )) == (trgPath.gschars() - 5)) )
                     scode = 2 ;
                  else if ( (vfcIsPerlSource ( trgPath )) )
                  {
                     scode = 3 ;
                  }
                  if ( scode > ZERO )
                  {
                     gString tmpPath ;
                     this->fmPtr->CreateTempname ( tmpPath ) ;
                     bool success = false ;
                     if ( scode == 1 )
                        success = this->vfcFormatXML_Source ( origPath, tmpPath ) ;
                     else if ( scode == 2 )
                        success = this->vfcFormatHTML_Source ( origPath, tmpPath ) ;
                     else if ( scode == 3 )
                        success = this->vfcFormatPERL_Source ( origPath, tmpPath ) ;
                     if ( success )
                     {
                        trgPath = tmpPath ;
                        fmtSrcCode = ansiColor = true ;
                     }
                  }

                  //* If target is a supported OpenDocument or   *
                  //* OpenXML file, capture and decode the text  *
                  //* data from the container archive.           *
                  else if ( (aType == atODOC) || (aType == atOXML) )
                     this->vfcDecodeOpenDoc ( origPath, trgPath, aType ) ;

                  //* For other plain text files, enable decoding *
                  //* of ANSI color escape sequences (if present).*
                  //* Currently, viewing the unprocessed ANSI     *
                  //* codes as plain text is not supported.       *
                  else
                     ansiColor = true ;
               }

               cmdBuff.compose( L"less -c \"%S\"", trgPath.gstr() ) ;

               if ( ansiColor )
               {
                  //* If specified, add the "-R" parameter to 'less' *
                  //* invocation, instructing it to interpret the    *
                  //* ANSI color sequences rather that displaying    *
                  //* them as ordinary text.                         *
                  cmdBuff.insert( L" -R", 4 ) ;
               }
               break ;
         }

         if ( status == OK )
         {
            // Programmer's Note: If user exits less with a CTRL+C, then the 
            // return value shows error, even though there may be no error.
            status = this->dPtr->ShellOut ( soX, cmdBuff.ustr() ) ;
         }
      }
      if ( status != OK && userCancel == false )
      {
         this->dPtr->RefreshWin () ;
         const char* sorryMsg[] = 
         {
            "  ALERT  ",
            " ",
            "Sorry, unable to view contents of specified file.",
            NULL
         } ;
         this->InfoDialog ( sorryMsg ) ;
      }

      //* If a temp file was created, delete it now.*
      fmFType ft ;
      if ( (viewFormat == vfcLINK || viewFormat == vfcBIN || viewFormat == vfcMETA ||
            viewFormat == vfcARCH || isOD_File || isOX_File || fmtSrcCode) 
           && (TargetExists ( trgPath, ft )) )
      {
         this->fmPtr->DeleteTempname ( trgPath ) ;
      }
   }
   return status ;

}  //* End ViewFileContents() *

//*************************
//*   vfcDecodeSymlink    *
//*************************
//******************************************************************************
//* Read and format the contents of a symbolic link file, then write it as a   *
//* plain text temporary file. Called only by ViewFileContents() method.       *
//*                                                                            *
//* Input  : trgPath : path/filename of target                                 *
//*          tmpPath : receives path/filename of temporary file to be displayed*
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*          ERR if file not found or unable to read file                      *
//******************************************************************************

short FileDlg::vfcDecodeSymlink ( const gString& trgPath, gString& tmpPath )
{
   const short hexCOLS = 49 ;
   short status = ERR ;

   tnFName slStats ;
   if ( (this->fmPtr->GetFileStats ( slStats, trgPath )) == OK )
   {
      UINT fSize = (UINT)slStats.fBytes ; // (narrowing, but OK)
      char slBuff[fSize + 2] ;
      unsigned slBytesIn = readlink ( trgPath.ustr(), slBuff, fSize ) ;
      //* If we have data, get a unique temp-file filespec.*
      if ( slBytesIn > ZERO &&
           ((this->fmPtr->CreateTempname ( tmpPath )) != false) )
      {
         ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
         if ( ofs.is_open() )
         {
            gString gsOut ;            // formatted output
            this->fmPtr->ExtractFilename ( gsOut, trgPath ) ;
            ofs << "Contents of Symbolic link file: '"
                << gsOut.ustr() << "'\n" << endl ;

            char hxBuff[32],           // hex output buffer
                 ccBuff[32] ;          // character output buffer
            unsigned slBytesOut = ZERO,// source index
                  oIndex ;             // output index
            short cols ;               // formatted columns

            while ( slBytesOut < slBytesIn )
            {
               cols = ZERO ;
               //* Initialize the output buffers *
               for ( oIndex = ZERO ; oIndex < 16 ; oIndex++ )
               {
                  hxBuff[oIndex] = ZERO ;
                  ccBuff[oIndex] = ' ' ;
               }
               //* Format and output a line of data *
               for ( short i = ZERO ; (i < 16) && (slBytesOut < slBytesIn) ; i++ )
               {
                  hxBuff[i] = slBuff[slBytesOut] ;
                  if ( isprint ( slBuff[slBytesOut] ) )
                     ccBuff[i] = slBuff[slBytesOut] ;
                  else
                     ccBuff[i] = '.' ;
                  ++slBytesOut ;
                  cols += 3 ;
                  if ( i == 8 )
                     cols += 2 ;
               }
               gsOut.compose( L"%02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX - "
                               "%02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX",
                              &hxBuff[ 0], &hxBuff[ 1], &hxBuff[ 2], &hxBuff[ 3], 
                              &hxBuff[ 4], &hxBuff[ 5], &hxBuff[ 6], &hxBuff[ 7],
                              &hxBuff[ 8], &hxBuff[ 9], &hxBuff[10], &hxBuff[11],
                              &hxBuff[12], &hxBuff[13], &hxBuff[14], &hxBuff[15] ) ;
               if ( cols < hexCOLS )      // partial (last) output line
               {
                  gsOut.limitCols( cols ) ;
                  gsOut.append( "                                               " ) ;
                  gsOut.limitCols( hexCOLS ) ;
               }
               ofs << gsOut.ustr() ;
               gsOut.compose( L" [%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c]",
                              &ccBuff[ 0], &ccBuff[ 1], &ccBuff[ 2], &ccBuff[ 3], 
                              &ccBuff[ 4], &ccBuff[ 5], &ccBuff[ 6], &ccBuff[ 7],
                              &ccBuff[ 8], &ccBuff[ 9], &ccBuff[10], &ccBuff[11],
                              &ccBuff[12], &ccBuff[13], &ccBuff[14], &ccBuff[15] ) ;
               ofs << gsOut.ustr() << endl ;
            }
            ofs << endl ;

            ofs.close() ;
            status = OK ;
         }
      }
   }
   return status ;

}  //* End vfcDecodeSymlink() *

//*************************
//*    vfcDecodeBinary    *
//*************************
//******************************************************************************
//* Read and format the contents of the target as binary output. Write the     *
//* results as a plain text temporary file. Called only by ViewFileContents(). *
//*                                                                            *
//* Input  : trgPath : path/filename of target                                 *
//*          tmpPath : receives path/filename of temporary file to be displayed*
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*          ERR if file not found or unable to read file                      *
//******************************************************************************

short FileDlg::vfcDecodeBinary ( const gString& trgPath, gString& tmpPath )
{
   const char* binHeader = 
   "OFFSET BINARY BYTES                                                     ASCII\n"
   "====== ==============================================================  =======" ;
   short status = ERR ;

   //* Get a unique temp-file filespec *
   if ( (this->fmPtr->CreateTempname ( tmpPath )) != false )
   {
      //* Open the temp file for writing *
      ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
      if ( ofs.is_open() )
      {
         gString gsOut ;            // formatted output
         ofs << binHeader << endl ; // write the header

         //* Open the target file for reading *
         ifstream ifs ( trgPath.ustr(), ifstream::in | ifstream::binary ) ;
         if ( ifs.is_open() )
         {  //* Read the file in 7-byte chunks until EOF. *
            // Note: chunk size is chosen to keep output within 80 columns.
            ULONG readCNT = 7, cnt, byteTotal = ZERO ;
            char inBuff[readCNT * 2] ;
            const char p = '.' ;    // placeholder character
            bool done = false ;
            while ( ! done )
            {
               ifs.read ( inBuff, readCNT ) ;
               if ( (cnt = ifs.gcount()) == readCNT )
               {
                  gsOut.compose(
                     L"%06lu "
                      "%.8hhb %.8hhb %.8hhb %.8hhb %.8hhb %.8hhb %.8hhb "
                      "|%c%c%c%c%c%c%c|",
                     &byteTotal,
                     &inBuff[0], &inBuff[1], &inBuff[2], &inBuff[3], 
                     &inBuff[4], &inBuff[5], &inBuff[6],  
                     (((inBuff[0] >= SPACE && inBuff[0] <= TILDE)) ? &inBuff[0] : &p),
                     (((inBuff[1] >= SPACE && inBuff[1] <= TILDE)) ? &inBuff[1] : &p),
                     (((inBuff[2] >= SPACE && inBuff[2] <= TILDE)) ? &inBuff[2] : &p),
                     (((inBuff[3] >= SPACE && inBuff[3] <= TILDE)) ? &inBuff[3] : &p),
                     (((inBuff[4] >= SPACE && inBuff[4] <= TILDE)) ? &inBuff[4] : &p),
                     (((inBuff[5] >= SPACE && inBuff[5] <= TILDE)) ? &inBuff[5] : &p),
                     (((inBuff[6] >= SPACE && inBuff[6] <= TILDE)) ? &inBuff[6] : &p) ) ;
                  ofs << gsOut.ustr() << endl ;
                  byteTotal += cnt ;
               }
               else if ( cnt > ZERO )
               {  //* Output the partial chunk *
                  gsOut.compose(
                     L"%06lu -------- -------- -------- -------- -------- -------- -------- |",
                     &byteTotal ) ;
                  gString gsBin ;
                  wchar_t  a ;
                  ULONG indx ;
                  for ( indx = ZERO ; indx < cnt ; ++indx )
                  {  //* Format binary byte *
                     gsBin.compose( L"%.8hhb", &inBuff[indx] ) ;
                     gsOut.replace( L"--------", gsBin.gstr() ) ;
                     //* Format ASCII equivalent *
                     a = (inBuff[indx] >= SPACE && inBuff[indx] <= TILDE) ? inBuff[indx] : p ;
                     gsOut.append( a ) ;
                  }
                  //* Pad until line is complete *
                  a = p ;
                  while ( indx++ < readCNT )
                     gsOut.append( a ) ;
                  gsOut.append( L'|' ) ;
                  ofs << gsOut.ustr() << endl ;
               }
               if ( cnt < readCNT )
                  done = true ;
            }
            ifs.close () ;          // close the target file
            status = OK ;           // return the good news
         }
         ofs.close() ;
      }
   }
   return status ;

}  //* End vfcDecodeBinary() *

//*************************
//*   vfcDecodeArchive    *
//*************************
//******************************************************************************
//* Read and format the contents of:                                           *
//*   a) tar archive file,                                                     *
//*   b) zip archive file,                                                     *
//*   c) OpenDocument and OpenXML files (special case of 'zip' archive)        *
//*   d) ePub document (special case of 'zip' archive)                         *
//* then write it as a plain text temporary file.                              *
//* Called only by ViewFileContents() method.                                  *
//*                                                                            *
//* Input  : trgPath : path/filename of target                                 *
//*          tmpPath : receives path/filename of temporary file to be displayed*
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*            Note that if the target is not a valid archive, then the        *
//*            appropriate error message will be written to the temp file, so  *
//*            this will not be reported to caller as an error.                *
//*          ERR if file not found or unable to read file                      *
//******************************************************************************

short FileDlg::vfcDecodeArchive ( const gString& trgPath, gString& tmpPath )
{
   short status = ERR ;

   if ( (this->fmPtr->CreateTempname ( tmpPath )) != false )
   {
      ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
      if ( ofs.is_open() )
      {
         gString gsOut ;            // formatted output
         this->fmPtr->ExtractFilename ( gsOut, trgPath ) ;
         ofs << "Contents of archive file: '"
             << gsOut.ustr() 
             << "'\n========================================"
                "====================" << endl ;
         ofs.close() ;
         char cmdBuff[gsMAXBYTES * 3] ;
         arcType aType = this->ArchiveTarget ( trgPath ) ;

         //* Escape any "special characters" which *
         //* might be misinterpreted by the shell. *
         gString escPath = trgPath ;
         this->EscapeSpecialChars ( escPath, escMIN ) ;

         //* 'zip' archives, ePub documents and OpenDocument/OpenXML files *
         if ( (aType == atZIP) || (aType == atEPUB) || 
              (aType == atODOC) || (aType == atOXML) )
         {
            snprintf ( cmdBuff, (gsMAXBYTES * 3), 
                       "unzip -lv \"%S\" 1>>\"%S\" 2>>\"%S\"", 
                       escPath.gstr(), tmpPath.gstr(), tmpPath.gstr() ) ;
         }

         //* 'tar' archives *
         else //if ( aType != NONE )
         {
            snprintf ( cmdBuff, (gsMAXBYTES * 3), 
                       "tar -tvvf \"%S\" 1>>\"%S\" 2>>\"%S\"", 
                       escPath.gstr(), tmpPath.gstr(), tmpPath.gstr() ) ;
         }
         this->fmPtr->Systemcall ( cmdBuff ) ;

         status = OK ;
      }
   }
   return status ;

}  //* End vfcDecodeArchive() *

//*************************
//*   vfcDecodeOpenDoc    *
//*************************
//******************************************************************************
//* Extract the text data from an OpenDocument or OpenXML (MS-Office) file,    *
//* and format it for viewing. Caller has verified that target is a supported  *
//* OpenDocument/OpenXML file, and type is indicated by the 'aType' parameter. *
//* Called only by ViewFileContents() method.                                  *
//*                                                                            *
//* Input  : srcPath : source filespec                                         *
//*          tmpPath : receives path/filename of temporary file to be displayed*
//*          aType   : archive type, either atODOC or atOXML                   *
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*          ERR if file not found or unable to decode file                    *
//******************************************************************************

short FileDlg::vfcDecodeOpenDoc ( const gString& srcPath, gString& tmpPath, arcType aType )
{
   gString contentSpec,             // raw XML text extracted to this file
           csPath ;                 // path of application's temp directory
   fmFType fType ;                  // file type code
   short status = ERR ;

   //* Create the target temp file *
   this->fmPtr->CreateTempname ( tmpPath ) ;
   //* Isolate the tempfile path *
   this->fmPtr->ExtractPathname ( csPath, tmpPath ) ;

   //* Extract the XML containing the OpenDocument text data *
   if ( (this->odExtractOD_Content ( srcPath, csPath, contentSpec, aType )) &&
        (this->TargetExists ( tmpPath, fType )) )
   {
      //* Scan the XML source, extract and format the text data.*
      if ( (this->odExtractOD_Text ( srcPath, contentSpec, tmpPath, aType )) != false )
         status = OK ;
   }
   this->DeleteFile ( contentSpec ) ;  // delete the XML source file

   return status ;

}  //* End vfcDecodeOpenDoc() *

//*************************
//*   vfcDecodeMetadata   *
//*************************
//******************************************************************************
//* Read and format the metadata for a supported audio or video file.          *
//* Formatted metadata are written to a plain text temporary file.             *
//* Called only by ViewFileContents() method.                                  *
//*                                                                            *
//* Metadata is the description of the file's media contents which is to be    *
//* formatted in human-readable form.                                          *
//*                                                                            *
//* Input  : srcPath : source filespec                                         *
//*          tmpPath : receives path/filename of temporary file to be displayed*
//*          mfType  : target media identifier                                 *
//*                                                                            *
//* Returns: OK if successful,                                                 *
//*            Note that if the target is not a supported media filetype then  *
//*            the appropriate error message will be written to the temp file, *
//*            so this will not be reported to caller as an error.             *
//*          ERR if file not found or unable to read file                      *
//******************************************************************************

short FileDlg::vfcDecodeMetadata ( const gString& srcPath, gString& tmpPath,
                                   MediafileType mfType )
{
   short status = ERR ;

   if ( (this->fmPtr->CreateTempname ( tmpPath )) != false )
   {
      ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
      if ( ofs.is_open() )
      {
         ofs << "Metadata for Media File\n"
                "=======================" << endl ;

         //* Determine whether this is a supported filetype *
         if ( mfType != mftNONE )
         {
            status = OK ; // any errors encountered will be embedded in the output
            if ( mfType == mftMP3 )
               this->ExtractMetadata_MP3 ( ofs, srcPath ) ;
            else if ( mfType == mftOGG )
               this->ExtractMetadata_OGG ( ofs, srcPath ) ;
            else if ( mfType == mftM4A )
               this->ExtractMetadata_M4A ( ofs, srcPath, tmpPath ) ;
         }

         if ( status != OK )
            ofs << "\n** Unsupported filetype OR Unable to extract metadata **\n" << endl ;

         ofs.close() ;
      }
   }
   return status ;

}  //* End vfcDecodeMetadata() *

//*************************
//* vfcFormatHTML_Source  *
//*************************
//******************************************************************************
//* Format an HTML source code document for display by the 'less' utility.     *
//* Add ANSI color syntax highlighting.                                        *
//*                                                                            *
//* Input  : srcPath: (by reference) file containing original HTML source code *
//*          tmpPath: (by reference) temp file to receive formatted source code*
//*                                                                            *
//* Returns: 'true' if operation is successful, else 'false'                   *
//******************************************************************************

bool FileDlg::vfcFormatHTML_Source ( const gString& srcPath, const gString& tmpPath )
{
   const wchar_t *MIMEtype = L"<!DOCTYPE HTML",
                  dQuote   = L'"',
                 *openCom  = L"<!--",
                 *closeCom = L"-->" ;
   const short    MIMEtypeLen = 14,
                  maxINBYTES = (gsMAXBYTES - 200) ; // max input line bytes
   const char rAngle   = '>' ;
              //* ANSI color codes *
   const char *bkColor = "\033[0;30m", // tag delimiters
              *maColor = "\033[0;35m", // HTML command
              *grColor = "\033[0;32m", // HTML command arguments
              *blColor = "\033[0;34m", // display text
              *reColor = "\033[1;31m" ;// HTML comments
   char inBuff[gsMAXBYTES] ;           // input buffer
   gString gsOut,                      // output buffer
           gsIn ;                      // input parsing
   short indx ;                        // index for parsing and formatting data
   bool status = false ;               // return value

   //* Open source and target files *
   ifstream ifs( srcPath.ustr(), ifstream::in ) ;
   ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
   if ( (ifs.is_open()) && (ofs.is_open()) )
   {
      //* Copy the MIME-type. (If test fails, not a valid HTML file.) *
      ifs.getline( inBuff, maxINBYTES, nckNEWLINE ) ;
      indx = (ifs.gcount() - 2) ;   // protect against Windoze CR/LF combo
      if ( (indx > ZERO) && (inBuff[indx] == '\r') )
         inBuff[indx--] = NULLCHAR ;
      gsOut = inBuff ;
      if ( (ifs.good() || (ifs.gcount() > ZERO)) &&
           ((gsOut.compare( MIMEtype, false, MIMEtypeLen )) == ZERO) &&
           ((inBuff[indx]) == rAngle) )
      {
         gsOut.insert ( maColor, 2 ) ;
         indx = ZERO ;
         while ( (indx = gsOut.find( dQuote, indx )) > ZERO )
         {
            gsOut.insert( grColor, indx++ ) ;
            indx += 7 ; // step over color sequence
            if ( (indx = gsOut.after( dQuote, indx )) > ZERO )
               gsOut.insert( bkColor, indx++ ) ;
         }
         if ( (indx = gsOut.find( L'>' )) > ZERO )
            gsOut.insert( bkColor, indx ) ;
         ofs << gsOut.ustr() << '\n' ;

         // Programmer's Note: We do not test for buffer overflow during capture 
         // and colorization. No real-world line even approaches 4,000 bytes.
         bool done = false ;
         while ( ! done )
         {
            ifs.getline( inBuff, maxINBYTES, nckNEWLINE ) ;
            if ( ifs.gcount() > ZERO )
            {
               indx = (ifs.gcount() - 2) ;   // protect against Windoze CR/LF combo
               if ( (indx > ZERO) && (inBuff[indx] == '\r') )
                  inBuff[indx--] = NULLCHAR ;
               gsIn = inBuff ;
               do
               {
                  if ( (htmlExtractTag ( gsIn, gsOut )) != false )
                  {
                     //* If tag is a comment *
                     if ( (gsOut.find( openCom )) == ZERO )
                     {
                        gsOut.insert( reColor, 4 ) ;
                        indx = gsOut.find( closeCom ) ;
                        gsOut.insert( bkColor, indx ) ;
                     }

                     //* Else tag is a command *
                     else
                     {
                        indx = 1 ;
                        if ( (gsOut.find( fSLASH )) == 1 )
                           ++indx ;
                        gsOut.insert( maColor, indx ) ;
                        indx += 7 ; // step over color sequence
   
                        while ( (indx = gsOut.find( dQuote, indx )) > ZERO )
                        {
                           gsOut.insert( grColor, indx++ ) ;
                           indx += 7 ; // step over color sequence
                           if ( (indx = gsOut.after( dQuote, indx )) > ZERO )
                              gsOut.insert( maColor, indx++ ) ;
                        }
                        if ( (indx = gsOut.find( L'>' )) > ZERO )
                        {
                           if ( gsOut.gstr()[indx - 1] == fSLASH )
                              --indx ;
                           gsOut.insert( bkColor, indx ) ;
                        }
                     }
                     ofs << gsOut.ustr() ;
                  }

                  //* If an incomplete tag OR non-tag data returned.*
                  else if ( (gsOut.gschars()) > 1 )
                  {
                     gsOut.insert( blColor ) ;
                     gsOut.append( bkColor ) ;
                     ofs << gsOut.ustr() ;
                  }
               }
               while ( (gsIn.gschars()) > 1 ) ;
               ofs << endl ;     // end-of-line
            }
            else
               done = true ;
         }  // while()
         status = true ;         // return success
      }  // valid MIME-type

      ifs.close() ;              // close the source file
      ofs.close() ;              // close the target file
   }

   return status ;

}  //* End vfcFormatHTML_Source() *

//*************************
//*    htmlExtractTag     *
//*************************
//******************************************************************************
//* Scan the raw input data and if a complete HTML tag "< ... >" is found,     *
//* move it to the output buffer.                                              *
//*                                                                            *
//* If input does not begin with a tag, then return all the data:              *
//*  a) through the end-of-tag character, or                                   *
//*  b) up to (but not including) the beginning of the next tag, or            *
//*  c) through end of input                                                   *
//*                                                                            *
//* Input  : gsIn   : (by reference) raw source input                          *
//*          gsOut  : (by reference) receives extracted tag data (if found)    *
//*                   if an HTML tag (command) is not the first thing in gsIn, *
//*                   then return the data _up to_ the start of the next tag   *
//*                                                                            *
//* Returns: 'true'  if a complete command is extracted                        *
//*          'false' if incomplete tag, no tag or no data extracted            *
//******************************************************************************

static bool htmlExtractTag ( gString& gsIn, gString& gsOut )
{
   const char lAngle   = '<',          // local char definitions
              rAngle   = '>' ;

   wchar_t w[gsMAXCHARS] ;             // work buffer
   short wIndex = ZERO,                // index into work buffer
         iChars = gsIn.gschars() ;     // number of input characters
   bool goodTag = false ;              // return value

   gsIn.copy( w, gsMAXCHARS ) ;        // make a working copy of input data

   //* Left angle bracket begins a tag.*
   if ( w[ZERO] == lAngle )
   {
      //* Find the bracket that ends the tag.*
      while ( (w[++wIndex] != rAngle) && (wIndex < iChars) ) ;
      if ( w[wIndex] == rAngle )
         w[++wIndex] = NULLCHAR ;
      gsOut = w ;

      if ( wIndex < iChars )     // shift the copied data out
      {
         gsIn.shiftChars( -(wIndex) ) ;
         goodTag = true ;
      }
      else                       // else, all data copied to output
         gsIn.clear() ;
   }

   //* Data does not begin with a tag.                *
   //* Scan to recognizable marker (see above notes). *
   else
   {
      while ( (w[wIndex] != lAngle) && (w[wIndex] != rAngle) && (wIndex < iChars) )
         ++wIndex ;
      if ( w[wIndex] == lAngle )
         w[wIndex] = NULLCHAR ;
      else if ( w[wIndex] == rAngle )
         w[++wIndex] = NULLCHAR ;
      else
         gsIn.clear() ;
      gsOut = w ;
      gsIn.shiftChars( -(wIndex) ) ;
   }

   return goodTag ;

}  //* End htmlExtractTag() *

//*************************
//*  vfcFormatXML_Source  *
//*************************
//******************************************************************************
//* Format an XML source code document for display by the 'less' utility.      *
//* For stand-alone XML documents (which already have linefeed formatting),    *
//* just add ANSI color-coding. However, if linefeeds are not found, add       *
//* linefeed formatting as well as ANSI color.                                 *
//*                                                                            *
//* Input  : srcPath: (by reference) file containing original XML source code  *
//*          tmpPath: (by reference) temp file to receive formatted source code*
//*                                                                            *
//* Returns: 'true' if operation is successful, else 'false'                   *
//******************************************************************************
//* Programmer's Note:                                                         *
//* The XML file extracted from an OpenDocument file does not use newline      *
//* Special Case: The "Content.xml" file extracted from OpenDocument files     *
//* does not use newline characters beyond the MIME-type line.                 *
//* Presumably this is done to save space, but file is pretty much unreadable  *
//* without line breaks. Therefore, we must insert them into the output stream.*
//* Note that the test this is not definitive, but viewing this type of file   *
//* is primarily for development, and users will seldom want to view this kind *
//* file anyway.                                                               *
//* -- How do browsers decide where linebreaks go when displaying XML files    *
//*    that don't have them? Is there a convention about breaking at the end   *
//*    of a command? Where else?                                               *
//*    In general, each command stands alone on a line, except when the        *
//*    command sequence encloses text data. A short sequence of                *
//*    <cmd>stuff>/cmd> is sometimes all on one line.                          *
//*                                                                            *
//* Other XML source files usually DO have linefeeds to format the source.     *
//* Note that Windoze-generated XML will (stupidly) have carriage-returns also.*
//*                                                                            *
//* XML viewers, such as browsers and XML (text) editors: How they do it:      *
//* 1) Each tag is on a line by itself. (with exceptions)                      *
//*    a) The '<' or '</' is black text.                                       *
//*    b) The first token of the tag is an outstanding color                   *
//*       (blue, magenta, etc.) which differs by viewer.                       *
//*    c) Second and subsequent tokens are black, EXCEPT text enclosed within  *
//*       quotations: <style:style style:name="P1" style:family="paragraph">   *
//*       where "P1" and "paragraph" (including the quotes) are an outstanding *
//*       color _different_ from the color of the first token.                 *
//*    d) The '>' or '/>' is black text.                                       *
//*                                                                            *
//*                                                                            *
//* Possible Enhancements:                                                     *
//* 1) Colorize comments: <!--This is a comment.-->  (grey?)                   *
//* 2) Colorize C-data sequences (literal strings): <![CDATA[...]]> (red?)     *
//*                                                                            *
//*                                                                            *
//* ANSI Colors (partial list):                                                *
//* =========== Foreground  Background   ============ Foreground  Background   *
//* Black       \033[0;30m  \033[0;40m   Dark Gray    \033[1;30m  \033[1;40m   *
//* Red         \033[0;31m  \033[0;41m   Bold Red     \033[1;31m  \033[1;41m   *
//* Green       \033[0;32m  \033[0;42m   Bold Green   \033[1;32m  \033[1;42m   *
//* Yellow      \033[0;33m  \033[0;43m   Bold Yellow  \033[1;33m  \033[1;43m   *
//* Blue        \033[0;34m  \033[0;44m   Bold Blue    \033[1;34m  \033[1;44m   *
//* Purple      \033[0;35m  \033[0;45m   Bold Purple  \033[1;35m  \033[1;45m   *
//* Cyan        \033[0;36m  \033[0;46m   Bold Cyan    \033[1;36m  \033[1;46m   *
//* Light Gray  \033[0;37m  \033[0;47m   White        \033[1;37m  \033[1;47m   *
//******************************************************************************

bool FileDlg::vfcFormatXML_Source ( const gString& srcPath, const gString& tmpPath )
{
   const char lAngle   = '<',          // local char definitions
              rAngle   = '>',
              ampChar  = '&',
              semiCol  = ';',
              *bkColor = "\033[0;30m", // ANSI color codes
              *grColor = "\033[0;32m",
//              *blColor = "\033[0;34m",    // (insufficient contrast)
              *maColor = "\033[0;35m",
              *reColor = "\033[1;31m" ;
   const wchar_t *MIMEtype = L"<?xml",
                  dQuote   = L'"' ;
   const short    MIMEtypeLen = 5 ;
   char inBuff[gsMAXBYTES] ;           // input buffer
   gString gsOut ;                     // output buffer
   short indx ;                        // index for parsing and formatting data
   bool status = false ;               // return value

   //* Special Case: The "Content.xml" file of OpenDocument files *
   //* does not use newline characters beyond the MIME-type line. *
   //* Therefore, we must insert them into the output stream.     *
   //* Note that this is not a definitive test, but viewing this  *
   //* type of file is primarily for development, not public view.*
   bool insertNewlines = false ;
   if ( ((indx = srcPath.findlast( fSLASH )) >= ZERO) &&
        ((srcPath.find( "content", ++indx)) >= ZERO) )
      insertNewlines = true ;

   //* Open source and target files *
   ifstream ifs( srcPath.ustr(), ifstream::in ) ;
   ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
   if ( (ifs.is_open()) && (ofs.is_open()) )
   {
      //* Copy the MIME-type. (If test fails, not a valid XML file.) *
      ifs.getline( inBuff, gsMAXBYTES, nckNEWLINE ) ;
      indx = (ifs.gcount() - 2) ;   // protect against Windoze CR/LF combo
      if ( (indx > ZERO) && (inBuff[indx] == '\r') )
         inBuff[indx--] = NULLCHAR ;
      gsOut = inBuff ;
      if ( (ifs.good() || (ifs.gcount() > ZERO)) &&
           ((gsOut.compare( MIMEtype, false, MIMEtypeLen )) == ZERO) && 
           ((inBuff[indx]) == rAngle) )
      {
         gsOut.insert ( maColor, 2 ) ;
         indx = ZERO ;
         while ( (indx = gsOut.find( dQuote, indx )) > ZERO )
         {
            gsOut.insert( grColor, indx++ ) ;
            indx += 7 ; // step over color sequence
            if ( (indx = gsOut.after( dQuote, indx )) > ZERO )
               gsOut.insert( bkColor, indx++ ) ;
         }
         if ( (indx = gsOut.find( L'>' )) > ZERO )
            gsOut.insert( bkColor, indx ) ;
         gsOut.append( L'\n' ) ;
         ofs.write( gsOut.ustr(), (gsOut.utfbytes() - 1) ) ;

         bool done = false ;
         while ( ! done )
         {
            indx = ZERO ;
            ifs.read( inBuff, 1 ) ;
            if ( ifs.gcount() == 1 )
            {
               //* If a tag identified, get remainder of the tag.*
               if ( inBuff[indx] == lAngle )
               {  // Programmer's Note: We do not test for buffer overflow 
                  // during capture of the tag. No real-world tag even approaches 
                  // 4,000 bytes. The longest we have seen is about 2,100 bytes 
                  // and that was in a malformed OpenDocument file.
                  while ( true )
                  {
                     ifs.read( &inBuff[++indx], 1 ) ;
                     if ( ifs.gcount() == 1 )
                     {
                        if ( inBuff[indx] == rAngle )    // end of tag
                        {  //* Complete tag captured *
                           inBuff[++indx] = NULLCHAR ;   // terminate the string
                           gsOut = inBuff ;

                           indx = ZERO ;                 // remove control chars
                           while ( (indx = gsOut.find( L'\r', indx )) >= ZERO )
                              gsOut.erase( L'\r' ) ;
                           indx = ZERO ;
                           while ( (indx = gsOut.find( L'\n', indx )) >= ZERO )
                              gsOut.erase( L'\n' ) ;

                           //* Delimit and colorize first token.*
                           //* Open tag : "<" or "</"
                           //* Token ends with: " " or ">" or "/>"
                           //* a) "<token:token " b) "<token:token>" c) "<token:token/>"
                           //* d) "</token:token " e) </token:token> f) "<
                           indx = 1 ;
                           indx = (gsOut.gstr()[1] == fSLASH) ? 2 : 1 ;
                           gsOut.insert ( maColor, indx ) ;
                           if ( (indx = gsOut.find( nckSPACE )) > ZERO )
                              gsOut.insert( bkColor, indx ) ;
                           else if ( (indx = gsOut.find( rAngle )) > ZERO )
                              gsOut.insert ( bkColor, indx ) ;

                           //* Scan for data between quotation marks.*
                           //* (quotes _should_ come in pairs)       *
                           indx = ZERO ;
                           while ( (indx = gsOut.find( dQuote, indx )) > ZERO )
                           {
                              //* Insert color sequence before opening quote.*
                              gsOut.insert( grColor, indx ) ;
                              //* Step over color sequence and the quote *
                              indx += strlen ( grColor ) + 1 ;
                              if ( (indx = gsOut.after( dQuote, indx )) > ZERO )
                                 gsOut.insert( bkColor, indx++ ) ;
                           }

                           //* Closing angle bracket is black, and if *
                           //* present, the "/>" sequence is black.   *
                           if ( (indx = gsOut.find( L'>' )) > ZERO )
                           {
                              if ( (indx > ZERO) && 
                                   ((gsOut.gstr()[indx - 1]) == fSLASH) )
                                 --indx ;
                              gsOut.insert( bkColor, indx ) ;
                           }

                           if ( insertNewlines )
                              gsOut.append( L'\n' ) ;
                           ofs.write( gsOut.ustr(), (gsOut.utfbytes() - 1) ) ;
                           break ;
                        }
                     }
                     else
                     {
                        done = true ;
                        break ;
                     }
                  }
               }

               //* Check for control characters *
               else if ( (inBuff[indx] >= ZERO) && (inBuff[indx] < nckSPACE) )
               {
                  if ( inBuff[ZERO] != '\r' )
                     ofs.put( inBuff[ZERO] ) ;
               }

               //* Else, assume text content and copy it to output.*
               else
               {
                  // Programer's Note: See note above about "predefined entities" 
                  // and literal values.
                  //* The '&' character begins "predefined entities".        *
                  //* We (foolishly) rely on the data to have correct syntax.* 
                  if ( inBuff[ZERO] == ampChar )
                  {
                     indx = ZERO ;
                     do
                     {
                        ifs.read( &inBuff[++indx], 1 ) ;
                     }
                     while ( (inBuff[indx] != semiCol) && (ifs.gcount() == 1) ) ;
                     inBuff[++indx] = NULLCHAR ;
                     gsOut = inBuff ;
                     gsOut.insert( reColor ) ;
                     gsOut.append( bkColor ) ;
                     ofs.write( gsOut.ustr(), (gsOut.utfbytes() - 1) ) ;
                  }
                  else           // not a special character
                     ofs.put( inBuff[ZERO] ) ;
               }
            }
            else
               done = true ;
         }
         status = true ;         // return success
      }

      ifs.close() ;              // close the source file
      ofs.close() ;              // close the target file
   }
   return status ;

}  //* End vfcFormatXML_Source() *

//*************************
//* vfcFormatPERL_Source  *
//*************************
//******************************************************************************
//* Format a Perl source code document with ANSI color coding for display by   *
//* the 'less' utility.                                                        *
//* (Caller has verified that target contains Perl code.)                      *
//*                                                                            *
//* Input  : srcPath: (by reference) file containing Perl source code          *
//*          tmpPath: (by reference) temp file to receive formatted source code*
//*                                                                            *
//* Returns: 'true' if operation is successful, else 'false'                   *
//******************************************************************************
//* Notes:                                                                     *
//* 1) Perl is written primarily by people who are too lazy and undisciplined  *
//*    to care about good programming practices or even abount proper syntax.  *
//*    For this reason, it is extremely difficult to parse Perl code           *
//*    accurately; therefore, the results of our scan are not particularly     *
//*    elegant. That said, Perl is a wonderful tool in the right hands, and    *
//*    it deserves respect and some special attention. Color coding seemed     *
//*    to be a good beginning.  -- Software Sam 2020-06-01                     *
//*                                                                            *
//* 2) Colorizing of comments was lifted directly from our SourceProfiler      *
//*    utility.                                                                *
//*                                                                            *
//* 3) Colorizing of non-comment tokens: Perl commands, operators, variable    *
//*    names, literals and other token types is modelled on the color coding   *
//*    used by the jEdit and Bluefish editors.                                 *
//*    a) Numeric constants are not identified here. Syntax for scalar values  *
//*       is simply too complex for such a simple parsing algorithm.           *
//*                                                                            *
//* 4) See list of ANSI color codes in 'vfcFormatXML_Source() above.           *
//*                                                                            *
//******************************************************************************

bool FileDlg::vfcFormatPERL_Source ( const gString& srcPath, const gString& tmpPath )
{
              //* ANSI color codes *
   const char *dfltColor = "\033[0;30m", // default color
              *commColor = "\033[0;33m", // comment color
              *cmdColor  = "\033[0;34m", // Perl command color
              *txtColor  = "\033[1;35m", // string literal color (bold purple)
//              *numColor  = "\033[1;31m", // numeric literals
              *varColor  = "\033[0;32m"; // variable name color
   const wchar_t wHASH    = L'#' ;     // hash character identifies comments
   const wchar_t wDOLLAR  = L'$' ;     // '$' indicates a variable name
   const wchar_t wAMPERE  = L'@' ;     // '@' indicates an array
   const wchar_t wDBLQ    = L'"' ;     // '"' string-data delimiter
   const wchar_t wUNSCORE = L'_' ;     // underscore character
   const wchar_t LBRACE   = (L'{') ;   // left curley brace
   const wchar_t RBRACE   = (L'}') ;   // right curley brace
   const wchar_t LBRACKET = (L'[') ;   // left curley brace
   const wchar_t RBRACKET = (L']') ;   // right curley brace
   const wchar_t LPAREN   = (L'(') ;   // left curley brace
   const wchar_t RPAREN   = (L')') ;   // right curley brace
   const wchar_t LANGLE   = (L'<') ;   // left angle bracket
   const wchar_t RANGLE   = (L'>') ;   // right angle bracket
   const wchar_t LDANGLE  = (L'«') ;   // left double-angle quotation mark
   const wchar_t RDANGLE  = (L'»') ;   // right double-angle quotation mark
   const wchar_t* OPENBLK = L"=begin comment" ;
   const wchar_t* CLOSEBLK1 = L"=end comment" ;
   const wchar_t* CLOSEBLK2 = L"=cut" ;

   char inBuff[gsMAXBYTES] ;           // input buffer
   gString gsOut,                      // output buffer
           gsIn,                       // input parsing
           gstmp ;                     // temp storage
   wchar_t bdelim,                     // index beginning delimiter for block comment
           edelim ;                    // index ending delimiter for block comment
   short indx ;                        // index for parsing and formatting data
   bool  open_comment = false,         // 'true' if block comment is open
         open_podcomment = false,      // 'true' if POD block comment is open
         status = false ;              // return value

   //* Open source and target files *
   ifstream ifs( srcPath.ustr(), ifstream::in ) ;
   ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
   if ( (ifs.is_open()) && (ofs.is_open()) )
   {
      bool done = false ;
      while ( ! done )
      {
         ifs.getline( inBuff, gsMAXBYTES, nckNEWLINE ) ;
         if ( ifs.gcount() > ZERO )
         {
            indx = (ifs.gcount() - 2) ;   // protect against Windoze CR/LF combo
            if ( (indx > ZERO) && (inBuff[indx] == '\r') )
               inBuff[indx--] = NULLCHAR ;
            gsIn = inBuff ;

            indx = gsIn.scan() ;

            //* Blank line or whitespace only? *
            if ( gsIn.gstr()[indx] == NULLCHAR )
               ofs << gsIn.ustr() << endl ;

            //* If an open comment block *
            else if ( open_comment )
            {
               ofs << commColor << gsIn.ustr() << dfltColor << endl ;
               if ( (gsIn.find( edelim, indx )) >= ZERO )
                  open_comment = false ;
            }

            //* If an open pod-comment block *
            else if ( open_podcomment )
            {
               ofs << commColor << gsIn.ustr() << dfltColor << endl ;
               if ( ((gsIn.find( CLOSEBLK1, indx )) >= ZERO) ||
                    ((gsIn.find( CLOSEBLK2, indx )) >= ZERO) )
               { open_podcomment = false ; }
            }

            //* Whole line is a comment _OR_ begin a delimited comment block *
            else if ( ((gsIn.gstr()[indx]) == wHASH ) )
            {
               bdelim = gsIn.gstr()[indx + 1] ; // character following the '#'
               if ( bdelim == LBRACE || bdelim == LBRACKET || bdelim == LPAREN || 
                    bdelim == LANGLE || bdelim == LDANGLE )
               {
                  //* Scan to end of current line for closing delimiter.*
                  //* If not found, then comment continues on next line.*
                  switch ( bdelim )
                  {
                     case LBRACE:   edelim = RBRACE ;    break ;
                     case LBRACKET: edelim = RBRACKET ;  break ;
                     case LPAREN:   edelim = RPAREN ;    break ;
                     case LANGLE:   edelim = RANGLE ;    break ;
                     case LDANGLE:  edelim = RDANGLE ;   break ;
                  }
                  if ( (gsIn.find( edelim, (indx + 2) )) < ZERO )
                     open_comment = true ;
               }
               ofs << commColor << gsIn.ustr() << dfltColor << endl ;
            }

            //* If opening a pod-comment block *
            else if ( (gsIn.find( OPENBLK, indx )) == indx )
            {
               ofs << commColor << gsIn.ustr() << dfltColor << endl ;
               open_podcomment = true ;   // and it continues to the next line
            }

            //* Parse each token on the line *
            else
            {
               //* Output any leading whitespace *
               if ( indx > ZERO )
               {  // Programmer's Note: all leading whitespace chars converted to ' '.
                  for ( short i = ZERO ; i < indx ; ++i )
                     ofs << ' ' ;
                  gsIn.shiftChars( -(indx) ) ;
                  indx = ZERO ;
               }

               while ( (vfcGetPerlToken ( gsIn, gsOut )) )
               {
                  //* If a comment which follow code on the line *
                  if ( (indx = gsOut.gstr()[ZERO]) == wHASH )
                  {
                     ofs << commColor << gsOut.ustr() 
                         << gsIn.ustr() << dfltColor ;
                     gsIn.clear() ;    // all data for this line written
                     break ;
                  }

                  //* If a literal string, it needs further parsing *
                  //* to locate any embedded variable names.        *
                  else if ( gsOut.gstr()[ZERO] == wDBLQ )
                  {
                     if ( (gsOut.find( wDOLLAR )) > ZERO )
                     {
                        wchar_t wchar ;
                        bool    varname = false ;
                        gstmp = txtColor ;

                        while ( gsOut.gschars() > 1 )
                        {
                           //* Extract a character *
                           wchar = gsOut.gstr()[ZERO] ;
                           gsOut.shiftChars( -1 ) ;

                           //* Copy plain text character to temp buffer.*
                           if ( !varname && (wchar != wDOLLAR) )
                              gstmp.append( wchar ) ;

                           //* Variable name identified *
                           else if ( wchar == wDOLLAR )
                           {
                              if ( gstmp.gschars() > 1 )
                                 ofs << gstmp.ustr() ;
                              gstmp.compose( "%s%C", varColor, &wchar ) ;
                              varname = true ;
                           }

                           //* If continuation of variable name *
                           else if ( varname )
                           {
                              if ( (wchar >= L'A' && wchar <= L'Z') ||
                                   (wchar >= L'a' && wchar <= L'z') ||
                                   (wchar >= L'0' && wchar <= L'9') ||
                                   (wchar == wUNSCORE) )
                              {
                                 gstmp.append( wchar ) ;
                              }
                              else
                              {
                                 if ( gstmp.gschars() > 1 )
                                    ofs << gstmp.ustr() ;
                                 gstmp.compose( "%s%C", txtColor, &wchar ) ;
                                 varname = false ;
                              }
                           }
                        }
                        gstmp.append( dfltColor ) ;
                        ofs << gstmp.ustr() ;
                     }
                     else
                        ofs << txtColor << gsOut.ustr() << dfltColor ;
                  }

                  //* If Perl command word *
                  else if ( (vfcIsPerlCommand ( gsOut )) )
                  {
                     ofs << cmdColor << gsOut.ustr() << dfltColor ;
                  }

                  //* If variable name or array name *
                  else if ( (gsOut.gstr()[ZERO] == wDOLLAR) ||
                            (gsOut.gstr()[ZERO] == wAMPERE) )
                  {
                     ofs << varColor << gsOut.ustr() << dfltColor ;
                  }

                  //* Unidentified token. Output with default color.*
                  else
                  {
                     ofs << dfltColor << gsOut.ustr() ;
                  }
               }     // while()

               //* Line has been formatted and written to target.*
               //* Terminate with newline and flush buffer.      *
               ofs << endl ;
            }
         }
         else     // end-of-file
            done = true ;
      }  // while()
      status = true ;         // return success

      ifs.close() ;              // close the source file
      ofs.close() ;              // close the target file
   }     // files open
   return status ;

}  //* End vfcFormatPERL_Source() *

//*************************
//*    vfcIsPerlSource    *
//*************************
//******************************************************************************
//* Non-member Method                                                          *
//* -----------------                                                          *
//* 1) Test the filename extension of the given filespec.                      *
//*    a) If it is one of the known Perl filename extensions, assume that it   *
//*       is Perl code _unless_ 'strict' flag is set.                          *
//*    b) If no filename extension identified, first line of file determines   *
//*       whether file contains Perl code.                                     *
//* 2) Test the first line of the source file to determine whether file        *
//*    actually contains Perl code.                                            *
//*    a) If 'strict' flag is set, then regardless of filename extension,      *
//*       first line of file determines whether file is actually Perl code.    *
//*    b) If no filename extension identified, test the first line of file.j   *
//*    The first line should be similar to:                                    *
//*                    #!/usr/bin/perl                                         *
//*    although the actual path may vary by installation.                      *
//*                                                                            *
//* Input  : trgPath : full filespec of target file                            *
//*          strict  : (optional, 'false' by default)                          *
//*                    if 'false', assume that a Perl filename extension       *
//*                                indicates Perl code                         *
//*                    if 'true',  verify that the first line of the file      *
//*                                indicates Perl code                         *
//*                                                                            *
//* Returns: 'true' if verified as Perl code, else 'false'                     *
//******************************************************************************

static bool vfcIsPerlSource ( const gString& trgPath, bool strict )
{
   char inBuff[gsMAXBYTES] ;     // input buffer
   gString gs ;                  // string search
   bool perlext = false,         // 'true' if recognized Perl filename extension
        noext = false,           // 'true' if no filename extension
        pearl = false ;          // return value

   //* Test the filename extension *
   if ( ((trgPath.find( ".pl" )) == (trgPath.gschars() - 4)) ||
         ((trgPath.find( ".pm" )) == (trgPath.gschars() - 4)) ||
         ((trgPath.find( ".t"  )) == (trgPath.gschars() - 3)) )
   {
      perlext = true ;
      if ( ! strict )
         pearl = true ;
   }
   else if ( (trgPath.findlast( L'.' )) < (trgPath.gschars() - 5) )
   {
      noext = true ;
   }

   //* Test the first line of the target file *
   else if ( strict && (perlext || noext) )
   {
      ifstream ifs( trgPath.ustr(), ifstream::in ) ;
      if ( ifs.is_open() )
      {
         ifs.getline( inBuff, gsMAXBYTES, nckNEWLINE ) ;
         if ( ifs.gcount() > ZERO )
         {
            inBuff[ifs.gcount()] = NULLCHAR ;
            gs = inBuff ;
            if ( ((gs.find( "#!" )) == ZERO) && ((gs.find( "/perl" )) > ZERO) )
               pearl = true ;
         }
         ifs.close() ;
      }
   }
   return pearl ;

}  //* End isPerlSource() *

//*************************
//*    vfcGetPerlToken    *
//*************************
//******************************************************************************
//* Non-member Method                                                          *
//* -----------------                                                          *
//* Extract the first token from 'gssrc' and place it into 'gstrg'.            *
//* The extracted token will include the trailing space characters (if any)    *
//*                                                                            *
//* Input  : gssrc : (by reference) contains source data                       *
//*          gstrg : (by reference) receives extracted token (if any)          *
//*                                                                            *
//* Returns: 'true'  if a token extracted from source                          *
//*          'false' if no remaining source tokens                             *
//******************************************************************************

static bool vfcGetPerlToken ( gString& gssrc, gString& gstrg )
{
   const wchar_t wDBLQ     = L'"' ;             // string-data delimiter
   const wchar_t wBKSTROKE = L'\\' ;            // escape character
   const wchar_t wUNSCORE  = L'_' ;             // underscore character
   const wchar_t wDOLLAR   = L'$' ;             // dollar-sign character
   const wchar_t wAMPERE   = L'@' ;             // ampersand character
   const wchar_t *wptr = gssrc.gstr() ;         // pointer to source data
   wchar_t wbuff[gsMAXCHARS] ;                  // temporary buffer
   short windex = ZERO ;                        // index
   wbuff[ZERO] = NULLCHAR ;                     // initialize temp buffer
   gstrg.clear() ;                              // initialize the target buffer

   while ( wptr[windex] != NULLCHAR )
   {
      wbuff[windex] = wptr[windex] ;            // copy a character

      if ( wbuff[windex] == SPACE )             // if end-of-token reached
      {  //* Copy spaces between tokens *
         while ( wptr[++windex] == SPACE )
            wbuff[windex] = wptr[windex] ;
         break ;
      }

      //* If first character indicates a variable name.        *
      //* Perl variable names use only a-z, A-Z, 0-9 and '_'   *
      else if ( (windex == ZERO) && 
                ((wbuff[windex] == wDOLLAR) || (wbuff[windex] == wAMPERE)) )
      {
         wchar_t nextwch = wptr[windex + 1] ;
         while ( ((nextwch >= L'a') && (nextwch <= L'z')) ||
                 ((nextwch >= L'A') && (nextwch <= L'Z')) ||
                 ((nextwch >= L'0') && (nextwch <= L'9')) ||
                 (nextwch == wUNSCORE) || (nextwch == SPACE) )
         {
            wbuff[windex] = wptr[windex] ;

            if ( wbuff[windex] == SPACE )             // if end-of-token reached
            {  //* Copy spaces between tokens *
               while ( wptr[++windex] == SPACE )
                  wbuff[windex] = wptr[windex] ;
               break ;
            }
            nextwch = wptr[++windex] ;
         }
         break ;
      }

      else if ( wbuff[windex] == wDBLQ )        // if a string literal, token
      {                                         // is the entire string
         do
         {
            ++windex ;
            wbuff[windex] = wptr[windex] ;
         }
         while ( ((wbuff[windex] != wDBLQ) || 
                  (wbuff[windex] == wDBLQ && wbuff[windex - 1] == wBKSTROKE)) && 
                 (wbuff[windex] != NULLCHAR) ) ;
         if ( wbuff[windex] == wDBLQ )          // copy space between tokens
         {
            ++windex ;
            while ( wptr[windex] == SPACE )
            {
               wbuff[windex] = wptr[windex] ;
               ++windex ;
            }
            wbuff[windex] = NULLCHAR ;
            break ;
         }
      }
      else if ( wbuff[windex] == NULLCHAR )     // end of source data
         break ;
      ++windex ;
   }
   wbuff[windex] = NULLCHAR ;                   // ensure string termination
   if ( windex > ZERO )                         // copy extracted data to target
      gstrg = wbuff ;
   gssrc.erase( gstrg.gstr(), ZERO, true ) ;    // remove extracted data

   return ( bool((gstrg.gschars()) > 1) ) ;

}  //* End vfcGetPerlToken() *

//*************************
//*   vfcIsPerlCommand    *
//*************************
//******************************************************************************
//* Non-member Method                                                          *
//* -----------------                                                          *
//* Compare the specified token to the list of known Perl function names.      *
//* Also tests for certain non-function-name reserved tokens.                  *
//*                                                                            *
//* Input  : gstoken : (by reference) contains a single token                  *
//*                                                                            *
//* Returns: 'true'  if a token is a Perl command, else 'false'                *
//******************************************************************************

static bool vfcIsPerlCommand ( const gString& gstoken )
{
   //* Perl functions list taken from:                 *
   //* "https://perldoc.perl.org/index-functions.html" *
   //* Most common functions are placed near the top of*
   //* the list for efficiency.                        *
   const short CMD_MAX = 258 ;         // number of recognized Perl commands
   const char* CmdNames[CMD_MAX] = 
   {
      "my",                // common function names
      "print",
      "printf",
      "sprintf",
      "colored",
      "if",
      "else",
      "elseif",
      "elsif",
      "for",
      "foreach",
      "while",
      "eof",
      "eq",                // comparisons
      "ge",
      "le",
      "ne",
      "gt",
      "lt",
      "getc",
      "eval",
      "evalbytes",
      "system",
      "use",
      "sub",
      "scalar",
      "grep",
      "hex",
      "index",
      "int",
      "exit",

      "AUTOLOAD",          // less common functions alphabetically
      "abs",         "accept",      "alarm",       "and",         "atan2",

      "BEGIN",       "bind",        "binmode",     "bless",       "break",

      "CHECK",       "caller",      "chdir",       "chmod",
      "chomp",       "chop",        "chown",       "chr",
      "chroot",      "close",       "closedir",    "cmp",
      "connect",     "continue",    "cos",         "crypt",

      "DESTROY",     "__DATA__",    "dbmclose",    "dbmopen",
      "default",     "defined",     "delete",      "die",
      "do",          "dump",

      "END",         "__END__",     "each",        "endgrent",
      "endhostent",  "endnetent",   "endprotoent", "endpwent",
      "endservent",  "exec",        "exists",      "exp",

      "__FILE__",    "fc",          "fcntl",       "fileno",
      "flock",       "fork",        "format",      "formline",

      "getgrent",    "getgrgid",    "getgmam",     "gethostbyaddr",
      "gethostbyname","gethostent", "getlogin",    "getnetbyaddr",
      "getnetbyname","getnetent",   "getpeername", "getpgrp",
      "getppid",     "getpriority", "getprotobyname", "getprotobynumber",
      "getprotoent", "getpwent",    "getpwnam",    "getpwuid",
      "getservbyname","getservbyport","getservent","getsockname",
      "getsockopt",   "given",      "glob",        "gmtime",      "goto",

      "INIT",        "import",      "ioctl",

      "join",

      "keys",        "kill",

      "__LINE__",    "last",        "lc",          "lcfirst",
      "length",      "link",        "listen",      "local",
      "localtime",   "lock",        "log",         "lstat",

      "m ",             // note: single-character command is dangerous
      "map",         "mkdir",       "msgctl",      "msgget",
      "msgrcv",      "msgsnd",

      "next",        "no",          "not",

      "oct",         "open",        "opendir",     "or",
      "ord",         "our",

      "__PACKAGE__", "pack",        "package",     "pipe",
      "pop",         "pos",         "prototype",   "push",

      "q ",             // note: single-character command is dangerous
      "qq",          "qr",          "quotemeta",   "qw",          "qx",

      "rand",        "read",        "readdir",     "readline",
      "readlink",    "readpipe",    "recv",        "redo",
      "ref",         "rename",      "require",     "reset",
      "return",      "reverse",     "rewinddir",   "rindex",      "rmdir",

      "__SUB__",
      "s ",             // note: single-character command is dangerous
      "say",         "seek",        "seekdir",     "select",
      "semctl",      "semget",      "semop",       "send",
      "setgrent",    "sethostent",  "setnetent",   "setpgrp",
      "setpriority", "setprotoent", "setpwent",    "setservent",
      "setsockopt",  "shift",       "shmctl",      "shmget",
      "shmread",     "shmwrite",    "shutdown",    "sin",
      "sleep",       "socket",      "socketpair",  "sort",
      "splice",      "split",       "sqrt",        "srand",
      "stat",        "state",       "study",       "substr",
      "symlink",     "syscall",     "sysopen",     "sysread",
      "sysseek",     "syswrite",

      "tell",        "telldir",     "tie",         "tied",
      "time",        "times",       "tr",          "truncate",

      "UNITCHECK",   "uc",          "ucfirst",     "umask",
      "undef",       "unless",      "unlink",      "unpack",
      "unshift",     "untie",       "until",       "utime",

      "values",      "vec",

      "wait",        "waitpid",     "wantarray",   "warn",
      "when",        "write",

      "-X",
      "x ",             // note: single-character command is dangerous
      "xor",

      "y ",             // note: single-character command is dangerous
   } ;
   const short COLOR_MAX = 16 ;
   const char* ColorGroup[COLOR_MAX] = 
   {
      "black",
      "red",
      "green",
      "yellow",
      "blue",
      "magenta",
      "cyan",
      "white",
      "bright_black",
      "bright_red",
      "bright_green",
      "bright_yellow",
      "bright_blue",
      "bright_magenta",
      "bright_cyan",
      "bright_white",
      // (Note: stand-alone "on_xxxx" varients ignored)
   } ;
   bool status = false ;

   for ( short i = ZERO ; i < CMD_MAX ; ++i )
   {
      if ( (gstoken.find( CmdNames[i] )) == ZERO )
      {
         status = true ;
         break ;
      }
   }
   if ( ! status )
   {
      short indx ;
      for ( short i = ZERO ; i < COLOR_MAX ; ++i )
      {
         indx = gstoken.find( ColorGroup[i] ) ;
         if ( (indx == ZERO) || ((indx == 1) && (gstoken.gstr()[ZERO] == L'\'')) )
         {
            status = true ;
            break ;
         }
      }
   }
   return status ;

}  //* End vfcIsPerlCommand() *

//**************************
//*  OpenExternalProgram   *
//**************************
//******************************************************************************
//* Open the currently-highlighted file. This could be a binary executable,    *
//* a script, a media file, a document or many other types.                    *
//*                                                                            *
//* Input  : cmdBuff: (optional, NULL* by default)                             *
//*                   if specified, return the command string to caller        *
//*                   instead of executing it locally. In this case, caller    *
//*                   should close the NcDialog object, stop the ncurses       *
//*                   engine, fork the process and exit the application by     *
//*                   calling 'execv' or a similar exec function.              *
//*                                                                            *
//* Returns: OK if: a) file launched successfully, or                          *
//*                 b) 'cmdBuff' contains a valid launch string                *
//*          ERR if file not found, unable to launch, or user aborted launch.  *
//******************************************************************************
//* NOTES:                                                                     *
//* ======                                                                     *
//* When the caller selects a file to be opened by the default application for *
//* that file type, there are many (hundreds) of scenarios                     *
//* for what will happen next. However, two basic scenarios present themselves:*
//* 1) An external, GUI application is invoked, in which case our application  *
//*    may continue uninterrupted.                                             *
//* 2) An external console application is invoked which wants the resources of *
//*    our terminal window, in which case we would need to:                    *
//*    a) put FileMangler into hibernation mode,                               *
//*    b) exit FileMangler and give control to the called application, or      *
//*    c) open a new terminal window in which to run the new application.      *
//*                                                                            *
//*  1. Determine whether the targeted file is 'executable'. It could be       *
//*     a compiled application, a shell script, Perl, Python, etc.             *
//*     a) If it is a GUI application, then it can be launched automatically   *
//*        as a new process without affecting our terminal window resources.   *
//*     b) If it is a script, an interpreted file or console application, we   *
//*        must ask the user whether to:                                       *
//*          i) run it in the current window, or                               *
//*         ii) open a new terminal window in which to launch it, or           *
//*        iii) if a script/interpreted file, edit the file.                   *
//*        (Unfortunately, the user may not know or may be unable to decide    *
//*        whether the file runs in a console window.)                         *
//*  2. If the target file is not an executable file, then it must be opened   *
//*     by an external application (GUI or console app).                       *
//*     a) There are certain classes of files which we have confidence will    *
//*        never be opened in a terminal window (e.g. JPEG images, videos).    *
//*        Therefore we test for images, certain media files, LibreOffice      *
//*        documents and miscellaneous other files, and if identified, we      *
//*        launch the default application for those files without the need to  *
//*        prompt the user.                                                    *
//*     b) The more difficult class of files are those which may be opened     *
//*        either by a GUI app _or_ by a console app, depending on the user's  *
//*        system configuration. Some examples would be audio files, and plain *
//*        text files (config files, source code, scripts, etc.).              *
//*        For these, we must ask the user for guidance.                       *
//*  3. Construct a command string for launching the operation, with optional  *
//*     user-supplied arguments.                                               *
//*  4. Protect our console I/O resources.                                     *
//*  5. Launch the external program.                                           *
//*  6. Continue with this application.                                        *
//*     a) Note that we _could_ offer to exit FileMangler and give the external*
//*        application control of the terminal window; however, GUI file       *
//*        managers don't do this, so we aren't sure whether it is wise to     *
//*        offer the user that choice. This would be done at a higher level by *
//*        returning the constructed command to the caller in the 'cmdBuff'    *
//*        argument.                                                           *
//*                                                                            *
//*                                                                            *
//* Execution Logic:                                                           *
//* ----------------                                                           *
//* There are three major choices for launch:                                  *
//*    a) launch from current window and continue immediately (GUI apps only)  *
//*    b) put the application to sleep and launch from current window,         *
//*      (GUI or console app)                                                  *
//*    c) launch from _new_ termwin (GUI or console app)                       *
//*                                                                            *
//* 1) If identified as GUI client, then launch as: "xdg-open client-filename" *
//* 2) If identified as non-GUI client file, then launch as:                   *
//*    "xdg-open non-gui-client-filename" from the specified terminal window.  *
//*          (cmdIndex == upiCUR) --> current termwin                          *
//*          (cmdIndex == upiNEW) --> new termwin                              *
//* 3) If binary executable (exeFlag && binFlag):                              *
//*     a) if (cmdIndex == upiGUI), then launch the application from the       *
//*        current terminal window: "xdg-open bin-exe-file"                    *
//*     b) if ( cmdIndex != upiGUI), then launch from specified terminal       *
//*        window.                                                             *
//* 4) If non-binary executable (exeFlag && !binFlag), i.e. shell scripts or   *
//*    interpreted files (Perl, Ruby, etc), it is assumed that it will execute *
//*    in a terminal window. There are two choices:                            *
//*     a) Execute in the specified terminal window.                           *
//*     b) Edit the file with the default editor for that type:                *
//*         i) (cmdIndex == upiGUI) editor is a GUI application.               *
//*        ii) (cmdIndex == upiGED) editor is a console application.           *
//*                                                                            *
//* Invoking the system's default application for the source file-type.        *
//* -------------------------------------------------------------------        *
//*  1. Media files: images, audio, video.                                     *
//*  2. Documents: pdf, odt, etc.                                              *
//*  This relies on the system script file 'xdg-open' to select the appropriate*
//*  external application. See LaunchDefaultApplication() method for details.  *
//*                                                                            *
//*  Because by default, we redirect stdout and stderr for the call to         *
//*  xdg-open, we retain control of output to the window; however, it is       *
//*  likely that a console application called from the active terminal window  *
//*  will be inaccessible to the user. While this itself is bad, what's worse  *
//*  is that we would be blamed for the mess.                                  *
//*                                                                            *
//*  To avoid this, we can offer to open a second terminal window              *
//*  (unfortuately not a new tab) from which to make the external call.        *
//*  See notes in oepNewTermwin().                                             *
//*                                                                            *
//******************************************************************************

short FileDlg::OpenExternalProgram ( char* cmdBuff )
{
   const char* dfltProg = "xdg-open" ;   // opens the default application

   gString fPath,                // target path string
           argList,              // optional arguments to launch of executable files
           gstmp ;               // general text formatting
   short cmdIndex = ERR,         // user instructions returned from setup dialog
         status = ERR ;          // return value
   bool  guiType = false,        // 'true' if extension matches a GUI type
         exeFlag = false,        // 'true' if user's 'x' bit is set
         binFlag = false,        // 'true' if a binary (compiled) executable
         autoReturn = false ;    // 'true' if close new termwin when called app closes

   //* Initialize caller's buffer *
   if ( cmdBuff != NULL )
      *cmdBuff = NULLCHAR ;

   //* If there is actually a file under the highlight *
   if ( this->fileCount > ZERO )
   {
      //* Retrieve the file stat information *
      tnFName  fStats ;             // copy of basic file info
      this->GetStats ( fStats ) ;   // stats for currently-highlighted file
      this->fmPtr->CatPathFname ( fPath, this->currDir, fStats.fName ) ;
      //* If target file is a symbolic link, check the link target *
      if ( fStats.fType == fmLINK_TYPE )
      {
         tnFName fnt ;        // filename target of symlink stats
         if ( (this->fmPtr->GetLinkTargetStats ( fPath, fnt, gstmp)) == OK )
         {
            fStats = fnt ;    // target found, copy stats and path
            fPath = gstmp ;
         }
      }

      //* Only 'Regular' files may be opened/executed *
      if ( fStats.fType != fmREG_TYPE )
      {
         const char* msgText[] = 
         {
            "  ALERT ALERT  ",
            " ",
            "Sorry, only \"Regular\" files may be opened.",
            " ",
            NULL
         } ;
         this->InfoDialog ( msgText ) ;
         //****************************
         //** Note the early return. **
         //****************************
         return status ;
      }

      //* Test whether filename extension is on our list of supported media    *
      //* and document files to be run by an external (GUI) application.       *
      if ( (guiType = oepGooeyType ( fPath, 
                               (bool)(this->cfgOptions.lCode == lcAutoAudio))) )
      {
         //* 'fPath' contains the filespec with which to *
         //* launch the default GUI application.         *
         status = OK ;
      }

      //* If target cannot be launched automagically,   *
      //* as a GUI client, test for executable filetype.*
      else  // ( ! guiType )
      {
         //* Test whether file has 'executable' permission bits set AND *
         //* whether user has permission to execute the file.           *
         exeFlag = this->fmPtr->ExecutePermission ( fPath, false ) ;
         if ( exeFlag )
            binFlag = oepBinexeTest ( fPath ) ;
      }

      //* If file is not a GUI client, and if config option  *
      //* instructs that programs should be launched from    *
      //* a new window.                                      *
      if ( ! guiType && (this->cfgOptions.lCode == lcSafeLaunch) )
      {
         cmdIndex = upiNEW ;
         status = OK ;
      }

      //* Otherwise, ask user how to proceed, either because *
      //* we can't determine what to do, _OR_ because user   *
      //* specified that we should always ask.               *
      else if ( ! guiType || (this->cfgOptions.lCode == lcManualLaunch) )
      {
         //* Value returned is index of selected Radiobutton: *
         //* (See meaning of return value in called method.)  *
         cmdIndex = this->oepUserPrompt ( fPath, argList, exeFlag, binFlag ) ;
         if ( cmdIndex >= ZERO )
            status = OK ;
      }

      //* If we have the needed information, continue. *
      if ( status == OK )
      {
         //* If opening a non-GUI, non-executable file in a terminal window.*
         //* This will invoke LaunchExternalApplication("xdg-open",filespec)*
         if ( ! exeFlag && ! guiType && (cmdIndex != upiGUI) )
         {  //* Execute 'xdg-open' with specified source file as its argument *
            argList = fPath ;
            fPath   = dfltProg ;
         }

         //* If editing a non-binary executable i.e. a script file,  *
         //* determine whether we should open the editor in a new    *
         //* terminal window.                                        *
         else if ( exeFlag && !binFlag && (cmdIndex == upiGED) )
         {  //* Execute 'xdg-open' in a new terminal window with     *
            //* specified source file as its argument.               *
            // Programmer's Note: Because we do not use a console text editor for 
            // anything, using xdg-open to invoke vim or another console editor 
            // has not been fully tested.
            argList = fPath ;
            fPath   = dfltProg ;
            cmdIndex = upiNEW ;     // force the editor to open in a new termwin
            autoReturn = true ;     // close termwin when app closes
         }

         //* If file is to be opened by the default GUI application.*
         //* This includes:                                         *
         //*  1) automatically-identified GUI client files          *
         //*  2) open any text file with the default editor for     *
         //*     that file type.                                    *
         //*  3) if EDITOR_LAUNCH==0, opening a script/interpreted  *
         //*     file with the default editor for that file type.   *
         if ( (guiType != false) ||
              ((cmdIndex == upiGUI) && (!exeFlag || (exeFlag && ! binFlag))) )
         {
            if ( (this->LaunchDefaultApplication ( fPath.ustr() )) >= ZERO )
            {
               dtbmData tbMsg( "  External Application Launched Successfully!  " ) ;
               this->dPtr->DisplayTextboxMessage ( this->mIndex, tbMsg ) ;
            }
            else  //* Report error in launching the program *
            {
               gstmp = fPath ;
               this->TrimPathString ( gstmp, (INFODLG_WIDTH - 3) ) ;
               const char* nodefaultMsg[] = 
               {
                  "  ALERT ALERT  ",
                  " ",
                  "   Unable to launch default application for:",
                  gstmp.ustr(),
                  " ",
                  NULL
                  
               } ;
               this->InfoDialog ( nodefaultMsg ) ;
               status = ERR ;       // return failure
            }
         }     // (guiType != false)

         //********************************************
         //* Launch according to user's instructions. *
         //********************************************
         else// if ( exeFlag != false )
         {
            //* Assumes GUI application which will not  *
            //* impact active terminal window.          *
            if ( cmdIndex == upiGUI )
            {
               if ( (this->LaunchExternalApplication ( fPath.ustr(), argList )) < ZERO )
               {  //* Report error in launching the program *
                  gstmp = fPath ;
                  this->TrimPathString ( gstmp, (INFODLG_WIDTH - 3) ) ;
                  argList.insert( "    " ) ;
                  const char* nolaunchMsg[] = 
                  {
                     "  ALERT ALERT  ",
                     " ",
                     "   Unable to launch external application.",
                     gstmp.ustr(),
                     argList.ustr(),
                     NULL
                  } ;
                  this->InfoDialog ( nolaunchMsg ) ;
                  status = ERR ;       // return failure
               }
            }

            //* Put FileMangler to sleep and launch     *
            //* application in current terminal window. *
            else if ( cmdIndex == upiCUR )
            {  //* If additional arguments, append them to filespec. *
               if ( argList.gschars() > 1 )
                  fPath.append( " %S", argList.gstr() ) ;
               this->dPtr->ShellOut ( soP, fPath.ustr() ) ;
            }

            //* Open a new terminal window and launch   *
            //* application in the new window.          *
            else  // ( cmdIndex == upiNEW )
            {
               this->fmPtr->CreateTempname ( gstmp ) ;
               DC_Emulator dce( gstmp, fPath, argList, autoReturn ) ;

               //* Scan the system configuration to define *
               //* the new terminal window.                *
               dce.DefineChildEmulator() ;
               this->fmPtr->DeleteTempname ( gstmp ) ; // delete the temp file

               //* Open the new terminal window *
               pid_t fpid = dce.OpenTermwin () ;

               //* The parent process continues execution here. Wait a moment *
               //* (in case the child PID writes an error message to display).*
               // Programmer's Note: We do not like calling the NCurses primitive 
               // here, but the just-launched terminal emulator may completely 
               // trash the display.
               chrono::duration<short, std::milli>aMoment( 250 ) ;
               this_thread::sleep_for( aMoment ) ;
               nc.ClearScreen () ;
               this->dPtr->RefreshWin () ;         // restore parent dialog

               dtbmData tbMsg( "  New Terminal Window Opened Successfully!  " ) ;
               if ( fpid <= ZERO )     // child window did not launch
               {
                  gstmp = "  ERROR: Unable To Open New Terminal Window!  " ;
                  gstmp.copy( tbMsg.textData, gsMAXCHARS ) ;
               }
               this->dPtr->DisplayTextboxMessage ( this->mIndex, tbMsg ) ;
            }  // (cmdIndex == upiNEW)
         }     // (executable != false)
      }        // if(status == OK)
   }           // if(this->fileCount > ZERO)

   return status ;

}  //* OpenExternalProgram() *

//*************************
//*     oepGooeyType      *
//*************************
//******************************************************************************
//* Compare the filename extension of the specified filespec to our list of    *
//* supported client filetypes which we are 'sure' will be run by GUI          *
//* applications.                                                              *
//* -- Note that audio media files are a special case in that the audio media  *
//*    player may, or may not be a GUI application. Therefore, we test for     *
//*    audio filetypes only if caller has verified that the default audio      *
//*    media player is a GUI application (guiAudio != false).                  *
//*                                                                            *
//* If the filespec indicates a file that will be opened by a GUI application, *
//* then the caller will not need to prompt the user for setup information.    *
//*                                                                            *
//* Input  : fPath    : filespec to be scanned                                 *
//*          guiAudio : indicates whether the default audio media player       *
//*                     is a GUI application.                                  *
//*                     'true'  if user has verified GUI audio player          *
//*                     'false' if type of audio player is a console app or if *
//*                             the type of player is unknown                  *
//*                                                                            *
//* Returns: 'true'  if matching filename extension found                      *
//*          'false' if no match found                                         *
//******************************************************************************
//* Notes:                                                                     *
//* ======                                                                     *
//* 1) Document filename extensions:                                           *
//*    a) We include standard LibreOffice / OpenOffice extensions as well as   *
//*       legacy extensions for earlier versions of the office suite including *
//*       the original StarOffice(tm) v:5 extensions.                          *
//*    b) As much as we detest Microsloth and all it stands for, we include    *
//*       the extensions for both the open-document-format versions and the    *
//*       proprietary version 6 extensions.                                    *
//*    c) Also included is the Portable Document Format (PDF) extension.       *
//*    d) We include the most common of the HTML extensions; HOWEVER, these    *
//*       may be removed because we don't actually know whether the user wants *
//*       to view them or edit them. Note also that there are some legacy      *
//*       text-only browsers out there, but we ignore that possibility.        *
//*                                                                            *
//* 2) Audio and video filename extensions are taken from Widipedia's lists of *
//*    common formats and filename extensions.                                 *
//*    a) Note that there is some overlap between audio and video formats as   *
//*       indicated by the filename extension. This is because the extension   *
//*       often indicates the _container format_ rather than the data it       *
//*       contains.                                                            *
//*    b) The intention is that for extensions that could indicate _either_    *
//*       audio or video, audio is assumed.                                    *
//*                                                                            *
//*                                                                            *
//******************************************************************************

static bool oepGooeyType ( const gString& fPath, bool guiAudio )
{
   const short docTYPES = 44 ;
   const wchar_t* DocExtensions[docTYPES]
   {  //* Document files *
      L".odt",    L".ods",    L".odp",    L".odg",    L".odf",    L".sxw",    
      L".sxc",    L".sxi",    L".sxd",    L".sxm",    L".doc",    L".docx",   
      L".docm",   L".dot",    L".dotx",   L".dotm",   L".docb",   L".ppt",    
      L".pot",    L".pps",    L".pptx",   L".pptm",   L".potx",   L".potm",   
      L".ppam",   L".ppsx",   L".ppsm",   L".sldx",   L".sldm",   L".xls",    
      L".xlt",    L".xlm",    L".xlsx",   L".xlsm",   L".xltx",   L".xltm",   
      L".xlsb",   L".xla",    L".xlam",   L".xll",    L".xlw",    L".pdf",    
      L".html",   L".htm",    // see note on HTML files above
   } ;
   const short audioTYPES = 38 ;
   const wchar_t* AudioExtensions[audioTYPES] = 
   {  //* Audio files *
      L".mp3",    L".ogg",    L".oga",    L".mogg",   L".m4a",    L".m4b",    
      L".m4p",    L".wav",    L".wma",    L".flac",   L".aiff",   L".ape",    
      L".aa",     L".aac",    L".aax",    L".act",    L".amr",    L".3gp",    
      L".au",     L".awb",    L".dct",    L".dss",    L".dvf",    L".gsm",    
      L".iklax",  L".isv",    L".mmf",    L".mpc",    L".msv",    L".opus",   
      L".ra",     L".rm",     L".raw",    L".sln",    L".tta",    L".vox",    
      L".wv",     L".8svx",       
   } ;
   const short videoTYPES = 46 ;
   const wchar_t* VideoExtensions[videoTYPES] = 
   {  //* Image and Video files *
      L".jpg",    L".jpeg",   L".jif",    L".jfif",   L".png",    L".gif",    
      L".pcd",    L".fpx",    L".jp2",    L".jpx",    L".j2k",    L".j2c",    
      L".tif",    L".tiff",   
      L".mp4",    L".m4v",    L".vob",    L".avi",    L".webm",   L".mkv",    
      L".ogv",    L".drc",    L".mov",    L".qt",     L".wmv",    L".yuv",    
      L".rmvb",   L".rm",     L".asf",    L".amv",    L".mpg",    L".mp2",    
      L".mpeg",   L".mpv",    L".m2v",    L".svi",    L".3gp",    L".3g2",    
      L".mxf",    L".roq",    L".nsv",    L".flv",    L".f4v",    L".f4p",    
      L".f4a",    L".f4b",    
   } ;


   short extIndex = fPath.findlast( L'.' ) ; // index the filename extension (if any)
   bool goodType = false ;                   // return value

   if ( extIndex >= ZERO )
   {
      for ( short i = ZERO ; i < docTYPES ; ++i )
      {
         if ( (fPath.compare( DocExtensions[i], false, gsMAXCHARS, extIndex )) == ZERO )
         {
            goodType = true ;
            break ;
         }
      }
      if ( ! goodType && guiAudio != false )
      {
         for ( short i = ZERO ; i < audioTYPES ; ++i )
         {
            if ( (fPath.compare( AudioExtensions[i], false, gsMAXCHARS, extIndex )) == ZERO )
            {
               goodType = true ;
               break ;
            }
         }
      }
      if ( ! goodType )
      {
         for ( short i = ZERO ; i < videoTYPES ; ++i )
         {
            if ( (fPath.compare( VideoExtensions[i], false, gsMAXCHARS, extIndex )) == ZERO )
            {
               goodType = true ;
               break ;
            }
         }
      }
   }
   return goodType ;

}  //* End oepGooeyType() *

//*************************
//*     oepBinexeTest     *
//*************************
//******************************************************************************
//* Called only when target file's 'x' (executable) bit is set.                *
//* Test for a binary executable file according to the "ELF" signature in the  *
//* "Executable and Linkable Format" specification.                            *
//*    This signature will be: 0x7F 0x45 0x4C 0x46                             *
//*                                 'E'  'L'  'F'                              *
//*                                                                            *
//* If not a binary executable, it may be assumed that the file is a shell     *
//* script or a source file for one of the interpreted programming languages.  *
//* See below for a list of likely script and interpreted files.               *
//*                                                                            *
//* Input  : fPath    : filespec of file to be scanned                         *
//*                                                                            *
//* Returns: 'true'  if file is identified as a binary executable file         *
//*          'false' otherwise                                                 *
//******************************************************************************
//* List of filename extensions for common scripting languages and source      *
//* files for interpreted languages.                                           *
//* The list is sorted arbitrarily, high-to-low, according to our              *
//* notion of how often they are encountered in a typical system.              *
//*                                                                            *
//* const short extCOUNT = 28 ;                                                *
//* const wchar_t* extList[extCOUNT] =                                         *
//* {//EXTENSION         SIGNATURE               COMMON NAME                   *
//*  //-------------  -------------------------  --------------                *
//*    L".bash",      // #!/bin/bash             Bash                          *
//*    L".sh",        // #!/bin/sh               Bourne                        *
//*    L".csh",       // #!/bin/csh              CShell                        *
//*    L".ksh",       // #!/usr/bin/ksh          Korn shell                    *
//*    L".dash",      // #!/bin/sh               Debian ash                    *
//*    L".pl",        // #!/usr/bin/perl         Perl                          *
//*    L".pm",        //                                                       *
//*    L".t",         //                                                       *
//*    L".py",        // #!/usr/bin/env python3  Python                        *
//*    L".pyw",       //                                                       *
//*    L".rb",        // #!/usr/bin/ruby -w      Ruby                          *
//*    L".rbw",       //                                                       *
//*    L".html",      // <!DOCTYPE html>         HTML                          *
//*    L".htm",       // <html>                                                *
//*    L".xml",       //                                                       *
//*    L".xhtml",     //                                                       *
//*    L".xht",       //                                                       *
//*    L".php",       // <?php                   PHP (incl. .php[3,4,5,6,7,s]) *
//*    L".phtml",     // <!DOCTYPE html>                                       *
//*    L".zsh",       // #!/bin/sh               ZShell                        *
//*    L".tcl",       // #!/usr/bin/tclsh        Tcl script                    *
//*    L".fish",      // #!/bin/fish             Friendly Int                  *
//*    L".tcsh",      // #!/bin/tcsh             TENEX CShell                  *
//*    L".rc",        // #!/usr/bin/rsh          RC                            *
//*    L".scsh",      // #!/usr/local/bin/scsh   Scheme                        *
//*    L".ash",       // #!/bin/sh               Almquist                      *
//*    L".ch",        //                         Ch shell                      *
//*    L".eshell",    //                         Emacs                         *
//* } ;                                                                        *
//*                                                                            *
//* Scripting languages, especially shell scripts often have no filename       *
//* extension, for ease in invoking them. Most other types of executables will *
//* have valid descriptive filename extensions.                                *
//* Practically speaking, three(3) characters are enough to identify a         *
//* signature (if it is properly constructed):                                 *
//*            "#!/". This is the "shebang + path" sequence.                   *
//* Note that although HTML files are interpreted scripts, they are not marked *
//* as executable, so we will never see them here.                             *
//******************************************************************************

static bool oepBinexeTest ( const gString& fPath )
{
   bool binFlag = false ;       // return value

   //* Open the file and extract the first few characters.        *
   //* Note that there is a high likelihood that this is a binary *
   //* (compiled) executable, so we open the file in binary mode. *
   ifstream ifs ( fPath.ustr(), ifstream::in | ifstream::binary ) ;
   if ( ifs.is_open() )
   {
      char ibuff[32] ;
      ifs.read( ibuff, 4 ) ;
      if ( (ibuff[0] == 0x7F) && (ibuff[1] == 'E') && 
                (ibuff[2] == 'L')  && (ibuff[3] == 'F') )
         binFlag = true ;
      ifs.close() ;
   }
   return binFlag ;

}  //* End oepBinexeTest() *

//*************************
//*     oepUserPrompt     *
//*************************
//******************************************************************************
//* Open a dialog for user instructions on how the external file should be     *
//* launched.                                                                  *
//*                                                                            *
//* Input  : fPath    : filespec to be launched                                *
//*          argList  : receives optional, user-specified arguments for        *
//*                     launch (executable files only)                         *
//*          exeFlag  : 'true' if file is executable, else 'false'             *
//*          binFlag  : 'true' if file is identified as a binary (compiled)    *
//*                     executable                                             *
//*              Note: If exeFlag != false && binFlag == false, then file is a *
//*                    shell script or interpreted file (Perl, Python, etc.).  *
//*                                                                            *
//* Returns: selected launch code:                                             *
//*          upiGUI : launch from current terminal window                      *
//*          upiCUR : put application to sleep and launch from current window  *
//*          upiNEW : create a new terminal window from which to launch program*
//*          upiGED : edit non-GUI executable (script) in a new termial window *
//*          (returns -1 if user abort)                                        *
//******************************************************************************
//* Programmer's Note: There is a logical hole in the choices presented to the *
//* user in this method. It is assumed that a script file or interpreted file  *
//* if it is to be executed, will always be executed in a terminal window.     *
//* As a practical matter, this is not a problem, but it is _theoretically_    *
//* possible that the script or interpreted file could be executed by a GUI    *
//* program. We do not support that theoretical possibility.                   *
//*                                                                            *
//******************************************************************************

short FileDlg::oepUserPrompt ( const gString& fPath, gString& argList,
                               bool exeFlag, bool binFlag )
{
   //* Conditional-compile flag for launch of default editor for non-binary    *
   //* executable files (shell scripts, Python, Perl, etc.)                    *
   //* true  : Initialize the upiGED Radiobutton to 'true', i.e. the default   *
   //*         editor for scripts is a GUI application which will be launched  *
   //*         from the current terminal window.                               *
   //* false : Initialize the upiGED Radiobutton to 'false', i.e. the default  *
   //*         editor for scripts is assumed to be a console applicaton        *
   //*         (vi/vim, pico, jed, etc.).                                      *
   const bool GUI_EDITOR = true ;


   const char* progType[][locLang] =
   {
      {  // English
         "This is a binary executable program.",
         "This is a shell script or interpreted program.",
         "This is a non-executable data file."
      },
      {  // Espanol
         "Este es un programa ejecutable binario.",
         "Este es un script de shell o programa interpretado.",
         "Este es un archivo de datos no ejecutable."
      },
      {  // Zhongwen
         "这是一个二进制可执行程序。",
         "这是一个命令脚本或解释程序。",
         "这是一个不可执行的数据文件。"
      },
   } ;
   const char* Labels[][11] = 
   {
      {  // English
         "  L^AUNCH  ",                                              // upiLAUNCH
         "  ^CANCEL  ",                                              // upiCANCEL
         "Launch the program as a GUI application OR open file\n"    // upiGUI
         "using the default (GUI) application for this file type.\n"
         "(FileMangler continues in the current window.)",
         "Put FileMangler to sleep and launch the program\n"         // upiCUR
         "in the current window. (then type EXIT to return.)",
         "Launch the program in a new terminal window.\n"            // upiNEW
         "(FileMangler continues in the current window.)",
         "Additional launch arguments (executable files only)",      // upiARGS
         "default editor is a GUI application",                      // upiGED
         "  Launch External Program  ",                              // dialog title
         "File: ",                                                   // upiHEADER
         "Edit the file using the default text editor.\n"            // upiEDIT
         "(FileMangler continues in the current window.)",
         "If you are unsure which option to choose, be safe, select\n" // upiQHELP
         "      \"Launch the program in a new terminal window.\"",
      },
      {  // Espanol
         "  L^ANZAR  ",                                              // upiLAUNCH
         " ^CANCELAR ",                                              // upiCANCEL
         "Abra el programa como una aplicación GUI, o abra el\n"     // upiGUI
         "archivo usando la aplicación predeterminada para este\n"
         "tipo de archivo.\n"
         "   (FileMangler continúa en la ventana actual.)",
         "Ponga FileMangler en suspensión e inicie el programa en\n" // upiCUR
         "la ventana actual. (luego escribe EXIT para regresar.)",
         "Inicie el programa en una nueva ventana de terminal.\n"    // upiNEW
         "   (FileMangler continúa en la ventana actual.)",
         "Argumentos de inicio adicionales (solo archivos ejecutables)", // upiARGS
         "editor predeterminado es una aplicación GUI",              // upiGED
         "  Lanzar Programa Externo  ",                              // dialog title
         "Archivo: ",                                                // upiHEADER
         "Edite el archivo usando el editor de texto predeter-\n"    // upiEDIT
         "minado. (FileMangler continúa en la ventana actual.)",
         "Si no está seguro de qué opción elegir, seleccione:\n" // upiQHELP
         "  \"Inicie el programa en una nueva ventana de terminal.\"",
      },
      {  // Zhongwen
         "   发起   ",                                                // upiLAUNCH
         "   取消   ",                                                // upiCANCEL
         "Launch the program as a stand-alone (GUI) application.\n"  // upiGUI
         "or Open file using default application for this file\n"
         "type. (FileMangler continues in the current window.)",
         "Put FileMangler to sleep and launch the program\n"         // upiCUR
         "in the current window. (then type EXIT to return.)",
         "Launch the program in a new terminal window.\n"            // upiNEW
         "(FileMangler continues in the current window.)",
         "Additional launch arguments (executable files only)",      // upiARGS
         "默认编辑器是一个 GUI 应用程序",                                // upiGED
         "   启动外部程序   ",                                         // dialog title
         "文件： ",                                                   // upiHEADER
         "使用默认的文本编辑器编辑文件。\n"                               // upiEDIT
         "(FileMangler 在当前窗口中继续。)",
         "If you are unsure which option to choose, be safe, select\n" // upiQHELP
         "      \"Launch the program in a new terminal window.\"",
      },
   } ;

   const short dlgROWS = 23,
               dlgCOLS = 64 ;
   const attr_t dColor = this->cs.sd,  // dialog background color
                hColor = this->cs.em,  // bold text
                rbnColor = dColor,     // radio-button colors
                rbfColor = this->cs.pf ;
   AppLang lang = this->cfgOptions.appLang ;
   short dlgY   = this->fulY,
         dlgX   = (this->fulX + this->fMaxX / 2) - dlgCOLS / 2,
         cIndex = -1 ;    // return value - assume user abort

   // Programmer's Note: This dialog is taller than the minimum application 
   // height. Application height must be >= (dlgROWS + 1), else the dialog 
   // will not open.
   if ( this->dRows < (dlgROWS + 1) )
   {
      dtbmData tbMsg( "Unable to open dialog! Please expand terminal window and try again." ) ;
      this->dPtr->DisplayTextboxMessage ( this->mIndex, tbMsg ) ;
      this->dPtr->UserAlert () ;
      //****************************
      //** Note the early return. **
      //****************************
      return cIndex ;
   }

   //* Adjust dialog position *
   if ( this->dRows < (dlgY + dlgROWS) )
      dlgY = 1 ;
   if ( dlgX < ZERO )
      dlgX = ZERO ;
   while ( ((dlgX + dlgCOLS) > this->dCols) && (dlgX > ZERO) )
      --dlgX ;

   argList.clear() ;       // initialize optional-arguments buffer

   InitCtrl ic[upiCONTROLS] = 
   {
   {  //* 'LAUNCH' Pushbutton - - - - - - - - - - - - - - - - - - -  upiLAUNCH *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      short(dlgROWS - 6),           // ulY:       upper left corner in Y
      short(dlgCOLS / 2 - 10 - 3),  // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      10,                           // cols:      control columns
      Labels[lang][upiLAUNCH],      // dispText
      this->cs.pn,                  // nColor:    non-focus color
      this->cs.pf,                  // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      "",                           // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[upiCANCEL]                // nextCtrl:  link in next structure
   },
   {  //* 'CANCEL' Pushbutton - - - - - - - - - - - - - - - - - - -  upiCANCEL *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      ic[upiLAUNCH].ulY,            // ulY:       upper left corner in Y
      short(ic[upiLAUNCH].ulX + ic[upiLAUNCH].cols + 4),// ulX: upper left corner in X
      1,                            // lines:     (n/a)
      10,                           // cols:      control columns
      Labels[lang][upiCANCEL],      // dispText
      this->cs.pn,                  // nColor:    non-focus color
      this->cs.pf,                  // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      "",                           // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[upiGUI]                   // nextCtrl:  
   },
   {  //* 'Launch-as-GUI' Radiobutton - - - - - - - - - - - - - - - - - upiGUI *
      dctRADIOBUTTON,               // type:      
      rbtS5s,                       // rbSubtype: standard, 3-wide
      false,                        // rbSelect:  initially set
      4,                            // ulY:       upper left corner in Y
      2,                            // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      rbnColor,                     // nColor:    non-focus color
      rbfColor,                     // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      Labels[lang][upiGUI],         // label
      ZERO,                         // labY:      
      6,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[upiCUR]                   // nextCtrl:  link in next structure
   },
   {  //* 'Launch-in-current window' Radiobutton  - - - - - - - - - - - upiCUR *
      dctRADIOBUTTON,               // type:      
      rbtS5s,                       // rbSubtype: standard, 3-wide
      false,                        // rbSelect:  initially set
      short(ic[upiGUI].ulY + 4),    // ulY:       upper left corner in Y
      ic[upiGUI].ulX,               // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      rbnColor,                     // nColor:    non-focus color
      rbfColor,                     // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      Labels[lang][upiCUR],         // label
      ZERO,                         // labY:      
      6,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[upiNEW]                   // nextCtrl:  link in next structure
   },
   {  //* 'Launch-in-new window' Radiobutton  - - - - - - - - - - - - - upiNEW *
      dctRADIOBUTTON,               // type:      
      rbtS5s,                       // rbSubtype: standard, 3-wide
      true,                         // rbSelect:  initially set
      short(ic[upiCUR].ulY + 3),    // ulY:       upper left corner in Y
      ic[upiCUR].ulX,               // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      rbnColor,                     // nColor:    non-focus color
      rbfColor,                     // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      Labels[lang][upiNEW],         // label
      ZERO,                         // labY:      
      6,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[upiARGS]                  // nextCtrl:  link in next structure
   },
   {  //* 'Arguments' Textbox  - - - - - - - - - - - - - - - - - - -   upiARGS *
      dctTEXTBOX,                   // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      short(ic[upiNEW].ulY + 4),    // ulY:       upper left corner in Y
      ic[upiNEW].ulX,               // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      short(dlgCOLS - 4),           // cols:      control columns
      "",                           // dispText:  initially empty
      this->cs.tn,                  // nColor:    non-focus color
      this->cs.tf,                  // fColor:    focus color
      #if LINUX_SPECIAL_CHARS != 0
      tbFileLinux,                  // filter: valid filename chars (incl. Linux "special")
      #else    // BASIC FILENAME FILTER
      tbFileName,                   // filter:    valid filename characters
      #endif   // BASIC FILENAME FILTER
      Labels[lang][upiARGS],        // label
      -1,                           // labY:      
      ZERO,                         // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      NULL                          // nextCtrl:  link in next structure
   },
   {  //* 'GUI Editor' Radiobutton  - - - - - - - - - - - - - - - - -   upiGED *
      dctRADIOBUTTON,               // type:      
      rbtS3p,                       // rbSubtype: parentheses, 3-wide
      GUI_EDITOR,                   // rbSelect:  initially set
      short(ic[upiARGS].ulY + 1),   // ulY:       upper left corner in Y
      short(ic[upiGUI].ulX + ic[upiGUI].labX), // ulX: upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      rbnColor,                     // nColor:    non-focus color
      rbfColor,                     // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      Labels[lang][upiGED],         // label
      ZERO,                         // labY:      
      4,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      false,                        // active:    allow control to gain focus
      NULL                          // nextCtrl:  link in next structure
   },
   } ;

   if ( ! exeFlag )     // supplimentary argument are only for executable files
      ic[upiARGS].active = false ;

   //* Executable scripts and interpreted files are assumed to be non-GUI.     *
   //* (Java source could be an exception, but .jar files are not executable.) *
   //* Adjust the label for the upiGUI radiobutton to offer edit of the file.  *
   //* Activate the upiGED Radiobutton, adjusting control positions.           *
   if ( exeFlag && !binFlag )
   {
      ic[upiGUI].label = Labels[lang][upiEDIT] ;
      ic[upiGED].active = true ;
      ic[upiARGS].nextCtrl = &ic[upiGED] ;
      --ic[upiCUR].ulY ;
      --ic[upiNEW].ulY ;
      --ic[upiARGS].ulY ;
      --ic[upiGED].ulY ;
   }

   //* Save the display for the parent dialog window *
   this->dPtr->SetDialogObscured () ;

   //* Initial parameters for dialog window *
   InitNcDialog dInit( dlgROWS,        // number of display lines
                       dlgCOLS,        // number of display columns
                       dlgY,           // Y offset from upper-left of terminal 
                       dlgX,           // X offset from upper-left of terminal 
                       NULL,           // dialog title
                       ncltSINGLE,     // border line-style
                       this->cs.bb,    // border color attribute
                       dColor,         // interior color attribute
                       ic              // pointer to list of control definitions
                     ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   //* Open the dialog window *
   if ( (dp->OpenWindow()) == OK )
   {
      dp->SetDialogTitle ( Labels[lang][upiCONTROLS], hColor ) ;

      //* Group the radiobuttons *
      short XorGroup[4] = { upiGUI, upiCUR, upiNEW, -1 } ;
      dp->GroupRadiobuttons ( XorGroup ) ;

      //* Draw the static text *
      gString gs = fPath ;
      winPos wp( 1, 2 ) ;
      wp = dp->WriteString ( wp, Labels[lang][upiHEADER], dColor ) ;
      this->TrimPathString ( gs, (dlgCOLS - wp.xpos - 2) ) ;
      dp->WriteString ( wp, gs, hColor ) ;
      ++wp.ypos ;
      short ptIndex = exeFlag ? (binFlag ? 0 : 1) : 2 ;
      dp->WriteString ( wp, progType[lang][ptIndex], dColor ) ;

      LineDef  hLine(ncltHORIZ, ncltDUAL, (dlgROWS - 4), ZERO, dlgCOLS, this->cs.bb) ;
      dp->DrawLine ( hLine ) ;
      dp->WriteParagraph ( (hLine.startY + 1), 3, Labels[lang][upiQHELP], dColor ) ;

      dp->RefreshWin () ;                 // make the text visible

      uiInfo   Info ;                     // user interface data returned here
      short    icIndex = ZERO ;           // index of control with input focus
      bool     done = false ;             // loop control
      while ( ! done )
      {
         //* If focus is currently on a Pushbutton   *
         if ( ic[icIndex].type == dctPUSHBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditPushbutton ( Info ) ;

            //* If a Pushbutton was pressed *
            if ( Info.dataMod != false )
            {
               if ( Info.ctrlIndex == upiLAUNCH )
               {
                  dp->GetTextboxText ( upiARGS, argList ) ;
                  cIndex = dp->GetRbGroupSelection ( upiGUI ) ;
                  if ( (cIndex == upiGUI) && ic[upiGED].active )
                  {
                     bool guiEditor ;
                     dp->GetRadiobuttonState ( upiGED, guiEditor ) ;
                     if ( ! guiEditor )   // signal edit with console editor
                        cIndex = upiGED ;
                  }
               }
               else if ( Info.ctrlIndex == upiCANCEL )   // abort the operation
                  cIndex = -1 ;
               done = true ;
            }
         }

         //* If focus is currently on a Radio Button *
         else if ( ic[icIndex].type == dctRADIOBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditRadiobutton ( Info ) ;
         }

         //* If focus is currently on a Textbox *
         else if ( ic[icIndex].type == dctTEXTBOX )
         {
            Info.viaHotkey = false ;
            icIndex = dp->EditTextbox ( Info ) ;
         }

         //* Navigate to next/previous control *
         if ( ! done && ! Info.viaHotkey )
         {
            if ( Info.keyIn == nckSTAB )
               icIndex = dp->PrevControl () ; 
            else
               icIndex = dp->NextControl () ;
         }
      }     // while()
   }
   if ( dp != NULL )                         // close the window
      delete ( dp ) ;

   //* Restore parent dialog window *
   this->dPtr->RefreshWin () ;

   return cIndex ;

}  //* End oepUserPrompt() *

//*****************************
//* LaunchExternalApplication *
//*****************************
//******************************************************************************
//* Launch a specific external application along with desired arguments.       *
//* See notes below for a detailed discussion of the arguments.                *
//*                                                                            *
//* Input  : appSpec  : filespec (or filename) of application                  *
//*        : argList  : arguments sent to target application as argv[] array   *
//*        : redirect : (optional, 'true' by default)                          *
//*                     if 'true',  redirect stdout and stderr to temp files   *
//*                     if 'false', allow stdout and stderr to write to the    *
//*                                 terminal window                            *
//*                                                                            *
//* Returns: a) the child process ID                                           *
//*          b) -1 if child process not created                                *
//******************************************************************************
//* Notes:                                                                     *
//* ------                                                                     *
//* 1) This method requires some intelligence on the part of the caller.       *
//*    Unlike the LaunchDefaultApplication() method above, this method must    *
//*    be able to parse the individual arguments into separate substrings      *
//*    with a pointer to each. This is what the called application will see as *
//*    the "char* argv[]" array.                                               *
//* 2) The 'appSpec' parameter will contain the target application's name,     *
//*    relative path or absolute path.                                         *
//*    a) If it is known that the application in on the path, this may be the  *
//*       name of the executable file only.                                    *
//*    b) If caller is not sure whether the executable is on the path, then    *
//*       'appSpec' parameter should be the absolute or relative filespec.     *
//*    c) Path search:                                                         *
//*        i) If 'appSpec' does not contain slash ( '/' ) characters,          *
//*           then 'execvp' will search the $PATH environment variable.        *
//*       ii) If 'appSpec' DOES contain slash ( '/' ) characters,              *
//*           then 'execvp' will assume that you have provided a valid         *
//*           filespec, and will not search the $PATH.                         *
//*    d) By convention, argv[0] is the application name. This method captures *
//*       the tail of 'appSpec', stripping any path elements and inserts the   *
//*       application name as argv[0].                                         *
//* 3) The 'argList' parameter contains a list of arguments to be passed to    *
//*    target application.                                                     *
//*    a) Each argument substring, except the last, must be terminated by a    *
//*       newline '\n' character to allow reliable parsing.                    *
//*       Example: "arg01\narg02\narg03\narg04"                                *
//*    b) Any arguments which contain whitespace must be properly quoted:      *
//*       Example:  "arg01\n'Waiting for the Rain.ogs'\narg03"                 *
//*       This method WILL NOT automatically delimit the substrings.           *
//*    c) Any characters which might be incorrectly interpreted as regexp      *
//*       expressions must be properly escaped to prevent the shell from       *
//*       trying to expand them. Examples would be: \[  \]  \*  \.             *
//*       Arguments which contain _intentional_ regexp need not be escaped.    *
//*       Note however, that regexp wildcard characters ABSOLUTELY WILL NOT BE *
//*       EXPANDED. For instance: grep -n 'Taggit::' *.cpp                     *
//*       will not work because the shell will not expand the '*' character.   *                                                                        *
//* 4) One additional argument will be automatically appended to the arg list: *
//*    The NULL pointer indicates the end of the argv[] array.                 *
//* 5) Redirection of 'stdout' and 'stderr':                                   *
//*    Please see discussion of redirection in the LaunchDefaultApplication()  *
//*    method, above.                                                          *
//******************************************************************************

int FileDlg::LaunchExternalApplication ( const char* appSpec, 
                                        const gString& argList, bool redirect )
{
   const short MAX_ARGS = 48 ;         // maximum number of argument tokens (arbitrary)
   const char* argv[MAX_ARGS + 1] ;    // argv[] array of arg pointers
   char argvStr[gsMAXBYTES] ;          // temp buffer for arguments
   char appName[gsMAXBYTES] ;          // application filename (no path)
   gString gs ;                        // text formatting
   int soDesc, seDesc ;                // descriptors for temp files
   short wCnt = argList.gschars(),     // number of characters in source
         windx = ZERO,                 // index into wbuff
         nindx = ZERO,                 // index of newline character
         subcnt = ZERO,                // number of characters in substring
         avindx = ZERO,                // index of argvStr array
         avCount = ZERO ;              // number of arguments i.e. 'argc'

   //* Create temp filnames for redirection of stdout and stderr.   *
   //* Unfortunately, the child process will not know how to delete *
   //* the temp files, so we must rely on the application destructor*
   //* to delete the temp files (which it does by default).         *
   gString stdoutTemp, stderrTemp ;
   if ( redirect )
   {
      this->fmPtr->CreateTempname ( stdoutTemp ) ;
      this->fmPtr->CreateTempname ( stderrTemp ) ;
   }

   //* Isolate the target application's name *
   gs = appSpec ;
   short sl = gs.findlast( fSLASH ) ;
   if ( sl >= ZERO )    // if a path specified and not just a filename
      gs.shiftChars( -(sl + 1) ) ;
   gs.copy( appName, gsMAXBYTES ) ;
   argv[avCount++] = appName ;

   //* Convert caller's arguments to an array of char pointers.*
   while ( (nindx = argList.find( L'\n', windx )) >= ZERO )
   {
      subcnt = nindx - windx ;
      argList.substr( gs, windx, subcnt ) ;
      windx = nindx + 1 ;
      argv[avCount++] = &argvStr[avindx] ;
      avindx += gs.copy( &argvStr[avindx], gsMAXBYTES ) ;
   }
   if ( windx < (wCnt - 1) )
   {
      subcnt = (wCnt - windx) - 1 ;
      argList.substr( gs, windx, subcnt ) ;
      argv[avCount++] = &argvStr[avindx] ;
      avindx += gs.copy( &argvStr[avindx], gsMAXBYTES ) ;
   }
   argv[avCount] = NULL ;

   pid_t fpid = vfork () ;
   if ( fpid == ZERO )
   {  //* The child process (if created) executes here. *
      if ( redirect )
      {
         //*  1) Open (create) each temp file.             *
         //*  2) Reassign (redirect) stdout and stderr     *
         //*     to the new descriptors.                   *
         //*  3) Close the temp files.                     *
         soDesc = open ( stdoutTemp.ustr(), O_WRONLY | O_CREAT | O_TRUNC, 
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH ) ;
         seDesc = open ( stderrTemp.ustr(), O_WRONLY | O_CREAT | O_TRUNC, 
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH ) ;
         dup2 ( soDesc, STDOUT_FILENO) ;
         dup2 ( seDesc, STDERR_FILENO) ;
         close ( soDesc ) ;
         close ( seDesc ) ;
      }

      execvp ( appSpec, (char* const*)argv ) ;

      //* In case the exec call fails: child process MUST NOT return.*
      _exit ( 0 ) ;
   }
   else
   {  //* The parent process continues execution here *
      if ( fpid > ZERO )      // child successfully launched
      {
         /* Currently, nothing to be done. */
      }
   }
   return fpid ;

}  //* End LaunchExternalApplication() *

//****************************
//* LaunchDefaultApplication *
//****************************
//******************************************************************************
//* Launch the system default application for handling the specified file type.*
//*                                                                            *
//* Input  : targFile : filespec of file to be processed by the external       *
//*                     application                                            *
//*          redirect : (optional, 'true' by default)                          *
//*                     if 'true',  redirect stdout and stderr to temp files   *
//*                     if 'false', allow stdout and stderr to write to the    *
//*                                 terminal window                            *
//*                                                                            *
//* Returns: a) the child process ID                                           *
//*          b) -1 if child process not created                                *
//******************************************************************************
//* Notes:                                                                     *
//* ------                                                                     *
//* 1) fork a new process: vfork()                                             *
//* 2) Redirection:                                                            *
//*    a) It is possible that the called application will write to the terminal*
//*       window, either during startup or with some later status message.     *
//*    b) If the called application is a console application, this is almost   *
//*       certain; while for external GUI applications, it is still likely to  *
//*       some extent.                                                         *
//*    c) If the parent application has relinquished control of stdout and     *
//*       stderr. i.e. the NCurses engine has been put into hibernation mode,  *
//*       then the called application can display messages without conflict.   *
//*    d) If, however, the parent application is still in active control of    *
//*       stdout and stderr, then resource conflicts are likely to occur.      *
//*       In this case, it is necessary to redirect the called application's   *
//*       output. Our solution is to redirect stdout and stderr to temporary   *
//*       files by default. These files will persist until the parent          *
//*       application closes. (To prevent this, call the method with           *
//*       'redirect' set to 'false'.) The Taggit-class destructor will delete  *
//*       these temporary files during application shutdown.                   *
//* 3) Use the new process to launch the application:                          *
//*    a) This method uses the 'xdg-open' system utility (shell script) which  *
//*       launches the application associated with the file type of 'targFile'.*
//*    b) execlp searches the path looking for 'xdg-open'                      *
//*       Because we are fairly safe in assuming that 'xdg-open' actually IS   *
//*       on the path, the first argument is the name of the script (no path). *
//*    c) The second argument is argv[0] i.e. the script name again.           *
//*    d) The third argument is argv[1] i.e. the filespec passed in by caller. *
//*    e) The fourth argument is argv[2] i.e. a null pointer                   *
//*    f) Note that xdg-open passes only the specified filespec to the target  *
//*       application, and cannot pass additional configuration arguments.     *
//*       To invoke the target application will additional setup/configuration *
//*       parameters, please see the LaunchExternalApplication() method, below.*
//*    g) The child process is not directly associated with the terminal       *
//*       window from which it was launched, so the window may be closed       *
//*       asynchronously.                                                      *
//*    h) The exec functions do not return. The child process terminates when  *
//*       the external application exits.                                      *
//* 4) primary process returns to caller                                       *
//*                                                                            *
//* Programmer's Note: This sequence assumes that multi-threading is not       *
//* active. For multi-threaded applications, use pthread_fork() instead.       *
//*                                                                            *
//*  --   --   --   --   --   --   --   --   --   --   --   --   --   --   --  *
//* The xdg group of utilities and the xdg standard are maintained by          *
//* freedesktop.org. <https://specifications.freedesktop.org/                  *
//*                                    mime-apps-spec/mime-apps-spec-1.0.html> *
//* This specifies how to setup the association between an application and     *
//* the MIME types it supports.                                                *
//*                                                                            *
//*  --   --   --   --   --   --   --   --   --   --   --   --   --   --   --  *
//* How xdg-open determines which application matches each MIME type.          *
//* -----------------------------------------------------------------          *
//* <https://wiki.archlinux.org/index.php/default_applications#Desktop_entries>*
//*                                                                            *
//* Each application is associated with a ".desktop" file.                     *
//*     Example:  libreoffice-writer.desktop                                   *
//* These files list which file types may be handled by the application;       *
//* however, it would be unreasonable to scan each of these ".desktop" files   *
//* every time to find the correct application for a particular MIME type.     *
//*                                                                            *
//* Instead, xdg-open calls xdg-mime which scans the database:                 *
//*     /usr/share/applications/mimeinfo.cache                                 *
//* and returns the MIME type associated with the filetype to be opened.       *
//* For example, invoking xdg-open for an 'rtf' file:                          *
//*     xdg-open grocery_list.rtf                                              *
//* Which results in a call:                                                   *
//*     xdg-mime query filetype grocery_list.rtf                               *
//* Which yields an entry from the database:                                   *
//*     application/rtf=libreoffice-writer.desktop;                            *
//* 'sed' is used to isolate the MIME type:                                    *
//*     application/rtf                                                        *
//*                                                                            *
//* Manual Exploration:                                                        *
//* -------------------                                                        *
//* # determine a file's MIME type                                             *
//* $ xdg-mime query filetype photo.jpeg                                       *
//* image/jpeg                                                                 *
//*                                                                            *
//* # determine the default application for a MIME type                        *
//* $ xdg-mime query default image/jpeg                                        *
//* gimp.desktop                                                               *
//*                                                                            *
//* # change the default application for a MIME type                           *
//* $ xdg-mime default feh.desktop image/jpeg                                  *
//*                                                                            *
//* # open a file with its default application                                 *
//* $ xdg-open photo.jpeg                                                      *
//*                                                                            *
//* Several applications may support a particular MIME type:                   *
//*     audio/x-mp3=totem.desktop;vlc.desktop;rhythmbox.desktop;               *
//*                                                                            *
//* The default application associated with the MIME type is then executed     *
//* with the source file as its argument.                                      *
//*                                                                            *
//*                                                                            *
//******************************************************************************

int FileDlg::LaunchDefaultApplication ( const char* targFile, bool redirect )
{
   gString stdoutTemp, stderrTemp ;       // stdout target file
   this->fmPtr->CreateTempname ( stdoutTemp ) ;  // stdout redirection file
   this->fmPtr->CreateTempname ( stderrTemp ) ;  // stderr redirection file
   int soDesc = ZERO,                     // stdout target descriptor
       seDesc ;                           // stderr target descriptor

   pid_t fpid = vfork () ;                // create the child process

   if ( fpid == ZERO )
   {  //* The child process (if created) executes here.                  *
      //* Create temp files as targets for redirecting stdout and stderr.*
      if ( redirect )
      {
         soDesc = open ( stdoutTemp.ustr(), O_WRONLY | O_CREAT | O_TRUNC, 
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH ) ;
         seDesc = open ( stderrTemp.ustr(), O_WRONLY | O_CREAT | O_TRUNC, 
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH ) ;
         dup2 ( soDesc, STDOUT_FILENO) ;
         dup2 ( seDesc, STDERR_FILENO) ;
         close ( soDesc ) ;
         close ( seDesc ) ;
      }

      execlp ( "xdg-open", "xdg-open", targFile, NULL ) ;

      //* In case the exec call fails: child process MUST NOT return.*
      _exit ( 0 ) ;
   }
   else
   {  //* The parent process continues execution here *
      if ( fpid > ZERO )      // child successfully launched
      {
         /* Currently, nothing to be done. */
      }
   }
   return fpid ;

}  //* End LaunchDefaultApplication() *

//*************************
//*  DefineChildEmulator  *
//*************************
//******************************************************************************
//* Member of DC_Emulator                                                      *
//* =====================                                                      *
//*                                                                            *
//* 1) Determine which terminal emulator and shell program are running in the  *
//*    current window.                                                         *
//* 2) Determine the size and position for a new terminal-emulator window.     *
//*    By default, the size and position are calculated based on the size and  *
//*    position of the parent window, but these calculations may be overridden *
//*    using the optional call parameters described below.                     *
//*                                                                            *
//* On entry, caller must have initialized the following data members:         *
//* 'tmpFile' : contains path of temporary file for collecting system info.    *
//* 'prgFile' : contains the filespec of program to be executed in to window.  *
//*             (May be initialized with the program name only if the program  *
//*              is known to be on the execution path.)                        *
//* 'argList' : contains additional arguments (if any) for program to be       *
//*             launched.                                                      *
//* 'autoReturn' : 'true' indicates that window should close immediately when  *
//*                       the launched program exits.                          *
//*                'false' indicates that window should remain open and return *
//*                       to command prompt after the launched program exits.  *
//* All other data members will be initialized here.                           *
//*                                                                            *
//* Input  : c : (optional, -1 by default)                                     *
//*              If non-default, specifies the number of columns               *
//*              for the new terminal-emulator window.                         *
//*        : r : (optional, -1 by default)                                     *
//*              If non-default, specifies the number of rows                  *
//*              for the new terminal-emulator window.                         *
//*        : x : (optional, -1 by default)                                     *
//*              If non-default, specifies the X offset (in columns)           *
//*              for the new terminal-emulator window.                         *
//*        : y : (optional, -1 by default)                                     *
//*              If non-default, specifies the Y offset (in rows)              *
//*              for the new terminal-emulator window.                         *
//*                                                                            *
//* Returns: value in dce.emuType                                              *
//*          gnomeType  : gnome-terminal (or unknown terminal)                 *
//*          konsoleType: konsole                                              *
//*          xtermType  : xterm                                                *
//******************************************************************************
//* Programmer's Note:                                                         *
//* This method is used to define a new terminal-emulator window which will be *
//* launched by a child process.                                               *
//*                                                                            *
//*         It is not necessary to understand what this method                 *
//*         is doing in order to understand the exec functions.                *
//*                                                                            *
//* 1) In previous releases, we did our best to query the X-server to determine*
//*    which terminal emulator we were running in and to launch a smaller      *
//*    version of the same emulator to gather the user's password.             *
//*    The X11 query algorithm was not foolproof; it was a simplistic way of   *
//*    identifying the terminal emulator in which we are running.              *
//*                                                                            *
//*    This method has been greatly simplified to work within the restrictions *
//*    imposed by Fedora's transition from X to the Wayland compositor.        *
//*    If you are curious about the X-server-query version of this method,     *
//*    please download FileMangler v:0.0.36, OR download our "Salmon" student  *
//*    exercise which discusses in detail how to programatically spawn a new   *
//*    terminal window.                                                        *
//*                                                                            *
//* 2) While there are at least twenty terminal emulator programs in general   *
//*    use, for our purposes we identify only three, and assume that at least  *
//*    one of them is available on any system.                                 *
//*      a) xterm                                                              *
//*      b) GNOME terminal                                                     *
//*      c) KDE konsole                                                        *
//*    If none of these emulators are installed on your system, then it is     *
//*    likely that mounting/unmounting devices requiring a password will fail. *
//*                                                                            *
//*    Also, there are many different shell programs in common use, so we      *
//*    rely on the $SHELL environment variable to tell us which one to use.    *
//*    If unable to get this information, we default to 'bash'.                *
//*                                                                            *
//* 3) Tales of Wayland the Obstructor (not exactly our favorite superhero):   *
//*    Fedora's transition from X11/Xorg to the Wayland compositor is a noble  *
//*    cause, but frustrating, annoying, buggy and often curse-worthy.         *
//*    In the interest of security, many things that are common practice in a  *
//*    standard X-windows system are forbidden under Wayland.                  *
//*    a) It is currently nearly impossible to determine what terminal emulator*
//*       our application is running under because the "ps" utility cannot get *
//*       the parent window's information ($WINDOWID is no longer an           *
//*       environment variable). The presence of other environment variables   *
//*       which might yield the needed information is similarly unreliable.    *
//*    b) Under Wayland, window positioning is forbidden, so the "geometry"    *
//*       offsets are ignored for gnome-terminal. Konsole no longer supports   *
//*       the "geometry" parameter at all, and instead requires info from a    *
//*       defined profile to indicate window size.                             *
//*    c) On the other hand, good-old xterm can still be relied upon for both  *
//*       size and position compliance. However, we must provide absolute      *
//*       positioning because Wayland will tell us neither the dimensions of   *
//*       the display nor the size or position of the parent window.           *
//*       (Note also that Xterm may no longer be installed by default on your  *
//*        system, which we consider to be the betrayal of a sacred trust!)    *
//*                                                                            *
//*                                                                            *
//* Opening a new terminal window (example):                                   *
//* ----------------------------------------                                   *
//*                                                                            *
//* gnome-terminal --geometry=80x8+400+300                                     *
//*        -e "bash -c \"grep -n \'COMPILE\' Makefile MakeChild ; exec bash\"" *
//*                                                                            *
//* a) Calling bash before and after the command will keep the terminal        *
//*    window open.                                                            *
//* b) Current-working-directory is default for GNOME --working-directory.     *
//*    Home directory is default for Konsole --workdir.                        *
//*    Current-working-directory is default for XTerm -working-directory.      *
//*    To explicity set the directory: --working-directory=$PWD                *
//*    To specify another directory  : --working-directory=$HOME/temp          *
//*    Keep in mind that the specification must be an absolute path, not a     *
//*    relative path an that the '~' shortcut character is not recognized.     *
//* c) Note that the example above describes the invocation of the Gnome       *
//*    terminal, and the bash shell. The equivalent command for Konsole or     *
//*    other terminals and for zsh and other shells may require a slightly     *
//*    modified command structure.                                             *
//*                                                                            *
//* Window dimensions and positioning:                                         *
//* ----------------------------------                                         *
//* Unfortunately, window size and position is not entirely under our control. *
//* Each terminal emulator insists of being different from its competitors,    *
//* and the environment they run in also must have its say in the proceedings. *
//* That said, here is what we would LIKE to see.                              *
//*                                                                            *
//* Position:                                                                  *
//* -- Ideally, the position will be one or two rows below and one column      *
//*    to the right of the X/Y anchor point of the current window.             *
//* -- Note that the Y offset is measured from the top of the tabs             *
//*    (if visible), otherwise the top of the user window _not_ from the top   *
//*    of the terminal software application.                                   *
//* Dimensions:                                                                *
//* -- Ideally, the width will be at least two columns less than the parent    *
//*    window width, and the height < current window height.                   *
//* -- A "standard" terminal window is 25 rows by 80 columns, so any           *
//*    spawned application should fit in a window that size.                   *
//* -- For launching a GUI application, or a console application with          *
//*    only a small amout of output, aesthetically, a smaller window           *
//*    is preferred. However, we can't determine this ahead of time.           *
//*                                                                            *
//******************************************************************************

EmuType DC_Emulator::DefineChildEmulator ( short c, short r, short x, short y )
{
   const short emuCOUNT = 3 ;
   const wchar_t* emuGroup[emuCOUNT] = 
   {
      L"xterm",
      L"gnome-terminal",
      L"konsole"
   } ;
   gString sysCmd,                     // generate system commands
           gs ;                        // text formatting
   char  inbuff[gsMAXBYTES] ;          // input buffer
   this->emuType = xtermType ;         // return value (xterm is default)
   gs = emuGroup[ZERO] ;               // default emulator program
   gs.copy( this->emuName, enSIZE ) ;

   //* Read the name of the shell program from the terminal environment *
   sysCmd.compose( "echo ${SHELL} 1>%s", this->tmpFile ) ;
   system ( sysCmd.ustr() ) ;

   for ( short e = ZERO ; e < emuCOUNT ; ++e )
   {
      sysCmd.compose( "which %S 1>>%s 2>/dev/null", emuGroup[e], this->tmpFile ) ;
      system ( sysCmd.ustr() ) ;
   }

   //* Open the file *
   ifstream ifs ( this->tmpFile, ifstream::in ) ;
   if ( ifs.is_open() )             // if input file open
   {
      //* Read the first line in the file (shell program) *
      ifs.getline ( inbuff, gsMAXBYTES, NEWLINE ) ;
      if ( (ifs.good()) || (ifs.gcount() > ZERO) )
      {
         //* Isolate the shell program filename *
         gs.loadChars( inbuff, ifs.gcount() ) ;
         short i = (gs.findlast( fSLASH )) + 1 ;
         if ( i > ZERO )
         {
            gs.shiftChars( -(i) ) ;
            gs.copy( this->shlName, enSIZE ) ;
         }

         //* Read as much of the file as necessary. If target  *
         //* emulator installed, response s/b something like:  *
         //*     /usr/bin/xterm                                *
         bool termFound = false ;
         for ( short e = ZERO ; (e < emuCOUNT) && !termFound ; ++e )
         {
            ifs.getline ( inbuff, gsMAXBYTES, NEWLINE ) ;
            if ( (ifs.good()) || (ifs.gcount() > ZERO) )
            {
               gs.loadChars( inbuff, ifs.gcount() ) ;
               short i = (gs.findlast( fSLASH )) + 1 ;
               if ( i > ZERO )
               {
                  for ( short f = ZERO ; f < emuCOUNT ; ++f )
                  {
                     if ( (gs.compare( emuGroup[f], true, gsMAXBYTES, i )) == ZERO )
                     {
                        gs = emuGroup[f] ;
                        gs.copy( this->emuName, enSIZE ) ;  // save emulator name
                        switch ( f )
                        {
                           case 1:  this->emuType = gnomeType ;   break ;
                           case 2:  this->emuType = konsoleType ; break ;
                           case 0:
                           default:
                              this->emuType = xtermType ;
                              break ;
                        }
                        termFound = true ;
                        break ;
                     }
                  }
               }
            }
         }
      }
      else     // file access error, return default shell
      {
         gs = "bash" ;
         gs.copy( this->shlName, enSIZE ) ;
      }
      ifs.close() ;                 // close the source file
   }

   //* Set the dimensions and position of the   *
   //* child window (geometry). See notes above.*
   if ( c == (-1) ) this->cols = 77 ;
   else             this->cols = c ;
   if ( r == (-1) ) this->rows = 23 ;
   else             this->rows = r ;
   if ( x == (-1) ) this->xpos = 150 ;
   else             this->xpos = x ;
   if ( y == (-1) ) this->ypos = 100 ;
   else             this->ypos = y ;

   //* Construct the geometry string and the command-sequence string.*
   if ( this->emuType == konsoleType )    // konsole terminal
   {
      // Programmer's Note: For newer versions of Konsole, "geometry" 
      // is no longer accepted as valid parameter.
      gs.compose( "--geometry=%dx%d+%d+%d", 
                  &this->cols, &this->rows, &this->xpos, &this->ypos ) ;
      sysCmd.compose( "%s -c \"\'%s\' %s ; exec %s\"",
         this->shlName, this->prgFile, this->argList, this->shlName ) ;
   }
   else if ( this->emuType == gnomeType ) // gnome terminal
   {
      gs.compose( "--geometry=%dx%d+%d+%d", 
                  &this->cols, &this->rows, &this->xpos, &this->ypos ) ;
      sysCmd.compose( "%s -c \"\'%s\' %s ; exec %s\"",
         this->shlName, this->prgFile, this->argList, this->shlName ) ;
   }
   else  // (this->emuType == xtermType) default
   {
      gs.compose( "%dx%d+%d+%d", 
                  &this->cols, &this->rows, &this->xpos, &this->ypos ) ;
      sysCmd.compose( "%s -c \"\'%s\' %s ; exec %s\"",
         this->shlName, this->prgFile, this->argList, this->shlName ) ;
   }
   gs.copy( this->geoData, enSIZE ) ;  // save the geometry sequence

   //* If windows is to close immediately after executing the *
   //* command sequence, remove the 'exec shell-prog' command.*
   if ( this->autoReturn )
   {
      short semiIndx = sysCmd.findlast( ';' ) - 1 ;
      sysCmd.replace( &sysCmd.gstr()[semiIndx], "\"", semiIndx ) ;
   }
   sysCmd.copy ( this->cmdData, gsMAXBYTES ) ; // save the command sequence

   return this->emuType ;

}  //* End DefineChildEmulator() *

//*************************
//*      OpenTermwin      *
//*************************
//******************************************************************************
//* Member of DC_Emulator                                                      *
//* =====================                                                      *
//*                                                                            *
//* Spawn a new process and open the previously-defined terminal window.       *
//*   1) Create the command for opening a new terminal window, incorporating   *
//*      the caller's commands.                                                *
//*   2) Launch the child process which will control the new window.           *
//*   3) The child process will open the window.                               *
//*   4) The parent process will return to caller.                             *
//*                                                                            *
//* Input  : dce  : (by reference) contains the setup parameters for launching *
//*                 the new terminal-emulator window.                          *
//*                                                                            *
//* Returns: child process ID,                                                 *
//******************************************************************************
//* Launching a command from a new terminal window.                            *
//* -----------------------------------------------                            *
//* Because this command is rather complex, let's go through it step-by-step.  *
//* Note that each argument to the execlp() function is a string. To put it    *
//* another way, each argument is interpreted as a const char*, and arguments  *
//* are separated by commas creating an array of pointers, i.e. the argv[]     *
//* array.                                                                     *
//* The arguments described are specific to the GNOME terminal and to the      *
//* 'bash' shell program but other terminals and shells follow the same        *
//* general pattern.                                                           *
//*                                                                            *
//* 1) 'execlp'                                                                *
//*    This function carries the arguments to the system.                      *
//* 2) 'emuName'                                                               *
//*    This is the name of the terminal emulation program, which is assumed    *
//*    to be on the path where execlp() can find it.                           *
//* 3) 'emuName'                                                               *
//*    This is the path-less emulator name which constitutes argv[0].          *
//* 4) 'geoData'                                                               *
//*    This string is the geometry i.e. the size and position of the new       *
//*    window.  --geometry=COLSxROWS+XOFFSET+YOFFSET                           *
//*    Example: --geometry=65x8+400+300                                        *
//*         Note that KDE Konsole has a long-standing bug which ignores this   *
//*         command, and/or interprets the dimensions as pixels counts rather  *
//*         than as columns and rows. Newer Konsole will not accept the        *
//*         "geometry" parameter, so we have removed it from the Konsole       *
//*         invocation.                                                        *
//*    Note that XTerm options begin with a single dash and do not allow the   *
//*    '=' to join the option and its argument.                                *
//*    Example: -geometry 65x8+400+300                                         *
//*             (this is two separate arguments)                               *
//* 5) 'hide-menubar'                                                          *
//*    Does what it says, prevents the menu bar from being displayed in the    *
//*    new window.                                                             *
//* 6) '--window' (gnome-terminal only): By default, gnome-terminal opens a    *
//*    new tab in the current window, so we must explicitly specify that the   *
//*    new terminal be opened in a new window.                                 *
//* 7) '-q' switch (gnome-terminal only): silent launch i.e. child process     *
//*    will not write to the parent window. The other emulators do not offer   *
//*    this option.                                                            *
//* 8) '-e' switch: This is the terminal emulator's "Execute" command. What    *
//*    follows it is the command to be executed inside the new terminal window.*
//*    The '-e' switch is used by all three emulators: gnome-terminal, konsole *
//*    and xterm,                                                              *
//* 9) 'cmdData'                                                               *
//*    This is the main argument. These are the commands to be executed within *
//*    the new terminal window.                                                *
//*    See the templates used to create the contents of the 'cmdData' string.  *
//*    a) Note that the entire argument is enclosed in quotation marks AND     *
//*       that the part of the argument that is the command to be executed is  *
//*       ALSO enclosed in (escaped) quotation marks.                          *
//*    b) 'bash -c'                                                            *
//*       This runs a copy of the 'bash' shell program inside the new terminal *
//*       window. The '-c' argument to the shell program indicates that the    *
//*       following argument is to be executed by the shell. Most shell        *
//*       programs use the "-c" command switch for this, but if you are using  *
//*       a shell program other than bash, zsh or csh, please verify the       *
//*       switch used by your shell.                                           *
//*       -- Note that if you do not require the window to remain open after   *
//*          execution, the shell specification may be omitted.                *
//*       -- Note that the command string is preceeded by an escaped           *
//*          double-quotation mark: ( \"commands... )                          *
//*       -- The matching escaped double-quotation mark follows (see item d)   *
//*          below:  ( ...exec bash\" )                                        *
//*       Placing the backslash before the quotation marks prevents the shell  *
//*       from removing them before the command is executed, allowing it to be *
//*       seen as a single argument.                                           *
//*       Similarly, escaped single-quotation marks must surrounding any       *
//*       argument which contains whitespace. The single-quotation marks will  *
//*       not be removed by the shell. Thus the token can be interpreted as a  *
//*       single argument. For example: \"grep -i \'Copyright Notice\' ....    *
//*    c) ';'                                                                  *
//*       The semicolon character separates the the main commands from the     *
//*       'exec' call, i.e. these are SEPARATE commands.                       *
//*    d) 'exec bash'                                                          *
//*       This command _replaces_ the currently running copy of 'bash' with a  *
//*       new copy. This may seem odd, but without the 'exec' two copies of    *
//*       the shell would be left running in the window.                       *
//*       -- Note that an escaped double-quotation mark follows this command   *
//*          which closes the escaped quotation pair (see item b) above.       *
//*       -- The _effect_ of this command is to keep the new window open and   *
//*          to return to the command prompt after all the specified commands  *
//*          have finished.                                                    *
//*       -- Note that a side-effect of this replacement is that the initial   *
//*          command(s) sent to the window will not be retained in the         *
//*          window's command history.                                         *
//*       -- If you want the window to close immediately when finished, then   *
//*          omit the 'exec bash'.                                             *
//* 10) NULL                                                                   *
//*    This is the NULL pointer indicating the end of the array of pointers.   *
//*               -----  -----  -----  -----  -----  -----  -----              *
//*                                                                            *
//* Working directory: For gnome-terminal and xterm the working directory is   *
//* taken from the parent's $PWD environment variable. However, Konsole        *
//* defaults to the directory specified in the profile, so we must explicitly  *
//* redirect it to the PWD using: "--workdir=."                                *
//******************************************************************************

pid_t DC_Emulator::OpenTermwin ( void )
{
   //* - - - - - - - - - - - - -*
   //* Create the child process *
   //* - - - - - - - - - - - - -*
   pid_t fpid = vfork () ;

   //* The child process (if created) executes here. *
   //*         (see notes in method header)          *
   if ( fpid == ZERO )
   {
      switch ( this->emuType )
      {
         case 0:     // gnome-terminal
            execlp ( this->emuName, this->emuName, this->geoData, "--hide-menubar", 
                     "--window", "-q", "-e", this->cmdData, NULL ) ;
            break ;
         case 1:     // konsole
            execlp ( this->emuName, this->emuName, 
                     "--hide-menubar", "--workdir=.", 
                     "-e", this->cmdData, NULL ) ;
            break ;
         case 2:     // xterm
            execlp ( this->emuName, this->emuName, "-geometry", this->geoData, 
                     "-fa", "Monospace", "-fs", "12",
                     "-e", this->cmdData, NULL ) ;
            break ;
      } ;
   
      //* In case the exec call fails: child process MUST NOT continue.*
      _exit ( 0 ) ;
   }

   return fpid ;

}  //* End OpenTermwin() *

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

