//******************************************************************************
//* File       : FileDlgTree.cpp                                               *
//* Author     : Mahlon R. Smith                                               *
//*              Copyright (c) 2005-2020 Mahlon R. Smith, The Software Samurai *
//*                  GNU GPL copyright notice located in FileMangler.hpp       *
//* Date       : 28-Jun-2020                                                   *
//* Version    : (see FileDlg Version string)                                  *
//*                                                                            *
//* Description:                                                               *
//* 1) This module contains methods related a graphical representation of      *
//*    the file-system tree structure.                                         *
//*                                                                            *
//* 2) This module also contains the implementation of the 'FIND' command      *
//*    which scans the directory tree for filenames that match a specified     *
//*    pattern.                                                                *
//*                                                                            *
//******************************************************************************
//* Version History (most recent first):                                       *
//*   See version history in FileDlg.cpp.                                      *
//******************************************************************************
//* Programmer's Notes:                                                        *
//* The tree-view functionality in this module could be significantly enhanced.*
//* - expand/collapse levels (stub commands in dftDisplayTree())               *
//* - display file statistics for each directory                               *
//* - read/write permissions                                                   *
//*                                                                            *
//* At this time (v:0.00.26), we think we have all that is necessary, so we'll *
//* wait for some user feedback.                                               *
//*                                                                            *
//******************************************************************************

#include "GlobalDef.hpp"
#include "FileDlg.hpp"

//******************************
//* Local definitions and data *
//******************************
//* Message text defined in FileDlgPrompt.cpp *
extern const char* fmtypeString[] ;
extern const char* fmtypeName[] ;
extern const char* readonlyFileSys[] ;
extern const char* notOntoSelf[] ;
extern const char* cannotOverrideWP[] ;
extern const char* noLinkOverwrite[] ;
extern const char* differentFileTypes[] ;
extern const char* cannotDeleteSpecial[] ;
extern const char* cannotOverwriteFifo[] ;
extern const char* cannotOverwriteDirectory[] ;
extern const char* warnSocketMod[] ;
extern const char* warnBdevCdevMod[] ;
extern const char* warnUnsuppMod[] ;
extern const char* fdNotImplemented[] ;

static const char* const noMatchMsg[] = 
{
   "   No matching directory name found.   ",
   "No more matching directory names found.",
} ;

const char *parentDir = ".. (parent dir)" ;

//* Dialog control definitions for Tree-view Mode *
enum ctrlNames : short { dummyTB = ZERO, pathTB, dftcontrolsDEFINED } ;

//********************
//* Local prototypes *
//********************
static void dftSortNode ( TreeNode* tnPtr, UINT count ) ;
static void dftMarkRestricted ( gString& dName ) ;
static void dftMarkExternal ( gString& dName ) ;
static bool dftNameSearch ( const dftDisplay& dftd, 
                            const gString ptrn, UINT& fIndex ) ;
static void dftNoMatch ( const winPos& spPos, const char* msg, attr_t dColor ) ;


//*************************
//*   DisplayFileTree     *
//*************************
//******************************************************************************
//* Display the file-system tree structure beginning with the 'current'        *
//* directory and traversing all subdirectories below current directory.       *
//*                                                                            *
//* User has the option of selecting a new current-working-directory' (CWD)    *
//* from the displayed data.                                                   *
//*                                                                            *
//* Input  : ul        : (by reference) upper left corner dialog position      *
//*          rows      : number of rows for dialog                             *
//*          cols      : number of columns for dialog                          *
//*                                                                            *
//* Returns: true if current working directory (CWD) has changed               *
//*          else false                                                        *
//******************************************************************************

bool FileDlg::DisplayFileTree ( winPos ul, short rows, short cols )
{
   gString origcwd ;                // CWD on entry
   this->GetPath ( origcwd ) ;      // where we are now
   gString newcwd = origcwd ;       // where we will be on return
   dtbmData msgData ;               // textbox messages to user
   dtbmData procMsg( "    Processing..." ) ;
   const TreeNode *constNode =      // access to scanned directory tree
                   this->fmPtr->GetBaseNode () ;
   TreeNode baseNode ;              // local copy of directory tree
   TreeNode tnAcc ;                 // directory-tree accumulator
   UINT actualNodes,                // number of nodes (directory names in tree
        actualFiles ;               // (unused)
   short status ;                   // value returned from user interaction
   bool newCWD = false ;            // return value

   //* Save the display for the parent dialog window *
   this->dPtr->SetDialogObscured () ;

   //* Enable full scan of root directory.              *
   bool old_sfr = this->fmPtr->ScanFromRoot ( true, true ) ;
   this->fmPtr->ScanFromRoot ( true ) ;

   //* If Tree-View invoked from the command line, tree *
   //* has not yet been scanned. If CWD is root, then   *
   //* tree scan was only for the top level. In these   *
   //* cases, we must rescan the tree before continuing.*
   if ( (constNode == NULL) || ((origcwd.compare( ROOT_PATH )) == ZERO) )
      this->RefreshCurrDir () ;

   //* Open the Tree-view dialog *
   attr_t iColor = this->cs.tn,     // item color
          hColor = this->cs.sd ;    // highlight color
   NcDialog* dp = this->dftOpenDialog ( ul, rows, cols, iColor ) ;

   //* Initialize the display parameters *
   dftDisplay dftd( ZERO, (rows - 6), (cols - 2), iColor, hColor ) ;

   //* Loop until either user makes a selection or *
   //* gets too tired to climb the tree.           *
   bool  done = (dp == NULL) ? true : false ;  // loop control
   while ( ! done )
   {
      dp->DisplayTextboxMessage ( pathTB, procMsg ) ;

      //* Get read-only access to the base of the directory tree and   *
      //* create a copy of the tree structure. (non-dir files ignored) *
      constNode = this->fmPtr->GetBaseNode () ;
      this->fmPtr->CopyDirTree ( constNode, baseNode, false ) ;

      //* Allocate space for the display data *
      this->fmPtr->TreeNodeSummary ( &baseNode, tnAcc ) ;
      tnAcc.GetActual ( actualNodes, actualFiles ) ;
      dftd.realloc( actualNodes + 1 ) ;

      //* Construct tree-view data from the directory tree data.*
      this->dftFormatTree ( dftd, &baseNode ) ;

      //* Display the current directory path.*
      dp->SetTextboxText ( pathTB, newcwd ) ;

      //* Display the data and interact with the user *
      //*     (as distasteful as that may be...)      *
      status = this->dftDisplayTree ( dp, dftd, newcwd ) ;

      //* Release local copy of directory tree *
      if ( baseNode.nextLevel != NULL )
         this->fmPtr->ReleaseDirTree ( &baseNode ) ;

      //* If user selected a new CWD, go there *
      if ( status >= ZERO )
      {
         //* Set new CWD and return to caller *
         if ( status == ZERO )
         {
            //* Close the Tree-view dialog *
            if ( dp != NULL )
               { delete ( dp ) ; dp = NULL ; this->dPtr->RefreshWin () ; }
            if ( (this->SetDirectory ( newcwd )) == OK )
            {
               msgData = " New working directory selected. " ;
               this->dPtr->DisplayTextboxMessage ( this->mIndex, msgData ) ;
               newCWD = true ;      // alert caller that directory has changed
            }
            else
            {  //* Error reading new directory. (Should not happen unless *
               //* user selected a directory with no read access.)        *
               msgData = " Access Error! Unable to attach to specified directory." ;
               this->dPtr->DisplayTextboxMessage ( this->mIndex, msgData ) ;
               chrono::duration<short>aWhile( 2 ) ;
               this_thread::sleep_for( aWhile ) ;
            }
            done = true ;
         }

         //* Scan the specified (parent or root) *
         //* directory and continue interaction. *
         else
         {
            this->dPtr->RefreshWin () ;         // restore the parent dialog
            this->SetDirectory ( newcwd ) ;     // scan new CWD
            this->dPtr->SetDialogObscured () ;  // save the renewed parent dialog
         }
      }

      //* Return to the original CWD. *
      else
      {
         //* Close the Tree-view dialog *
         if ( dp != NULL )
            { delete ( dp ) ; dp = NULL ; this->dPtr->RefreshWin () ; }

         this->SetDirectory ( origcwd ) ;
         done = true ;
      }
   }     // while(!done)

   //* Close the Tree-view dialog (in case it didn't happen earlier) *
   if ( dp != NULL )
   {
      delete ( dp ) ;
      this->dPtr->RefreshWin () ;      // restore the parent dialog
   }

   //* If scan-from-root was not active on entry, *
   //* disable full scan of root directory.       *
   if ( ! old_sfr )
      this->fmPtr->ScanFromRoot ( false ) ;

   return newCWD ;

}  //* End DisplayFileTree() *

//*************************
//*     dftOpenDialog     *
//*************************
//******************************************************************************
//* Open the Tree-view dialog window and return a pointer to it.               *
//*                                                                            *
//*                                                                            *
//* Input  : ul    : (by reference) upper left corner dialog position          *
//*          rows  : number of rows for dialog                                 *
//*          cols  : number of columns for dialog                              *
//*          iColor: dialog interior color                                     *
//*                                                                            *
//* Returns: pointer to open dialog window                                     *
//*          OR NULL pointer if dialog did not open                            *
//******************************************************************************

NcDialog* FileDlg::dftOpenDialog ( const winPos& ul, short rows, 
                                   short cols, attr_t iColor )
{
const char* const userInstructions =
   "Navigation: cursor keys or mouse. Parent Dir: ALT+UpArrow. Root Dir: CTRL+R.\n"
   "Search: CTRL+F/CTRL+G. New Cwd: ENTER key. Close: CTRL+Q. Help: CTRL+H." ;

attr_t bColor = this->cs.bb ;    // border color

InitCtrl ic[dftcontrolsDEFINED] = 
{
   {  //* 'dummy' textbox  - - - - - - - - - - - - - - - - - - - - -   dummyTB *
      dctTEXTBOX,                   // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      2,                            // ulY:       upper left corner in Y
      1,                            // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      1,                            // cols:      control columns
      "",                           // dispText:  initially empty
      bColor,                       // nColor:    non-focus color
      bColor,                       // fColor:    focus color
      #if LINUX_SPECIAL_CHARS != 0
      tbPathLinux,                  // filter: valid filespec chars (incl. Linux "special")
      #else    // BASIC FILESPEC FILTER
      tbPathName,                   // filter:    valid filename characters
      #endif   // BASIC FILESPEC FILTER
      NULL,                         // label:     
      ZERO,                         // labY:      
      ZERO,                         // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[pathTB],                  // nextCtrl:  link in next structure
   },
   {  //* 'new target path' textbox  - - - - - - - - - - - - - - - -    pathTB *
      dctTEXTBOX,                   // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      1,                            // ulY:       upper left corner in Y
      1,                            // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      short(cols - 2),              // cols:      control columns
      "",                           // dispText:  initially empty
      this->cs.tf,                  // nColor:    non-focus color
      (this->cs.pf & ~ncrATTR),     // fColor:    focus color
      #if LINUX_SPECIAL_CHARS != 0
      tbPathLinux,                  // filter: valid filespec chars (incl. Linux "special")
      #else    // BASIC FILESPEC FILTER
      tbPathName,                   // filter:    valid filename characters
      #endif   // BASIC FILESPEC FILTER
      NULL,                         // label:     
      ZERO,                         // labY:      
      ZERO,                         // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      false,                        // active:    view only
      NULL                          // nextCtrl:  link in next structure
   },
} ;

   //* Initial parameters for dialog window *
   InitNcDialog dInit( rows,           // number of display lines
                       cols,           // number of display columns
                       ul.ypos,        // Y offset from upper-left of terminal 
                       ul.xpos,        // X offset from upper-left of terminal 
                       NULL,           // dialog title
                       ncltSINGLE,     // border line-style
                       bColor,         // border color attribute
                       iColor,         // interior color attribute
                       ic              // pointer to list of control definitions
                     ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;
   if ( (dp->OpenWindow()) == OK )
   {
      //* Draw the static data *
      dp->WriteChar ( 0, 0, wcsLTEE, bColor ) ;
      dp->WriteChar ( 0, (cols - 1), wcsRTEE, bColor ) ;
      dp->ClearLine ( 2, bColor ) ;
      dp->ClearLine ( 3, bColor ) ;
      dp->WriteParagraph ( 2, 2, userInstructions, bColor ) ;
      LineDef  lDefH(ncltHORIZ, ncltSINGLE, 4, ZERO, cols, bColor) ;
      dp->DrawLine ( lDefH ) ;
      dp->RefreshWin () ;
   }
   return dp ;

}  //* End dftOpenDialog() *

//*************************
//*    dftDisplayTree     *
//*************************
//******************************************************************************
//* Display a graphic representation of the tree structure in the open area    *
//* of the specified dialog.                                                   *
//* User may optionally select a new CWD.                                      *
//*                                                                            *
//* Input  : dp    : pointer to open Tree-view dialog window                   *
//*          dftd  : (by reference) display-data control info                  *
//*          newcwd: (by reference)                                            *
//*                  - on entry contains full path of current-working directory*
//*                  - if user selects a new current-working-directory path,   *
//*                    then path is returned here, else, contents undefined    *
//*                                                                            *
//* Returns: -1   : new CWD not selected ('newcwd' uninitialized)              *
//*           0   : set selected directory as new normal-mode CWD              *
//*           1   : move to parent (or root) directory and create new tree     *
//******************************************************************************
//* Programmer's Note:                                                         *
//* For left-shift and right-shift: The sense of the arrows is intended to be  *
//* user-oriented. That is, if user presses right arrow, the _highlight_       *
//* appears to move to the right (the data move left). This is the opposite of *
//* programmer thinking where 'go left' means that the data should go left.    *
//*                                                                            *
//* Also note that because characters and columns are not the same thing,      *
//* when shifted left, the first character displayed on a line may occasionally*
//* be different from the one you might expect. In general, characters require *
//* either one or two columns for display. In the case where the first         *
//* character to be displayed in a string is a two-column character beginning  *
//* on an ODD column offset, it may be skipped. No harm done though.           *
//*                                                                            *
//******************************************************************************

short FileDlg::dftDisplayTree ( NcDialog* dp, dftDisplay& dftd, gString& newcwd )
{
   gString oldcwd = newcwd,      // CWD on entry
           gsName,               // filename formatting
           gsPath,               // target-path construction
           gstmp,                // general text formatting
           sPattern ;            // user's search criteria
   UINT  sIndex = ZERO,          // search index
         sMatch = ZERO ;         // index of search match
   short firstDLine = 5, y,      // tracking for data redraw
         fnIndx ;                // filename formatting
   attr_t tColor ;               // line-item color attribute
   wkeyCode wk ;                 // key input data
   bool  redraw   = true,        // if 'true' redraw display data
         done     = false,       // loop control
         newCWD   = false,       // true if new CWD selected
         toParent = false ;      // true if new CWD is parent directory

   //* Establish position of active window (for mouse input) *
   short rows, cols ;
   dp->GetDialogDimensions ( rows, cols ) ;
   short btop = ZERO + 4,        // top border
         bbot = rows - 1,        // bottom border
         blef = ZERO,            // left border
         brgt = cols - 1 ;       // right border


   do
   {
      if ( redraw )
      {
         //* Clear the display space *
         y = firstDLine ;
         for ( UINT i = ZERO ; i < dftd.rows ; i++ )
            dp->ClearLine ( y++ ) ;
         y = firstDLine ;

         //* Display the data *
         for ( UINT di = dftd.first ; di <= dftd.last ; di++ )
         {
            //* Set text color attribute for this line.*
            tColor = (di == dftd.curr) ? dftd.hColor : dftd.iColor ;
            gstmp = dftd.data[di].n ;
            if ( dftd.lshift > ZERO )     // if data are left-shifted
            {
               //* Shift left by specified amount *
               gstmp.shiftCols( -(dftd.lshift) ) ;

               //* If currently-highlighted item has been shifted out of    *
               //* display area, mark the spot so user doesn't get confused.*
               //* If entire (non-highlighted line) has been shifted out of *
               //* display area, nothing to display (gstmp is empty string).*
               if ( di == dftd.curr ) 
                  gstmp = L"<<" ;
            }
            gstmp.limitCols( dftd.cols ) ;
            dp->WriteString ( y++, 1, gstmp, tColor ) ;
         }

         //* Display the full path of highlighted directory name *
         gsName = dftd.data[dftd.curr].n ;
         if ( (gsName.compare( parentDir )) == ZERO )
         {
            this->fmPtr->ExtractPathname ( gsPath, oldcwd ) ;
         }
         else
         {
            gsName.strip() ;           // strip leading and trailing whitespace
            //* If tagged with one of the special indicators, *
            //* strip the indicator.                          *
            if ( (fnIndx = gsName.find( extFileSys )) > ZERO )
               gsName.limitChars( fnIndx ) ;
            else if ( (fnIndx = gsName.find( noAccess )) > ZERO )
               gsName.limitChars( fnIndx ) ;

            //* If highlighted name is base directory     *
            //* a) dftd.curr == ZERO, then no parent item *
            //* b) dftd.curr == 1, and parent item exists *
            gstmp = dftd.data[ZERO].n ;
            if ( (dftd.curr == ZERO) || 
                 ((dftd.curr == 1) && ((gstmp.compare( parentDir )) == ZERO)) )
            {
               gsPath = oldcwd ;
            }

            //* Construct the path of highlighted directory name *
            else
            {
               gsPath = oldcwd ;
               if ( (oldcwd.compare( ROOT_PATH )) != ZERO )
                  gsPath.append( fSLASH ) ;
               gsPath.append( gsName.gstr() ) ;
               fnIndx = gsPath.findlast( fSLASH ) ;
               UINT di = dftd.curr,
                    prevlev ;
               short tmpIndx ;
               while ( (prevlev = dftd.data[di].l) > 1 )
               {
                  while ( dftd.data[--di].l >= prevlev ) ;
                  gstmp = dftd.data[di].n ;
                  gstmp.strip() ;
                  if ( (tmpIndx = gstmp.find( extFileSys )) > ZERO )
                     gstmp.limitChars( tmpIndx ) ;
                  else if ( (tmpIndx = gstmp.find( noAccess )) > ZERO )
                     gstmp.limitChars( tmpIndx ) ;
                  gstmp.insert( fSLASH ) ;
                  gsPath.insert( gstmp.gstr(), fnIndx ) ;
                  prevlev = dftd.data[di].l ;
               }
            }
         }
         //* Shift text to the left so the rightmost char will be visible *
         gstmp = gsPath ;
         if ( gstmp.gscols() > (cols - 2) )
            gstmp.shiftCols( (cols - 2) - gstmp.gscols() ) ;
         dp->SetTextboxText ( pathTB, gstmp ) ;

         redraw = false ;
      }
      dp->RefreshWin () ;        // make changes visible

      //*************************
      //** Get input from user **
      //*************************
      dp->GetKeyInput ( wk ) ;

      //* Convert any unconverted mouse input within the scrolling window *
      //* to keycodes. Emulates conversions within the NcDialog API.      *
      if ( (wk.type == wktMOUSE) && !wk.mevent.conv && (this->dPtr->meMouseEnabled ()) )
      {
         //* Test whether the event lies within our scrolling window *
         if ( wk.mevent.ypos >= btop && wk.mevent.xpos >= blef &&
              wk.mevent.ypos <= bbot && wk.mevent.xpos <= brgt )
         {
            //* ScrollWheel upward or click in top border *
            if ( (wk.mevent.meType == metSW_D) ||
                 ((wk.mevent.meType == metB1_S || wk.mevent.meType == metB1_D)
                  && (wk.mevent.ypos == btop)) )
            { wk.type = wktFUNKEY ; wk.key  = nckUP ; }

            //* ScrollWheel downward or click in bottom border *
            else if ( (wk.mevent.meType == metSW_U) ||
                      ((wk.mevent.meType == metB1_S || wk.mevent.meType == metB1_D)
                       && (wk.mevent.ypos == bbot)) )
            { wk.type = wktFUNKEY ; wk.key  = nckDOWN ; }

            //* Single-click and double-click events on the left *
            //* and right borders shift the data left and right. *
            else if ( ((wk.mevent.meType == metB1_S || wk.mevent.meType == metB1_D))
                      && (wk.mevent.xpos == blef) )
            { wk.type = wktFUNKEY ; wk.key  = nckLEFT ; }
            else if ( ((wk.mevent.meType == metB1_S || wk.mevent.meType == metB1_D))
                      && (wk.mevent.xpos == brgt) )
            { wk.type = wktFUNKEY ; wk.key  = nckRIGHT ; }

            //* Single-click and double-click events inside the *
            //* scrolling area yield selection of a new target  *
            //* directory (or a move to parent directory).      *
            else if ( ((wk.mevent.ypos > btop) && (wk.mevent.ypos < bbot) &&
                       (wk.mevent.xpos > blef) && (wk.mevent.xpos < brgt))
                      &&
                      ((wk.mevent.meType == metB1_S) || (wk.mevent.meType == metB1_D)) )
            {
               //* Calculate the new position for the highlight *
               //* and push the ENTER key into the input stream.*
               dftd.curr = dftd.first + UINT(wk.mevent.ypos - btop - 1) ;
               wk.type = wktFUNKEY ; wk.key = nckENTER ;
               dp->UngetKeyInput ( wk ) ;
               redraw = true ;
               continue ;
            }
         }
      }     // mouse input

      if ( dp->IsScrollKey ( wk ) )    // scroll through the data
      {
         UINT prevCurr = dftd.curr ;
         short prevLshift = dftd.lshift ;
         switch ( wk.key )
         {
            case nckDOWN:
               if ( dftd.curr < dftd.last )
                  ++dftd.curr ;
               else if ( dftd.last < (dftd.total - 1) )
               {
                  ++dftd.first ;
                  ++dftd.last ;
                  ++dftd.curr ;
               }
               break ;
            case nckUP:
               if ( dftd.curr > dftd.first )
                  --dftd.curr ;
               else if ( dftd.first > ZERO )
               {
                  --dftd.first ;
                  --dftd.last ;
                  --dftd.curr ;
               }
               break ;
            case nckPGDOWN:
               if ( dftd.curr < dftd.last )
                  dftd.curr = dftd.last ;
               else if ( dftd.last < (dftd.total - 1) )
               {
                  dftd.last = dftd.curr = 
                     ((dftd.last + dftd.rows) < dftd.total) ? 
                      (dftd.last + dftd.rows) : (dftd.total - 1) ;
                  dftd.first = dftd.last - dftd.rows + 1 ;
               }
               break ;
            case nckPGUP:
               if ( dftd.curr > dftd.first )
                  dftd.curr = dftd.first ;
               else if ( dftd.first > ZERO )
               {
                  dftd.first = dftd.curr = 
                   (dftd.first <= dftd.rows) ? ZERO : (dftd.first - dftd.rows) ;
                  dftd.last = dftd.first + dftd.rows - 1 ;
               }
               break ;
            case nckHOME:
               if ( dftd.curr > ZERO || dftd.lshift > ZERO )
               {
                  dftd.first = dftd.curr = ZERO ;
                  dftd.last  = (dftd.total < dftd.rows ?
                                 dftd.total : dftd.rows)  - 1 ;
                  dftd.lshift = ZERO ;
               }
               break ;
            case nckEND:
               if ( dftd.curr < (dftd.total - 1) || dftd.lshift > ZERO )
               {
                  dftd.last = dftd.curr = dftd.total - 1 ;
                  dftd.first = (dftd.total <= dftd.rows) ? 
                                ZERO : (dftd.last - dftd.rows + 1) ;
                  dftd.lshift = ZERO ;
               }
               break ;
            case nckRIGHT:
               if ( ((dftd.width - dftd.lshift) > dftd.cols) &&
                    (dftd.lshift < (dftd.levels * 2)) )
               {
                  dftd.lshift += 2 ;
               }
               break ;
            case nckLEFT:
               if ( dftd.lshift > ZERO )
               {
                  dftd.lshift -= 2 ;
                  if ( dftd.lshift < ZERO )  dftd.lshift = ZERO ; // (unlikely)
               }
               break ;
         }
         if ( dftd.curr != prevCurr || dftd.lshift != prevLshift )
            redraw = true ;
      }

      //* Request to navigate to parent directory *
      //*       (if not already at root)          *
      else if ( ((oldcwd.compare( ROOT_PATH )) != ZERO) && 
                (wk.type == wktFUNKEY) &&
                (((dftd.curr == ZERO) &&
                 ((wk.key == nckENTER) || (wk.key == nckpENTER)))
                 ||
                 ((wk.key == nckBKSP) || (wk.key == nckAUP))) )
      {
         this->fmPtr->ExtractPathname ( newcwd, oldcwd ) ;
         done = newCWD = toParent = true ;
      }

      //* Request to navigate to root directory   *
      //*       (if not already at root)          *
      else if ( ((wk.type == wktFUNKEY) && (wk.key == nckC_R)) &&
                ((oldcwd.compare( ROOT_PATH )) != ZERO) )
      {
         newcwd = ROOT_PATH ;
         done = newCWD = toParent = true ;
      }

      //* Request to navigate to specified (lower-level) directory *
      //* This selection will cause exit from Tree-View mode.      *
      else if ( (wk.type == wktFUNKEY && 
                 (wk.key == nckENTER || wk.key == nckpENTER)) || 
                (wk.type == wktEXTEND && wk.key == nckA_Q) )
      {
         newcwd = gsPath ;
         done = newCWD = true ;
      }

      //* Exit Tree-View mode without making a selection *
      //*   (returned path will be ignored by caller)    *
      else if ( ((wk.type == wktFUNKEY && 
                 (wk.key == nckC_Q || wk.key == nckESC))) || 
                (wk.type == wktPRINT && 
                 (wk.key == 'q' || wk.key == 'Q')) )
      {
         done = true ;
      }

      //* Display filesystem info for highlighted directory name.*
      else if ( wk.type == wktFUNKEY && wk.key == nckC_Y )
      {
         this->DisplayFilesystemInfo ( gsPath.ustr() ) ;
      }

      //* Search the list for a filename matching user's criteria.*
      else if ( ((wk.type == wktFUNKEY) && 
                 ((wk.key == nckC_F) || (wk.key == nckC_G)))
                ||
                ((wk.type == wktPRINT) && 
                 ((wk.key == fSLASH) || (wk.key == L'n'))) )
      {  //* Duplicate 'less' search keys *
         if ( wk.key == fSLASH )  wk.key = nckC_F ;
         else if ( wk.key == L'n' )  wk.key = nckC_G ;

         winPos spPos( (firstDLine + 1), (cols / 2) ) ;
         bool newSearch = false ;
         if ( wk.key == nckC_F || sPattern.gschars() == 1 )
         {  //* Ask user for search pattern, and search for a match *
            dp->SetDialogObscured () ;
            if ( (newSearch = this->dftNamePrompt ( spPos, sPattern )) != false )
            {
               sIndex = ZERO ;
               sMatch = ZERO ;
               newSearch = true ;
            }
            dp->RefreshWin () ;
         }
         if ( newSearch || wk.key == nckC_G ) // 'find' OR 'find again'
         {
            if ( dftd.curr < sIndex )
               sIndex = ZERO ;
            else if ( !newSearch && (sIndex < (dftd.total-2)) )
               ++sIndex ;
            if ( (sIndex < (dftd.total-1)) &&
                 (dftNameSearch ( dftd, sPattern, sIndex )) != false )
            {
               dftd.curr = sIndex ;    // new highlighted item
               //* Scroll the item into view *
               while ( dftd.curr < dftd.first && dftd.first > ZERO )
               {
                  --dftd.first ;
                  --dftd.last ;
               }
               while ( dftd.curr > dftd.last && dftd.last < (dftd.total-1) )
               {
                  ++dftd.first ;
                  ++dftd.last ;
               }
               sMatch = 1 ;
               redraw = true ;
            }
            else
            {  //* Report no-match *
               dp->SetDialogObscured () ;
               dftNoMatch ( spPos, noMatchMsg[sMatch], this->cs.sb ) ;
               dp->RefreshWin () ;
               sIndex = ZERO ;   // force next search to top of list
            }
         }
      }

      //* Invoke context help for Tree View *
      else if ( (wk.type == wktFUNKEY) && 
                ((wk.key == nckSF01) || (wk.key == nckC_H)) )
      {
         dp->SetDialogObscured () ;    // save the dialog display data
         this->CallContextHelp ( cxtTREE_VIEW ) ;
         dp->RefreshWin () ;           // restore dialog display data
      }

      #if 0    // Potential future enhancements
      else if ( (wk.type == wktFUNKEY && wk.key == nckC_C) )
      {  //* Collapse current node of the tree view - hide lower branches *
      }
      else if ( (wk.type == wktFUNKEY && wk.key == nckC_E) )
      {  //* Expand current node of the tree view - show lower branches *
      }
      else if ( (wk.type == wktFUNKEY && wk.key == nckAC_C) )
      {  //* Collapse the entire tree view - show only top two levels *
      }
      else if ( (wk.type == wktFUNKEY && wk.key == nckAC_E) )
      {  //* Expand the entire tree view - show all branches *
      }
      #endif   // U/C
   }
   while ( ! done ) ;

   short status = newCWD == false ? (-1) : toParent == false ? ZERO : 1 ;
   return status ;

}  //* End dftDisplayTree() *

//*************************
//*    dftFormatTree      *
//*************************
//******************************************************************************
//* Create a two-dimensional (wide) character array and format the directory   *
//* information for display.                                                   *
//*                                                                            *
//* Input  : dftd    : (by reference) initialized display control data         *
//*                    dftd.data[] receives the formatted display/index data   *
//*          baseNode: base node of source directory tree                      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Programmer's Note: Formatting and display of the data use a two-column     *
//* indentation for each successive directory-tree level AND note that         *
//* formatting of the data will fail (low-level branches may not be fully      *
//* visible) if any string is longer than dftITEM_LENGTH.                      *
//*                                                                            *
//******************************************************************************

void FileDlg::dftFormatTree ( dftDisplay& dftd, TreeNode* baseNode )
{
   gString gt ;         // for formatting target node's name for display
   UINT di = ZERO ;     // index into top-level subdirectories

   // If there is a parent directory (basePath != root) *
   gt = baseNode->dirStat.fName ;
   if ( (gt.compare( ROOT_PATH )) != ZERO )
   {
      gt = parentDir ;
      gt.copy( dftd.data[di++].n, dftITEM_LEN ) ;
      dftd.width = gt.gscols() ; // set baseline column count
   }

   dftd.curr = di ;              // initially-highlighted item

   //* Base Directory Name: If directory name is tagged as no-read-access *
   //* or external-filesystem. Note that there will be no more than one   *
   //* indicator per line item with 'noAccess' having precedence.         *
   gt = baseNode->dirStat.fName ;
   if ( ! baseNode->dirStat.readAcc )
      dftMarkRestricted ( gt ) ;
   else if ( ! baseNode->dirStat.fsMatch )
      dftMarkExternal ( gt ) ;
   gt.copy( dftd.data[di++].n, dftITEM_LEN ) ;
   if ( gt.gscols() > dftd.width )  dftd.width = gt.gscols() ;

   //* Recursively initialize the display data *
   di += this->dftFormatNode ( dftd, baseNode, di ) ;

   //* Define initial block of items to display.*
   dftd.first = ZERO ;
   dftd.last  = (dftd.total < dftd.rows ? dftd.total : dftd.rows) - 1 ;
   dftd.lshift = ZERO ;

}  //* End dftFormatTree() *

//*************************
//*     dftFormatNode     *
//*************************
//******************************************************************************
//* Format for display the specified node on the tree.                         *
//* Caller has verified that there is at least one subdirectory attached       *
//* to this node.                                                              *
//*                  NOTE: This is a recursive method.                         *
//*                                                                            *
//* Input  : dftd  : (by reference) initialized display control data           *
//*                  dftd.data[] receives the formatted display and index data *
//*          tnPtr : pointer to array of TreeNode objects to be processed      *
//*          di    : index of next free item in dftd.data[] array              *
//*                                                                            *
//* Returns: next free item in dftd.data[] array                               *
//******************************************************************************

UINT FileDlg::dftFormatNode ( dftDisplay& dftd, TreeNode* tnPtr, UINT di )
{
   gString gt ;
   UINT parentIndex = di - 1, 
        actualNodes, actualFiles ;
   tnPtr->GetActual ( actualNodes, actualFiles ) ;
   short indent ;

   for ( UINT ni = ZERO ; (ni < actualNodes) && (di < dftd.total) ; ++ni )
   {
      //* Sort the TreeNode array alphabetically *
      dftSortNode ( tnPtr->nextLevel, actualNodes ) ;

      //* Format the directory name *
      gt.compose( "%s", tnPtr->nextLevel[ni].dirStat.fName ) ;
      indent = tnPtr->nextLevel[ni].level * 2 ;
      while ( indent-- > ZERO )
         gt.insert( L' ' ) ;

      //* If restricted-access directory *
      if ( ! tnPtr->nextLevel[ni].dirStat.readAcc )
         dftMarkRestricted ( gt ) ;    // mark directory as read-only

      //* If mountpoint for external filesystem *
      else if ( ! tnPtr->nextLevel[ni].dirStat.fsMatch )
         dftMarkExternal ( gt ) ;      // mark directory as external filesystem

      //* Save the formatted entry *
      gt.copy( dftd.data[di].n, dftITEM_LEN ) ;
      dftd.data[di].p = parentIndex ;
      dftd.data[di++].l = tnPtr->nextLevel[ni].level ;

      //* Track the widest entry *
      if ( gt.gscols() > dftd.width )  dftd.width = gt.gscols() ;

      //* Process lower levels of the tree *
      if ( tnPtr->nextLevel[ni].nextLevel != NULL )
         di = this->dftFormatNode ( dftd, &tnPtr->nextLevel[ni], di ) ;

      //* Track the depth of the tree *
      if ( (ni == ZERO) && (tnPtr->nextLevel[ni].level > dftd.levels) )
         dftd.levels = tnPtr->nextLevel[ni].level ;
   }

   return di ;       // index of next free display item

}  //* End dftFormatNode() *

//*************************
//*     dftNamePrompt     *
//*************************
//******************************************************************************
//* Ask user for all or part of the subdirectory name for which to search.     *
//*                                                                            *
//* Input  : spPos : (by reference) position of dialog window                  *
//*          ptrn  : (by reference) receives the user's search-pattern text    *
//*                                                                            *
//* Returns: 'true'  if search pattern entered, else 'false'                   *
//******************************************************************************

bool FileDlg::dftNamePrompt ( const winPos& spPos, gString& ptrn )
{
   enum dCtrls : short { patTB, okPB, canPB, controlsDEFINED } ;
   const short dialogROWS = 8,
               dialogCOLS = 48,
               dialogULY  = spPos.ypos,
               dialogULX  = spPos.xpos - dialogCOLS / 2 ;
   const attr_t dColor = this->cs.sb ;
   InitCtrl ic[controlsDEFINED] =      // array of dialog control info
   {
      {  //* 'Pattern" Text Box  - - - - - - - - - - - - - - - - - - -   patTB *
         dctTEXTBOX,                   // type:      
         rbtTYPES,                     // rbSubtype: (n/a)
         false,                        // rbSelect:  (n/a)
         3,                            // ulY:       upper left corner in Y
         2,                            // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         short(dialogCOLS - 4),        // cols:      control columns
         ptrn.ustr(),                  // dispText:  
         this->cs.tn,                  // nColor:    non-focus color
         this->cs.tf,                  // fColor:    focus color
         #if LINUX_SPECIAL_CHARS != 0
         tbFileLinux,                  // filter: valid filename chars (incl. Linux "special")
         #else    // BASIC FILENAME FILTER
         tbFileName,                   // filter:    valid filename characters
         #endif   // BASIC FILENAME FILTER
         "Enter all or part of subdirectory name:", // label:     
         -1,                           // labY:      
         ZERO,                         // labX       
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[okPB],                    // nextCtrl:  link in next structure
      },
      {  //* 'OK' pushbutton - - - - - - - - - - - - - - - - - - - - -    okPB *
         dctPUSHBUTTON,                // type:      
         rbtTYPES,                     // rbSubtype: (n/a)
         false,                        // rbSelect:  (n/a)
         short(dialogROWS - 2),        // ulY:       upper left corner in Y
         short(dialogCOLS / 2 - 11),   // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         8,                            // cols:      control columns
         "   OK   ",                   // dispText:  
         this->cs.pn,                  // nColor:    non-focus color
         this->cs.pf,                  // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "",                           // label:     (n/a)
         ZERO,                         // labY:      (n/a)
         ZERO,                         // labX       (n/a)
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         &ic[canPB],                   // nextCtrl:  link in next structure
      },
      {  //* 'Cancel' pushbutton - - - - - - - - - - - - - - - - - -   cancelPush *
         dctPUSHBUTTON,                // type:      
         rbtTYPES,                     // rbSubtype: (n/a)
         false,                        // rbSelect:  (n/a)
         ic[okPB].ulY,                 // ulY:       upper left corner in Y
         short(ic[okPB].ulX + 11),     // ulX:       upper left corner in X
         1,                            // lines:     (n/a)
         8,                            // cols:      control columns
         " CANCEL ",                   // dispText:  
         this->cs.pn,                  // nColor:    non-focus color
         this->cs.pf,                  // fColor:    focus color
         tbPrint,                      // filter:    (n/a)
         "",                           // label:     (n/a)
         ZERO,                         // labY:      (n/a)
         ZERO,                         // labX       (n/a)
         ddBoxTYPES,                   // exType:    (n/a)
         1,                            // scrItems:  (n/a)
         1,                            // scrSel:    (n/a)
         NULL,                         // scrColor:  (n/a)
         NULL,                         // spinData:  (n/a)
         true,                         // active:    allow control to gain focus
         NULL,                         // nextCtrl:  link in next structure
      },
   } ;

   bool search = false ;            // return value

   //* Initial parameters for dialog window *
   InitNcDialog dInit( dialogROWS,     // number of display lines
                       dialogCOLS,     // number of display columns
                       dialogULY,      // Y offset from upper-left of terminal 
                       dialogULX,      // X offset from upper-left of terminal 
                       NULL,           // dialog title
                       ncltSINGLE,     // border line-style
                       dColor,         // border color attribute
                       dColor,         // interior color attribute
                       ic              // pointer to list of control definitions
                     ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   //* Open the dialog window *
   if ( (dp->OpenWindow()) == OK )
   {
      dp->SetDialogTitle ( "  NAME SEARCH  ", this->cs.em ) ;
      dp->WriteString ( ic[patTB].ulY + 1, ic[patTB].ulX + 7,
                        "(search IS NOT case sensitive)", dColor ) ;
      dp->RefreshWin () ;                    // make the text visible

      uiInfo   Info ;                     // user interface data returned here
      short    icIndex = ZERO ;           // index of control with input focus
      bool     done = false ;             // loop control
      while ( ! done )
      {
         if ( ic[icIndex].type == dctPUSHBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditPushbutton ( Info ) ;
            if ( Info.dataMod != false )
            {
               if ( Info.ctrlIndex == okPB )
               {
                  if ( (dp->GetTextboxText ( patTB, ptrn )) > 1 )
                  {
                     search = true ;
                     done = true ;
                  }
                  else
                  {
                     dp->ClearLine ( 5, false ) ;
                     dp->WriteString ( 5, (dialogCOLS / 2 - 21), 
                        " No directory name specified! - Try again. ", 
                        dColor | ncbATTR, true ) ;
                     dp->NextControl () ;
                  }
               }
               else     // 'Cancel'
               {
                  ptrn.clear() ;
                  done = true ;
               }
            }
         }
         else if ( ic[icIndex].type == dctTEXTBOX )
         {
            Info.viaHotkey  = false ;     // ignore hotkey data
            icIndex = dp->EditTextbox ( Info ) ;
            dp->ClearLine ( 5, false ) ;
         }
         //* Move focus to appropriate control *
         if ( !done && !Info.viaHotkey )
         {
            if ( Info.keyIn == nckSTAB )
               icIndex = dp->PrevControl () ; 
            else
               icIndex = dp->NextControl () ;
         }
      }     // while(!done)
   }
   if ( dp != NULL )
      delete ( dp ) ;                        // close the window
   return search ;

}  //* End dftNamePrompt() *


//* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  *
//* - - - - - - - - -  Context-help Invocation Group  - - - - - - - - - - - -  *
//* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  *

//*************************
//*    CallContextHelp    *
//*************************
//******************************************************************************
//* Invoke the 'Info' system to display contextual help information.           *
//*                                                                            *
//* -- This method uses the NcDialog-class 'ShellOut' method                   *
//* -- The ShellOut() method saves the application dialog (if caller hasn't    *
//*    already saved it).                                                      *
//* -- Note that after returning from the shell the application dialog is      *
//*    refreshed, which releases the saved display data; so if caller has set  *
//*    the application dialog as 'obscured', then caller must save it AGAIN    *
//*    on return.                                                              *
//* -- Since we don't know about any sub-dialogs that may be open, it is       *
//*    assumed that caller has saved the display data for those dialogs and    *
//*    will restore them on return.                                            *
//*                                                                            *
//* Example sequence:                                                          *
//*   this->dPtr->SetDialogObscured () ;                                       *
//*     .  .  . (do stuff) .  .  .                                             *
//*   dp->SetDialogObscured () ;                                               *
//*   this->CallContextHelp ( cxtMOD_STATS ) ;                                 *
//*   this->dPtr->SetDialogObscured () ;                                       *
//*   dp->RefreshWin () ;                                                      *
//*                                                                            *
//*                                                                            *
//* Input  : ccode  : member of contextCode indicating the context of call     *
//*                                                                            *
//* Returns:  OK if successful, ERR if error                                   *
//******************************************************************************
//* Programmer's Notes:                                                        *
//*                                                                            *
//*                                                                            *
//*                                                                            *
//******************************************************************************

short FileDlg::CallContextHelp ( contextCode ccode )
{
   //* For development only. The development version of the application *
   //* is not in the installation directory, so compensate.             *
   //*         REMEMBER TO DISABLE THIS BEFORE POSTING A RELEASE.       *
   #define DEBUG_CallContextHelp (0)

   short status = OK ;        // return value

   gString  helpPath( "%s/filemangler.info", this->cfgOptions.appPath ),
            shellCmd ;         // construct the command

   #if DEBUG_CallContextHelp != 0
   helpPath.insert( "/Texinfo", helpPath.findlast( fSLASH ) ) ;
   #endif   // DEBUG_CallContextHelp

   //* Be sure the help file exists *
   fmFType ft ;
   if ( (this->TargetExists ( helpPath, ft )) != false )
   {
      if ( ccode == cxtMAIN )
         shellCmd.compose( "info -f %S", helpPath.gstr() ) ;
      else if ( ccode == cxtMOD_STATS )
         shellCmd.compose( "info -f %S -n 'modify stats'", helpPath.gstr() ) ;
      else if ( ccode == cxtARCHIVE_C )
         shellCmd.compose( "info -f %S -n 'Creating an Archive'", helpPath.gstr() ) ;
      else if ( ccode == cxtARCHIVE_U )
         shellCmd.compose( "info -f %S -n 'Updating an Archive'", helpPath.gstr() ) ;
      else if ( ccode == cxtARCHIVE_E )
         shellCmd.compose( "info -f %S -n 'Expanding an Archive'", helpPath.gstr() ) ;
      else if ( ccode == cxtTREE_VIEW )
         shellCmd.compose( "info -f %S -n 'Tree-view Mode'", helpPath.gstr() ) ;
      else                       // invalid option
         status = ERR ;
   }
   else
      status = ERR ;

   if ( status == OK )
      this->dPtr->ShellOut ( soX, shellCmd.ustr() ) ;

   if ( status != OK )
   {  //* Write a warning to the status window *
      dtbmData msgData = "  Unable to access the help file.  " ;
      this->dPtr->DisplayTextboxMessage ( this->mIndex, msgData ) ;
   }
   return status ;

}  //* End CallContextHelp() *

//* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  *
//* - - - - - - - - -   Non-member Methods  - - - - - - - - - - - - - - - - -  *
//* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  *

//*************************
//*      dftSortNode      *
//*************************
//******************************************************************************
//* Sort the items in specified TreeNode array.                                *
//*                                                                            *
//* Input  : tnPtr  : pointer to array of TreeNode objects                     *
//*          count  : number of items in the array                             *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

int snq ( const TreeNode* itemA, const TreeNode* itemB )
{  //* qsort callback function *
   gString A( itemA->dirStat.fName ),
           B( itemB->dirStat.fName ) ;
   int compared = A.compare( B.gstr() ) ;
   return compared ;
}

static void dftSortNode ( TreeNode* tnPtr, UINT count )
{
   if ( count > 1 )
   {
      //* Sort the array items by directory name *
      qsort ( tnPtr, count, sizeof(TreeNode), (comparison_fn_t)snq ) ;
   }
}  //* End dftSortNode() *

//*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  - *
//*  -  -  -  -  Non-member methods for DisplayFileTree group. -  -  -  -  -   *
//*  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  - *

//*************************
//*   dftMarkRestricted   *
//*************************
//******************************************************************************
//* If read access to named directory is restricted, then append an indicator. *
//*                                                                            *
//* Input  : dName : (by reference) display data including directory name      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void dftMarkRestricted ( gString& dName )
{

   dName.append( noAccess ) ;

}  //* End dftMarkRestricted)() *

//*************************
//*    dftMarkExternal    *
//*************************
//******************************************************************************
//* If named directory is on a different filesystem from its parent, then      *
//* append an indicator.                                                       *
//*                                                                            *
//* Input  : dName : (by reference) display data including directory name      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void dftMarkExternal ( gString& dName )
{

   dName.append( L"%S", extFileSys ) ;

}  //* End dftMarkExternal() *

//*************************
//*     dftNameSearch     *
//*************************
//******************************************************************************
//* Search the list of subdirectory names for a match to caller's string.      *
//* Search is CASE INSENSITIVE.                                                *
//*                                                                            *
//* Input  : dftd  : (by reference) initialized display control data           *
//*          ptrn  : text pattern for which to search                          *
//*          fIndex: index of first record to search (search proceeds downward)*
//*                                                                            *
//* Returns: 'true' if a matching entry found                                  *
//*            fIndex will reference matched record                            *
//*          else, 'false' (fIndex unchanged)                                  *
//******************************************************************************

static bool dftNameSearch ( const dftDisplay& dftd, const gString ptrn, UINT& fIndex )
{
   gString gsItem ;
   bool matchFound = false ;     // return value

   for ( UINT di = fIndex ; !matchFound && di < dftd.total ; ++di )
   {
      gsItem = dftd.data[di].n ;
      if ( (gsItem.find( ptrn.gstr() )) >= ZERO )
      {
         fIndex = di ;
         matchFound = true ;
      }
   }
   return matchFound ;

}  //* End dftNameSearch() *

//*************************
//*      dftNoMatch       *
//*************************
//******************************************************************************
//* Reports that no matching sufdirectory name (or no additional names) found. *
//*                                                                            *
//* Input  : spPos : (by reference) position of dialog window                  *
//*          msg   : message to be displayed                                   *
//*          dColor: background color of dialog window                         *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void dftNoMatch ( const winPos& spPos, const char* msg, attr_t dColor )
{
   const short dialogROWS = 3,
               dialogCOLS = 44,
               dialogULY  = spPos.ypos,
               dialogULX  = spPos.xpos - dialogCOLS / 2 ;
   //* Initial parameters for dialog window *
   InitNcDialog dInit( dialogROWS,     // number of display lines
                       dialogCOLS,     // number of display columns
                       dialogULY,      // Y offset from upper-left of terminal 
                       dialogULX,      // X offset from upper-left of terminal 
                       NULL,           // dialog title
                       ncltSINGLE,     // border line-style
                       dColor,         // border color attribute
                       dColor,         // interior color attribute
                       NULL            // pointer to list of control definitions
                     ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   //* Open the dialog window *
   if ( (dp->OpenWindow()) == OK )
   {
      dp->WriteString ( 1, 2, msg, dColor, true ) ;
      chrono::duration<short, std::milli>aMoment( 1600 ) ;
      this_thread::sleep_for( aMoment ) ;
   }
   if ( dp != NULL )
      delete ( dp ) ;                        // close the window

}  //* End dftNoMatch

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

