//******************************************************************************
//* File       : FmCommand.cpp                                                 *
//* Author     : Mahlon R. Smith                                               *
//*              Copyright (c) 2005-2020 Mahlon R. Smith, The Software Samurai *
//*                  GNU GPL copyright notice located in FileMangler.hpp       *
//* Date       : 30-Jul-2020                                                   *
//* Version    : (see AppVersion string)                                       *
//*                                                                            *
//* Description: This is a support module for the FileMangler application.     *
//* This module, along with FmInterface.cpp contains the non-menu (command-key)*
//* portion of the User Interface which gathers key input and executes the     *
//* user's commands.                                                           *
//*                                                                            *
//* Development Tools: see notes in FileMangler.cpp.                           *
//******************************************************************************
//* Programmer's Notes:                                                        *
//*                                                                            *
//******************************************************************************

//* Header Files *
#include "FileMangler.hpp"    // FileMangler definitions and data

//*****************
//** Definitions **
//*****************

//****************
//** Prototypes **
//****************

//****************
//** Local Data **
//****************

//* Threshold for "very large" storage devices (in bytes). *
//* Used to determine whether to scan the full directory   *
//* tree or only the top-level data. See CmdMount().       *
//* This value was selected to exclude most USB flash      *
//* drives and common SD cards. Intended primarily to      *
//* identify external SATA hard drives.                    *
const UINT64 HUGE_EXT_DRIVE = 64000000000 ;  // (64GB)


//***********************
//*     CmdTrash        *
//***********************
//******************************************************************************
//* Perform the specified operation on the system trashcan.                    *
//* The trashcan used is either the GNOME/KDE/ETC. Trashcan, OR AN ALTERNATE   *
//* trashcan directory specified during startup.                               *
//*                                                                            *
//* Note that if desktop trashcan not validated AND if no alternate Trashcan   *
//* location was specified in the config file (or it is mis-configured),       *
//* then Trashcan operations will have been disabled. We check for this before *
//* trying to trash the files and alert the user about the trashcan error.     *
//*                                                                            *
//* Input  : mCode: menu code                                                  *
//*                 mcFcmdMB_TRASH  : Move files to Trashcan.                  *
//*                 mcUtilMB_UTRASH : Restore most recent move to Trashcan     *
//*                 mcUtilMB_MTRASH : Empty the Trashcan                       *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Note on sending files to trashcan:                                         *
//*   If in Dual-Window mode AND if the inactive window is anywhere on the path*
//*   of files being deleted, then after the operation, the path of inactive   *
//*   window will no longer exist and must be updated to a directory that      *
//*   actually does exist.                                                     *
//*                                                                            *
//* NOTE: There is a certain schizophrenia about calling the Empty/Restore     *
//*       dialog for either an Empty or Restore operation. It's really the     *
//*       same call sequence, but the flag is different. However, the user     *
//*       can change his/her/its mind about the operation while in the dialog, *
//*       and at this level, we would not know it.                             *
//******************************************************************************

void FileMangler::CmdTrash ( MenuCode mCode )
{
   //* If the CWD is the root directory, disallow file operations.*
   if ( (this->isRoot ()) )
   { this->RootWarning () ;  return ; }   // NOTE THE EARLY RETURN


   if ( this->fdPtr->TrashcanConfigured () )
   {
      //* Move selected files to Trashcan *
      if ( mCode == mcFcmdMB_TRASH )
      {
         //* If in Dual-Win mode, clear any file *
         //* selections in the inactive window   *
         if ( (this->ClearAltFileListSelections ()) == false )
            this->RestoreQuickHelpMsg () ;

         //* Move files that have been marked as 'selected' * 
         //* (or currently-highlighted file) to trashcan    *
         UINT     selFiles,            // number of 'selected' files
                  movFiles ;           // number file files successfully moved to trash
         movFiles = this->fdPtr->TrashcanSelectedFiles ( selFiles ) ;

         //* If in Dual-Win mode AND alternate window is the same as source    *
         //* dir (or an affected child dir), update the alternate window also. *
         if ( (this->DualwinMode ()) )
         {
            fmFType apType ;
            gString srcPath, trgPath ;
            this->fdPtr->GetPath ( srcPath ) ;
            this->altPtr->GetPath ( trgPath ) ;
            //* If alt window's target dir has disappeared (see note above) *
            if ( !(this->altPtr->TargetExists ( trgPath, apType )) )
            {
               trgPath = srcPath ;
               if ( (this->altPtr->SetDirectory ( trgPath )) == OK )
                  this->altPtr->RedrawCurrDir ( false ) ;
               else
               { /* (this is unlikely) */ }
            }
            else if ( trgPath == srcPath )
               this->RefreshAltFileList () ;
         }

         dtbmData msgData ;            // message display
         gString gs ;                  // output formatting
         if ( movFiles == selFiles )
            gs.compose( L"  %u files moved to trashcan", &movFiles ) ;
         else
         {
            UINT errFiles = selFiles - movFiles ;
            gs.compose( L"  %u moved successfully, but ERROR moving %u file(s)", 
                        &movFiles, &errFiles ) ;
         }
         gs.copy( msgData.textData, gsMAXCHARS ) ;
         this->dWin->DisplayTextboxMessage ( dwMsgsTB, msgData ) ;
      }  // (mCode == mcFcmdMB_TRASH)

      //* Restore item(s) from Trashcan *
      else if ( mCode == mcUtilMB_UTRASH )
      {  //* Undo the most recent move-to-trash operation targeting *
         //* our trashcan. It doesn't matter if the operation was   *
         //* performed by our application or by another application.*

         //* Give user a clue *
         attr_t color = dtbmNFcolor ;
         dtbmData msgData( " Undo recent move-to-trash operation. ", &color ) ;
         this->dWin->DisplayTextboxMessage ( dwMsgsTB, msgData ) ;

         this->fdPtr->ManageTrashcan ( true ) ; // restore items

         //* Active window will be automagically refreshed by the call, but    *
         //* if in Dual-win Mode, we explicitly update the inactive window     *
         //* also because user may have restored files to it.                  *
         this->RefreshAltFileList ( true ) ;

      }  // (mCode == mcUtilMB_UTRASH)

      //* Empty the Trashcan *
      else if ( mCode == mcUtilMB_MTRASH )
      {
         attr_t color = dtbmFcolor ;
         dtbmData    msgData( " Manage the Trashcan ", &color ) ;
         this->dWin->DisplayTextboxMessage ( dwMsgsTB, msgData ) ;

         //* Active window will be automagically refreshed by the call, but    *
         //* if in Dual-win Mode, we explicitly update the inactive window     *
         //* also because user may have restored files to it.                  *
         this->fdPtr->ManageTrashcan () ;
         this->RefreshAltFileList ( true ) ;
         this->RestoreQuickHelpMsg () ;
      }  // (mCode == mcUtilMB_MTRASH)

      else
         { /* THIS WOULD BE AN APPLICATION ERROR! */ }
   }     // if(TrashcanConfigured)
   else
   {
      const char* msgText[] = 
      {
         "  ALERT! ALERT!  ",
         "Trashcan operations disabled.",
         " ",
         "Either the System Trashcan could not be found,",
         " OR",
         "specified Alt Trash location improperly configured.",
         NULL
      } ;
      this->dWin->SetDialogObscured () ;
      this->fdPtr->InfoDialog ( msgText ) ;
      this->dWin->RefreshWin () ;
   }

}  //* End CmdTrash() *

//*************************
//*     CmdFavorites      *
//*************************
//******************************************************************************
//* Invoke a dialog in which the user can select a new target directory from   *
//* the favorites list.                                                        *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void FileMangler::CmdFavorites ( void )
{
const char* dirVanished[] = 
{
   "  ALERT  ",
   "The selected target directory does not exist or is",
   "currently inaccessable. Please verify that the path",
   "was entered correctly.",
   " ",
   "If path refers to an external mountable device or",
   "to a network drive, please mount it and try again.",
   NULL
} ;

const short 
   dialogRows = (MIN_ROWS - 3),  // dialog size and position
   dialogCols = (MIN_COLS - 2),
   ulY  = (this->actualRows / 2) - (dialogRows / 2),
   ulX  = (this->actualCols / 2) - (dialogCols / 2),
   pathCols = dialogCols - 2 ;   // max length of path display string
   gString newDir ;              // new (or current) path
   attr_t dColor = this->cfgOpt.cScheme.sd ; // dialog base color attribute
   bool  newCWD = false ;        // if true, move to user-selected directory

   //* Formatting for the favorites list display data *
   gString fgs[MAX_FAVDIRS] ; 
   const char* seboxData[MAX_FAVDIRS] ;
   attr_t seboxColors[MAX_FAVDIRS] ;
   ssetData sData( seboxData, seboxColors, ZERO, ZERO, true ) ;
   for ( short i = ZERO ; i < MAX_FAVDIRS ; i++ )
   {
      fgs[i] = this->FavoriteDirs[i] ;
      this->fdPtr->TrimPathString ( fgs[i], pathCols ) ;
      seboxData[i] = fgs[i].ustr() ;
      if ( (fgs[i].gschars()) > 1 )
         ++sData.dispItems ;
      seboxColors[i] = this->cfgOpt.cScheme.tf ;
   }
   if ( sData.dispItems == ZERO )
   {
      fgs[ZERO] = "  none specified  " ;
      seboxData[ZERO] = fgs[ZERO].ustr() ;
      ++sData.dispItems ;
   }

   //* Create a list of needed controls *
   enum dpCtrls { favSE, okPB, canPB, dpcCOUNT } ;
   InitCtrl ic[dpcCOUNT] =            // control definitions
   {
   { //* 'Favorite Directories' scroll ext - - - - - - - - - - - - - -   favSE *
      dctSCROLLEXT,                 // type:      define a scrolling-data control
      rbtTYPES,                     // rbSubtype: (na)
      false,                        // rbSelect:  (n/a)
      ZERO,                         // ulY:       upper left corner in Y
      ZERO,                         // ulX:       upper left corner in X
      10,                           // lines:     control lines
      dialogCols,                   // cols:      control columns
      NULL,                         // dispText:  n/a
      dColor,                       // nColor:    non-focus border color
      dColor,                       // fColor:    focus border color
      tbPrint,                      // filter:    (n/a)
      " Favorite Directories ",     // label:     
      ZERO,                         // labY:      offset from control's ulY
      ZERO,                         // labX       offset from control's ulX
      ddBoxTYPES,                   // exType:    (n/a)
      MAX_FAVDIRS,                  // scrItems:  number of elements in text/color arrays
      ZERO,                         // scrSel:    (n/a)
      seboxColors,                  // scrColor:  single-color data display
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[okPB],                    // nextCtrl:  link in next structure
   },
   {  //* 'OK' pushbutton   - - - - - - - - - - - - - - - - - - - - - -   okPB *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      short(dialogRows - 2),        // ulY:       upper left corner in Y
      short(dialogCols / 2 - 10),   // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      8,                            // cols:      control columns
      "   OK   ",                   // dispText:  
      this->cfgOpt.cScheme.pn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[canPB],                   // nextCtrl:  link in next structure
   },
   {  //* 'CANCEL' pushbutton - - - - - - - - - - - - - - - - - - - - -  canPB *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      ic[okPB].ulY,                 // ulY:       upper left corner in Y
      short(ic[okPB].ulX + 12),     // ulX:
      1,                            // lines:     (n/a)
      8,                            // cols:      control columns
      " CANCEL ",                   // dispText:  
      this->cfgOpt.cScheme.pn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      NULL,                         // nextCtrl:  link in next structure
   },
   } ;

   //* Save the application's display data *
   this->dWin->SetDialogObscured () ;

   //* Initial parameters for dialog window, passed to sub-methods.*
   InitNcDialog dInit
   ( 
      dialogRows,           // number of display lines
      dialogCols,           // number of display columns
      ulY,                  // Y offset from upper-left of terminal 
      ulX,                  // X offset from upper-left of terminal 
      NULL,                 // dialog title (see below)
      ncltSINGLE,           // border line-style
      dColor,               // border color attribute
      dColor,               // interior color attribute
      ic                    // pointer to list of control definitions
   ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   if ( (dp->OpenWindow()) == OK )
   {
      //* Make visual connections to dialog border and     *
      //* initialize display data for dctSCROLLEXT control *
      cdConnect cdConn ;
      cdConn.connection = true ;
      cdConn.ll2Left = cdConn.lr2Right = true ;
      dp->ConnectControl2Border ( favSE, cdConn ) ;
      dp->SetScrollextText ( favSE, sData ) ;

      //* Draw the static text *
      dp->WriteParagraph ( 10, 9, 
               " Highlight the desired target directory, then select 'OK'.\n"
               "     Select 'CANCEL' to remain in current directory.\n"
               "\n"
               "       Note: To edit the Favorites list, go to the\n"
               "             'Util' menu and select 'Configure'.", dColor ) ;

      dp->RefreshWin () ;           // make everything visible

      //*************************************
      //* Talk to the animals, Dr. Dolittle *
      //*************************************
      uiInfo Info ;                 // user interface data returned here
      short  icIndex = ZERO ;       // index of control with input focus
      bool   done = false ;         // loop control
      while ( ! done )
      {
         //* If focus is currently on the Scroll Box *
         if ( ic[icIndex].type == dctSCROLLEXT )
         {
            for ( short i = ZERO ; i < MAX_FAVDIRS ; i++ )
               seboxColors[i] = this->cfgOpt.cScheme.tf ;
            dp->RefreshScrollextText ( favSE ) ;

            Info.viaHotkey = false ;      // ignore hotkey data
            icIndex = dp->EditScrollext ( Info ) ;

            for ( short i = ZERO ; i < MAX_FAVDIRS ; i++ )
               seboxColors[i] = this->cfgOpt.cScheme.tn ;
            dp->RefreshScrollextText ( favSE ) ;
         }

         //* If focus is currently on a Pushbutton   *
         else if ( ic[icIndex].type == dctPUSHBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditPushbutton ( Info ) ;

            if ( Info.dataMod != false )
            {
               if ( Info.ctrlIndex == okPB )
               {
#if 0    // TEMP - SCREENSHOT
dp->CaptureDialog ( "capturedlg.txt" ) ;
dp->CaptureDialog ( "capturedlg.html", true, false, 
                    "infodoc-styles.css", 4, false, nc.blR ) ;
#endif   // SCREENSHOT

                  //* Get a copy of the target string *
                  newDir = this->FavoriteDirs[dp->GetScrollextSelect ( favSE )] ;
   
                  //* Determine whether the selected entry is actually accessible.*
                  fmFType fileType ;
                  bool fileExists = this->fdPtr->TargetExists ( newDir, fileType ) ;
                  if ( fileExists && fileType == fmDIR_TYPE )
                  {  //* Set the new target AFTER dialog has closed *
                     newCWD = done = true ;
                  }
                  else
                  {  //* Selected directory does not exist, inform user *
                     dp->SetDialogObscured () ;
                     this->fdPtr->InfoDialog ( dirVanished ) ;
                     dp->PrevControl () ;
                     dp->RefreshWin () ;
                  }
               }
               else if ( Info.ctrlIndex == canPB ) // abort selection
                  done = true ;
            }
         }

         if ( done == false && Info.viaHotkey == false )
         {
            if ( Info.keyIn == nckSTAB )
               icIndex = dp->PrevControl () ; 
            else
               icIndex = dp->NextControl () ;
         }
      }  // while(!done)
   }
   else           // unable to open window (likely memory-allocation error)
      ;

   if ( dp != NULL )
      delete ( dp ) ;                        // close the window

   //* Restore application's display data *
   this->dWin->RefreshWin () ;

   //* If user has selected a new target directory, *
   //* set it as the new Current Working Directory. *
   if ( newCWD )
   {
      //* Get the size of the target filesystem. If target directory   *
      //* is the mountpoint for a very large storage device, read only *
      //* the top-level data.                                          *
      //* Else, assume that target is small enough to be read quickly. *
      // Programmer's Note: This test will identify the "/" (root) directory of 
      // an external drive as a mountpoint. Unfortunately, this test does not 
      // capture the "/home" directory of such a drive because it is not a mountpoint.
      bool recurse = true ;
      fileSystemStats fsStats ;
      this->fdPtr->GetFilesystemStats ( newDir, fsStats ) ;
      if ( (fsStats.isMountpoint) &&
           (fsStats.blockTotal * fsStats.blockSize) >= HUGE_EXT_DRIVE )
         recurse = false ;

      if ( (this->fdPtr->SetDirectory ( newDir, recurse )) != OK )
      {  //* Unable to move to new target directory *
         this->dWin->SetDialogObscured () ;
         this->fdPtr->InfoDialog ( dirVanished ) ;
         this->dWin->RefreshWin () ;
      }
   }

}  //* End CmdFavorites() *

//***********************
//*      CmdMount       *
//***********************
//******************************************************************************
//* Invoke a dialog in which the user can select the filesystem to be          *
//* mounted or unmounted.                                                      *
//*                                                                            *
//* Input  : mCode : menu code (either mcFileMB_MOUNT or mcFileMB_UNMOUNT      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Mount / Unmount Sequence                                                   *
//* ========================                                                   *
//*  - Because mounting and unmounting often require superuser privlege,       *
//*    we create a small terminal window in which to launch the mount call.    *
//*                                                                            *
//*  - The recommended way to find currently-mounted filesystem is 'findmnt'.  *
//*       findmnt --kernel --real --list -o TARGET,FSTYPE,SOURCE               *
//*    Our hack, 'SelectMount' uses this command to get the raw data, and then *
//*    filters out the obviouly-not-block-storage entries.                     *
//*                                                                            *
//*    We then use a combination of 'lsusb' and 'gio' to identify the attached *
//*    virtual filesystems (MTP/GVfs).                                         *
//*                                                                            *
//*    So, if we combine the results of pre-defined mounts in 'fstab' with     *
//*    the list of currently-mounted and currently-attached devices, PLUS the  *
//*    mounts specified in the config file (minus duplicates), we should have  *
//*    a solid list of devices which can be mounted/unmounted.                 *
//*       findmnt --fstab -o TARGET,FSTYPE,SOURCE                              *
//*       findmnt --kernel --real --list -o TARGET,FSTYPE,SOURCE               *
//*                                                                            *
//*  - Parsing the 'fstab' file for "known" filesystems.                       *
//*    - To find out what filesystems the system knows about, see /dev/disk.   *
//*      This directory contains three subdirectories with symbolic links      *
//*      according to:                                                         *
//*        /dev/disk/by-id            (this is pretty useless)                 *
//*        /dev/disk/by-uuid          (reports uuid and _relative_ device)     *
//*        /dev/disk/by-label         (reports label and _relative_ device)    *
//*        /dev/mapper                (members of LVM)                         *
//*      These directories contain symlinks to specific devices, (block        *
//*      special device drivers).                                              *
//*      Example: ls -l /dev/disk/by-label                                     *
//*       F20BACKUP32 -> ../../mmcblk0p1                                       *
//*       linuxboot -> ../../sda1                                              *
//*       LVM_home -> ../../dm-2                                               *
//*       LVM_root -> ../../dm-1                                               *
//*       LVM_swap -> ../../dm-0                                               *
//*                                                                            *
//*    - sudo blkid           list all block-device IDs                        *
//*      blkid -L <label>     ex: blkid /LVM_home  (reports device)            *
//*      blkid -U <uuid>                                                       *
//*      blkid <device>       ex: blkid /dev/sda1  (reports all, see '-o')     *
//*      blkid -o list                             (formatted report)          *
//*      device  fs_type  label  mountpoint  UUID                              *
//*      ------------------------------------------                            *
//*      x       x        x      x           x                                 *
//*      Note that the '-o list' option doesn't necessarily list the UUID for  *
//*      all devices, but does give the mountpoint and device for all _mounted_*
//*      filesystems.                                                          *
//*                                                                            *
//*    - The /etc/fstab file contains the rules for mounting known filesystems.*
//*        a) UUID  : "UUID=..."                                               *
//*        b) Label : "/dev/mapper/Fedora20-home /home ..."                    *
//*        c) Device: "/dev/sdb ..."                                           *
//*      The UUID entries are the most reliable because they are the most      *
//*      likely to uniquely identify the mount.                                *
//*                                                                            *
//*    - If the /etc/fstab entry for a filesystem includes the word 'user',    *
//*      then it can be mounted by any user.                                   *
//*      Example: UUID=e8ccb905-969b-44de-8e84-7a3a09d121c2                    *
//*                /run/media/sam/Ubuntu16_root  ext4                          *
//*                rw,suid,dev,exec,noauto,user,async 0 0                      *
//*    - Unless the device is 'auto' mounted, the 'sudo' command may be        *
//*      required for the invocation.                                          *
//*                                                                            *
//*    - Smartphone access:                                                    *
//*      lsusb command for USB device list                                     *
//*      gio mount --list --detail for additional info                         *
//*      Additional resources:                                                 *
//*         sudo install gmtp                                                  *
//*         sudo install mtp-tools                                             *
//*         jmtpfs (FUSE - POSIX)                                              *
//*                                                                            *
//*  - /etc/mtab and /proc/mounts                                              *
//*    These files contain a (mostly) up-to-date list of currently-mounted     *
//*    filesystems.                                                            *
//*                                                                            *
//*  - Commands:                                                               *
//*    mount -t TYPE DEVICE MOUNTPOINT                                         *
//*    mount /MOUNTPOINT  _or_  mount --target /MOUNTPOINT                     *
//*          This assumes that the associated device will be found in 'fstab'. *
//*    Note that if the target mountpoint does not exist, Nautilus will create *
//*    it before attempting the mount. For instance, our external Ubuntu drive *
//*    is mounted at '/run/media/sam/Ubuntu16_root' HOWEVER, that directory    *
//*    does not persist past a reboot, so it must be re-created each time the  *
//*    drive is mounted. (We are not sure how Nautilus handles this.)          *
//*                                                                            *
//*  - See '/proc/filesystems' for a list of supported filesystem types.       *
//*    The list of physical types will look something like this:               *
//*               ext3, ext2, ext4, fuseblk, vfat                              *
//*    'tmpfs' will often be seen in the tree, but this is not an actual       *
//*    filesystem.                                                             *
//*    Other entries are probably virtual types, e.g. 'proc' 'fusectl' etc.    *
//*                                                                            *
//******************************************************************************

//* Callback method used by CmdMount().*
static short cmCallback ( const short currIndex, const wkeyCode wkey, bool firstTime = false ) ;
//* Dialog pointer used by call-back method for CmdMount().*
static NcDialog* cmcbPtr = NULL ;
//* Pointer to filesystem mount information *
static const fsInfo* cmcbFsi = NULL ;

void FileMangler::CmdMount ( MenuCode mCode )
{
   const char *mntLabel  = "  MOUNT  " ;
   const short dialogRows = (MIN_ROWS - 3),  // dialog size and position
               dialogCols = (MIN_COLS - 2),
               ulY  = (this->actualRows / 2) - (dialogRows / 2),
               ulX  = (this->actualCols / 2) - (dialogCols / 2),
               pathCols = dialogCols - 21 ;  // limit display-string length
   attr_t dColor = this->cfgOpt.cScheme.sd,  // dialog base color attribute
          mntColor = nc.gr,      // color attribute of string for 'mounted'
          nmntColor = nc.br ;    // color attribute of string for 'not mounted'
   gString trgMnt ;              // target mountpoint
   short seSelect ;              // index if selected mount target
   bool    newMNT = false,       // if true, new filesystem mounted
           setCWD = false,       // if true, make target the new CWD
           superPriv = false ;   // if true, execute operation at superuser privilege

   //*Create a list of mounted (or mountable) storage devices (filesystems).*
   short fsCount = ZERO ;
   fsInfo* fsiPtr = this->cmScan4Filesystems ( fsCount ) ;
   gString* fgs = new gString[fsCount + 1] ;
   const char* seboxData[fsCount + 1] ;
   attr_t seboxColors[fsCount + 1] ;
   ssetData sData( seboxData, seboxColors, ZERO, ZERO, true ) ;


   //* Initialize the display data.*
   for ( short i = ZERO ; i < fsCount ; ++i )
   {
      fgs[i] = fsiPtr[i].fsMountpt ;
      this->fdPtr->TrimPathString ( fgs[i], pathCols ) ; // make display text fit
      seboxColors[i] = fsiPtr[i].fsMounted ? mntColor : nmntColor ;
      seboxData[i] = fgs[i].ustr() ;
      ++sData.dispItems ;        // non-null-string display items
   }

   if ( sData.dispItems == ZERO )
   {
      fgs[ZERO] = "  none specified  " ;
      seboxData[ZERO] = fgs[ZERO].ustr() ;
      seboxColors[ZERO] = nmntColor ;
      ++sData.dispItems ;
   }

   //* Create a list of needed controls *
   enum dpCtrls { mntSE, mntPB, canPB, cwdRB, superRB, dpcCOUNT } ;
   InitCtrl ic[dpcCOUNT] =            // control definitions
   {
   { //* 'Mount Commands' scroll ext   - - - - - - - - - - - - - - - -   mntSE *
      dctSCROLLEXT,                 // type:      define a scrolling-data control
      rbtTYPES,                     // rbSubtype: (na)
      false,                        // rbSelect:  (n/a)
      ZERO,                         // ulY:       upper left corner in Y
      ZERO,                         // ulX:       upper left corner in X
      10,                           // lines:     control lines
      dialogCols,                   // cols:      control columns
      NULL,                         // dispText:  n/a
      dColor,                       // nColor:    non-focus border color
      dColor,                       // fColor:    focus border color
      tbPrint,                      // filter:    (n/a)
      "  Mount Filesystems  ",      // label:     
      ZERO,                         // labY:      offset from control's ulY
      ZERO,                         // labX       offset from control's ulX
      ddBoxTYPES,                   // exType:    (n/a)
      MAX_MNTCMDS,                  // scrItems:  number of elements in text/color arrays
      ZERO,                         // scrSel:    (n/a)
      seboxColors,                  // scrColor:  single-color data display
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[mntPB],                   // nextCtrl:  link in next structure
   },
   {  //* 'MOUNT' ('UNMOUNT') pushbutton  - - - - - - - - - - - - - - -  mntPB *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      short(dialogRows - 4),        // ulY:       upper left corner in Y
      short(dialogCols / 2 - 12),   // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      9,                            // cols:      control columns
      mntLabel,                     // dispText:  
      this->cfgOpt.cScheme.pn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[canPB],                   // nextCtrl:  link in next structure
   },
   {  //* 'CANCEL' pushbutton - - - - - - - - - - - - - - - - - - - - -  canPB *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      ic[mntPB].ulY,                // ulY:       upper left corner in Y
      short(ic[mntPB].ulX + 14),    // ulX:
      1,                            // lines:     (n/a)
      8,                            // cols:      control columns
      " CANCEL ",                   // dispText:  
      this->cfgOpt.cScheme.pn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[cwdRB]                    // nextCtrl:  link in next structure
   },
   {  //* 'Display Mount Target' radiobutton  - - - - - - - - - - - - -  cwdRB *
      dctRADIOBUTTON,               // type:      
      rbtS3s,                       // rbSubtype: standard, 3-wide
      true,                         // rbSelect:  initially set
      short(ic[mntPB].ulY + 1),     // ulY:       upper left corner in Y
      short(ic[mntPB].ulX),         // ulX: upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      this->cfgOpt.cScheme.sd,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      "Display mount target data immediately.", // label:
      ZERO,                         // labY:      
      4,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[superRB]                  // nextCtrl:  link in next structure
   },
   {  //* 'Superuser' radiobutton   - - - - - - - - - - - - - - - - -  superRB *
      dctRADIOBUTTON,               // type:      
      rbtS3s,                       // rbSubtype: standard, 3-wide
      false,                        // rbSelect:  initially reset
      short(ic[cwdRB].ulY + 1),     // ulY:       upper left corner in Y
      ic[cwdRB].ulX,                // ulX: upper left corner in X
      1,                            // lines:     (n/a)
      0,                            // cols:      (n/a)
      NULL,                         // dispText:  (n/a)
      this->cfgOpt.cScheme.sd,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      "Execute operation with superuser privilege.", // label:
      ZERO,                         // labY:      
      4,                            // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      NULL                          // nextCtrl:  link in next structure
   },
   } ;

   //* Save the application's display data *
   this->dWin->SetDialogObscured () ;

   //* Initial parameters for dialog window, passed to sub-methods.*
   InitNcDialog dInit
   ( 
      dialogRows,           // number of display lines
      dialogCols,           // number of display columns
      ulY,                  // Y offset from upper-left of terminal 
      ulX,                  // X offset from upper-left of terminal 
      NULL,                 // dialog title (see below)
      ncltSINGLE,           // border line-style
      dColor,               // border color attribute
      dColor,               // interior color attribute
      ic                    // pointer to list of control definitions
   ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   if ( (dp->OpenWindow()) == OK )
   {
      //* Make visual connections to dialog border and     *
      //* initialize display data for dctSCROLLEXT control *
      cdConnect cdConn ;
      cdConn.connection = true ;
      cdConn.ll2Left = cdConn.lr2Right = true ;
      dp->ConnectControl2Border ( mntSE, cdConn ) ;
      dp->SetScrollextText ( mntSE, sData ) ;

      //* Draw the static text *
      winPos wp( 10, 2 ) ;
      dp->WriteParagraph ( wp, 
               "    Highlight the desired target filesystem, then tab to '(UN)MOUNT'.\n"
               "\n"
               "\n"
               "\n"
               "Legend:\n"
               "        mounted\n"
               "        not mounted", dColor ) ;
      dp->WriteString ( wp.ypos + 5, wp.xpos, " Green ", nc.gr ) ;
      dp->WriteString ( wp.ypos + 6, wp.xpos, " Brown ", nc.br ) ;

      //* Establish a callback to monitor user activity.*
      cmcbPtr = dp ;
      cmcbFsi = fsiPtr ;
      dp->EstablishCallback ( &cmCallback ) ;

      wp.ypos += 1 ;  wp.xpos = 2 ;
      dp->RefreshWin () ;           // make everything visible

      //*************************************
      //* Talk to the animals, Dr. Dolittle *
      //*************************************
      uiInfo Info ;                 // user interface data returned here
      short  icIndex = ZERO ;       // index of control with input focus
      bool   done = false ;         // loop control
      while ( ! done )
      {
         //* If focus is currently on the Scroll Box *
         if ( ic[icIndex].type == dctSCROLLEXT )
         {
            Info.viaHotkey = false ;      // ignore hotkey data
            icIndex = dp->EditScrollext ( Info ) ;

            //* If an MTP URI, and if device is not mounted.*
            dp->ClearLine ( wp.ypos + 1 ) ;
            if ( fsiPtr[Info.selMember].fsMtp && !fsiPtr[Info.selMember].fsMounted )
            {
               dp->WriteString ( wp.ypos + 1, wp.xpos,
                  "    Verify that screen of target device is unlocked.", 
                  this->cfgOpt.cScheme.em, true ) ;
            }
         }

         //* If focus is currently on a Pushbutton   *
         else if ( ic[icIndex].type == dctPUSHBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditPushbutton ( Info ) ;
            if ( Info.dataMod != false )
            {
               if ( Info.ctrlIndex == mntPB )
               {
                  chrono::duration<short, std::milli>aMoment( 1500 ) ;
                  winPos wpm = wp ;
                  dp->ClearLine ( wp.ypos, false ) ;

                  //* Get a copy of the target string *
                  seSelect = dp->GetScrollextSelect ( mntSE ) ;
                  trgMnt = fsiPtr[seSelect].fsMountpt ;
                  //* Execute with superuser privilege? *
                  dp->GetRadiobuttonState ( superRB, superPriv ) ; 

                  //**************************************
                  //* If target is not mounted, mount it.*
                  //**************************************
                  if ( fsiPtr[seSelect].fsMounted == false )
                  {
                     //* Give user a clue *
                     wpm = dp->WriteString ( wpm, " MOUNTING : ", 
                                            this->cfgOpt.cScheme.tf ) ;
                     wpm = dp->WriteString ( wpm, fgs[seSelect], 
                                            this->cfgOpt.cScheme.tf, true ) ;

                     dp->SetDialogObscured () ;    // save the dialog display

                     short status = 
                        this->cmMountFilesystem ( fsiPtr[seSelect], dp, superPriv ) ;

                     //* If a gvfs mount has failed, perform an automatic  *
                     //* retry. (See notes in cmMountFilesystem().)        *
                     if ( (status == -1) && fsiPtr[seSelect].fsMtp )
                     {
                        //* Signal the retry *
                        dp->WriteString ( wpm, " -RETRY", this->cfgOpt.cScheme.em, true ) ;

                        //* Re-scan the target device for current *
                        //* device path: /dev/bus/usb/00n/00?     *
                        gString gstmp ;
                        if ( (this->fdPtr->USB_DeviceID ( 
                                 fsiPtr[seSelect].fsMountpt, gstmp )) )
                        {
                           gstmp.copy( fsiPtr[seSelect].fsDevice, FS_BYTES ) ;
                           status = 
                              this->cmMountFilesystem ( fsiPtr[seSelect], 
                                                        dp, superPriv ) ;
                        }
                     }

                     if ( status == OK )           // successfully mounted
                     {
                        dp->WriteString ( wpm, " -OK   ", 
                                          this->cfgOpt.cScheme.pf, true ) ;
                        dp->GetRadiobuttonState ( cwdRB, setCWD ) ;
                        newMNT = true ;
                     }
                     else if ( status == (-1) )    // key input - user gave up waiting
                     {
                        dp->WriteString ( wpm, " -ERROR", 
                                          this->cfgOpt.cScheme.pf, true ) ;
                     }
                     else if ( status == (-2) )    // specified target not a directory
                     {
                        dp->WriteString ( wp, 
                           " Specified Mount Point Is Not a Directory! ", 
                           this->cfgOpt.cScheme.wr, true ) ;
                     }
                  }

                  //************************************
                  //* If target is mounted, unmount it.*
                  //************************************
                  else  // (fsiPtr[seSelect].fsMounted != false)
                  {
                     //* If application's CWD is on the target device, change  *
                     //* CWD before attempting to dismount. Otherwise, dismount*
                     //* will fail because the device is "busy". The logical   *
                     //* target is the user's 'home' directory, e.g. /home/sam.*
                     //* Exception: If '/home' is the target mountpoint,       *
                     //* (unlikely), we set CWD to a universally-available     *
                     //* directory '/' (root).                                 *
                     //* If in dual-window mode, check inactive window also.   *
                     gString currPath,             // CWD
                             newPath( "%s", std::getenv( "HOME" ) ) ;
                     if ( (newPath.find( trgMnt, ZERO, true, 
                              (trgMnt.gschars() - 1) )) == ZERO )
                        newPath = "/" ;
                     gString cmpPath( trgMnt ) ;   // comparison path
                     if ( (trgMnt.find( mtpURI )) == ZERO )
                        this->fdPtr->Uri2Filespec ( cmpPath, trgMnt.ustr() ) ;

                     if ( (this->DualwinMode ()) )
                     {
                        this->altPtr->GetPath ( currPath ) ;
                        if ( (currPath.find( cmpPath, ZERO, true, 
                               (cmpPath.gschars() - 1))) == ZERO )
                        {
                           dp->SetDialogObscured () ;  // save our dialog
                           this->dWin->RefreshWin () ; // restore parent dialog
                           this->altPtr->SetDirectory ( newPath ) ; // scan the new CWD
                           this->altPtr->RedrawCurrDir ( false ) ;  // remove highlight
                           this->dWin->SetDialogObscured () ; // save parent dialog
                           dp->RefreshWin () ;         // restore our dialog
                        }
                     }
                     this->fdPtr->GetPath ( currPath ) ;
                     if ( (currPath.find( cmpPath, ZERO, true, 
                            (cmpPath.gschars() - 1))) == ZERO )
                     {
                        dp->SetDialogObscured () ;  // save our dialog
                        this->dWin->RefreshWin () ; // restore parent dialog
                        this->fdPtr->SetDirectory ( newPath ) ; // scan the new CWD
                        this->dWin->SetDialogObscured () ; // save parent dialog
                        dp->RefreshWin () ;         // restore our dialog
                     }

                     //* Give user a clue *
                     wpm = dp->WriteString ( wpm, " UNMOUNTING : ", 
                                            this->cfgOpt.cScheme.tf ) ;
                     wpm = dp->WriteString ( wpm, fgs[seSelect], 
                                            this->cfgOpt.cScheme.tf, true ) ;

                     dp->SetDialogObscured () ;    // save the dialog display

                     short status = this->cmDismountFilesystem ( trgMnt, dp, superPriv ) ;

                     if ( status == OK )           // successfully unmounted
                     {
                        dp->WriteString ( wpm, " -OK   ", 
                                          this->cfgOpt.cScheme.pf, true ) ;
                     }
                     else if ( status == (-1) )    // key input - user gave up waiting
                     {
                        dp->WriteString ( wpm, " -ERROR", 
                                          this->cfgOpt.cScheme.pf, true ) ;
                     }
                  }  // (fsiPtr[seSelect].fsMounted != false)

                  //* Give user time to read message *
                  this_thread::sleep_for( aMoment ) ;
               }  // (Info.ctrlIndex == mntPB)

               done = true ;
            }     // dataMod
         }        // (ic[icIndex].type == dctPUSHBUTTON)

         //* If focus is currently on a Radio Button *
         else if ( ic[icIndex].type == dctRADIOBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditRadiobutton ( Info ) ;
         }

         if ( done == false && Info.viaHotkey == false )
         {
            if ( Info.keyIn == nckSTAB )
               icIndex = dp->PrevControl () ; 
            else
               icIndex = dp->NextControl () ;
         }
      }  // while(!done)
   }
   else           // unable to open window (likely memory-allocation error)
      ;

   if ( dp != NULL )
      delete ( dp ) ;                  // close the window
   cmcbPtr = NULL ;                    // reset callback pointers
   cmcbFsi = NULL ;

   //* Restore application's display data *
   this->dWin->RefreshWin () ;

   //* If the filesystem was mounted and it is accessible, go there *
   if ( newMNT && setCWD )
   {
      //* For virtual filesystems, convert the URI to the full filespec.*
      bool mtpFilesys = false ;
      if ( (mtpFilesys = ((trgMnt.find( mtpURI )) == ZERO)) )
         this->fdPtr->Uri2Filespec ( trgMnt, fsiPtr[seSelect].fsMountpt ) ;

      //* Get the size of the target filesystem.                       *
      //* For very large storage devices, read only the top-level data.*
      //* Else, assume that target is small enough to be read quickly. *
      bool recurse = true ;
      fileSystemStats fsStats ;
      this->fdPtr->GetFilesystemStats ( trgMnt, fsStats ) ;
      if ( (fsStats.blockTotal * fsStats.blockSize) >= HUGE_EXT_DRIVE )
         recurse = false ;

      if ( (this->fdPtr->SetDirectory ( trgMnt, recurse )) != OK )
      {  //* Unable to move to new target directory *
         this->dWin->SetDialogObscured () ;
         const char* dirVanished[] = 
         {
            "  ALERT  ",
            "File system was mounted successfully, but it is not",
            "currently accessible. Please verify that you have",
            "read-access for the target.",
            NULL
         } ;
// UNDER CONSTRUCTION - FOR GVfs FILESYSTEMS, REVISIT THE MESSAGE
         const char* badHandshake[] = 
         {
            "  ALERT  ",
            "      Handshake with smartphone device failed.",
            "Check the connection, verify that screen is unlocked",
            "and retry the mount operation.",
            NULL
         } ;
         this->fdPtr->InfoDialog ( mtpFilesys ? badHandshake : dirVanished ) ;
         this->dWin->RefreshWin () ;
      }
   }
   delete[]fsiPtr ;        // release dynamically-allocated data
   delete[] fgs ;

}  //* End CmdMount() *

//*************************
//*      cmCallback       *
//*************************
//******************************************************************************
//* NON-MEMBER METHOD.                                                         *
//* This is a call-back method for CmdMount.                                   *
//* It is called from within the NcDialog class during the user-input loop.    *
//*                                                                            *
//* Input  : currIndex: index of control that currently has focus              *
//*          wkey     : user's key input data                                  *
//*          firstTime: the EstablishCallback() method calls this method once  *
//*                     with firstTime==true, to perform any required          *
//*                     initialization. Subsequently, the NcDialog class       *
//*                     always calls with firstTime==false.                    *
//* Returns: OK                                                                *
//******************************************************************************
//* Programmer's Note: This callback makes some intimate assumptions about     *
//* the target dialog. It is important to keep the reference data synchronized.*
//*                                                                            *
//******************************************************************************

static short cmCallback ( const short currIndex, const wkeyCode wkey, bool firstTime )
{
   static short mntSE = ZERO,       // index of Scrollext control
                mntPB = 1 ;         // index of "Mount" Pushbutton control
   static bool prevMounted = false ;// 'true' if previous call indexed a mounted item
                                    // (initial text is "MOUNT")

   //* If the Scrollext control currently has focus.*
   if ( currIndex == mntSE )
   {
      const char *mntLabel  = "  MOUNT  ",
                 *umntLabel = " UNMOUNT " ;
      short mntseIndex = cmcbPtr->GetScrollextSelect ( mntSE ) ;
      bool  isMounted = cmcbFsi[mntseIndex].fsMounted ;

      if ( isMounted != prevMounted )
      {
         cmcbPtr->SetPushbuttonText ( mntPB, (isMounted ? umntLabel : mntLabel) ) ;
         cmcbPtr->RefreshWin () ;
         prevMounted = isMounted ;
      }
   }

   return OK ;

}  //* End cmCallback() *

//*************************
//*   cmMountFilesystem   *
//*************************
//******************************************************************************
//* Called only by CmdMount(). Mount the specified filesystem.                 *
//*  1) Caller has verified that the filesystem is not mounted at the          *
//*     specified position. The target directory may, or may not exist and     *
//*     may, or may not be a directory file.                                   *
//*  2) The system commands are executed from a child terminal window.         *
//*     This window is defined and launched using the DC_Emulator class.       *
//*     The child window will close immediately after the mount command has    *
//*     been executed.                                                         *
//*  3) Some mounts require super-user privelege, but we have no idea which    *
//*     ones do, so the mount command includes the 'sudo' security which       *
//*     prompts for the user password.                                         *
//*  4) Mounting a filesystem such as a USB hard drive can take several        *
//*     seconds, so we wait for the target device to become available before   *
//*     returning to caller.                                                   *
//*  5) Timing issue with virtual (MTP/GVfs) filesystems:                      *
//*     When a phone is first attached to the USB port, there is a delay       *
//*     during which a handshake is attempted between the host and the target  *
//*     device. The device will prompt the user to acknowledge whether the     *
//*     host is allowed to access device storage.                              *
//*     If the handshake fails (or is denied), then the mountpoint directory   *
//*     may appear for a moment and then disappear. When this happens, caller  *
//*     must alert the user that the handshake has failed and that he/she      *
//*     should attempt the mount again.                                        *
//*     An interesting side issue is that when this happens, the device number *
//*     assigned to the device may be incremented.                             *
//*           /dev/bus/usb/003/006  BECOMES  /dev/bus/usb/003/007              *
//*     While the change in device number may be transparent to the user, it   *
//*     is quite annoying to those of us who care about the underlying process.*
//*     See the auto-retry for this situation in CmdMount() method.            *
//*  6) There is always the possibility that the mount will fail, either       *
//*     because the device is not physically connected, or because the 'fstab' *
//*     file does not fully specify the device (or the user is just having a   *
//*     bad day. For this reason, we return a very simple set of codes to the  *
//*     caller indicating the status of the operation.                         *
//*                                                                            *
//* Input  : fsi      : information about filesystem to be mounted             *
//*          subDlg   : handle to sub-dialog window                            *
//*                     (we must refresh after launch)                         *
//*          super    : 'true' operation is executed with superuser privilege  *
//*                     'false' operation is executed with user privilege      *
//*                                                                            *
//* Returns: OK     if filesystem successfully mounted                         *
//*          -1     if user key input received                                 *
//*                   (probably mount failed and user got tired of waiting)    *
//*          -2     if target mountpoint exists but is not a directory         *
//*                                                                            *
//******************************************************************************

short FileMangler::cmMountFilesystem ( const fsInfo& fsi, 
                                       NcDialog* subDlg, bool super )
{
   gString newCwd( fsi.fsMountpt ), // mountpoint path (or mountpoint directory)
           fPath,
           argList,
           tmpFile ;
   fileSystemStats fsStats ;     // determine whether the target is mounted
   fmFType fType ;               // determine whether the target exists
   short status = OK ;           // return value (see above)
   bool createMp = false ;       // 'true' if mountpoint dir does not exist
   bool mtpFilesys = bool(((newCwd.find( mtpURI )) == ZERO) ? true : false) ;

   //* Determine whether target exists AND if so, whether it is a directory.  *
   //* For virtual filesystems, the parent directory is read-only for the     *
   //* user, but mountpoint directory is automagically created.               *
   if ( ! mtpFilesys )
   {
      if ( !(this->fdPtr->TargetExists ( newCwd, fType )) )
         createMp = true ;
      else if ( fType != fmDIR_TYPE )
         status = (-2) ;
   }

   //* If target is a virtual (mtp/gvfs) filesystem *
   //* to be mounted using the 'gio' utility.       *
   if ( mtpFilesys )
   {
      //* Convert the activation_root" to a mountpoint filespec.*
      this->fdPtr->Uri2Filespec ( newCwd, fsi.fsMountpt ) ;

      if ( super != false )   // superuser privilege
      {
         fPath = "sudo" ;
         argList.compose( "gio mount --device='%s' \"%s\"", 
                          fsi.fsDevice, fsi.fsMountpt ) ;
      }
      else                    // user privilege
      {
         fPath = "gio" ;
         argList.compose( "mount --device='%s' \"%s\"", 
                          fsi.fsDevice, fsi.fsMountpt ) ;
      }
   }

   //* Target is a "real" (physical) filesystem *
   //* to be mounted by the OS kernel.          *
   else
   {
      if ( super != false )   // superuser privilege
      {
         fPath = "sudo" ;
         argList.compose( "mount --target '%S'", newCwd.gstr() ) ;
      }
      else                    // user privilege
      {
         fPath = "mount" ;
         argList.compose( "--target '%S'", newCwd.gstr() ) ;
      }
   
      if ( createMp )
      {
         gString gs( "mkdir '%S' ; ", newCwd.gstr() ) ;
         argList.insert( gs.gstr() ) ;
      }
   }

   if ( status == OK )
   {
      this->CreateTempname ( tmpFile ) ;  // create a temp file
      DC_Emulator dce( tmpFile, fPath, argList, true ) ; // console window definition
      wkeyCode wk ;           // user key input
      //* Max iterations of the while() loop. Time elapsed: *
      //*   four(4) seconds for "real" filesystems,         *
      //*   eight(8) seconds for virtual filesystems        *
      short deadman = 4 ;
      bool done = false ;     // loop control

      //* Scan the system configuration to define *
      //* the new terminal window.                *
      //* Specify window size (small is cute here)*
      dce.DefineChildEmulator ( 56, 8 ) ;

      //* Open the new terminal window *
      dce.OpenTermwin () ;

      //* Wait a moment in case child process writes to the parent window. *
      chrono::duration<short, std::milli>aMoment( 1000 ) ;
      this_thread::sleep_for( aMoment ) ;
      this->dWin->ClearTerminalWindow () ;
      this->dWin->RefreshWin () ;         // restore parent dialog
      this->dWin->SetDialogObscured () ;  // re-save parent dialog
      subDlg->RefreshWin () ;             // restore the sub-dialog

      //* Monitor the status of the target device until *
      //* the status changes OR until key input.        *
      done = false ;
      while ( ! done )
      {
         if ( (this->dWin->KeyPeek ( wk )) != ERR )
         {
            this->dWin->GetKeyInput ( wk ) ; // discard the keystroke
            status = (-1) ;      // mount _probably_ failed, user has given up
            done = true ;
         }
         else
         {
            for ( short i = (mtpFilesys ? 2 : 1) ; i > ZERO ; --i )
               this_thread::sleep_for( aMoment ) ;
            this->fdPtr->GetFilesystemStats ( newCwd, fsStats ) ;

            if ( fsStats.isMounted )
            {
               status = OK ;     // mount was successful, alert caller to set CWD
               done = true ;
            }
            else if ( --deadman <= ZERO )
            {
               status = (-1) ;
               done = true ;
            }
         }
      }
      this->fdPtr->DeleteFile ( tmpFile ) ; // delete the temp file
   }

   return status ;

}  //* End cmMountFilesystem() *

//*************************
//* cmDismountFilesystem  *
//*************************
//******************************************************************************
//* Called only by CmdMount(). Un-mount the specified filesystem.              *
//*  1) Caller has verified that the target directory is a mountpoint AND that *
//*     a storage device is mounted at that mountpoint.                        *
//*  2) Because a device should not be unmounted when someone is using it,     *
//*     caller has tested whether the application's CWD was on the target      *
//*     device. If it was, a new CWD has been set to ensure that there is no   *
//*     conflict.                                                              *
//*     Another application or process may be using data on the target device, *
//*     but we can't determine that ahead of time. For this reason, we use     *
//*     the '--lazy' option which performs the unmount immediately, but waits  *
//*     to do the /etc/mtab cleanup until the filesystem is no longer busy.    *
//*                                                                            *
//* Input  : trgMnt   : filespec of mountpoint (for "real" filesystems), or    *
//*                     the URI (for virtual MTP/BVFS filesystems)             *
//*          subDlg   : handle to sub-dialog window                            *
//*                     (we must refresh after launch)                         *
//*          super    : 'true' operation is executed with superuser privilege  *
//*                     'false' operation is executed with user privilege      *
//*                                                                            *
//* Returns: OK     if filesystem successfully unmounted                       *
//*          -1     if user key input received                                 *
//*                   (probably unmount failed and user got tired of waiting)  *
//******************************************************************************

short FileMangler::cmDismountFilesystem ( const gString& trgMnt, 
                                          NcDialog* subDlg, bool super )
{
   fileSystemStats fsStats ;     // determine whether the target is mounted
   short status = (-1) ;         // return value (see above)

   gString fPath,
           argList,
           tmpFile ;

   //* If a virtual filesystem URI specified *
   if ( (trgMnt.find( mtpURI )) == ZERO )
   {
      if ( super != false )   // superuser privilege
      {
         fPath = "sudo" ;
         argList.compose( "gio mount --unmount '%S'", trgMnt.gstr() ) ;
      }
      else                    // user privilege
      {
         fPath = "gio" ;
         argList.compose ( "mount --unmount '%S'", trgMnt.gstr() ) ;
      }
   }

   //* Else assume target is a "real" filesystem *
   else
   {
      if ( super != false )   // superuser privilege
      {
         fPath = "sudo" ;
         argList.compose( "umount --lazy '%S'", trgMnt.gstr() ) ;
      }
      else                    // user privilege
      {
         fPath = "umount" ;
         argList.compose ( "--lazy '%S'", trgMnt.gstr() ) ;
      }
   }

   this->CreateTempname ( tmpFile ) ;  // Create a temp file to capture sys info
   DC_Emulator dce( tmpFile, fPath, argList, true ) ;

   //* Scan the system configuration to define *
   //* the new terminal window.                *
   //* Specify window size (small is cute here)*
   dce.DefineChildEmulator ( 56, 8 ) ;
   this->fdPtr->DeleteFile ( tmpFile ) ; // delete the temp file

   //* Open the new terminal window *
   dce.OpenTermwin () ;

   //* Wait a moment in case child process writes to the parent window.*
   // Programmer's Note: We do not like calling the NCurses primitive here, 
   // but the just-launched terminal emulator may completely trash the display.
   chrono::duration<short, std::milli>aMoment( 1000 ) ;
   this_thread::sleep_for( aMoment ) ;
   nc.ClearScreen () ;
   this->dWin->RefreshWin () ;         // restore parent dialog
   this->dWin->SetDialogObscured () ;  // re-save parent dialog
   subDlg->RefreshWin () ;             // restore the sub-dialog

   //* Monitor the status of the target device until *
   //* the status changes OR until key input.        *
   wkeyCode wk ;
   bool done = false ;
   while ( ! done )
   {
      this_thread::sleep_for( aMoment ) ;
      if ( (this->dWin->KeyPeek ( wk )) != ERR )
      {
         this->dWin->GetKeyInput ( wk ) ; // discard the keystroke
         status = (-1) ;      // unmount _probably_ failed, user has given up
         done = true ;
      }
      else
      {
         this->fdPtr->GetFilesystemStats ( trgMnt, fsStats ) ;
         if ( ! fsStats.isMounted )
         {
            status = OK ;     // unmount was successful
            done = true ;
         }
      }
   }

   return status ;

}  //* End cmDismountFilesystem() *

//*************************
//*   Scan4Filesystems    *
//*************************
//******************************************************************************
//* Called only by CmdMount().                                                 *
//*                                                                            *
//* Scan the system for "real" (physical storage) filesystems.                 *
//* 1) Scan the system for mounted filesystems. (/proc/self/mountinfo)         *
//* 2) Scan 'fstab' for defined mountpoints.    (/etc/fstab)                   *
//* 3) Add the mountpoints from config file, then eliminate duplicate entries. *
//* 4) Stat each entry to determine whether it is actually mounted.            *
//*                                                                            *
//* Scan the system for mtp/gvfs virtual filesystems.                          *
//* 1) Scan all USB devices and select "mtp" devices.                          *
//* 2) Gather information on each of the selected devices.                     *
//*                                                                            *
//* Input  : fsCount : (by reference, initial value ignored)                   *
//*                                                                            *
//* Returns: pointer to dynamically-allocated array of fsInfo objects          *
//*          It is caller's responsibility to free the dynamic allocation.     *
//******************************************************************************
//* Programmer's Notes:                                                        *
//* -- "Real" filesystems are those which are physical (or logical) devices.   *
//*    We must explicitly filter out the logical devices such as 'tmpfs',      *
//*    'cgroup', 'selinuxfs' etc, as well as the '/' (root) directory which    *
//*    cannot be mounted or unmounted anyway.                                  *
//*                                                                            *
//* -- There is a bug in 'findmnt' which ignores the "--real" option when the  *
//*    "--list" formatting option is used.                                     *
//* -- Note however that the "--df" option which _implies_ both the "--list"   *
//*    and "--real" options actually does filter out the pseudo-filesystems.   *
//* -- In the first pass at this functionality, we pre-allocate a fixed number *
//*    of 'fsInfo' objects so we don't have to scan the raw data twice. If     *
//*    this becomes a problem, we can count the records before allocating the  *
//*    'fsInfo' array.                                                         *
//* ---  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  --- *
//*                                                                            *
//* Possible Enhancements:                                                     *
//* ----------------------                                                     *
//* -- To report attached, but unmounted filesystems we could use:             *
//*       sudo blkid -o list                                                   *
//*    However, we can't run a root command from inside the application unless *
//*    we shell out or run it in another window. In user mode:                 *
//*       blkid -o list                                                        *
//*    we get the information, but it is mixed with some garbage such as       *
//*    unallocated clusters on a hard drive, etc.                              *
//*    If the mountpoint column says "(not mounted" AND if we have at least    *
//*    _some_ of the data for fstype, label, UUID, then it's probably good:    *
//*         DEVICE     FSTYPE  LABEL     MOUNTPOINT     UUID                   *
//*         /dev/sdb1  vfat    KINGSTON  (not mounted)  6C55-431E              *
//*    To identify the target mountpoint, we would need some additional magic. *
//*    Note that the documentation says that this option is deprecated in      *
//*    favor of 'lsblk'.                                                       *
//* -- lsblk --fs   could be used, but the information would be difficult to   *
//*    parse.                                                                  *
//*       lsblk --fs --list -o MOUNTPOINT,LABEL,UUID,FSTYPE,NAME               *
//*    Using the above options, we would get an entry for the attached-but-    *
//*    not-mounted filesystem.                                                 *
//*    MOUNTPOINT     LABEL       UUID        FSTYPE   NAME                    *
//*    "              KINGSTON    6C55-431E   vfat     sdb1"                   *
//*    The mountpoint should be listed first, but because there is no          *
//*    mountpoint, the first field is blank. The logic would then be that      *
//*    if no mountpoint, but we DO have a filesystem type and a label and/or   *
//*    a UUID, then it is probably an attached-but-not-mounted filesystem.     *
//*    We would need a label in order to create a mountpoint.                  *
//*    Add the '--noheadings' option to simplify things.                       *
//*    Of course, this is all useless because we would have to specify a       *
//*    mountpoint, presumably: /run/media/$USER/LABEL                          *
//*    This is possible, but tricky.                                           *
//* ---  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  --- *
//*                                                                            *
//******************************************************************************

fsInfo* FileMangler::cmScan4Filesystems ( short& fsCount )
{
   #define MAX_RECORDS (64)      // (see note above)
   #define VERIFY_MOUNTS (0)     // optionally verify mounts (see note below)

   fsInfo* fsiPtr = new fsInfo[MAX_RECORDS] ; // array of mountpoint records
   fmFType fType ;               // determine whether the target exists
   fileSystemStats fsStats ;     // determine whether the target is mounted
   gString tmpPath,              // filespec of temporary file
           cmdBuff,              // command string buffer
           gs ;                  // text formatting
   char  lineBuff[gsMAXBYTES] ;  // input buffer
   short fsiIndex = ZERO,        // index into fsiPtr array
         srcIndx,                // source text index
         trgIndx ;               // target text index

         fsCount = ZERO ;        // initialize caller's variable
   bool  done = false ;          // loop control

   //*******************************************************************
   //* Place user's mountpoints from configuration file at top of list.*
   //*******************************************************************
   for ( short i = ZERO ; i < this->mntcmdCount ; ++i )
   {
      gs = this->MountCmds[i] ;
      gs.copy( fsiPtr[fsiIndex].fsMountpt, gsMAXBYTES ) ;
      if ( (this->fdPtr->TargetExists ( fsiPtr[fsiIndex].fsMountpt, fType )) 
            && fType == fmDIR_TYPE )
      {  // Programmer's Note: The presence of the mountpoint directory is not 
         // a guarantee that the drive is actually mounted, so we dig deeper.
         if ( ((this->fdPtr->GetFilesystemStats ( gs, fsStats )) == OK)
              && fsStats.isMounted && fsStats.isMountpoint )
         {
            gs = fsStats.device ;
            gs.copy( fsiPtr[fsiIndex].fsDevice, gsMAXBYTES ) ;
            gs = fsStats.fsType ;
            gs.copy( fsiPtr[fsiIndex].fsType, gsMAXBYTES ) ;
            fsiPtr[fsiIndex].fsMounted = true ;
         }
      }
      ++fsiIndex ;   // increment the record count
   }

   //* Create a temporary file to capture the raw data.*
   this->CreateTempname ( tmpPath ) ;

   //**************************************************************
   //* Scan the /proc/self/mountinfo file for currently-mounted   *
   //* physical storage devices, (SATA, SSD, USB, SD cards, etc.).*
   //**************************************************************
   const char* fsTemplate = 
   "findmnt --df --noheadings -o SOURCE,FSTYPE,TARGET "
   "| grep -v '^tmpfs' "            // (exclude "tmpfs" filesystems)
   "| grep -v '[[:space:]]/sys' "   // (exculde filesystems mounted under '/sys')
   "| grep -v '[[:space:]]/dev' "   // (exculde filesystems mounted under '/dev')
   "| grep -v '[[:space:]]/var/' "  // (exculde filesystems mounted under '/var') 
   "| grep -v '[[:space:]]/$' "     // (exclude root directory)                   
   "1>>\"%s\" 2>>\"%s\"" ;             
   cmdBuff.compose( fsTemplate, tmpPath.ustr(), tmpPath.ustr() ) ;
   this->Systemcall ( cmdBuff.ustr() ) ;

   //* Open the temp file and extract the interesting data.*
   ifstream ifs( tmpPath.ustr(), ifstream::in ) ; // open the file
   if ( ifs.is_open() )                           // if file opened
   {
      while ( ! done && (fsiIndex < MAX_RECORDS) )
      {
         ifs.getline ( lineBuff, gsMAXBYTES, NEWLINE ) ; // read a source line
         if ( ifs.good() || ifs.gcount() > ZERO )        // if a good read
         {
            //* Extract device filespec.*
            srcIndx = trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsDevice[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsDevice[trgIndx] = NULLCHAR ;
            while ( (srcIndx < gsMAXBYTES) && (lineBuff[srcIndx] == SPACE) )
               ++srcIndx ;
            //* Extract device type.*
            trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsType[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsType[trgIndx] = NULLCHAR ;
            while ( (srcIndx < gsMAXBYTES) && (lineBuff[srcIndx] == SPACE) )
               ++srcIndx ;
            //* Extract mountpoint filespec.*
            trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsMountpt[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsMountpt[trgIndx] = NULLCHAR ;

            // Programmer's Note: For real (block) devices, we can assume that 
            // all listed filesystems are currently mounted because we got them 
            // from the system's list of mounted filesystems; however, we can 
            // optionally verify that they are mounted.
            #if VERIFY_MOUNTS // OPTIONAL MOUNT VERIFICATION
            if ( (this->fdPtr->TargetExists ( fsiPtr[fsiIndex].fsMountpt, fType )) 
                  && fType == fmDIR_TYPE )
            {
               if ( ((this->fdPtr->GetFilesystemStats ( 
                       fsiPtr[fsiIndex].fsMountpt, fsStats )) == OK)
                    && fsStats.isMounted && fsStats.isMountpoint )
               {
                  fsiPtr[fsiIndex].fsMounted = true ;
               }
            }
            #else    // Assume that filesystem is mounted
            fsiPtr[fsiIndex].fsMounted = true ;
            #endif   // VERIFY_MOUNTS

            ++fsiIndex ;   // increment the record count
         }
         else           // end-of-file found
            done = true ;
      }
      ifs.close() ;           // close the file
   }

   //********************************************************
   //* Scan the /etc/fstab file for mountpoint definitions, *
   //* (may, or may not be currently mounted).              *
   //********************************************************
   //* Discard current contents of temp file.*
   ofstream ofs( tmpPath.ustr(), ofstream::out | ofstream::trunc ) ;
   if ( ofs.is_open() )
      ofs.close() ;

   const char* const fstabTemplate = 
         "findmnt --fstab --noheadings -o SOURCE,FSTYPE,TARGET "
         "| grep -v 'none$' "             // (exclude swap partition)
         "| grep -v '[[:space:]]/$' "     // (exclude root directory)                   
         "1>>\"%s\" 2>>\"%s\"" ;
   cmdBuff.compose( fstabTemplate, tmpPath.ustr(), tmpPath.ustr() ) ;
   this->Systemcall ( cmdBuff.ustr() ) ;
   //* Open the temp file and extract the interesting data.*
   ifs.open( tmpPath.ustr(), ifstream::in ) ; // open the file
   if ( ifs.is_open() )                           // if file opened
   {
      done = false ;

      while ( ! done && (fsiIndex < MAX_RECORDS) )
      {
         ifs.getline ( lineBuff, gsMAXBYTES, NEWLINE ) ; // read a source line
         if ( ifs.good() || ifs.gcount() > ZERO )        // if a good read
         {
            //* Extract device filespec. (may be a UUID or other identifier) *
            srcIndx = trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsDevice[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsDevice[trgIndx] = NULLCHAR ;
            while ( (srcIndx < gsMAXBYTES) && (lineBuff[srcIndx] == SPACE) )
               ++srcIndx ;
            //* Extract device type.*
            trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsType[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsType[trgIndx] = NULLCHAR ;
            while ( (srcIndx < gsMAXBYTES) && (lineBuff[srcIndx] == SPACE) )
               ++srcIndx ;
            //* Extract mountpoint filespec.*
            trgIndx = ZERO ;
            while ( (srcIndx < gsMAXBYTES) && (trgIndx < gsMAXBYTES) && 
                    (lineBuff[srcIndx] != SPACE) && (lineBuff[srcIndx] != NULLCHAR) )
               fsiPtr[fsiIndex].fsMountpt[trgIndx++] = lineBuff[srcIndx++] ;
            fsiPtr[fsiIndex].fsMountpt[trgIndx] = NULLCHAR ;

            //* Determine whether target filesystem is currently mounted.*
            if ( (this->fdPtr->TargetExists ( fsiPtr[fsiIndex].fsMountpt, fType )) 
                  && fType == fmDIR_TYPE )
            {
               if ( ((this->fdPtr->GetFilesystemStats ( 
                       fsiPtr[fsiIndex].fsMountpt, fsStats )) == OK)
                    && fsStats.isMounted && fsStats.isMountpoint )
               {
                  fsiPtr[fsiIndex].fsMounted = true ;
               }
            }
            ++fsiIndex ;   // increment the record count
         }
         else           // end-of-file found
            done = true ;
      }
      ifs.close() ;           // close the file
   }

   //********************************************************
   //* Scan for mtp/gvfs filesystems attached via USB.      *
   //* (may, or may not be currently mounted).              *
   //********************************************************
   short usbCount = ZERO ;
   usbDevice *usbdPtr = this->fdPtr->USB_DeviceStats ( usbCount, true ) ;
   for ( short usbIndex = ZERO ; usbIndex < usbCount ; ++usbIndex )
   {
      gs = usbdPtr[usbIndex].devpath ;
      gs.copy( fsiPtr[fsiIndex].fsDevice, FS_BYTES ) ;
      gs = usbdPtr[usbIndex].uri ;
      gs.copy( fsiPtr[fsiIndex].fsMountpt, FS_BYTES ) ;
      fsiPtr[fsiIndex].fsMounted = usbdPtr[usbIndex].mounted ;
      gs = usbdPtr[usbIndex].fstype ;
      gs.copy( fsiPtr[fsiIndex].fsType, FS_BYTES ) ;
      fsiPtr[fsiIndex].fsMtp = true ;     // indicates MTP/GVfs filesystem

      ++fsiIndex ;    // filesystem record initialized
   }
   if ( usbdPtr != NULL )     // release the dynamic memory allocation
      delete [] usbdPtr ;

   //* Eliminate duplicate entries.*
   fsCount = fsiIndex ;       // initialize the return value
   for ( fsiIndex = (fsCount - 1) ; fsiIndex > ZERO ; --fsiIndex )
   {
      gs = fsiPtr[fsiIndex].fsMountpt ;
      for ( short indx = ZERO ; indx < fsiIndex ; ++indx )
      {
         //* If entries are identical, shift all following records upward.*
         if ( (gs.compare( fsiPtr[indx].fsMountpt )) == ZERO )
         {
            trgIndx = fsiIndex ;
            srcIndx = trgIndx + 1 ;
            while ( srcIndx < fsCount )
               fsiPtr[trgIndx++] = fsiPtr[srcIndx++] ;
            --fsCount ;
            break ;
         }
      }
   }

   this->fdPtr->DeleteFile ( tmpPath ) ;  // delete the temporary file

   return fsiPtr ;      // return pointer to dynamic allocation

}  //* End cmScan4Filesystems() *

//*************************
//*    CmdLocateFile      *
//*************************
//******************************************************************************
//* Given the first character of the filename, scan the list of displayed      *
//* filenames, and if a match is found, move highlight to that file.           *
//*   (Not to be confused with CmdFind() which searches the directory tree.)   *
//*                                                                            *
//* Input  : firstChar: user-provided first character of filename              *
//*                     (target may, or may not exist)                         *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Concepts:                                                                  *
//* ---------                                                                  *
//* A) Be helpful, but non-obtrusive.                                          *
//*    -- If not a first-character match, return immediately.                  *
//*    -- Do not open the interactive dialog unless it is actually needed.     *
//* B) Leave non-search characters in the queue for caller.                    *
//*    -- Remove a character from the input queue only if we can use it.       *
//*    -- If input is not ours, search is terminated and caller will process   *
//*       the input.                                                           *
//*                                                                            *
//* Logic:                                                                     *
//* 1) If caller's character is a printing character i.e. a potential filename *
//*    character, continue to step 2.                                          *
//* 2) Test caller's character against the first character of each filename    *
//*    in the list.                                                            *
//*    a) If a match is found, move the highlight to the matching file and     *
//*       go to step 3.                                                        *
//*    b) If no match, then RETURN with an audible boink but without a screen  *
//*       update. (To the user, this will look like invalid input.)            *
//* 3) If a first-character match is found, wait for the next character to     *
//*    arrive in the input queue. Peek at the character.                       *
//*    a) If it is a printing character, extract the character from the queue. *
//*       i) If there is a two-character filename match found, move the        *
//*          highlight to the matching file and go to step 4.                  *
//*       ii) If not a two-character match, boink the user and RETURN to       *
//*           caller.                                                          *
//*    b) If not a printing character, leave it in the queue and RETURN to     *
//*       caller.                                                              *
//* 4) A two-character match has been found and highlight is on matching file. *
//*    a) Open the interactive dialog and display the first two characters of  *
//*       user input.                                                          *
//* 5) Continue to gather input:                                               *
//*    a) printing characters are appended to the search string                *
//*    b) DELETE/BACKSPACE remove the most-recent character from the string.   *
//*    c) ESCAPE aborts the search. Highlight is left at its most-recent       *
//*       position                                                             *
//*    d) All other input is left in the queue, search is terminated, and      *
//*       highlight is left at its most-recent position.                       *
//* 6) When search is complete:                                                *
//*    a) the interactive dialog is closed                                     *
//*    b) highlight is left on most recent matching filename                   *
//*    c) caller may or may not have key input waiting in the queue.           *
//*                                                                            *
//* -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -   -  *
//* Undocumented Feature:                                                      *
//* This method is documented as a case-insensitive search; however, once      *
//* the dialog is open, the INSERT key toggles case-sensitivity for string     *
//* comparisons. This feature is not user friendly: If we have made a          *
//* case-insensitive match, and then toggle to case-sensitive, the previous    *
//* match may no longer be valid. Besides, we couldn't document this           *
//* functionality without confusing the crap out of the user.                  *
//*                                                                            *
//******************************************************************************

void FileMangler::CmdLocateFile ( wkeyCode& firstChar )
{
   if ( (firstChar.type == wktPRINT) && iswprint ( firstChar.key ) )
   {
      short currCtrl  = this->fdPtr == this->fWin ? dwFileSE : dwFileSEr,
            pathCtrl  = currCtrl == dwFileSE ? dwPathTB : dwPathTBr,
            currIndex ;
      bool  matchFound = false ;

      //* Scroll to the next file which begins with the specified character *
      gString substr ;
      substr.compose( L"%C", &firstChar ) ;
      matchFound = this->fdPtr->Scroll2MatchingFile ( substr, currIndex ) ;

      //* If a first-character match was found, wait for the next character.*
      if ( matchFound )
      {
         const short INTERVAL = 250 ;     // key wait interval (mS)
         chrono::duration<short, std::milli>aMoment( INTERVAL ) ;
         wkeyCode wk ;              // key input
         matchFound = false ;       // reset the match flag

         while ( true )
         {
            if ( (this->dWin->KeyPeek ( wk )) != wktERR )
            {
               if ( wk.type == wktPRINT )
               {
                  this->dWin->GetKeyInput ( wk ) ;
                  substr.append( wk.key ) ;
                  matchFound = this->fdPtr->Scroll2MatchingFile ( substr, currIndex ) ;
                  if ( ! matchFound )
                  {
                     //* Handle SPACE character as a command key.*
                     if ( wk.key == nckSPACE )
                        this->dWin->UngetKeyInput ( wk ) ;
                     else
                        this->dWin->UserAlert () ;
                  }
               }
               else
                  ;     // Data is not for us. Leave it for caller.
               break ;
            }
            else
               this_thread::sleep_for( aMoment ) ;    // take a nap
         }

         //* If a second printing character was found AND   *
         //* if we have a two-character match, then open an *
         //* interactive sub-dialog to get additional data. *
         if ( matchFound )
         {
            attr_t   dColor = this->cfgOpt.cScheme.tf,   // text color
                     bColor = dColor | ncrATTR ;         // background color
            short ssWidth = this->ic[pathCtrl].cols / 2, // substring display area
                  ssOffset ;                             // substring display offset
            bool sensi = false ;                         // 'true' for case-sensitive compare
            winPos wp( ZERO, ssWidth ) ;                 // substring display position
            InitNcDialog dInit
             ( 1,                         // number of display lines
               this->ic[pathCtrl].cols,   // number of display columns
               this->ic[pathCtrl].ulY,    // Y offset
               this->ic[pathCtrl].ulX,    // X offset 
               NULL,                      // n/a (no title)
               ncltSINGLE,                // n/a (no border)
               dColor,                    // n/a (no border)tribute
               bColor,                    // interior color attribute
               NULL                       // n/a (no controls)
             ) ;

            NcDialog* dp = new NcDialog ( dInit ) ;
            if ( (dp->OpenWindow ()) == OK )
            {  //* Indicate that filename search is in progress *
               dtbmData  msgData( "  Filename Search: Printing characters, BKSP/DEL or ESCAPE keys.  " ) ;
               this->dWin->DisplayTextboxMessage ( dwMsgsTB, msgData ) ;
               dp->WriteString ( wp.ypos, (wp.xpos - 10), "Filename:", bColor ) ;
               bool  done = false ;       // loop control

               while ( !done )
               {  //* Clear the display area.                        *
                  //* It is possible (though unlikely) that user has *
                  //* entered a search string that is wider than our *
                  //* display area. If so, adjust the offset.        *
                  dp->ClearLine ( ZERO, false, wp.xpos ) ;
                  ssOffset = ZERO ;
                  if ( (substr.gscols()) >= ssWidth )
                  {  // Programmer's Note: Yes, this is a lot of work for
                     // small reward, but user-friendliness R Us.
                     short swidth = substr.gscols(), ci ;
                     const short* cwidth = substr.gscols( ci ) ;
                     for ( short i = ZERO ; (swidth >= ssWidth) && i < ci ; ++i )
                     {
                        ++ssOffset ;
                        swidth -= cwidth[i] ;
                     }
                  }
                  dp->WriteString ( wp, &substr.gstr()[ssOffset], dColor, true ) ;

                  //* If data in the key-input queue *
                  if ( (dp->KeyPeek ( wk )) != wktERR )
                  {
                     //* Only printing characters, BKSP/DEL and ESCAPE are *
                     //* processed. Anything else is returned to caller.   *
                     if ( wk.type == wktPRINT )       // printing character
                     {
                        dp->GetKeyInput ( wk ) ;      // retrieve data from queue
                        substr.append( wk.key ) ;     // append to search string
                        matchFound = 
                           this->fdPtr->Scroll2MatchingFile ( substr, currIndex, sensi ) ;
                        if ( ! matchFound )           // no match, audible boink
                        {  //* Erase most-recent character and boink the user *
                           substr.limitChars( (substr.gschars() - 2) ) ;

                           //* Handle SPACE character as a command key.*
                           if ( wk.key == nckSPACE )
                           {
                              dp->UngetKeyInput ( wk ) ;
                              done = true ;  // data belongs to caller
                           }
                           else
                              dp->UserAlert () ;
                        }
                     }
                     else if ( (wk.type == wktFUNKEY) && // one of the special keys
                               ((wk.key == nckESC) ||
                                (wk.key == nckBKSP) ||
                                (wk.key == nckDELETE) ||
                                (wk.key == nckINSERT)) )
                     {
                        dp->GetKeyInput ( wk ) ;      // retrieve data from queue
                        if ( wk.key == nckESC )       // search aborted
                           done = true ;
                        else if ( wk.key == nckINSERT )
                        {
                           sensi = sensi ? false : true ;
                           wchar_t indicator = sensi ? L'*' : L' ' ;
                           dp->WriteChar ( wp.ypos, (wp.xpos - 11), 
                                           indicator, bColor, true ) ;
                        }
                        else
                        {  //* Erase most-recent character from search string,*
                           //* and if string is now empty, we're done.        *
                           substr.limitChars( (substr.gschars() - 2) ) ;
                           if ( (substr.gschars()) > 1 )
                           {  //* Return to previous matching filename *
                              this->fdPtr->Scroll2MatchingFile ( substr, currIndex, sensi ) ;
                           }
                           else     // search string is empty
                              done = true ;
                        }
                     }
                     else     // data belongs to caller
                        done = true ;
                  }
                  else
                  {
                     this_thread::sleep_for( aMoment ) ;    // take a nap
                  }
               }     //while()
            }
            if ( dp != NULL )                // close the dialog
               delete dp ;
            this->dWin->RefreshWin () ;      // restore path control
            this->RestoreQuickHelpMsg () ;   // restore application Help message
         }     // two-char match found
      }        // first char match found
      else     // no match, boink the user
         this->dWin->UserAlert () ;
   }           // first char == wktPRINT

}  //* End CmdLocateFile() *

//*************************
//*    CmdKeyBindings     *
//*************************
//******************************************************************************
//* View a list of the current key bindings.                                   *
//*                                                                            *
//* Current key bindings are listed in the format:                             *
//* CUSTOM KEY  DEFAULT KEY  DESCRIPTION                                       *
//* ========================================================================== *
//*     --      nckC_A       select/deselect All files in the current window   *
//*     --      nckC_B       unassigned                                        *
//*     --      nckC_C       Copy selected files to target directory           *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void FileMangler::CmdKeyBindings ( void )
{
   const short ulY      = this->ic[dwMsgsTB].ulY,  // dialog Y position
               ulX      = ZERO,                    // dialog X position
               dlgROWS  = this->actualRows - ulY,  // dialog height
               dlgCOLS  = KEYMAP_DLGX ;            // dialog width
   const attr_t dColor = this->cfgOpt.cScheme.sd,  // dialog base color attribute
                eColor = this->cfgOpt.cScheme.em,  // emphasis color
                seFocus = this->cfgOpt.cScheme.wr; // scrollext focus color
   const char* scrollText[KEYMAP_KEYS] ;           // scrollext text data
   attr_t scrollColors[KEYMAP_KEYS] ;              // scrollext color data
   gString gs, gsEntry ;                           // text formatting and search

   enum Controls : short { dpCLOSE, dpSEARCH, dpSCROLL, dpCONTROLS } ;

   InitCtrl ic[dpCONTROLS] =          // control definitions
   {
   {  //* 'CLOSE' pushbutton  - - - - - - - - - - - - - - - - - - - -  dpCLOSE *
      dctPUSHBUTTON,                // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      short(dlgROWS - 2),           // ulY:       upper left corner in Y
      2,                            // ulX:       upper left corner in X
      1,                            // lines:     (n/a)
      9,                            // cols:      control columns
      "  CLOSE  ",                  // dispText:  
      this->cfgOpt.cScheme.pn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.pf,      // fColor:    focus color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:     (n/a)
      ZERO,                         // labY:      (n/a)
      ZERO,                         // labX       (n/a)
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[dpSEARCH]                 // nextCtrl:  link in next structure
   },
   {  //* 'Search' Textbox - - - - - - - - - - - - - - - - - - - -    dpSEARCH *
      dctTEXTBOX,                   // type:      
      rbtTYPES,                     // rbSubtype: (n/a)
      false,                        // rbSelect:  (n/a)
      ic[dpCLOSE].ulY,              // ulY:       upper left corner in Y
      short(ic[dpCLOSE].ulX + ic[dpCLOSE].cols + 7), // ulX: upper left corner in X
      1,                            // lines:     (n/a)
      18,                           // cols:      control columns
      NULL,                         // dispText:  initially empty
      this->cfgOpt.cScheme.tn,      // nColor:    non-focus color
      this->cfgOpt.cScheme.tf | ncrATTR, // fColor:    focus color
      tbPrint,                      // filter:    all printing characters
      NULL,                         // label:     
      ZERO,                         // labY:      
      ZERO,                         // labX       
      ddBoxTYPES,                   // exType:    (n/a)
      1,                            // scrItems:  (n/a)
      1,                            // scrSel:    (n/a)
      NULL,                         // scrColor:  (n/a)
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      &ic[dpSCROLL]                 // nextCtrl:  link in next structure
   },
   { //* 'Bindings' scrollext  - - - - - - - - - - - - - - - - - - -  dpSCROLL *
      dctSCROLLEXT,                 // type:      define a scrolling-data control
      rbtTYPES,                     // rbSubtype: (na)
      false,                        // rbSelect:  (n/a)
      1,                            // ulY:       upper left corner in Y
      ZERO,                         // ulX:       upper left corner in X
      short(dlgROWS - 3),           // lines:     control lines
      dlgCOLS,                      // cols:      control columns
      NULL,                         // dispText:  n/a
      dColor,                       // nColor:    non-focus border color
      seFocus,                      // fColor:    focus border color
      tbPrint,                      // filter:    (n/a)
      NULL,                         // label:
      ZERO,                         // labY:      offset from control's ulY
      ZERO,                         // labX       offset from control's ulX
      ddBoxTYPES,                   // exType:    (n/a)
      KEYMAP_KEYS,                  // scrItems:  number of elements in text/color arrays
      ZERO,                         // scrSel:    (n/a)
      scrollColors,                 // scrColor:  array of color attributes
      NULL,                         // spinData:  (n/a)
      true,                         // active:    allow control to gain focus
      NULL,                         // nextCtrl:  link in next structure
   },
   } ;
   //* Save the application's display data *
   this->dWin->SetDialogObscured () ;

   //* If the keymap object has not yet been initialized, format    *
   //* the display text and initialize the default command keycodes.*
   //* Note: If there are user-defined keycodes, then the data      *
   //*       will have been initialized during startup.             *
#if 0    // TEMP - TESTING ONLY
this->keyMap.initialize() ;
for ( short i = ZERO ; i < KEYMAP_KEYS ; ++i )
   this->keyMap.keydef[i].kuser = this->keyMap.keydef[i].kdflt ;
this->keyMap.kmInit = false ;
this->keyMap.initialize( false, true ) ;
#elif 0  // TEMP - TESTING ONLY
this->keyMap.keydef[2].kuser.type = wktFUNKEY ;
this->keyMap.keydef[2].kuser.key  = nckC_B ;
this->keyMap.keydef[11].kuser.type = wktFUNKEY ;
this->keyMap.keydef[11].kuser.key  = nckC_P ;
this->keyMap.kmInit = false ;
this->keyMap.initialize( false, true ) ;
#else    // PRODUCTION
   if ( ! this->keyMap.kmInit )
      this->keyMap.initialize() ;
#endif   // PRODUCTION

   //* Initialize text-pointer and color attribute arrays for dctSCROLLEXT.*
   for ( short i = ZERO ; i < KEYMAP_KEYS ; ++i )
   {
      scrollText[i] = this->keyMap.keydef[i].ktext ;
      scrollColors[i] = dColor ;
   }


   //* Initial parameters for dialog window *
   InitNcDialog dInit( dlgROWS,        // number of display lines
                       dlgCOLS,        // number of display columns
                       ulY,            // Y offset from upper-left of terminal 
                       ulX,            // X offset from upper-left of terminal 
                       NULL,           // dialog title
                       ncltSINGLE,     // border line-style
                       dColor,         // border color attribute
                       dColor,         // interior color attribute
                       ic              // pointer to list of control definitions
                     ) ;

   //* Instantiate the dialog window *
   NcDialog* dp = new NcDialog ( dInit ) ;

   //* Open the dialog window *
   if ( (dp->OpenWindow()) == OK )
   {
      dp->SetDialogTitle ( "  CUSTOM KEY   - DEFAULT KEY - DESCRIPTION "
                           "------------------------------ ", eColor ) ;

      //* Make a visual connection for the dctSCROLLEXT control that *
      //* overlaps the dialog window's border.                       *
      cdConnect cdConn ;
      cdConn.ul2Left = true ;
      cdConn.ll2Left = true ;
      cdConn.ur2Right = true ;
      cdConn.lr2Right = true ;
      cdConn.connection = true ;
      dp->ConnectControl2Border ( dpSCROLL, cdConn ) ;

      //* Set the dctSCROLLEXT display data *
      ssetData sData( scrollText, scrollColors, KEYMAP_KEYS, ZERO, false ) ;
      dp->SetScrollextText ( dpSCROLL, sData ) ;

      winPos scntPos( ic[dpSEARCH].ulY, (ic[dpSEARCH].ulX - 6) ) ;
      dp->WriteString ( ic[dpSEARCH].ulY, (ic[dpSEARCH].ulX + ic[dpSEARCH].cols + 1),
                        "Enter search text or a command code.", eColor ) ;

      dp->RefreshWin () ;           // make everything visible

      uiInfo   Info ;                     // user interface data returned here
      short    icIndex = ZERO ;           // index of control with input focus
      bool     done = false ;             // loop control
      while ( ! done )
      {
         if ( ic[icIndex].type == dctPUSHBUTTON )
         {
            if ( Info.viaHotkey )
               Info.HotData2Primary () ;
            else
               icIndex = dp->EditPushbutton ( Info ) ;

            if ( Info.dataMod != false )
            {
               done = true ;
            }
         }

         else if ( ic[icIndex].type == dctSCROLLEXT )
         {
            Info.viaHotkey = false ;      // ignore hotkey data
            icIndex = dp->ViewScrollext ( Info ) ;
         }

         else if ( ic[icIndex].type == dctTEXTBOX )
         {
            //* Manually capture user input for the 'search' textbox control.*
            //* Update the scrollext data to indicate matching entries.      *
            short match = ZERO,                       // number of matching search items
                  firstMatch = -1,                    // index of first matching search item
                  lastRow = ic[dpSCROLL].lines - 2 ;  // last visible scrollext row
            Info.dataMod = false ;                    // reset data-match flag
            while ( true )
            {
               dp->GetKeyInput ( Info.wk ) ;
               if ( (Info.wk.type == wktFUNKEY) &&
                    ((Info.wk.key == nckTAB) || (Info.wk.key == nckDOWN)
                      || (Info.wk.key == nckENTER) || (Info.wk.key == nckSTAB) 
                      || (Info.wk.key == nckUP) || (Info.wk.key == nckESC)) )
               {
                  Info.keyIn = Info.wk.key ; // indicate focus direction
                  break ;                    // edit complete
               }
               else if ( Info.wk.type == wktPRINT )
               {
                  dp->PrevControl () ;
                  dp->GetTextboxText ( dpSEARCH, gs ) ;
                  gs.append( Info.wk.key ) ;
                  dp->SetTextboxText ( dpSEARCH, gs ) ;
                  Info.dataMod = true ;
                  dp->NextControl () ;
               }
               else if ( (Info.wk.type == wktFUNKEY) &&
                         ((Info.wk.key == nckDELETE) || (Info.wk.key == nckBKSP)) )
               {
                  dp->PrevControl () ;
                  dp->GetTextboxText ( dpSEARCH, gs ) ;
                  if ( gs.gschars() >= 2 )
                  {
                     gs.limitChars( gs.gschars() - 2 ) ;
                     dp->SetTextboxText ( dpSEARCH, gs ) ;
                     Info.dataMod = true ;
                  }
                  else
                     dp->UserAlert () ;
                  dp->NextControl () ;
               }
               else
               {  //* Keycode is potentially a command-key sequence. *
                  //* Scan the list(s) of valid command keys.        *
                  //* 1) Scan user-defined list (if any).            *
                  //* 2) Scan the default command-key list.          *
                  gs = "INVALID KEYCODE!" ;
                  if ( this->keyMap.kmActive )
                  {
                     for ( short i = ZERO ; i < KEYMAP_KEYS ; ++i )
                     {
                        if ( Info.wk == this->keyMap.keydef[i].kuser )
                        {
                           // Programmer's Note: Capturing the substring for which
                           // to search begins with the first character in the 
                           // display element and ends before the vertical bar.
                           // Leading and trailing spaces are discarded.
                           // Because the field is narrower than the widest 
                           // substring, it is possible, but unlikely that the 
                           // search will identify multiple matches.
                           dp->PrevControl () ;
                           gs = this->keyMap.keydef[i].ktext ;
                           short offset = gs.find( wcsVERT) ;
                           gs.limitChars( offset ) ;
                           gs.strip() ;
                           dp->SetTextboxText ( dpSEARCH, gs ) ;
                           Info.dataMod = true ;
                           dp->NextControl () ;
                           break ;
                        }
                     }
                  }
                  if ( ! Info.dataMod )
                  {
                     for ( short i = ZERO ; i < KEYMAP_KEYS ; ++i )
                     {
                        if ( Info.wk == this->keyMap.keydef[i].kdflt )
                        {
                           // Programmer's Note: Capturing the substring for which 
                           // to search includes one leading, and one trailing space.
                           // This avoids multiple matches with the human-readable 
                           // version of the command code.
                           dp->PrevControl () ;
                           gs = this->keyMap.keydef[i].ktext ;
                           short offset = (gs.find( '|' )) + 1 ;
                           gs.shiftChars( -(offset) ) ;
                           offset = gs.find( L' ', 1 ) ;
                           gs.limitChars( offset + 1 ) ;
                           dp->SetTextboxText ( dpSEARCH, gs ) ;
                           Info.dataMod = true ;
                           dp->NextControl () ;
                           break ;
                        }
                     }
                     if ( ! Info.dataMod )   // user's command not found in table
                        dp->UserAlert () ;
                  }
               }

               //* Rescan the display data for matching text *
               if ( Info.dataMod != false )
               {
                  match = ZERO ;
                  for ( short i = ZERO ; i < KEYMAP_KEYS ; ++i )
                  {
                     gsEntry = this->keyMap.keydef[i].ktext ;
                     if ( (gsEntry.find( gs.gstr() )) >= ZERO )
                     {
                        scrollColors[i] = seFocus ;
                        //* Be sure first matched item is visible *
                        if ( ++match == 1 )
                        {
                           firstMatch = i ;
                           if ( firstMatch < lastRow )
                              firstMatch = ZERO ;
                        }
                     }
                     else
                        scrollColors[i] = dColor ;
                  }

                  dp->WriteString ( scntPos, "     ", dColor ) ;
                  if ( match > ZERO )
                  {
                     gs.compose( "(%3hd)", &match ) ;
                     dp->WriteString ( scntPos, gs, seFocus ) ;
                  }
                  dp->SetScrollextSelect ( dpSCROLL, firstMatch ) ;
                  dp->RefreshScrollextText ( dpSCROLL, false ) ;
                  dp->RefreshWin () ;
                  Info.dataMod = false ;
               }     // dataMod != false
            }        // while()
         }           // dctTEXTBOX
         
         //* Move focus to appropriate control *
         if ( ! done && ! Info.viaHotkey )
         {
            if ( Info.keyIn == nckSTAB )
               icIndex = dp->PrevControl () ; 
            else
               icIndex = dp->NextControl () ;
         }
      }     // while(!done)
   }
   if ( dp != NULL )
      delete ( dp ) ;                        // close the window
   this->dWin->RefreshWin () ;               // restore application window

}  //* CmdKeyBindings() *

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

