//******************************************************************************
//* File       : FileDlgMedia.cpp                                              *
//* Author     : Mahlon R. Smith                                               *
//*              Copyright (c) 2016-2020 Mahlon R. Smith, The Software Samurai *
//*                  GNU GPL copyright notice located in FileMangler.hpp       *
//* Date       : 21-Jul-2020                                                   *
//* Version    : (see FileDlgVersion string in FileDlg.cpp)                    *
//*                                                                            *
//* Description:                                                               *
//* This module of the FileDlg class contains decoding algorithms for          *
//* extracting metadata from media files (audio/video).                        *
//*                                                                            *
//* Supported Formats:                                                         *
//* ------------------                                                         *
//*  MPEG-3 Audio (MP3)                                                        *
//*  OGG/Vorbis Audio (OGG, OGA)                                               *
//*  MPEG-4 Audio (M4A)                                                        *
//*                                                                            *
//* Development Tools: see notes in FileMangler.cpp.                           *
//******************************************************************************
//* Version History (most recent first):                                       *
//*   See version history in FileDlg.cpp.                                      *
//******************************************************************************
//* Programmer's Notes:                                                        *
//* Audio/video metadata can be extracted from the media file and displayed    *
//* as part of the View-file-contents sequence.                                *
//*                                                                            *
//* Because the media data are binary, there is nothing to see there; however, *
//* each media-file format contains a mechanism for storing human-readable     *
//* data for Artist, Composer, Producer, Copyright, etc. These data fields are *
//* generically known as "tags."                                               *
//*                                                                            *
//* While MP3 audio is by far the most popular, other audio media formats are  *
//* in general use. Specifically the open-standard OGG/Vorbis audio encoding.  *
//*                                                                            *
//* Please note that the formatting standards are not well implemented by      *
//* media players (VLC and others), so a certain tolerance for technically     *
//* invalid formatting is built into these methods.                            *
//*                                                                            *
//* Support for other media formats will be added as requested.                *
//*                                                                            *
//* FileMangler does not support editing of the metadata because it would      *
//* stray from the basic mission of file management. Other open-source         *
//* applications are available for editing media metadata. If these turn out   *
//* to be clunky, may we recommend our own Taggit application for editing tag  *
//* data.                                                                      *
//*                                                                            *
//******************************************************************************

#include "GlobalDef.hpp"
#include "FileDlg.hpp"

//******************************
//* Local definitions and data *
//******************************

//******************************************************************************
//**               Definitions for MPEG-3 (MP3) audio files                   **
//******************************************************************************

//* Codes used to identify frames which contain text. *
const short TEXT_FRAMES = 39 ;      // number of text frame types
const char* const TextFrameID[TEXT_FRAMES] = 
{
"TALB",    // #TALB Album/Movie/Show title
"TBPM",    // #TBPM BPM (beats per minute)
"TCOM",    // #TCOM Composer
"TCON",    // #TCON Content type
"TCOP",    // #TCOP Copyright message
"TDAT",    // #TDAT Date
"TDLY",    // #TDLY Playlist delay
"TENC",    // #TENC Encoded by
"TEXT",    // #TEXT Lyricist/Text writer
"TFLT",    // #TFLT File type
"TIME",    // #TIME Time
"TIT1",    // #TIT1 Content group description
"TIT2",    // #TIT2 Title/songname/content description
"TIT3",    // #TIT3 Subtitle/Description refinement
"TKEY",    // #TKEY Initial key
"TLAN",    // #TLAN Language(s)
"TLEN",    // #TLEN Length
"TMED",    // #TMED Media type
"TOAL",    // #TOAL Original album/movie/show title
"TOFN",    // #TOFN Original filename
"TOLY",    // #TOLY Original lyricist(s)/text writer(s)
"TOPE",    // #TOPE Original artist(s)/performer(s)
"TORY",    // #TORY Original release year
"TOWN",    // #TOWN File owner/licensee
"TPE1",    // #TPE1 Lead performer(s)/Soloist(s)
"TPE2",    // #TPE2 Band/orchestra/accompaniment
"TPE3",    // #TPE3 Conductor/performer refinement
"TPE4",    // #TPE4 Interpreted, remixed, or otherwise modified by
"TPOS",    // #TPOS Part of a set
"TPUB",    // #TPUB Publisher
"TRCK",    // #TRCK Track number/Position in set
"TRDA",    // #TRDA Recording dates
"TRSN",    // #TRSN Internet radio station name
"TRSO",    // #TRSO Internet radio station owner
"TSIZ",    // #TSIZ Size
"TSRC",    // #TSRC ISRC (international standard recording code)
"TSSE",    // #TSEE Software/Hardware and settings used for encoding
"TYER",    // #TYER Year
"TXXX",    // #TXXX User defined text information frame
} ;
//* Human-friendly equivalents to the TextFrameID list above *
const char* const TextFrameDesc[TEXT_FRAMES] = 
{
"Album Title ------------------ ",
"BPM (beats per minute) ------- ",
"Composer --------------------- ",
"Content type ----------------- ",
"Copyright © ------------------ ",
"Date ------------------------- ",
"Playlist delay --------------- ",
"Encoded by ------------------- ",
"Lyricist/Text writer --------- ",
"File type -------------------- ",
"Time ------------------------- ",
"Content Group ---------------- ",
"Title ------------------------ ",
"Subtitle --------------------- ",
"Initial Key ------------------ ",
"Language --------------------- ",
"Length ----------------------- ",
"Media type ------------------- ",
"Original Title --------------- ",
"Original Filename ------------ ",
"Original lyricist ------------ ",
"Original artist -------------- ",
"Original Release ------------- ",
"File owner ------------------- ",
"Lead performer(s) ------------ ",
"Accompaniment ---------------- ",
"Conductor/performer ---------- ",
"Remixed by ------------------- ",
"Part of a set ---------------- ",
"Publisher -------------------- ",
"Track number/Position -------- ",
"Recording dates -------------- ",
"Internet radio station name -- ",
"Internet radio station owner - ",
"Size ------------------------- ",
"ISRC ------------------------- ",
"Software/Hardware ------------ ",
"Year ------------------------- ",
"User Defined ----------------- ",
} ;

//* "Picture Type" descriptions as defined for id3v2 standard.  *
//* (Used for both MP3 and OGG decoding.)                       *
const short PIC_TYPES = 21 ;     // number of descriptions per sub-array
const char* const pType[PIC_TYPES] = 
{  // English
  "Other",
  "32x32 pixels 'file icon' (PNG only)",
  "Other file icon",
  "Cover (front)",
  "Cover (back)",
  "Leaflet page",
  "Media (e.g. label side of CD)",
  "Lead artist/lead performer/soloist",
  "Artist/performer",
  "Conductor",
  "Band/Orchestra",
  "Composer",
  "Lyricist/text writer",
  "Recording Location",
  "During recording",
  "During performance",
  "Movie/video screen capture",
  "A bright coloured fish",
  "Illustration",
  "Band/artist logotype",
  "Publisher/Studio logotype",
} ;

//* According to Wikipedia, the following text-encoding markers are valid.     *
//* Note: As of Dec 2016, many media players cannot decode ID3v2.4.            *
enum TextEncode : char
{
   ENCODE_ASCII   = 0x00,  // ISO-8859-1 (0x20 - 0xFF plus 0x0A)
                           // (LATIN-1, Identical to ASCII for values smaller than 0x80).
   ENCODE_UTF16   = 0x01,  // UTF-16, (big-endian) encoded with BOM, ID3v2.2 and ID3v2.3.
                           // (Formerly UCS-2 (ISO/IEC 10646-1:1993, obsolete))
   ENCODE_UTF16BE = 0x02,  // UTF-16BE (big-endian) encoded without BOM, in ID3v2.4.
   ENCODE_UTF8    = 0x03,  // UTF-8 encoded in ID3v2.4.
   ENCODE_TXTERR           // Invalid text-frame encoding
} ;

//* Definitions for encoding and decoding UTF-16 data.*
const UINT usxMIN20  = 0x0010000 ;  // minimum value requiring 20-bit conversion
const UINT usxMAX20  = 0x010FFFF ;  // maximum value for 20-bit UTF-16
const UINT usxMSU16  = 0x000D800 ;  // MS unit mask value
                                    // and beginning of 16-bit reserved sequence
const UINT usxMSU16e = 0x000DBFF ;  // end of MS unit range
const UINT usxLSU16  = 0x000DC00 ;  // LS unit mask value
const UINT usxLSU16e = 0x000DFFF ;  // end of LS unit range
const UINT usxHIGH16 = 0x00E000 ;   // beginning of high-range 16-bit codepoints
const UINT usxMAX16  = 0x00FFFD ;   // maximum value for 16-bit codepoint
const UINT usxMASK10 = 0x0003FF ;   // 10-bit mask

//* If a text frame begins with ENCODE_UTF16, then the next two  *
//* bytes must be either "UTF_MSB UTF_LSB" or "UTF_LSB UTF_MSB". *
//* If not, it is not a valid Unicode-16 string.                 *
const char UTF_MSB = 0xFE ;
const char UTF_LSB = 0xFF ;

//* A valid Unicode-16 encoded string must be at least this long: *
//*             "01 FF FE xx xx" or "01 FE FF xx xx               *
//*             "02 xx xx" or "02 xx xx                           *
const int  UTF16_MIN = 5 ;
const int  UTF16BE_MIN = 3 ;

//* A valid ASCII or UTF-8 encoded string must be at least this long: *
//*             "00 xx"                                               *
const int  ASCII_MIN = 2 ;

//* Map of ISO8859-1, single-byte characters between 0x7F and 0xFF.   *
const short iso8859FIRST = 0x00A0,  // first character in array
            iso8859LAST  = 0x00FF,  // last character in array
            iso8859COUNT = 96 ;     // number of elements in array

const ULONG framehdrCNT = 10 ;  // size of frame header
class id3v2_framehdr
{
   public:

   id3v2_framehdr ( void )
   { this->reset () ; }
   ~id3v2_framehdr ( void ) {}
   void reset ( void )
   {
      this->frame_id[0] = this->frame_id[1] = this->frame_id[2] = 
      this->frame_id[3] = this->frame_id[4] = this->frame_id[5] = NULLCHAR ;
      this->frame_size = ZERO ;
      this->status_flags = this->encode_flags = ZERO ;
      this->flag_tag_pres = this->flag_file_pres = this->flag_readonly = 
      /*this->flag_d = this->flag_e = this->flag_f = this->flag_g = this->flag_h =*/ false ;
      this->flag_compress = this->flag_encrypt   = this->flag_grouped  = 
      /*this->flag_l = this->flag_m = this->flag_n = this->flag_o = this->flag_p =*/ false ;
      this->encoding = ENCODE_ASCII ;
      this->big_endian = false ;
      this->decomp = ZERO ;   // if compressed, Zlib compression is used
      this->crypto = ZERO ;   // if active, s/b a value greater than 0x80 (see ENCR frame)
      this->group_id = ZERO ; // if active, s/b a value greater than 0x80 (see GRID frame)
   }

   //* Convert a 4-byte sequence (MSB at offset 0) to an integer value. *
   //* Programmer's Note: This clunky construct avoids the C library's  *
   //* "helpful" automatic sign extension.                              *
   int intConv ( const UCHAR* ucp )
   {
      int i =    (UINT)ucp[3] 
              | ((UINT)ucp[2] << 8)
              | ((UINT)ucp[1] << 16)
              | ((UINT)ucp[0] << 24) ;
      return i ;
   }

   //* Convert a 32-bit integer into a big-endian bytes stream.   *
   //* (used for both 'frame-size' and 'decomp')                  *
   short intConv ( int ival, char* obuff )
   {
      const int bMASK = 0x000000FF ;
      short indx = ZERO ;     // return value

      obuff[indx++] = (char)((ival >> 24) & bMASK) ;
      obuff[indx++] = (char)((ival >> 16) & bMASK) ;
      obuff[indx++] = (char)((ival >>  8) & bMASK) ;
      obuff[indx++] = (char)(ival & bMASK) ;

      return indx ;
   }

   //* Convert raw byte data to gString (UTF-8) format.           *
   //* -- Assumes that 'frame_count' has been initialized.        *
   //* -- First byte of source indicates encoding:                *
   //*      ENCODE_ASCII or ENCODE_UTF16 (with BOM) or            *
   //*      ENCODE_UTF16BE (no BOM) or ENCODE_UTF8                *
   //* -- ENCODE_UTF16 requires frame_size >= 5 because"          *
   //*        "01 FF FE xx xx" or "01 FE FF xx xx"                *
   //*    is the shortest possible valid UTF-16 string.           *
   //* -- ENCODE_UTF16BE requires frame_size >= 3 because:        *
   //*        "02 xx xx"                                          *
   //*    is the shortest possible UTF16BE string.                *
   //* -- ENCODE_ASCII and ENCODE_UTF8 require frame_size >= 2    *
   //*       because: 0n xx is the shorted possible string        *
   //*                                                            *
   //* Note that if UTF-16 were written as a byte stream as it    *
   //* should have been, then there wouldn't be an issue with     *
   //* 'endian-ness'. However, we must accomodate the encoders    *
   //* who write the data as 16-bit integers on either big-endian *
   //* or little-endian hardware without regard to what they are  *
   //* actually doing.                                            *
   //* Be aware that when we _encode_, we always encode byte-wise *
   //* as big-endian.                                             *
   //*                                                            *
   //* Because the folks at the VLC project and elsewhere have    *
   //* little regard for the standard, we must do some defensive  *
   //* programming:                                               *
   //* a) Verify any claim that the source data are ASCII.        *
   //*    Note that we do not support the Latin-1 extensions      *
   //*    (ISO-8859-1). It is either pure ASCII or it is not.     *
   //* b) Verify that a UTF-16 string actually is.                *
   //* c) Verify the endian-ness of UTF-16 data.                  *
   //* d) For id3v2.4 (but not id3v2.3) text frames may consist   *
   //*    of multiple strings which are delimited by the null     *
   //*    character (00h(00h)). We believe that this is a huge    *
   //*    mistake and a bone-headed "enhancement." However, since *
   //*    they didn't consult us, we have to parse for it, and if *
   //*    found, we substitute a forward slash '/' to concatenate *
   //*    the strings for display.                                *
   //*         "This is a \0 concatenated string."                *
   //*    becomes:                                                *
   //*         "This is a / concatenated string."                 *
   //* e) If encoding error, then attempt to decode as UTF-8,     *
   //*    which has been routinely used in versions prior to      *
   //*    id3v2.4 even though it was not part of the standard.    *
   //* -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  - *
   //* Input  : rawPtr  : pointer to source byte stream           *
   //*          gsTrg   : (by reference) receives converted text  *
   //*                                                            *
   //* Return: number of wchar_t characters converted             *
   //*         or (-1) if conversion error                        *
   short txtConv ( const char* rawPtr, gString& gsTrg )
   {
      short convChars = ZERO ;         // return value

      this->encoding = rawPtr[0] ;     // text encoding

      if ( this->encoding == ENCODE_UTF16 || this->encoding == ENCODE_UTF16BE )
      {
         this->big_endian = true ;     // assume big-endian input
         if ( this->encoding == ENCODE_UTF16 )
         {
            //* Check the byte-order mark.        *
            //* If the BOM was stored backward,   *
            //* then it is little-endian encoding.*
            if ( rawPtr[1] == UTF_LSB )
               this->big_endian = false ;
         }

         if ( ((this->encoding == ENCODE_UTF16) && (this->frame_size >= UTF16_MIN)
                && (rawPtr[1] == UTF_MSB || rawPtr[1] == UTF_LSB)
                && (rawPtr[2] == UTF_MSB || rawPtr[2] == UTF_LSB))
              ||
              ((this->encoding == ENCODE_UTF16BE) && (this->frame_size >= UTF16BE_MIN)) )
         {
            short offset = this->encoding == ENCODE_UTF16 ? 3 : 1 ;
            convChars = this->utf16Conv ( &rawPtr[offset], gsTrg ) ;
         }
         else  // invalid UTF-16 encoding
         {
            convChars = this->utf8Conv ( &rawPtr[1], gsTrg ) ;
         }
      }

      else if ( this->encoding == ENCODE_ASCII )
         convChars = this->ascConv ( &rawPtr[1], gsTrg ) ;

      else     // ENCODE_UTF8 or unsupported format
         convChars = this->utf8Conv ( &rawPtr[1], gsTrg ) ;

      return convChars ;
   }  //* End txtConv() *

   //* Public Method.                                             *
   //* Encode the wchar_t (wide text) array into a form that MP3  *
   //* text frames can understand.                                *
   //*  Note that we always include the NULL terminator to the    *
   //*  text data. The standard neither demands nor forbids it.   *
   //*                                                            *
   //* Input  : enc   : inticates the type of encoding:           *
   //*                   a) ENCODE_ASCII                          *
   //*                   b) ENCODE_UTF8                           *
   //*                   c) ENCODE_UTF16 (big-endian with BOM)    *
   //*                   d) ENCODE_UTF16BE (big-endian, no BOM)   *
   //*          idIndx: index into the TextFrameID[] array (above)*
   //*          fData : receives the encoded data                 *
   //*          src   : wchar_t-encoded source text data          *
   //*                                                            *
   //* Returns: number of bytes in formatted output               *
   //*          or ZERO if encoding error                         *
   short txtEncode ( TextEncode enc, short idIndx, char* fData, const gString& src )
   {
      //* Initialize the data members *
      this->reset () ;

      //* Caller's choice of encoding (validate) *
      if ( (enc == ENCODE_ASCII)   || (enc == ENCODE_UTF16) || 
           (enc == ENCODE_UTF16BE) || (enc == ENCODE_UTF8) )
         this->encoding = enc ;
      else
         this->encoding = ENCODE_UTF16 ;
      this->big_endian = true ;           // 'true' but ignored

      //* Validate the frame ID *
      if ( (idIndx >= ZERO) && (idIndx < TEXT_FRAMES) )
      {
         gString gsid( TextFrameID[idIndx] ) ; // 4-byte Tnnn frame ID
         gsid.copy( this->frame_id, 6 ) ;

         //* Encode the text data.                                       *
         //* -- For ASCII and UTF-8, simply copy the source to target.   *
         //*    'frame_size' == number of text bytes + encoding byte     *
         //* -- For UTF-16 varients, convert from wchar_t to 16-bit data.*
         //*    'frame_size' == number of text bytes + encoding byte     *
         if ( (this->encoding == ENCODE_ASCII) || (this->encoding == ENCODE_UTF8) )
         {
            short fIndex = ZERO ;   // index into output array
            fData[fIndex++] = this->encoding ;
            src.copy( &fData[fIndex], src.utfbytes() ) ;
            this->frame_size = src.utfbytes() + 1 ;
         }
         else
            this->frame_size = this->utf16Encode ( enc, fData, src ) ;
      }
      return (short)this->frame_size ;

   }  //* End txtEncode() *

   #if 0    // FOR DEBUGGING ONLY
   //* Encode source data as UTF-16, little-endian.               *
   //* Note that for debugging purposes, we have implemented code *
   //* that can encode in a little-endian format, but this option *
   //* is not useful in production mode.                          *
   //* Note that only 'frame_size' member is initialized.         *
   short utf16leEncode ( char* fText, const gString& src )
   {
      short fIndex = ZERO ;   // index into output array

      fText[fIndex++] = ENCODE_UTF16 ;    // type of encoding
      fText[fIndex++] = (char)0x0FF ;     // LSB of Byte-Order-Mark
      fText[fIndex++] = (char)0x0FE ;     // MSB of Byte-Order-Mark

      //* Establish the source array of wchar_t (wide) characters *
      short wcnt ;   // number of source characters (including the null terminator)
      const wchar_t *wstr = src.gstr( wcnt ) ;
      UINT cx ;      // source character

      for ( short wIndex = ZERO ; wIndex < wcnt ; ++wIndex )
      {
         cx = wstr[wIndex] ;     // get a character from the input stream

         //* If the character can be encoded with a single, 16-bit value *
         if ( ((cx >= ZERO) && (cx < usxMSU16)) || 
              ((cx >= usxHIGH16) && (cx <= usxMAX16)) )
         {  //* Encode the bytes in little-endian order *
            fText[fIndex++] = cx & 0x000000FF ;
            fText[fIndex++] = (cx >> 8) & 0x000000FF ;
         }

         //* Else, character requires a pair of 16-bit values *
         else
         {
            UINT msUnit = ZERO, lsUnit = ZERO ;
            if ( (cx >= usxMIN20) && (cx <= usxMAX20) )
            {
               msUnit = ((cx - usxMIN20) >> 10) | usxMSU16 ;
               lsUnit = (cx & usxMASK10) | usxLSU16 ;
               fText[fIndex++] = msUnit & 0x000000FF ;
               fText[fIndex++] = (msUnit >> 8) & 0x000000FF ;
               fText[fIndex++] = lsUnit & 0x000000FF ;
               fText[fIndex++] = (lsUnit >> 8) & 0x000000FF ;
            }
            //* Character cannot be encoded as UTF-16. *
            //* Encode a question mark character.      *
            else
            {
               msUnit = L'?' ;
            }
         }
      }
      return ( this->frame_size = fIndex ) ;

   }  //* End utf16leEncode() *
   #endif   // FOR DEBUGGING ONLY

   //***************************************************************
   //* Methods that require controlled setup sequences are private.*
   //***************************************************************
   private:

   //* Private Method.                                            *
   //* Convert a raw ISO 8859-1 string to gString format.         *
   //* -- Assumes that 'frame_count' has been initialized.        *
   //* -- ISO 8859-1 is a group of single-byte characters.        *
   //* -- Note that the input strings are not always terminated.  *
   //* -- Note that Linux systems do not use the "Latin-1"        *
   //*    extensions of ISO8859-1 by default, so we must match    *
   //*    the Danish, German, Spanish etc. special alphabetical   *
   //*    characters (A0-FFh) explicitly.                         *
   short ascConv ( const char* rawPtr, gString& gsTrg )
   {
      gsTrg.clear() ;                        // initialize the target buffer
      int rawBytes = this->frame_size - 1 ;  // number of bytes in raw string
      wchar_t wc ;                           // 32-bit character
      for ( short ti = ZERO ; ti < rawBytes ; ++ti )
      {
         //* Printing characters of the ISO 8859-1 standard *
         if (   (((UCHAR)(rawPtr[ti]) >= 0x20) && (((UCHAR)rawPtr[ti]) <= 0x7E))
             || (rawPtr[ti] == '\n')
             || ((UCHAR)(rawPtr[ti]) >= iso8859FIRST && (UCHAR)(rawPtr[ti]) <= iso8859LAST)
            )
         {
            wc = (wchar_t)((UCHAR)(rawPtr[ti])) ;
            gsTrg.append( wc ) ;
         }
         //* If a null terminator is encountered BEFORE end-of-text,         *
         //* concatenate the separate strings in the field. (see note above) *
         else if ( (rawPtr[ti] == NULLCHAR) && (ti < (rawBytes - 1)) && (rawPtr[ti - 1] != ' ') )
         {
            gsTrg.append( fSLASH ) ;
         }
         //* Control character OR a character not defined in ISO 8859-1.*
         else if ( rawPtr[ti] != NULLCHAR ) { wc = L'?' ; gsTrg.append( wc ) ; }
      }
      return ( gsTrg.gschars() ) ;
   }  //* End ascConv() *

   //* Private Method.                                            *
   //* Convert an (assumed) UTF-8 string to gString format.       *
   //* Note that UTF-8 is not supported by id3v2.3 but is         *
   //* supported by id3.v2.4.                                     *
   //* -- Assumes that 'frame_count' has been initialized.        *
   //* -- Assumes that 'rawPtr' points to head of text (not       *
   //*    encoding byte)                                          *
   //* -- The string may or may not be terminated.                *
   //* -- Returns number of wchar_t (wide) characters created.    *

   short utf8Conv ( const char* rawPtr, gString& gsTrg )
   {
      char cbuff[this->frame_size] ;
      int ti,                                // source index
          rawBytes = this->frame_size - 1 ;  // number of bytes in raw string
      gsTrg.clear() ;                        // initialize the target buffer
      for ( ti = ZERO ; ti < rawBytes ; ++ti ) // replace non-terminal terminators
      {
         if ( rawPtr[ti] != '\0' )
            cbuff[ti] = rawPtr[ti] ;
         else           // null terminator encountered before end-of-text
         {              // concatenate the strings (see note above)
            if ( ti < (rawBytes - 1) )
               cbuff[ti] = '/' ;
            else        // end of text
               break ;
         }
      }
      cbuff[ti] = '\0' ;                     // be sure string is terminated
      gsTrg = cbuff ;
      return ( gsTrg.gschars() ) ;
   }  //* End utf8Conv() *

   //* Private Method:                                            *
   //* Convert a Unicode-16 string (big-endian or little-endian)  *
   //* to gString format.                                         *
   //* -- Assumes that 'frame_count' has been initialized.        *
   //* -- Note that the frame may contain multiple strings, but   *
   //*    if so, we concatenate them. It is assumed that 2nd and  *
   //*    subsequent strings have the same byte order as the 1st. *
   //* -- Unicode-16 strings are generally not terminated (except *
   //*    for empty strings), but if multiple strings, then       *
   //*    obviously, all but the last must be terminated.         *
   //* -- See additional information on decoding UTF-16 in the    *
   //*    mptDecodetextFrame() method header.                     *
   //* -- Note that because the wchar_t type is a signed integer, *
   //*    the compiler wants to sign-extend. Don't allow this.    *
   //* -- Returns number of wchar_t (wide) characters created.    *

   short utf16Conv ( const char* rawPtr, gString& gsTrg )
   {
      int ti = ZERO ;                        // source index
      gsTrg.clear() ;                        // initialize the target buffer
      int rawBytes = this->frame_size        // number of bytes in raw string
                     - ((this->encoding == ENCODE_UTF16) ? 3 : 1) ;
      UINT msUnit, lsUnit,                   // for converting unit pairs
           cx ;                              // undecoded 16-bit character
      while ( rawBytes > ZERO )
      {
         if ( this->big_endian )
         {
            cx = ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
            cx |= (UINT(rawPtr[ti++]) & 0x000000FF) ;
         }
         else  // (little-endian)
         {
            cx = (UINT(rawPtr[ti++]) & 0x000000FF) ;
            cx |= ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
         }
         rawBytes -= 2 ;
         //* If the character is fully represented by 16 bits *
         //* (most characters are)                            *
         if ( ((cx >= ZERO) && (cx < usxMSU16)) || 
              ((cx >= usxHIGH16) && (cx <= usxMAX16)) )
         {
            if ( cx != 0 )
               gsTrg.append( cx ) ;
            else if ( rawBytes > ZERO ) // if interim NULLCHAR found
            {
               gsTrg.append( fSLASH ) ; // string concatenation (see note above)
               //* Note that the following string may also have a *
               //* byte-order-mark (BOM). If so, step over it.    *
               if ( (rawBytes >= 2) &&
                    ((rawPtr[ti] == UTF_MSB && rawPtr[ti + 1] == UTF_LSB)
                     ||
                     (rawPtr[ti] == UTF_LSB && rawPtr[ti + 1] == UTF_MSB)) )
               {
                  rawBytes -= 2 ;
                  ti += 2 ;
               }
            }
         }

         //* Character is represented by 32 bits    *
         //* (20 bits actually used), 16 MSBs first.*
         else
         {
            msUnit = cx ;
            lsUnit = ZERO ;
            if ( this->big_endian )
            {
               lsUnit = ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
               lsUnit |= (UINT(rawPtr[ti++]) & 0x000000FF) ;
            }
            else     // (little-endian)
            {
               lsUnit = (UINT(rawPtr[ti++]) & 0x000000FF) ;
               lsUnit |= ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
            }
            rawBytes -= 2 ;

            //* Validate the range *
            if ( ((msUnit >= usxMSU16) && (msUnit <= usxMSU16e))
                 &&
                 ((lsUnit >= usxLSU16) && (msUnit <= usxLSU16e)) )
            {
               cx = usxMIN20 + ((msUnit & usxMASK10) << 10) ;
               cx |= (lsUnit & usxMASK10) ;
               if ( (cx == 0) && rawBytes > ZERO )
               {
                  cx = fSLASH ;   // string concatenation (see note above)
                  //* Note that the following string may also have a *
                  //* byte-order-mark (BOM). If so, step over it.    *
                  if ( (rawBytes >= 2) &&
                       ((rawPtr[ti] == UTF_MSB && rawPtr[ti + 1] == UTF_LSB)
                        ||
                        (rawPtr[ti] == UTF_LSB && rawPtr[ti + 1] == UTF_MSB)) )
                  {
                     rawBytes -= 2 ;
                     ti += 2 ;
                  }
               }
            }
            else                 // invalid UTF-16 codepoint
               cx = L'?' ;
            gsTrg.append( cx ) ; // add the chararacter to output buffer
         }
      }
      return ( gsTrg.gschars() ) ;

   }  //* End utf16Conv() *

   //* Private Method.                                            *
   //* Convert the wchar_t string to a UTF-16-encoded             *
   //* (big-endian) byte stream.                                  *
   //*                                                            *
   //* Input  : enc   : inticates the type of encoding:           *
   //*                   a) ENCODE_UTF16 (big-endian with BOM)    *
   //*                   b) ENCODE_UTF16BE (big-endian, no BOM)   *
   //*          fText : receives the encoded data                 *
   //*          src   : wchar_t-encoded source text data          *
   //*                                                            *
   //* Returns: number of bytes in formatted output               *
   //*          or (-1) if encoding error                         *
   //* -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  - *
   //* Note: UTF-16 may be encoded on either big-endian or        *
   //* little-endian hardware; however, we encode the UTF-16 as   *
   //* a byte stream, so our output is always encoded as          *
   //* big-endian.                                                *
   short utf16Encode ( TextEncode enc, char* fText, const gString& src )
   {
      short fIndex = ZERO ;   // index into output array

      if ( enc == ENCODE_UTF16 )
      {
         fText[fIndex++] = ENCODE_UTF16 ;    // type of encoding
         fText[fIndex++] = (char)0x0FE ;     // MSB of Byte-Order-Mark
         fText[fIndex++] = (char)0x0FF ;     // LSB of Byte-Order-Mark
      }
      else
         fText[fIndex++] = ENCODE_UTF16BE ;  // type of encoding

      //* Establish the source array of wchar_t (wide) characters *
      short wcnt ;   // number of source characters (including the null terminator)
      const wchar_t *wstr = src.gstr( wcnt ) ;
      UINT cx ;      // source character

      for ( short wIndex = ZERO ; wIndex < wcnt ; ++wIndex )
      {
         cx = wstr[wIndex] ;     // get a character from the input stream

         //* If the character can be encoded with a single, 16-bit value *
         if ( ((cx >= ZERO) && (cx < usxMSU16)) || 
              ((cx >= usxHIGH16) && (cx <= usxMAX16)) )
         {  //* Encode the bytes in big-endian order *
            fText[fIndex++] = (cx >> 8) & 0x000000FF ;
            fText[fIndex++] = cx & 0x000000FF ;
         }

         //* Else, character requires a pair of 16-bit values *
         else
         {
            UINT msUnit = ZERO, lsUnit = ZERO ;
            if ( (cx >= usxMIN20) && (cx <= usxMAX20) )
            {
               msUnit = ((cx - usxMIN20) >> 10) | usxMSU16 ;
               lsUnit = (cx & usxMASK10) | usxLSU16 ;
               fText[fIndex++] = (msUnit >> 8) & 0x000000FF ;
               fText[fIndex++] = msUnit & 0x000000FF ;
               fText[fIndex++] = (lsUnit >> 8) & 0x000000FF ;
               fText[fIndex++] = lsUnit & 0x000000FF ;
            }
            //* Character cannot be encoded as UTF-16. *
            //* Encode a question mark character.      *
            else
            {
               msUnit = L'?' ;
            }
         }
      }
      return ( this->frame_size = fIndex ) ;

   }  //* End utf16Encode() *


   //********************************
   //* All data members are public. *
   //********************************
   public:
   char  frame_id[6] ;     // S/B one of the 4-character names defined by the standard
   int   frame_size ;      // size of the frame in bytes
   char  status_flags ;    // status bits %abc0.0000
   char  encode_flags ;    // encoding flags %ijk0.0000
   bool  flag_tag_pres ;   // if 'true',  preserve frame if tag is modified
                           // if 'false', discard frame if tag is modified
   bool  flag_file_pres ;  // if 'true',  preserve frame if non-tag data modified
                           // if 'false', discard frame if non-tag data is modified
   bool  flag_readonly ;   // if 'true',  frame is read-only unless software know what it's doing
   bool  flag_compress ;   // if 'true',  frame is compressed
                           //    (4 bytes "decompressed size" added to frame header)
                           // if 'false', frame is not compressed
   bool  flag_encrypt ;    // if 'true',  frame is encrypted
                           //    (1 byte "encryption type"" added to frame header)
                           // if 'false', frame is not encrypted
   bool  flag_grouped ;    // if 'true',  frame is a member of a frame group
                           //    (1 byte "group identifier" added to frame header)
                           // if 'false', frame is not grouped
   char  encoding ;        // for text frames only: text encoding
                           // 00h == ISO8859-1 (ASCII, i.e. Latin-1: 0x20 - 0xFF)
                           // 01h == UTF-16 (with BOM)
                           // 02h == UTF-16 big-endian (no BOM) (id3v2.4 only)
                           // 03h == UTF-8 encoding (id3v2.4 only)
   bool  big_endian ;      // for UTF-16 text frames only
                           // if 'true', big-endian Unicode-16
                           // if 'false', little-endian Unicode-16
   int   decomp ;          // for compressed frames only: expanded size
   char  crypto ;          // for encrypted frames only: type of encryption
   char  group_id ;        // for grouped frames only: group ID

} ;   // id3v2_framehdr

//* Flag byte for id3v2_taghdr class *
//* Four bits are defined: Bits 7-4  *
//*  bit7 unsynch
//*  bit6 extended header follows
//*  bit5 experimental tag
//*  bit4 tag footer terminates the tag (defined in id3v2.4)
//*       (bits 6 and 4 are mutually exclusive)
//*  bits 3-0 currently undefined (always reset)
// Programmer's Note: We could have the same functionality 
// using a union with bit fields, but this is more fun.
class taghdrFlags
{
   public:
   ~taghdrFlags ( void ) {}         // destructor
   taghdrFlags ( void )             // default constructor
   { this->reset () ; }
   taghdrFlags ( UCHAR init )       // initialization constructor
   { this->reset () ; this->Flags = init ; }
   void reset ( void )
   { this->Flags = ZERO ; }
   UCHAR setflags ( UCHAR reinit )  // Initialize all flags (requires intelligence)
   { return ( (UCHAR)((this->Flags = UINT(reinit & 0x000000F0))) ) ; }
   UCHAR getflags ( void ) const    // Returns flag byte
   { return (UCHAR)(this->Flags) ; }

   bool unsynch ( void ) const      // Bit7 - unsynch
   { return bool(this->Flags & 0x080) ; }
   void unsynch ( bool set )
   { 
      if ( set ) this->Flags |= 0x080 ;
      else       this->Flags &= 0x07F ;
   }

   bool exthdr ( void ) const       // Bit6 - exthdr
   { return bool(this->Flags & 0x040) ; }
   void exthdr ( bool set )
   { 
      if ( set ) this->Flags |= 0x040 ;
      else       this->Flags &= 0x0BF ;
   }

   bool exper ( void ) const        // Bit5 - exper
   { return bool(this->Flags & 0x020) ; }
   void exper ( bool set )
   { 
      if ( set ) this->Flags |= 0x020 ;
      else       this->Flags &= 0x0DF ;
   }

   bool footer ( void ) const       // Bit4 - footer
   { return bool(this->Flags & 0x010) ; }
   void footer ( bool set )
   { 
      if ( set ) this->Flags |= 0x010 ;
      else       this->Flags &= 0x0EF ;
   }

   //* Data Members *
   UINT Flags ;
} ;

//* Decoded data from an ID3v2.x Tag Header.           *
//* The tag header lives at the top of every MP3 file. *
const UINT taghdrCNT = 10 ;   // size of tag header
const UINT exthdrCNT = 6 ;    // size of extended header (if present)
class id3v2_taghdr
{
   public:
   id3v2_taghdr ( void )
   { this->reset () ; }
   ~id3v2_taghdr ( void ) {}
   void reset ( void )
   {
      this->file_id[0] = this->file_id[1] = this->file_id[2] = this->file_id[3] = 
      this->major = this->rev = ZERO ;
      this->flags.reset() ;
      this->tag_size = ZERO ;
      this->exBytes = this->padBytes = this->tagCRC = ZERO ;
      this->exFlags = ZERO ;
      this->crcFlag = false ;
   }

   void setHeader ( void )
   {
      this->file_id[0] = 'I' ;
      this->file_id[1] = 'D' ;
      this->file_id[2] = '3' ;
      this->file_id[3] = NULLCHAR ;
      this->major      = 3 ;     // id3v2.3.0
      this->rev        = 0 ;
   }

   //* Convert a 4-byte sequence (MSB at offset 0) to an integer value. *
   //* Integer data are stored MSBit first in the bytes a,d MSByte first*
   //* in multi-byte integers, i.e. big-endian.                         *
   //* Programmer's Note: This clunky construct avoids the C library's  *
   //* "helpful" automatic sign extension.                              *
   int intConv ( const UCHAR* ucp ) const
   {
      int i =    (UINT)ucp[3] 
              | ((UINT)ucp[2] << 8)
              | ((UINT)ucp[1] << 16)
              | ((UINT)ucp[0] << 24) ;
      return i ;
   }

   //* Convert a 32-bit integer into an 4-byte, big-endian binary stream *
   int intConv ( int val, UCHAR* ucp ) const
   {
      int i = ZERO ;
      ucp[i++] = (UCHAR)((val >> 24) & 0x000000FF) ;
      ucp[i++] = (UCHAR)((val >> 16) & 0x000000FF) ;
      ucp[i++] = (UCHAR)((val >>  8) & 0x000000FF) ;
      ucp[i++] = (UCHAR)((val      ) & 0x000000FF) ;
      return i ;
   }

   //* Convert a 2-byte sequence (MSB at offset 0) to a short integer value. *
   short intConv16 ( const UCHAR* ucp ) const
   {
      short s =    (short)ucp[1]
                | ((short)ucp[0] << 8) ;
      return s ;
   }

   //* Convert a 16-bit integer into an 2-byte, big-endian binary stream *
   int intConv ( short val, UCHAR* ucp ) const
   {
      int i = ZERO ;
      ucp[i++] = (UCHAR)((val >>  8) & 0x00FF) ;
      ucp[i++] = (UCHAR)((val      ) & 0x00FF) ;
      return i ;
   }

   //* Decode the formatted tag size, a 28-bit value.         *
   //* Four(4) raw hex bytes with bit7 of each byte reset.    *
   //* Example: $00 00 02 01 == 257 byte (256 + 1) tag length.*
   int decodeTagSize ( const char* rawPtr )
   {
      this->tag_size = ZERO ;
      if ( !(rawPtr[0] & 0x80) && !(rawPtr[1] & 0x80) &&
           !(rawPtr[2] & 0x80) && !(rawPtr[3] & 0x80) )
      {
         this->tag_size =    int(rawPtr[3])
                          + (int(rawPtr[2]) << 7)
                          + (int(rawPtr[1]) << 14)
                          + (int(rawPtr[0]) << 21) ;
      }
      return this->tag_size ;
   }

   //* Convert the integer tag size (byte count) to a byte stream.*
   int encodeTagSize ( char* obuff ) const
   {
      const int bMASK = 0x0000007F ;
      short indx = ZERO ;
      obuff[indx++] = (this->tag_size >> 21) & bMASK ;
      obuff[indx++] = (this->tag_size >> 14) & bMASK ;
      obuff[indx++] = (this->tag_size >>  7) & bMASK ;
      obuff[indx++] = this->tag_size & bMASK ;
      return indx ;
   }  //* End encodeTagSize() *

   char  file_id[4] ;      // S/B "ID3", if not, then no metadata
   char  major ;           // ID3 major version (hex)
   char  rev ;             // ID3 revision (hex)
   taghdrFlags flags ;     // flag byte

   //* Size of the tag record EXCLUDING the 10-byte tag header *
   int   tag_size ;

   //* Extended Header (if present) *
   int   exBytes ;         // (32-bit int) (indicated header size excludes this value)
                           // currently: six(6) bytes without CRC or
                           //            ten(10) bytes with CRC 
   int   padBytes ;        // bytes of reserved tag space i.e. the padding
   int   tagCRC ;          // 32-bit CRC value (if CRC used)
                           // Note: CRC is calculated over the range from AFTER the 
                           // extended header to the beginning of the padding i.e.
                           // only the actual frame data (not headers and not padding)
   short exFlags ;         // (all except MSB currently unused)
   bool  crcFlag ;         // true if CRC is used
   bool  spare ;           // (unused)

} ;   // id3v2_taghdr

//* From the ID3v2.4 specification: section 3.4:                               *
//* "To speed up the process of locating an ID3v2 tag when searching from      *
//*  the end of a file, a footer can be added to the tag. It is REQUIRED       *
//*  to add a footer to an appended tag, i.e. a tag located after all          *
//*  audio data. The footer is a copy of the header, but with a different      *
//*  identifier."                                                              *
//*                  ID3v2 identifier           "3DI"                          *
//*                  ID3v2 version              $04 00                         *
//*                  ID3v2 flags                %abcd0000                      *
//*                  ID3v2 size             4 * %0xxxxxxx                      *
//*  a) The size of the footer is the same as the header: taghdrCNT (10 bytes).*
//*  a) A tag footer is the same as a tag header except with a different       *
//*     identifier ("3DI" vs. "ID3").                                          *
//*  b) The 'size' is the same 28-bit format as the header                     *
//*  c) There must be no tag padding if a footer is present.                   *
//*  d) If a "SEEK" frame is found in the prepended tag, then use it to scan   *
//*     for additional tag(s).                                                 *
//*  e) If the continuation flag is set in the extended header, it indicates   *
//*     a continuation of the tag data.                                        *
//*                                                                            *
//*                                                                            *
//*                                                                            *

class id3v2_tagfooter
{
   public:
   ~id3v2_tagfooter ( void ) {}
   id3v2_tagfooter ( void )
   {
      this->reset() ;
   }
   void reset ( void )
   {
      this->foot_id[0] = this->foot_id[1] = this->foot_id[2] = this->foot_id[3] = 
      this->major = this->rev = /*this->flags = */ZERO ;
      this->flags.reset() ;
      this->tag_size = ZERO ;
   }

   void setFootID ( void )
   {
      this->foot_id[0] = '3' ;
      this->foot_id[1] = 'D' ;
      this->foot_id[2] = 'I' ;
      this->foot_id[3] = NULLCHAR ;
   }

   //* Decode the formatted tag size. *
   int decodeTagSize ( const char* rawPtr )
   {
      this->tag_size = ZERO ;
      if ( !(rawPtr[0] & 0x80) && !(rawPtr[1] & 0x80) &&
           !(rawPtr[2] & 0x80) && !(rawPtr[3] & 0x80) )
      {
         this->tag_size =    int(rawPtr[3])
                          + (int(rawPtr[2]) << 7)
                          + (int(rawPtr[1]) << 14)
                          + (int(rawPtr[0]) << 21) ;
      }
      return this->tag_size ;
   }

   //* Convert the integer tag size (byte count) to a byte stream.*
   int encodeTagSize ( char* obuff ) const
   {
      const int bMASK = 0x0000007F ;
      short indx = ZERO ;
      obuff[indx++] = (this->tag_size >> 21) & bMASK ;
      obuff[indx++] = (this->tag_size >> 14) & bMASK ;
      obuff[indx++] = (this->tag_size >>  7) & bMASK ;
      obuff[indx++] = this->tag_size & bMASK ;
      return indx ;
   }  //* End encodeTagSize() *

   char  foot_id[4] ;      // S/B "3DI"
   char  major ;           // ID3 major version (hex)
   char  rev ;             // ID3 revision (hex)
   taghdrFlags flags ;     // flag byte

   //* Size of the tag record EXCLUDING the 10-byte tag header *
   //* -- four(4) raw hex bytes with bit7 of each byte ignored.*
   //*    Thus $00 00 02 01 == 257 byte (256 + 1) tag length.  *
   //* -- This value is the DECODED tag size.                  *
   int   tag_size ;
} ;   // id3v2_tagfooter

//* From the ID3v2.4 specification: section 4.15                               *
//* Data contained in an 'APIC' image frame.                                   *
//* <Header for 'Attached picture', ID: "APIC">                                *
//* Text encoding   $xx                                                        *
//* MIME type       <text string> $00                                          *
//* Picture type    $xx                                                        *
//* Description     <text string according to encoding> $00 (00)               *
//* Picture data    <binary data>                                              *
const short imgMAX_TYPE = 64 ;         // Max bytes in type strings
const short imgMAX_DESC = 256 ;        // Max bytes in description string

class id3v2_image
{
   public:
   ~id3v2_image ( void ) {}
   id3v2_image ( void )
   {
      this->reset() ;
   }
   void reset ( void )
   {
      //* NOTE: If 'picPath' points to a temp file, *
      //*       the reset orphans the file.         *
      this->picPath[0] = this->txtDesc[0] = 
      this->mimType[0] = this->picExpl[0] = NULLCHAR ;
      this->picSize = ZERO ;
      this->picType = 0x00 ;           // default to "Other"
      this->encoding = ENCODE_ASCII ;  // default to ASCII
   }

   //* Convert raw byte data to gString (UTF-8) format.           *
   //*                                                            *
   //* Programmer's Note: For the image description, we take a    *
   //* shortcut in decoding ISO8859-1 text. If the data contain   *
   //* "Latin-1" extension characters, we will incorrectly decode *
   //* them because we assume pure ASCII or UTF-8 encoding.       *
   //* For full decoding of ISO8859-1 characters, see the         *
   //* id3v2_framehdr class definition.                           *
   //*                                                            *
   //* Input  : rawPtr  : pointer to source byte stream           *
   //*                    Note that string _should be_ null       *
   //*                    terminated, but we verify.              *
   //*          enc     : contains the text encoding type.        *
   //*                                                            *
   //* Return: number of wchar_t characters converted             *
   //*         or (-1) if conversion error                        *
   short txtConv ( const char* rawPtr, TextEncode enc )
   {
      short convChars = ZERO ;         // return value

      //* Range check and set text encoding *
      if ( (enc == ENCODE_ASCII) || (enc == ENCODE_UTF16) || 
           (enc == ENCODE_UTF16BE) || (enc == ENCODE_UTF8) )
         this->encoding = enc ;
      else
         this->encoding = ENCODE_UTF8 ;

      //* If ASCII encoding was specified, verify that the *
      //* data are actually ASCII. If not, assume UTF-8.   *
      if ( this->encoding == ENCODE_ASCII )
      {
         gString gs( rawPtr, imgMAX_DESC ) ;
         if ( !(gs.isASCII()) )
            this->encoding = ENCODE_UTF8 ;
      }

      if ( (this->encoding == ENCODE_UTF16) || (this->encoding == ENCODE_UTF16BE) )
      {
         // Programmer's Note: If the text encoding byte is wrong or if the BOM
         // is missing or otherwise invalid, we may produce garbage output, but
         // we rely on the presence of the null terminator to stop the madness.
         int ti = ZERO ;               // source index
         gString gs ;                  // temp buffer
         UINT msUnit, lsUnit,          // for converting unit pairs
              cx ;                     // undecoded 16-bit character
         bool big_endian = true ;      // assume big-endian input
         if ( this->encoding == ENCODE_UTF16 )
         {
            //* Check the byte-order mark.        *
            //* If the BOM was stored backward,   *
            //* then it is little-endian encoding.*
            if ( rawPtr[0] == UTF_LSB )
               big_endian = false ;
            if ( (rawPtr[0] == UTF_MSB || rawPtr[0] == UTF_LSB) &&
                 (rawPtr[1] == UTF_MSB || rawPtr[1] == UTF_LSB) )
               ti += 2 ;
         }
         do
         {
            if ( big_endian )
            {
               cx = ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
               cx |= (UINT(rawPtr[ti++]) & 0x000000FF) ;
            }
            else
            {
               cx = (UINT(rawPtr[ti++]) & 0x000000FF) ;
               cx |= ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
            }

            //* If the character is fully represented by 16 bits *
            //* (most characters are)                            *
            if ( ((cx >= ZERO) && (cx < usxMSU16)) || 
                 ((cx >= usxHIGH16) && (cx <= usxMAX16)) )
            {
               if ( cx != ZERO )    // add the chararacter to output buffer
                  gs.append( cx ) ;
            }
            //* Character is represented by 32 bits    *
            //* (20 bits actually used), 16 MSBs first.*
            else
            {
               msUnit = cx ;
               lsUnit = ZERO ;
               if ( big_endian )
               {
                  lsUnit = ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
                  lsUnit |= (UINT(rawPtr[ti++]) & 0x000000FF) ;
               }
               else     // (little-endian)
               {
                  lsUnit = (UINT(rawPtr[ti++]) & 0x000000FF) ;
                  lsUnit |= ((UINT(rawPtr[ti++]) << 8) & 0x0000FF00) ;
               }
   
               //* Validate the range *
               if ( ((msUnit >= usxMSU16) && (msUnit <= usxMSU16e))
                    &&
                    ((lsUnit >= usxLSU16) && (msUnit <= usxLSU16e)) )
               {
                  cx = usxMIN20 + ((msUnit & usxMASK10) << 10) ;
                  cx |= (lsUnit & usxMASK10) ;
               }
               else                 // invalid UTF-16 codepoint
                  cx = L'?' ;
               if ( cx != ZERO )    // add the chararacter to output buffer
                  gs.append( cx ) ;
            }
         }
         while ( cx != ZERO ) ;
         convChars = gs.gschars() ;
         gs.copy( this->txtDesc, imgMAX_DESC ) ;
      }

      else if ( (this->encoding == ENCODE_ASCII) || (this->encoding == ENCODE_UTF8) )
      {
         gString gs( rawPtr, imgMAX_DESC ) ;
         convChars = gs.gschars() ;
         gs.copy( this->txtDesc, imgMAX_DESC ) ;
      }
      return convChars ;
   }  //* End txtConv() *

   //* Public Method.                                             *
   //* Encode the wchar_t (wide text) array into a form that MP3  *
   //* text frames can understand.                                *
   //*  Note that we always include the NULL terminator to the    *
   //*  text data. The standard neither demands nor forbids it.   *
   //*                                                            *
   //* Input  : enc   : inticates the type of encoding:           *
   //*                   a) ENCODE_ASCII                          *
   //*                   b) ENCODE_UTF8                           *
   //*                   c) ENCODE_UTF16 (big-endian with BOM)    *
   //*                   d) ENCODE_UTF16BE (big-endian, no BOM)   *
   //*          fData : receives the encoded data                 *
   //*          src   : wchar_t-encoded source text data          *
   //*                                                            *
   //* Returns: number of bytes in formatted output               *
   //*          or ZERO if encoding error                         *
   short txtEncode ( TextEncode enc, char* fData, const gString& src )
   {
      gString srcx = src ;
      srcx.limitChars( imgMAX_DESC ) ; // standard limits size to 64 "characters"
      short fIndex = ZERO ;

      //* Caller's choice of encoding (validate) *
      if ( !(enc == ENCODE_ASCII)   && !(enc == ENCODE_UTF16) && 
           !(enc == ENCODE_UTF16BE) && !(enc == ENCODE_UTF8) )
         enc = ENCODE_UTF16 ;

      if ( (this->encoding == ENCODE_ASCII) || (this->encoding == ENCODE_UTF8) )
      {
         srcx.copy( &fData[fIndex], srcx.utfbytes() ) ;
         fIndex = srcx.utfbytes() ;
      }
      else
      {
         if ( enc == ENCODE_UTF16 )
         {
            fData[fIndex++] = (char)0x0FE ;     // MSB of Byte-Order-Mark
            fData[fIndex++] = (char)0x0FF ;     // LSB of Byte-Order-Mark
         }
         short wcnt ;   // number of source chars (incl. null)
         const wchar_t *wstr = srcx.gstr( wcnt ) ;
         UINT cx ;      // source character

         for ( short wIndex = ZERO ; wIndex < wcnt ; ++wIndex )
         {
            cx = wstr[wIndex] ;     // get a character from the input stream

            //* If the character can be encoded with a single, 16-bit value *
            if ( ((cx >= ZERO) && (cx < usxMSU16)) || 
                 ((cx >= usxHIGH16) && (cx <= usxMAX16)) )
            {  //* Encode the bytes in big-endian order *
               fData[fIndex++] = (cx >> 8) & 0x000000FF ;
               fData[fIndex++] = cx & 0x000000FF ;
            }

            //* Else, character requires a pair of 16-bit values *
            else
            {
               UINT msUnit = ZERO, lsUnit = ZERO ;
               if ( (cx >= usxMIN20) && (cx <= usxMAX20) )
               {
                  msUnit = ((cx - usxMIN20) >> 10) | usxMSU16 ;
                  lsUnit = (cx & usxMASK10) | usxLSU16 ;
                  fData[fIndex++] = (msUnit >> 8) & 0x000000FF ;
                  fData[fIndex++] = msUnit & 0x000000FF ;
                  fData[fIndex++] = (lsUnit >> 8) & 0x000000FF ;
                  fData[fIndex++] = lsUnit & 0x000000FF ;
               }
               //* Character cannot be encoded as UTF-16. *
               //* Encode a question mark character.      *
               else
               {
                  cx = L'?' ;
                  fData[fIndex++] = (cx >> 8) & 0x000000FF ;
                  fData[fIndex++] = cx & 0x000000FF ;
               }
            }
         }     // for(;;)
      }
      return fIndex ;

   }  // txtEncode()

   //* Convert the image info to a byte stream.      *
   //* Note that caller will append the binary image *
   //* data at the returned insertion point.         *
   short encode ( char* obuff )
   {
      short oindx = ZERO ;
      gString gs( this->txtDesc ) ;
      this->encoding = (gs.isASCII()) ? ENCODE_ASCII : ENCODE_UTF16 ;
      obuff[oindx++] = this->encoding ;
      for ( short i = ZERO ; i < imgMAX_TYPE ; ++i )
      {
         obuff[oindx++] = this->mimType[i] ;
         if ( this->mimType[i] == NULLCHAR )
            break ;
      }
      obuff[oindx++] = this->picType ;
      oindx += this->txtEncode( this->encoding, &obuff[oindx], gs ) ;

      return oindx ;
   }

   //* Encode the image header data into a byte stream.*

   //***********************
   //* Public Data Members *
   //***********************
   char picPath[gsMAXBYTES] ;       // path/filename of external image file
   char txtDesc[imgMAX_DESC] ;      // text description of image (optional)
   char mimType[imgMAX_TYPE] ;      // MIME type: generally "image/jpg" or 
                                    //   "image/png" or "image/" (default)
   char picExpl[imgMAX_DESC] ;      // explanation of 'picType'. Note: Array of
                                    // pre-defined description strings: pType[][x]
   int  picSize ;                   // image size in bytes
   UCHAR picType ;                  // picture type code
   TextEncode encoding ;            // text encoding (member of enum TextEncode)

} ;   // id3v2_image

const char* const POP_TAG = "POPM" ;   // Tag header code for a Popularimeter.
const char* const CNT_TAG = "PCNT" ;   // Tag header code for a Play Counter.

//* Tag data for the POPM 'Popularimeter' and PCNT 'Play Counter'
class popMeter
{
   public:
   popMeter ( void )
   {
      this->reset () ;
   }
   void reset ( void )
   {
      this->popEmail[0] = NULLCHAR ;
      this->popCount = this->playCount = ZERO ;
      this->popStar = ZERO ;
      this->popdata = this->cntdata = this->pcMod = false ;
   }

   //* Convert a 4-byte sequence (MSB at offset 0) to an integer value. *
   //* Programmer's Note: This clunky construct avoids the C library's  *
   //* "helpful" automatic sign extension.                              *
   UINT intConv ( const UCHAR* ucp ) const
   {
      UINT i =    (UINT)ucp[3] 
               | ((UINT)ucp[2] << 8)
               | ((UINT)ucp[1] << 16)
               | ((UINT)ucp[0] << 24) ;
      return i ;
   }

   //* Convert a 32-bit integer into a big-endian byte stream.    *
   //* (used for both 'frame-size' and 'decomp')                  *
   short intConv ( UINT uval, char* obuff ) const
   {
      const int bMASK = 0x000000FF ;
      short indx = ZERO ;     // return value

      obuff[indx++] = (char)((uval >> 24) & bMASK) ;
      obuff[indx++] = (char)((uval >> 16) & bMASK) ;
      obuff[indx++] = (char)((uval >>  8) & bMASK) ;
      obuff[indx++] = (char)(uval & bMASK) ;

      return indx ;
   }

   char  popEmail[gsMAXBYTES] ;  // User's email address, etc.
   UINT  popCount ;     // Popularimeter (personal) play counter
   UINT  playCount ;    // Play-count counter
   UCHAR popStar ;      // Popularimeter value (star rating)
   bool  popdata ;      // 'true' if POPM data read from source file OR entered by user
   bool  cntdata ;      // 'true' if PCNT data read from source file OR entered by user
   bool  pcMod ;        // 'true' if user has modified the data
} ;   // popMeter

//*****************************
//* Local, non-member methods *
//*****************************
static short mptDecodeFrameHeader ( ifstream& ifs, id3v2_framehdr& fhdr, 
                                    id3v2_tagfooter& tftr ) ;
static UINT  mptDecodeTextFrame ( ifstream& ifs, id3v2_framehdr& fhdr, 
                                  gString& gsOut, TextEncode& enc ) ;
static UINT  mptDecodeImageFrame ( ifstream& ifs, const id3v2_framehdr& fhdr, 
                                  id3v2_image& imgData ) ;
static UINT  mptDecodePopFrame ( ifstream& ifs, const id3v2_framehdr& fhdr, 
                                 popMeter& popData ) ;



//******************************************************************************
//**                Definitions for Ogg/Vorbis audio files                    **
//******************************************************************************
//* Notes on Ogg/Vorbis formatting:                                            *
//* 1) Page Header (https://xiph.org/vorbis/doc/framing.html)                  *
//*     00-03   "OggS"                                                         *
//*     04      stream structure version (always 00h)                          *
//*     05      header-type flags:                                             *
//*             0x01: unset = fresh packet                                     *
//*                     set = continued packet                                 *
//*             0x02: unset = not first page of logical bitstream              *
//*                     set = first page of logical bitstream (bos)            *
//*             0x04: unset = not last page of logical bitstream               *
//*                     set = last page of logical bitstream (eos)             *
//*             (remaining bits reserved)                                      *
//*     06-13   absolute granule position (little-endian, LSByte first)        *
//*     14-17   stream serial number (little-endian)                           *
//*     18-21   page sequence number (little-endian)                           *
//*     22-25   page checksum (little-endian)                                  *
//*     26      page segment count (0-255 segments)                            *
//*     27 ...  segment table, one byte for each segment, i.e. table is        *
//*             page-segment-count bytes long                                  *
//*     -- Page size is the size of the header plus the size of the segment    *
//*        table plus the size of the data segments described in the segment   *
//*        table.                                                              *
//*     -- Length of last segment is always < 255 bytes.                       *
//*                                                                            *
//*  2) The first page of an Ogg stream contains the Identification header.    *
//*       (https://xiph.org/vorbis/doc/Vorbis_I_spec.html#x1-620004.2.1)       *
//*     This page is exactly 58 bytes in length.                               *
//*     00-03   "OggS"                                                         *
//*     04      stream structure version (always 00h)                          *
//*     05      header type (always 02h i.e. begin stream)                     *
//*     06-13   granpos (always zero)                                          *
//*     14-17   stream serial number                                           *
//*     18-21   page sequence number (zero)                                    *
//*     22-25   page checksum                                                  *
//*     26      segment count (01h)                                            *
//*     27      segment table (number of bytes in ID header: 1Eh)              *
//*     28      packet type (01h for Identification header)                    *
//*     29-34   'v' 'o' 'r' 'b' 'i' 's'                                        *
//*     35-38   vorbis version (00h for Vorbis I)                              *
//*     39      audio channels                                                 *
//*     40-43   sample rate                                                    *
//*     44-47   bitrate maximum                                                *
//*     48-51   bitrate nominal                                                *
//*     52-55   bitrate minimum                                                *
//*     56      blocksize 0 (top 4 bits)                                       *
//*             blocksize 1 (bottom 4 bits)                                    *
//*     57      framing flag (always 80h)                                      *
//*                                                                            *
//*                                                                            *
//******************************************************************************

const char* Ogg_Tag = "OggS" ;      // Tag indicating an Ogg-format file
const int OGG_TAG_BYTES = 4 ;       // Length of Ogg_Tag
const int OGG_PAGE1_BYTES = 58 ;    // Length of first Ogg page
const int OGG_PAGEHDR_BYTES = 27 ;  // Minimum number of bytes in a page header
                                    // 0-255 'segments' follow
const char* const Ogg_PktHdr = "vorbis" ; // Packet Header ID
const int OGG_PKTHDR_BYTES = 7 ;    // Length of packet header (see Ogg_Packet)
const int OGG_SEGSIZE = 0x0FF ;     // Maximum byte count in segment table entry (255)
                                       // AND maximum number of segments
const UCHAR FRAMING_BIT = 0x01 ;    // The framing bit (byte) at the end of a pac
typedef u_int32_t UINT32 ;          // alias a 32-bit integer type
//* Embedded image comment name (APIC in MP3 world) *
const short opcBYTES = 22 ;   // bytes in oggPicComment string (not incl. NULLCHAR)
const char *oggPicComment = "METADATA_BLOCK_PICTURE", // standard comment name
           *oggPicCommentOld = "COVERART" ;           // ad-hoc and obsolete comment name

//* Each OGG page begins with a page header.                *
//* For reading the metadata, we don't care about the page  *
//* header contents except to verify that it is valid data. *
class Ogg_PageHdr
{
   public:

   Ogg_PageHdr ( void )
   { this->reset () ; }
   ~Ogg_PageHdr ( void ) {}
   void reset ( void )
   {
      *this->tag = NULLCHAR ;
      this->version = this->ptype = this->segcount = ZERO ;
      this->granpos = ZERO ;
      this->streamser = this->pageseq = this->cksum = this->segBytes = ZERO ;
   }

   //* Convert a 4-byte sequence (LSB at offset 0) to an integer value. *
   int intConv ( const UCHAR* ucp ) const
   {
      int i =    (UINT)ucp[0] 
              | ((UINT)ucp[1] << 8)
              | ((UINT)ucp[2] << 16)
              | ((UINT)ucp[3] << 24) ;
      return i ;
   }

   //* Convert a 32-bit integer into an 4-byte, little-endian binary stream *
   void intConv ( int val, UCHAR* ucp ) const
   {
      ucp[0] = (UCHAR)(val & 0x000000FF) ;
      ucp[1] = (UCHAR)((val >>  8) & 0x000000FF) ;
      ucp[2] = (UCHAR)((val >> 16) & 0x000000FF) ;
      ucp[3] = (UCHAR)((val >> 24) & 0x000000FF) ;
   }

   //* Convert an 8-byte sequence (LSB at offset 0) to an int64 value. *
   long long int intConv64 ( const UCHAR* ucp ) const
   {
      long long int i =
              (UINT64)ucp[0] 
            | ((UINT64)ucp[1] << 8)
            | ((UINT64)ucp[2] << 16)
            | ((UINT64)ucp[3] << 24)
            | ((UINT64)ucp[4] << 32)
            | ((UINT64)ucp[5] << 40)
            | ((UINT64)ucp[6] << 48)
            | ((UINT64)ucp[7] << 56) ;
      return i ;
   }

   //* Convert a 64-bit integer into an 8-byte, little-endian binary stream *
   void intConv64 ( long long int val64, UCHAR* ucp ) const
   {
      ucp[0] = (UCHAR)(val64 & 0x00000000000000FF) ;
      ucp[1] = (UCHAR)((val64 >>  8) & 0x00000000000000FF) ;
      ucp[2] = (UCHAR)((val64 >> 16) & 0x00000000000000FF) ;
      ucp[3] = (UCHAR)((val64 >> 24) & 0x00000000000000FF) ;
      ucp[4] = (UCHAR)((val64 >> 32) & 0x00000000000000FF) ;
      ucp[5] = (UCHAR)((val64 >> 40) & 0x00000000000000FF) ;
      ucp[6] = (UCHAR)((val64 >> 48) & 0x00000000000000FF) ;
      ucp[7] = (UCHAR)((val64 >> 56) & 0x00000000000000FF) ;
   }

   char   tag[5] ;         // should be 'OggS'
   UCHAR  version ;        // should be 0x00
   UCHAR  ptype ;          // may be 0x01, 0x02, 0x04 or an OR value from these
   UCHAR  segcount ;       // number of segments in page (range: 0-255)
   UINT64 granpos ;        // encoded data used by decoder (8 bytes)
   UINT   streamser ;      // uniquely identifies the stream (4 bytes)
   UINT   pageseq ;        // page sequence number (4 bytes)
   UINT   cksum ;          // CRC checksum (4 bytes)
   UINT   segBytes ;       // total bytes in page segments
} ;

//* Each of the three headers begins with the sequence:  *
//*             nn 'v' 'o' 'r' 'b' 'i' 's'               *
//* where 'nn' is the packet type:                       *
//*             01h==Identification header               *
//*             03h==Comment header                      *
//*             05h==Setup header                        *
class Ogg_Packet
{
   public:

   Ogg_Packet ( void )
   { this->reset () ; }
   ~Ogg_Packet ( void ) {}
   void reset ( void )
   {
      this->ptype = ZERO ;
      this->pid[0] = this->pid[1] = this->pid[2] = this->pid[3] = 
      this->pid[4] = this->pid[5] = this->pid[6] = NULLCHAR ;
   }
   void populate ( UINT page )
   {
      this->reset () ;
      this->ptype = page == 1 ? 1 : page == 2 ? 3 : 5 ;
      this->pid[0] = 'v' ; this->pid[1] = 'o' ; this->pid[2] = 'r' ; 
      this->pid[3] = 'b' ; this->pid[4] = 'i' ; this->pid[5] = 's' ; 
   }

   UCHAR ptype ;     // packet type
   char  pid[7] ;    // packet identifier
} ;

class Ogg_ID
{
   public:

   Ogg_ID ( void )
   { this->reset () ; }
   ~Ogg_ID ( void ) {}
   void reset ( void )
   {
      this->version = this->blksize0 = this->blksize1 = ZERO ;
      this->samprate = this->bitratemax = this->bitratenom = this->bitratemin = ZERO ;
      this->channels = ZERO ;
      this->frameflag = false ;
   }

   //* Convert a 4-byte sequence (LSB at offset 0) to an integer value. *
   //* Programmer's Note: This clunky construct avoids the C library's  *
   //* "helpful" automatic sign extension.                              *
   int intConv ( const UCHAR* ucp )
   {
      int i =    (UINT)ucp[0] 
              | ((UINT)ucp[1] << 8)
              | ((UINT)ucp[2] << 16)
              | ((UINT)ucp[3] << 24) ;
      return i ;
   }

   UINT  version ;      // vorbis I version (must be zero)
   int   samprate ;     // audio sample rate (must be > zero)
   int   bitratemax ;   // bitrate maximum
   int   bitratenom ;   // bitrate nominal
   int   bitratemin ;   // bitrate minimum
   UINT  blksize0 ;     // exponent for power of 2 (4 bits only)
   UINT  blksize1 ;     // exponent for power of 2 (4 bits only)
   UCHAR channels ;     // number of audio channels (must be > zero)
   bool  frameflag ;    // framing flag (true==GOOD, false==ERROR)
} ;
const int OGG_ID_BYTES = sizeof(Ogg_ID) ;

//* Individual comment *
class Ogg_Comment
{
   public:

   Ogg_Comment ( void )       // constructor
   { this->reset () ; }
   ~Ogg_Comment ( void ) {}   // destructor

   void reset ( void )
   {
      this->clen = ZERO ;
      for ( short i = ZERO ; i < gsMAXBYTES ; ++i )
         this->ctxt[i] = NULLCHAR ;
   }

   //* Convert a 4-byte sequence (LSB at offset 0) to an integer value. *
   //* Programmer's Note: This clunky construct avoids the C library's  *
   //* "helpful" automatic sign extension.                              *
   int intConv ( const UCHAR* ucp )
   {
      int i =    (UINT)ucp[0] 
              | ((UINT)ucp[1] << 8)
              | ((UINT)ucp[2] << 16)
              | ((UINT)ucp[3] << 24) ;
      return i ;
   }

   //* Convert a 32-bit unsigned integer to a little-endian byte stream.      *
   //* NOTE: In a 32-bit or 64-bit little-endian system, 'u' and 'ucp'  will  *
   //*       be the same, but we can't make assumptions about user's hardware.*
   void intConv ( UINT u, UCHAR* ucp ) const
   {
      ucp[0] = (UCHAR)(u & 0x000000FF) ;
      ucp[1] = (UCHAR)((u >>  8) & 0x000000FF) ;
      ucp[2] = (UCHAR)((u >> 16) & 0x000000FF) ;
      ucp[3] = (UCHAR)((u >> 24) & 0x000000FF) ;
   }

   UINT  clen ;               // number of bytes in comment (excl. NULLCHAR)
   char  ctxt[gsMAXBYTES] ;   // null terminated text comment
   // NOTE: The specification says that a comment may have a length of up to
   //       (2 to the 32nd) - 1. However it also suggests that a comment be
   //       no more than a short paragraph. For this reason, we discard any
   //       part of the comment beyond four(4) kbytes.
   //       The specification also allows comments to extend into the next 
   //       logical page (page 3); however we make the assumption that this
   //       will never happen because page 2 can hold 256-squared bytes. 
   //       If comments DO extend into page 3, then the comment output from
   //       page 3 will be corrupted.
   //       This will become a problem if embedded images are encoded into 
   //       the comment header. Stay tuned for an update from Taggit....
} ;

//*****************************
//* Local, non-member methods *
//*****************************
static bool  oggDecodeImageVector ( const char* ibuff, Ogg_Comment& ogc ) ;


//** - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  **
//** - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  **

//*************************
//*    MediafileTarget    *
//*************************
//******************************************************************************
//* Compare the filename extension of the specified file against the list of   *
//* supported media files from which we can extract the file's metadata.       *
//*                                                                            *
//*                                                                            *
//* Input  : trgPath  : filespec of file to be tested.                         *
//*                                                                            *
//* Returns: member of enum MediafileType                                      *
//******************************************************************************

MediafileType FileDlg::MediafileTarget ( const gString& trgPath )
{
   gString trgExt ;                    // filename extension
   this->fmPtr->ExtractExtension ( trgExt, trgPath ) ;
   MediafileType mftype = mftNONE ;    // return value

   if ( (trgExt.compare( L"mp3" )) == ZERO )
      mftype = mftMP3 ;
   else if (    ((trgExt.compare( L"oga" )) == ZERO) 
             || ((trgExt.compare( L"ogg" )) == ZERO) )
      mftype = mftOGG ;
   else if ( (trgExt.compare( L"m4a" )) == ZERO )
      mftype = mftM4A ;

   return mftype ;

}  //* End MediafileTarget() *

//*************************
//*  ExtractMetadata_MP3  *
//*************************
//******************************************************************************
//* Read the media file and write the metadata to the temp file.               *
//* Note: Certain data are not displayed:                                      *
//*       a) images                                                            *
//*       b) compressed data                                                   *
//*       c) encrypted data                                                    *
//*       d) non-text frames are displayed only in verbose mode                *
//*                                                                            *
//* Input  : ofs    : open output stream to temporary file                     *
//*          srcPath: filespec of media file to be read                        *
//*          vebose : (optional, 'false' by default)                           *
//*                   if 'false', display all text-frame records               *
//*                   if 'true', ALSO display extended technical data and      *
//*                              non-text frames (used primarily for debugging)*
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void FileDlg::ExtractMetadata_MP3 ( ofstream& ofs, const gString& srcPath, bool verbose )
{
   #define DEBUG_EMDATA (0)      // Set to non-zero for debugging only

   const char* ImgTemplate = "  Image: %S bytes, MIME:%s, Type:%s, Desc:%s" ;
   const char* PopTemplate = "  Popularity: %hhu of 255, (%u) Message:'%s' Play-count: %S" ;
   // Programmer's Note: There is no practical limit to the number of images 
   // that may be embedded in an audio file; however, we _arbitrariy_ limit
   // the number of images we actually report.
   // (Each instance of the id3v2_image class is about 4800 bytes of data.)
   const short MAX_IMAGES = 16 ;

   gString gsOut, gstmp ;        // text formatting
   id3v2_image* img = new id3v2_image[MAX_IMAGES] ; // image data (dynamic allocation)
   short imgCount = ZERO ;       // image counter
   popMeter popData ;            // Popularimeter and Play Counter data

   #if DEBUG_EMDATA != 0
   const char* tbrTemplate = "** Tag Bytes Remaining: %S\n" ;
   const char* padTemplate = "Uninitialized space (padding): %S\n" ;
   verbose = true ;              // force verbose output
   #endif   // DEBUG_EMDATA

   tnFName fn ;
   this->fmPtr->GetFileStats ( fn, srcPath, true ) ;
   gstmp.formatInt( fn.fBytes, 11, true ) ;

   short fni = srcPath.findlast( fSLASH ) + 1 ;
   ofs << "FILE NAME  : \"" << &srcPath.ustr()[fni] 
       << "\"  (" << gstmp.ustr() << " bytes)\n\n" ;

   //* Open the source file *
   ifstream ifs( srcPath.ustr(), ifstream::in ) ;
   if ( ifs.is_open() )
   {
      char ibuff[gsMAXBYTES + 4] ;  // input buffer
      UINT Tag_eod = ZERO ;         // unread bytes in tag
      id3v2_taghdr thdr ;

      //* Read the file header and verify that we have valid MP3 data.*
      UINT cnt ;                 // for validating header record
      bool valid_hdr = false ;

      //* Read header data from source file *
      ifs.read ( ibuff, taghdrCNT ) ;
      cnt = ifs.gcount() ;

      //* A valid ID3 tag header will be of the form: *
      //*    $49 44 43 yy yy xx zz zz zz zz           *
      //* where yy < FFh, xx==flags, zz < 80h         *
      if ( cnt == taghdrCNT )
      {
         thdr.file_id[0]   = ibuff[0] ;
         thdr.file_id[1]   = ibuff[1] ;
         thdr.file_id[2]   = ibuff[2] ;
         thdr.file_id[3]   = NULLCHAR ;
         thdr.major        = ibuff[3] ;
         thdr.rev          = ibuff[4] ;
         thdr.flags.setflags( ibuff[5] ) ;

         //* Tag size value is a 28-bit value encoded *
         //* in 32 bits: bytes 6, 7, 8, 9.            *
         thdr.decodeTagSize ( &ibuff[6] ) ;

         //* Validate the remainder of the header *
         gString gstmp( thdr.file_id ) ;
         if ( (gstmp.compare( "ID3" )) == ZERO &&
              (thdr.major != 0xFF && thdr.rev != 0xFF) )
            valid_hdr = true ;
      }

      if ( valid_hdr )
      {
         Tag_eod = thdr.tag_size ;     // tag data byte count
         short ttCount = ZERO ;        // number of text tags scanned

         #if DEBUG_EMDATA != 0
         gstmp.formatInt( thdr.tag_size, 11, true ) ;
         bool funsynch = thdr.flags.unsynch(),
              fexthdr  = thdr.flags.exthdr(),
              fexper   = thdr.flags.exper(),
              ffooter  = thdr.flags.footer() ;
         gsOut.compose( "ID3v2 Tag Header:\n"
                        " File ID: %s\n"
                        " Version: 2.%hhX.%hhX\n"
                        " Flags  : %hhd %hhd %hhd %hhd\n"
                        " TagSize: %S\n"
                        " Valid  : true\n",
                        thdr.file_id, &thdr.major, &thdr.rev, 
                        &funsynch, &fexthdr, &fexper, &ffooter, 
                        gstmp.gstr() ) ;
         ofs << gsOut.ustr() << endl ;
         #endif   // DEBUG_EMDATA

         //* If an "extended header" is defined *
         if ( thdr.flags.exthdr() )
         {
            //* MSB of flags set indicates presence of CRC error detection *
            const short CRC_MASK = 0x8000 ;
            short bytesRead = ZERO ;   // return value

            //* Get size of extended header *
            ifs.read ( ibuff, 4 ) ;
            bytesRead += ifs.gcount() ;
            thdr.exBytes = thdr.intConv( (UCHAR*)&ibuff[0] ) ;

            //* Read the extended header *
            ifs.read ( ibuff, thdr.exBytes ) ;
            bytesRead += ifs.gcount() ;

            // Decode the flags (16 bits, big-endian) *
            thdr.exFlags = thdr.intConv16( (UCHAR*)(&ibuff[0]) ) ;
            thdr.crcFlag = (bool)(thdr.exFlags & CRC_MASK ) ;

            //* Decode the size of the padding (32 bits, big-endian) *
            thdr.padBytes = thdr.intConv( (UCHAR*)&ibuff[2] ) ;

            //* Decode the CRC value if present (32 bits, big-endian) *
            if ( thdr.crcFlag )
               thdr.tagCRC = thdr.intConv( (UCHAR*)&ibuff[6] ) ;

            #if DEBUG_EMDATA != 0
            gstmp.formatInt( thdr.padBytes, 11, true ) ;
            gsOut.compose( L"Extended Header:\n"
                            "  exBytes: %d\n"
                            "  exFlags: %-#hb\n"
                            "  crcFlag: %hhd\n"
                            " padBytes: %S\n"
                            "   tagCRC: %08X\n",
                            &thdr.exBytes, &thdr.exFlags, &thdr.crcFlag, 
                            gstmp.gstr(), &thdr.tagCRC ) ;
            ofs << gsOut.ustr() << endl ;
            #endif   // DEBUG_EMDATA

            Tag_eod -= bytesRead ;
         }

         #if DEBUG_EMDATA != 0
         gstmp.formatInt( Tag_eod, 11, true ) ;
         gsOut.compose( tbrTemplate, gstmp.gstr() ) ;
         ofs << gsOut.ustr() << endl ;
         #endif   // DEBUG_EMDATA

         //* Read each frame and report its contents *
         id3v2_framehdr fhdr ;
         id3v2_tagfooter tftr ;
         TextEncode enc ;
         do
         {
            //* Get type of frame and its size *
            Tag_eod -= mptDecodeFrameHeader ( ifs, fhdr, tftr ) ;

            #if DEBUG_EMDATA != 0
            if ( verbose &&
                  ((*fhdr.frame_id >= 'A' && *fhdr.frame_id <= 'Z') ||
                   (*fhdr.frame_id >= '0' && *fhdr.frame_id <= '9')) )
            {
               gstmp.formatInt( fhdr.frame_size, 11, true ) ;
               gsOut.compose( L"Frame %s\n"
                               " size     : %S\n"
                               " flags    : %-#hhb %-#hhb\n",
                               fhdr.frame_id, gstmp.gstr(), 
                               &fhdr.status_flags, &fhdr.encode_flags ) ;
               //* Optional data appended to frame header *
               if ( fhdr.decomp > ZERO )     // decompressed frame-data size
               {
                  gstmp.formatInt( fhdr.decomp, 11, true ) ;
                  gsOut.append( " expanded : %S\n", gstmp.gstr() ) ;
               }
               if ( fhdr.crypto != ZERO )
                  gsOut.append( " crypto   : %hhX\n", &fhdr.crypto ) ;
               if ( fhdr.group_id != ZERO )
                  gsOut.append( " group id : %hhX\n", &fhdr.group_id ) ;

               ofs << gsOut.ustr() ;
            }
            #endif   // DEBUG_EMDATA

            if ( *fhdr.frame_id == 'T' )        // text frames
            {
               //* Get the text-frame text *
               Tag_eod -= mptDecodeTextFrame ( ifs, fhdr, gsOut, enc ) ;

               //* Determine which field the tag belongs to.  *
               //* Walk through the list of field identifiers *
               //* until a match is found, then write the text*
               //* to the corresponding tag field.            *
               bool goodID = false ;
               gstmp = fhdr.frame_id ;
               for ( short i = ZERO ; i < TEXT_FRAMES ; ++i )
               {
                  if ( (gstmp.compare( TextFrameID[i] )) == ZERO )
                  {
                     #if DEBUG_EMDATA != 0
                     if ( verbose )
                     {
                        gsOut.compose( " encoding : %02hhX (%s)", &enc,
                           (enc == ENCODE_ASCII   ? "ASCII"    : 
                            enc == ENCODE_UTF16   ? "UTF-16"   :
                            enc == ENCODE_UTF16BE ? "UTF-16BE" :
                            enc == ENCODE_UTF8    ? "UTF-8"    : "UNKNOWN" ) ) ;
                        ofs << gsOut.ustr() << endl ;
                     }
                     #endif   // DEBUG_EMDATA

                     ofs << TextFrameDesc[i] << gsOut.ustr() << '\n' ;

                     #if DEBUG_EMDATA != 0
                     gstmp.formatInt( Tag_eod, 11, true ) ;
                     gsOut.compose( tbrTemplate, gstmp.gstr() ) ;
                     ofs << gsOut.ustr() << endl ;
                     #endif   // DEBUG_EMDATA

                     ++ttCount ;
                     goodID = true ;
                     break ;
                  }
               }
               if ( ! goodID )
               {  //* Text-frame ID is not recognized.   *
                  //* Ignore the non-standard text frame.*
                  #if DEBUG_EMDATA != 0
                  gsOut.compose( "Warning: Non-standard text-frame ID: '%S'\n"
                                 "         (frame discarded)", gstmp.gstr() ) ;
                  ofs << gsOut.ustr() << endl ;
                  #endif   // DEBUG_EMDATA
               }
            }

            //* Collect image-frame description (Step over the image data.) *
            else if ( (fhdr.frame_id[0] == 'A') && (fhdr.frame_id[1] == 'P') &&
                      (fhdr.frame_id[2] == 'I') && (fhdr.frame_id[3] == 'C') )
            {
               if ( imgCount < MAX_IMAGES )
                  Tag_eod -= mptDecodeImageFrame ( ifs, fhdr, img[imgCount++] ) ;
               else  // scan and discard excess images (see note above)
               {
                  id3v2_image eximg ;
                  Tag_eod -= mptDecodeImageFrame ( ifs, fhdr, eximg ) ;
               }
            }

            //* Scan for POPM (Popularimeter) and PCNT (Play Counter) frames. *
            //* Because only one of each is allowed, we can report it _after_ *
            //* the text-frame and image-frame data.                          *
            else if ( (fhdr.frame_id[0] == 'P') &&
                      (((fhdr.frame_id[1] == 'O') && (fhdr.frame_id[2] == 'P') && 
                        (fhdr.frame_id[3] == 'M'))
                       ||
                       ((fhdr.frame_id[1] == 'C') && (fhdr.frame_id[2] == 'N') && 
                        (fhdr.frame_id[3] == 'T'))) )
            {
               Tag_eod -= mptDecodePopFrame ( ifs, fhdr, popData ) ;
               popData.pcMod = true ;
            }

            //* Non-text frames are not decoded.*
            else if (   (*fhdr.frame_id >= 'A' && *fhdr.frame_id <= 'Z')
                     || (*fhdr.frame_id >= '0' && *fhdr.frame_id <= '9') )
            {
               int loop = fhdr.frame_size / gsMAXBYTES, 
                   remainder = fhdr.frame_size % gsMAXBYTES ;

               //* Read the raw data *
               for ( int i = ZERO ; i < loop ; ++i )
               {
                  ifs.read( ibuff, gsMAXBYTES ) ;
                  Tag_eod -= gsMAXBYTES ;
               }
               if ( remainder > ZERO )
               {
                  ifs.read( ibuff, remainder ) ;
                  Tag_eod -= remainder ;
               }

               #if DEBUG_EMDATA != 0
               if ( verbose )
               {
                  gstmp.formatInt( Tag_eod, 11, true ) ;
                  gsOut.compose( tbrTemplate, gstmp.gstr() ) ;
                  ofs << " (non-text frame: '"
                      << fhdr.frame_id
                      << "' not decoded)\n"
                      << gsOut.ustr() << endl ;
               }
               #endif   // DEBUG_EMDATA
            }
            else
            {
               // Programmer's Note: In real life, padding is routinely added
               // to the tag without bothering to mention it (no extended header).
               // Although this is technically an error, there's not much we 
               // can do about it at this level.
               // Therefore, an invalid frame ID means we're done.
               #if DEBUG_EMDATA != 0
               int padCount = Tag_eod + framehdrCNT ;
               gstmp.formatInt( padCount, 11, true ) ;
               gsOut.compose( padTemplate, gstmp.gstr() ) ;
               ofs << gsOut.ustr() ;
               #endif   // DEBUG_EMDATA

               Tag_eod = ZERO ;     // discard remainder of tag
            }
         }
         while ( Tag_eod > ZERO ) ;

         //* If one or more images were captured *
         if ( imgCount > ZERO )
         {
            for ( short i = ZERO ; i < imgCount ; ++i )
            {
               gstmp.formatInt( img[i].picSize, 13, true ) ;
               gsOut.compose( ImgTemplate, gstmp.gstr(), img[i].mimType, 
                              &img[i].picExpl, img[i].txtDesc ) ;
               ofs << gsOut.ustr() << endl ;
               
            }
         }
         //* If Popularimeter and/or PlayCounter was captured *
         if ( popData.pcMod != false )
         {
            gstmp.formatInt( popData.playCount, 11, true ) ;
            gsOut.compose( PopTemplate, &popData.popStar, &popData.popCount,
                           popData.popEmail, gstmp.gstr() ) ;
            ofs << gsOut.ustr() << endl ;
         }

         //* If a tag footer is defined *
         if ( thdr.flags.footer() )
         {  // Programmer's Note: We do not decode this because it is of no 
            // interest to the user AND because it means we are at the end 
            // of the tag data.
            //mptDecodeTagFooter ( ifs, tftr ) ;
         }

         if ( ttCount == ZERO )
            ofs << "File contains no ID3v2 text frames.\n\n" ;
      }
      else
         ofs << "No ID3v2 metadata available.\n\n" ;
      ifs.close() ;           // close the source file
   }
   else
      ofs << "Sorry, unable to open file for reading.\n" ;
   ofs << endl ;

   if ( img != NULL )
      delete [] img ;   // release the dynamic allocation

   #undef DEBUG_EMDATA

}  //* End ExtractMetadata_MP3() *

//***************************
//*  mptDecodeFrameHeader   *
//***************************
//******************************************************************************
//* Read and decode the MP3 frame header.                                      *
//* A frame header details the kind of data contained in the frame.            *
//*                                                                            *
//*                                                                            *
//* Input  : ifs     : handle for open input stream                            *
//*          fhdr    : (by reference) receives the decoded header record       *
//*          tftr    : (by reference) if a Tag Footer is identified INSTEAD OF *
//*                    a frame header, this object recieves its decoded data   *
//*                                                                            *
//* Returns: number of bytes read from the input stream                        *
//******************************************************************************
//* Notes:                                                                     *
//* 1) For id3v2.4 an optional "Tag Footer" is defined.                        *
//*    a) In order to avoid corrupting it if found, we do a special test of    *
//*       the frame_id.                                                        *
//*    b) If the tag-footer ID is found, then 'tftr' will receive the decoded  *
//*       data. For efficiency, we do not reset the 'tftr' members if input    *
//*       data are not a tag footer. It is the caller's responsibility to      *
//*       notice if the members have been initialized.                         *
//*    c) By lucky chance, framehdrCNT == taghdrCNT, so the input stream       *
//*       remains synchronized.                                                *
//******************************************************************************

static short mptDecodeFrameHeader ( ifstream& ifs, id3v2_framehdr& fhdr, 
                                    id3v2_tagfooter& tftr )
{
   const UCHAR usefulFLAGS = 0xE0 ;
   char ibuff[gsMAXBYTES] ;   // input buffer
   short bytesRead = ZERO,    // return value
         i = ZERO ;           // input index

   fhdr.reset() ;       // clear previous data

   ifs.read ( ibuff, framehdrCNT ) ;   // read the frame header
   bytesRead += ifs.gcount() ;

   //* Test for Tag Footer (see note above) *
   if (    (ibuff[0] == '3')
        && ((ibuff[1] == 'D') || (ibuff[1] == 'd'))
        && ((ibuff[2] == 'I') || (ibuff[2] == 'i')) )
   {
      tftr.reset() ;
      tftr.foot_id[0] = ibuff[i++] ;
      tftr.foot_id[1] = ibuff[i++] ;
      tftr.foot_id[2] = ibuff[i++] ;
      tftr.major       = ibuff[i++] ;
      tftr.rev         = ibuff[i++] ;
      tftr.flags.setflags( ibuff[i++] ) ;
      tftr.decodeTagSize( &ibuff[i] ) ;
   }
   else        // frame header
   {
      fhdr.frame_id[0] = ibuff[i++] ;
      fhdr.frame_id[1] = ibuff[i++] ;
      fhdr.frame_id[2] = ibuff[i++] ;
      fhdr.frame_id[3] = ibuff[i++] ;
      fhdr.frame_size = fhdr.intConv( (UCHAR*)&ibuff[i] ) ;
      i += 4 ;    // step over the integer
      fhdr.status_flags = ibuff[i++] & usefulFLAGS ;
      fhdr.encode_flags = ibuff[i] & usefulFLAGS ;
      fhdr.flag_tag_pres  = (fhdr.status_flags & 0x80) ? true : false ;
      fhdr.flag_file_pres = (fhdr.status_flags & 0x40) ? true : false ;
      fhdr.flag_readonly  = (fhdr.status_flags & 0x20) ? true : false ;
      fhdr.flag_compress  = (fhdr.encode_flags & 0x80) ? true : false ;
      fhdr.flag_encrypt   = (fhdr.encode_flags & 0x40) ? true : false ;
      fhdr.flag_grouped   = (fhdr.encode_flags & 0x20) ? true : false ;

      //* Simple test to verify that frame_id is valid data *
      if ( ((*fhdr.frame_id >= 'A' && *fhdr.frame_id <= 'Z') ||
            (*fhdr.frame_id >= '0' && *fhdr.frame_id <= '9')) )
      {
         // Programmer's Note: The following optional additions are not 
         // included in frame header size, but are included in frame size.
         //* If 'compress' flag, 4 bytes added to header *
         //* indicating de-compressed size.              *
         if ( fhdr.flag_compress )
         {
            ifs.read( ibuff, 4 ) ;
            bytesRead += ifs.gcount() ;
            fhdr.decomp = fhdr.intConv( (UCHAR*)&ibuff[0] ) ;
         }
         //* If 'encrypt' flag,  1 byte added to header *
         //* indicating type of encryption.             *
         if ( fhdr.flag_encrypt )
         {
            ifs.read( ibuff, 1 ) ;
            bytesRead += ifs.gcount() ;
            fhdr.crypto = ibuff[0] ;
         }
         //* If 'grouped' flag,  1 byte added to header indicating group ID.*
         if ( fhdr.flag_grouped )
         {
            ifs.read( ibuff, 1 ) ;
            bytesRead += ifs.gcount() ;
            fhdr.group_id = ibuff[0] ;
         }
      }
   }
   return bytesRead ;

}  //* End mptDecodeFrameHeader() *

//***************************
//*   mptDecodeTextFrame    *
//***************************
//******************************************************************************
//* Read and decode an MP3 text frame.                                         *
//*   1) Read the raw data.                                                    *
//*   2) Convert the raw text to wchar_t (gString).                            *
//*                                                                            *
//*                                                                            *
//* Input  : ifs     : handle for open input stream                            *
//*          fhdr    : (by reference) receives the decoded header record       *
//*          gsOut   : (by reference) receives the decoded text string         *
//*          enc     : (by reference) receives text-encoding code              *
//*                                                                            *
//* Returns: number of bytes read from the input stream                        *
//******************************************************************************
//* Notes:                                                                     *
//* ======                                                                     *
//* A text frame is encoded as:   (see 'enum TextEncode')                      *
//*   a) ISO8859-1 (ASCII + Latin-1 extensions) or                             *
//*   b) UTF-16 with Byte-Order-Mark (BOM)  id3v2.2 and id3v2.3                *
//*      -- big-endian, read as   : 0xFE 0xFF, or                              *
//*      -- little-endian, read as: 0xFF 0xFE).                                *
//*   c) UTF-16BE (no BOM) id3v2.4                                             *
//*   d) UTF-8   id3v2.4                                                       *
//*                                                                            *
//******************************************************************************

static UINT mptDecodeTextFrame ( ifstream& ifs, id3v2_framehdr& fhdr, 
                                 gString& gsOut, TextEncode& enc )
{
   char ibuff[gsMAXBYTES] ;   // input buffer
   UINT bytesRead = ZERO ;    // return value

   if ( fhdr.frame_size < gsMAXBYTES )
   {
      //* Read the raw text *
      ifs.read( ibuff, fhdr.frame_size ) ;
      bytesRead += ifs.gcount() ;
      enc = (TextEncode)ibuff[0] ;
   }
   else
   {
      // Programmer's Note: If text frame is larger than our buffer,
      // (unlikely), we will display a truncated text string.
      // Also, declaring the data captured as less than a full buffer
      // prevents buffer overrun during Unicode conversion.
      ifs.read( ibuff, gsMAXBYTES ) ;
      char ctmp[gsMAXBYTES] ;
      int loop = (fhdr.frame_size - gsMAXBYTES) / gsMAXBYTES, 
          remainder = (fhdr.frame_size - gsMAXBYTES) % gsMAXBYTES ;
      for ( int i = ZERO ; i < loop ; ++i )
      {
         ifs.read( ctmp, gsMAXBYTES ) ;
         if ( i == ZERO )
            enc = (TextEncode)ibuff[0] ;
      }
      if ( remainder > ZERO )
         ifs.read( ctmp, remainder ) ;
      bytesRead += fhdr.frame_size ;
      fhdr.frame_size = gsMAXBYTES - 3 ;
   }

   //* Convert the raw byte stream to text *
   fhdr.txtConv( ibuff, gsOut ) ;

   return bytesRead ;

}  //* End mptDecodeTextFrame() *

//***************************
//*   mptDecodeImageFrame   *
//***************************
//******************************************************************************
//* Read and decode an MP3 image frame.                                        *
//*   1) Read the setup and description data and save it for later.            *
//*   2) Read the binary image data and,                                       *
//*      a) save it to the specified file, or                                  *
//*      b) discard it
//*                                                                            *
//* Input  : ifs     : handle for open input stream                            *
//*          fhdr    : (by reference) receives the decoded header record       *
//*          imgData : (by reference) receives the decoded image data          *
//*                                                                            *
//* Returns: number of bytes read from the input stream                        *
//******************************************************************************

static UINT mptDecodeImageFrame ( ifstream& ifs, const id3v2_framehdr& fhdr, 
                                  id3v2_image& imgData )
{
//   //* "Picture Type" descriptions as defined for id3v2 standard.  *
//   const short PIC_TYPES = 21 ;     // number of descriptions per sub-array
//   const char* const pType[PIC_TYPES] = 
//   {  // English
//     "Other",
//     "32x32 pixels 'file icon' (PNG only)",
//     "Other file icon",
//     "Cover (front)",
//     "Cover (back)",
//     "Leaflet page",
//     "Media (e.g. label side of CD)",
//     "Lead artist/lead performer/soloist",
//     "Artist/performer",
//     "Conductor",
//     "Band/Orchestra",
//     "Composer",
//     "Lyricist/text writer",
//     "Recording Location",
//     "During recording",
//     "During performance",
//     "Movie/video screen capture",
//     "A bright coloured fish",
//     "Illustration",
//     "Band/artist logotype",
//     "Publisher/Studio logotype",
//   } ;

   char ibuff[gsMAXBYTES] ;   // input buffer
   gString gstmp ;            // data formatting
   ofstream ofs ;             // temp-file access
   bool toFile = (*imgData.picPath != NULLCHAR ? true : false) ;
   UINT bytesRead = ZERO ;    // return value

   //* Read and decode the text data *
   ifs.read( ibuff, 1 ) ;           // text-encoding indicator
   bytesRead += ifs.gcount() ;
   imgData.encoding = (TextEncode)(char)ibuff[0] ;

   for ( short i = ZERO ; i < imgMAX_TYPE ; ++i )   // MIME type
   {
      ifs.read( &ibuff[i], 1 ) ;
      bytesRead += ifs.gcount() ;
      if ( ibuff[i] == NULLCHAR )
         break ;
   }
   gstmp = ibuff ;
   gstmp.copy( imgData.mimType, imgMAX_TYPE ) ;

   ifs.read( ibuff, 1 ) ;           // picture-type indicator
   bytesRead += ifs.gcount() ;
   imgData.picType = ibuff[0] ;
   gstmp = pType[imgData.picType] ;
   gstmp.copy( imgData.picExpl, imgMAX_TYPE ) ;

   //* Read the description string according to text encoding *
   if ( (imgData.encoding == ENCODE_ASCII) || (imgData.encoding == ENCODE_UTF8) )
   {
      for ( short i = ZERO ; i < imgMAX_DESC ; ++i )
      {
         ifs.read( &ibuff[i], 1 ) ;
         bytesRead += ifs.gcount() ;
         if ( ibuff[i] == NULLCHAR )
            break ;
         ibuff[i + 1] = NULLCHAR ;     // ensure that string is terminated
      }
   }
   else     // one of the UTF-16 flavours
   {
      for ( short i = ZERO ; i < imgMAX_DESC ; i += 2 )
      {
         ifs.read( &ibuff[i], 2 ) ;
         bytesRead += ifs.gcount() ;
         if ( ibuff[i] == ZERO && ibuff[i+1] == ZERO )
            break ;
         ibuff[i + 2] = ibuff[i + 3] = NULLCHAR ; // ensure that string is terminated
      }
   }
   imgData.txtConv( ibuff, imgData.encoding ) ;

   imgData.picSize = fhdr.frame_size - bytesRead ; // bytes of image data
   int loop = imgData.picSize / gsMAXBYTES, 
       remainder = imgData.picSize % gsMAXBYTES ;

   if ( toFile )     // open the temp file
      ofs.open( imgData.picPath, (ofstream::out | ofstream::trunc) ) ;

   for ( int i = ZERO ; i < loop ; ++i )
   {
      ifs.read( ibuff, gsMAXBYTES ) ;
      bytesRead += ifs.gcount() ;
      if ( toFile && (ofs.is_open()) )
         ofs.write( ibuff, ifs.gcount() ) ;
   }
   if ( remainder > ZERO )
   {
      ifs.read( ibuff, remainder ) ;
      bytesRead += ifs.gcount() ;
      if ( toFile && (ofs.is_open()) )
         ofs.write( ibuff, ifs.gcount() ) ;
   }
   if ( toFile )     // close the temp file
      ofs.close() ;

   return bytesRead ;

}  //* End mptDecodeImageFrame() *

//***************************
//*    mptDecodePopFrame    *
//***************************
//******************************************************************************
//* Read and decode an MP3 POPM or PCNT frame.                                 *
//*                                                                            *
//* a) Note that even though the standard allows for multiple popularimeters   *
//*    (POPM), we will only retain the one most recently read. This should not *
//*    be a problem since only a complete fool would encode more than one.     *
//* b) Only one play counter (PCNT) is allowed by the standard, but if         *
//*    multiple instances are encountered, only the most recent will be        *
//*    retained.                                                               *
//*                                                                            *
//* Input  : ifs     : handle for open input stream                            *
//*          fhdr    : (by reference) receives the decoded header record       *
//*          popData : (by reference) receives the decoded frame data          *
//*                                                                            *
//* Returns: number of bytes read from the input stream                        *
//******************************************************************************
//* From the id3v2.3.0 standard:                                               *
//* ============================                                               *
//* <Header for 'Play counter', ID: "PCNT">                                    *
//* Counter         $xx xx xx xx (xx ...)                                      *
//*                                                                            *
//* "This is simply a counter of the number of times a file has been played.   *
//*  The value is increased by one every time the file begins to play.         *
//*  There may only be one "PCNT" frame in each tag. When the counter reaches  *
//*  all one's, one byte is inserted in front of the counter thus making the   *
//*  counter eight bits bigger. The counter must be at least 32-bits long to   *
//*  begin with."                                                              *
//* This explanation _implies_ that the value is a big-endian integer.         *
//*                                                                            *
//* <Header for 'Popularimeter', ID: "POPM">                                   *
//* Email to user   <text string> $00                                          *
//* Rating          $xx                                                        *
//* Counter         $xx xx xx xx (xx ...)                                      *                                                          *
//*                                                                            *
//* "The purpose of this frame is to specify how good an audio file is.        *
//*  Many interesting applications could be found to this frame such as a      *
//*  playlist that features better audiofiles more often than others or it     *
//*  could be used to profile a person's taste and find other 'good' files     *
//*  by comparing people's profiles. The frame is very simple. It contains     *
//*  the email address to the user, one rating byte and a four byte play       *
//*  counter, intended to be increased with one for every time the file is     *
//*  played. The email is a terminated string.                                 *
//*  The rating is 1-255 where 1 is worst and 255 is best. 0 is unknown.       *
//*  If no personal counter is wanted it may be omitted. When the counter      *
//*  reaches all one's, one byte is inserted in front of the counter thus      *
//*  making the counter eight bits bigger in the same way as the play          *
//*  counter ("PCNT"). There may be more than one "POPM" frame in each tag,    *
//*  but only one with the same email address.                                 *
//* This explanation _implies_ that not only is the integer a big-endian value,*
//* but that the email address (if any) is an ASCII (or UTF-8) string.         *
//*                                                                            *
//******************************************************************************

static UINT mptDecodePopFrame ( ifstream& ifs, const id3v2_framehdr& fhdr, 
                                popMeter& popData )
{
   char ibuff[gsMAXBYTES] ;            // input buffer
   gString gstmp( fhdr.frame_id ) ;    // data formatting
   UINT bytesRead = ZERO ;             // return value

   //* Read the frame data *
   ifs.read( ibuff, fhdr.frame_size ) ;
   bytesRead = ifs.gcount() ;

   if ( (gstmp.compare( POP_TAG )) == ZERO )
   {
      //* Reinitialize the target fields (possible previous tag field data) *
      popData.popEmail[0] = NULLCHAR ;
      popData.popStar = ZERO ;
      popData.popCount = ZERO ;

      //* Scan the email address(if any). If valid, it will be null terminated.*
      short i = ZERO ;
      while ( i < fhdr.frame_size )
      {
         popData.popEmail[i] = ibuff[i] ;
         if ( popData.popEmail[i++] == NULLCHAR )
            break ;
         else if ( i >= (MAX_FNAME - 2) )  // prevent buffer overrun
            break ;
      }
      popData.popEmail[i] = NULLCHAR ;    // be sure string is terminated

      if ( i < fhdr.frame_size )
      {
         popData.popStar = ibuff[i++] ;
         if ( i <= (fhdr.frame_size - 4) )
         {  //* This is a big-endian integer of at least 32 bits (4 bytes). *
            //* Data beyond 32 bits is silently discard.                    *
            popData.popCount = popData.intConv( (UCHAR*)&ibuff[fhdr.frame_size - 4] ) ;
         }
      }
      popData.popdata = true ;            // set data flag
   }
   else     // CNT_TAG
   {  //* This is a big-endian integer of at least 32 bits (4 bytes). *
      //* Data beyond 32 bits is silently discard.                    *
      popData.playCount = popData.intConv( (UCHAR*)&ibuff[fhdr.frame_size - 4] ) ;
      popData.cntdata = true ;            // set data flag
   }
   return bytesRead ;

}  //* End mptDecodePopFrame() *


//******************************************************************************
//*********************    OGG/Vorbis Audio Files    ***************************
//******************************************************************************

//*************************
//*  ExtractMetadata_OGG  *
//*************************
//******************************************************************************
//* Read the media file and write the metadata to the temp file.               *
//* The original '.ogg' and the audio-only '.oga' filename extensions indicate *
//* Ogg/Vorbis audio files.                                                    *
//*                                                                            *
//* Ogg/Vorbis 'I' files contain three metadata headers:                       *
//*  1) Identification - identifies the file as Ogg/Vorbis format.             *
//*     We decode this to verify that the file can be decoded.                 *
//*     This header is reported only in verbose mode OR if format is invalid.  *
//*  2) Comments - text fields describing the audio data                       *
//*     We decode and report the fields within the Comments header.            *
//*  3) Setup - codec setup and bitstream definitions                          *
//*     This is encoded binary data which is not decoded here.                 *
//*                                                                            *
//*                                                                            *
//* The fields of the Identification header:                                   *
//*     [vorbis_version] = read 32 bits as unsigned integer            (0)     *
//*     [audio_channels] = read 8 bit integer as unsigned              (>0)    *
//*     [audio_sample_rate] = read 32 bits as unsigned integer         (>0)    *
//*     [bitrate_maximum] = read 32 bits as signed integer                     *
//*     [bitrate_nominal] = read 32 bits as signed integer                     *
//*     [bitrate_minimum] = read 32 bits as signed integer                     *
//*       a) The fields are meaningful only when greater than zero.            *
//*          None set indicates the encoder does not care to speculate.        *
//*       b) All three fields set to the same value implies a fixed rate, or   *
//*          tightly bounded, nearly fixed-rate bitstream.                     *
//*       c) Only nominal set implies a VBR or ABR stream that averages the    *
//*          nominal bitrate.                                                  *
//*       d) Maximum and or minimum set implies a VBR bitstream that obeys the *
//*          bitrate limits.                                                   *
//*     [blocksize_0] = power-of-2 (read 4 bits as unsigned integer)   (<=bs_1)*
//*                     [64 | 128 | 256 | 512 | 1024 | 2048 | 4096 | 8192      *
//*     [blocksize_1] = power-of-2 (read 4 bits as unsigned integer)   (>bs_0) *
//*     [framing_flag] = read one bit (see note below on framing flag)         *
//*   Note that the integer fields are defined as little-endian (LSByte first).*
//*                                                                            *
//* The Comments header begins with:                                           *
//*     a) [vendor\_length] = unsigned 32-bit integer                          *
//*     b) [vendor\_string] = UTF-8 vector of [vendor\_length] bytes.          *
//*     c) [number of comment fields] = unsigned 32-bit integer                *
//*                  The number of fields is limited to 2 to the 32nd - 1.     *
//*     d) Zero or more comment fields as specified by (c)                     *
//*         i) [field length] = unsigned 32-bit integer (little-endian)        *
//*                         This (apparently) includes the FIELDNAME, 0x3D     *
//*                         and the UTF-8 text.                                *
//*        ii) [field contents] = FIELDNAME + '=' + UTF-8(unterminated) string *
//*     e) [framing\_bit] = read a single bit as boolean                       *
//*        If ( [framing\_bit] unset or end-of-packet ) then ERROR             *
//*        Note that as stated, this is completely insane because a single bit *
//*        cannot be read from a byte stream. This is the problem with         *
//*        letting morons write the documentation. In practice, this is the    *
//*        LSbit of the byte.                                                  *
//*     f) End of Comments header                                              *
//*                                                                            *
//* The individual fields contained the Comments header are of the form:       *
//*           nnnnFIELDNAME=<field contents>                                   *
//*     a) 'nnnn' is a 4-byte, little-endian integer indicating the length of  *
//*        the text record.                                                    *
//*     b) The 'FIELDNAME' may be any 7-bit ASCII sequence.                    *
//*        The field name is case-insensitive and may consist of ASCII 0x20    *
//*        through 0x7D, 0x3D (’=’) excluded. ASCII 0x41 through 0x5A inclusive*
//*        (characters A-Z) is to be considered equivalent to ASCII 0x61       *
//*        through 0x7A inclusive (characters a-z).                            *
//*                                                                            *
//*        The basic list includes:                                            *
//*         TITLE : Track/Work name                                            *
//*         VERSION : The version field may be used to differentiate multiple  *
//*                   versions of the same track title in a single collection. *
//*                  (e.g. remix info)                                         *
//*         ALBUM : The collection name to which this track belongs            *
//*         TRACKNUMBER : The track number of this piece if part of a specific *
//*                       larger collection or album.                          *
//*         ARTIST : The artist generally considered responsible for the work. *
//*                  In popular music this is usually the performing band or   *
//*                  singer. For classical music it would be the composer.     *
//*                  For an audio book it would be the author of the original  *
//*                  text.                                                     *
//*         PERFORMER : The artist(s) who performed the work. In classical     *
//*                     music this would be the conductor, orchestra, soloists.*
//*                     In an audio book it would be the actor who did the     *
//*                     reading. In popular music this is typically the same   *
//*                     as the ARTIST and is omitted.                          *
//*         COPYRIGHT : Copyright attribution, e.g., ’2001 Nobody’s Band’ or   *
//*                     ’1999 Jack Moffitt’                                    *
//*         LICENSE : License information, eg, ’All Rights Reserved’,          *
//*                   ’Any Use Permitted’, a URL to a license such as a        *
//*                   Creative Commons license                                 *
//*                   (”www.creativecommons.org/blahblah/license.html”) or the *
//*                   EFF Open Audio License (’distributed under the terms of  *
//*                   the Open Audio License.                                  *
//*                   see http://www.eff.org/IP/Open_licenses/eff_oal.html     *
//*                   for details’), etc.                                      *
//*         ORGANIZATION : Name of the organization producing the track (i.e.  *
//*                        the ’record label’)                                 *
//*         DESCRIPTION : A short text description of the contents             *
//*         GENRE : A short text indication of music genre                     *
//*         DATE : Date the track was recorded                                 *
//*         LOCATION : Location where track was recorded                       *
//*         CONTACT : Contact information for the creators or distributors of  *
//*                   the track. This could be a URL, an email address, the    *
//*                   physical address of the producing label.                 *
//*         ISRC : International Standard Recording Code for the track; see    *
//*                the ISRC intro page for more information on ISRC numbers.   *
//*     b) The '=' character (0x3D) terminates the 'FIELDNAME'.                *
//*     c) The 'field contents' are UTF-8 encoded text (NOT null terminated).  *
//*        Length is limited to 2 to the 32nd - 1 bytes; HOWEVER, the          *
//*        recommendation is a length of no more than a short paragraph (about *
//*        1,000 bytes).                                                       *
//*                                                                            *
//*                                                                            *
//* Input  : ofs    : open output stream to temporary file                     *
//*          srcPath: filespec of media file to be read                        *
//*          vebose : (optional, 'false' by default)                           *
//*                   if 'false', display all text-frame records               *
//*                   if 'true', ALSO display extended technical data and      *
//*                              non-text frames (used primarily for debugging)*
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void FileDlg::ExtractMetadata_OGG ( ofstream& ofs, const gString& srcPath, bool verbose )
{
   #define DEBUG_EMDATA (0)      // Set to non-zero for debugging only

   #if DEBUG_EMDATA != 0
   verbose = true ;              // force verbose output
   #endif   // DEBUG_EMDATA

   const char* BadOgg = "Error: Invalid Ogg/Vorbis format.\n\n" ;
   const char* noComment =  "No Ogg metadata available.\n\n" ;
   const short CNAME_COLS = 18 ; // output column alignment

   gString gsOut, gstmp ;     // output formatting
   tnFName fn ;               // source file stats
   this->fmPtr->GetFileStats ( fn, srcPath, true ) ;
   gstmp.formatInt( fn.fBytes, 11, true ) ;

   short fni = srcPath.findlast( fSLASH ) + 1 ;
   ofs << "FILE NAME  : \"" << &srcPath.ustr()[fni] 
       << "\"  (" << gstmp.ustr() << " bytes)\n\n" ;

   //* Open the source file *
   ifstream ifs( srcPath.ustr(), ifstream::in ) ;
   if ( ifs.is_open() )
   {
      char ibuff[gsMAXBYTES] ;   // input buffer
      Ogg_ID oid ;               // Identification header data
      UINT cnt ;                 // number of bytes read from source file
      bool valid_hdr = true ;    // true if valid header record, else false

      //* Read the Ogg ID header *
      Ogg_PageHdr oph ;
      ifs.read ( ibuff, OGG_PAGE1_BYTES ) ;
      if ( (cnt = ifs.gcount()) == OGG_PAGE1_BYTES )
      {
         gstmp = ibuff ;   // (version number should also be the null terminator)
         gstmp.copy( oph.tag, OGG_TAG_BYTES ) ;
         oph.version = ibuff[4] ;
         oph.ptype   = ibuff[5] ;
                                       // granpos
                                       // streamser
                                       // pageseq
                                       // cksum
         oph.segcount = ibuff[26] ;

         //* Count the total segment bytes (s/b 30 bytes) *
         for ( UCHAR i = 27 ; i < (oph.segcount + 27) ; ++i )
            oph.segBytes += (UCHAR)(ibuff[i]) ;

         //* Read packet header *
         Ogg_Packet opkt ;
         opkt.ptype  = ibuff[28] ;
         opkt.pid[0] = ibuff[29] ;
         opkt.pid[1] = ibuff[30] ;
         opkt.pid[2] = ibuff[31] ;
         opkt.pid[3] = ibuff[32] ;
         opkt.pid[4] = ibuff[33] ;
         opkt.pid[5] = ibuff[34] ;

         //* Read the identification header *
         oid.version = oid.intConv( (UCHAR*)&ibuff[35] ) ;
         oid.channels = (UCHAR)ibuff[39] ;
         oid.samprate = oid.intConv( (UCHAR*)&ibuff[40] ) ;
         oid.bitratemax = oid.intConv( (UCHAR*)&ibuff[44] ) ;
         oid.bitratenom = oid.intConv( (UCHAR*)&ibuff[48] ) ;
         oid.bitratemin = oid.intConv( (UCHAR*)&ibuff[52] ) ;
         oid.blksize0 = (((UINT)ibuff[56]) & 0x0000000F) ;
         oid.blksize0 = ((UINT)0x00000001) << oid.blksize0 ;
         oid.blksize1 = (((UINT)ibuff[56]) & 0x000000F0) >> 4 ;
         oid.blksize1 = ((UINT)0x00000001) << oid.blksize1 ;
         oid.frameflag = (bool)(ibuff[57] & 0x01) ;

         //* Validate the page header data *
         if ( (gstmp.compare( Ogg_Tag ) == ZERO) &&
              (oph.version == ZERO) && (oph.ptype == 0x02) && 
              (oph.segcount &= 1) && (oph.segBytes == 30)
            )
         {
            //* Validate the ID header *
            if ( (oid.version != ZERO) ||
                 (oid.channels <= ZERO) ||
                 (oid.samprate <= ZERO) ||
                 !((oid.blksize0 == 64)   || (oid.blksize0 == 128) || 
                   (oid.blksize0 == 256)  || (oid.blksize0 == 512) || 
                   (oid.blksize0 == 1024) || (oid.blksize0 == 2048) || 
                   (oid.blksize0 == 4096) || (oid.blksize0 == 8192)) ||
                 !((oid.blksize1 == 64)   || (oid.blksize1 == 128) || 
                   (oid.blksize1 == 256)  || (oid.blksize1 == 512) || 
                   (oid.blksize1 == 1024) || (oid.blksize1 == 2048) || 
                   (oid.blksize1 == 4096) || (oid.blksize1 == 8192)) ||
                 (oid.frameflag == false) )
               valid_hdr = false ;
         }
         else
            valid_hdr = false ;

         #if DEBUG_EMDATA != 0
         gsOut.compose( "Packet Header : %02hhX '%s'\n"
                        "Ogg/Vorbis tag:'%S'\n"
                        "Version       : %02hhX\n"
                        "Page Type     : %02hhXh\n"
                        "Segment Count : %hhu\n"
                        "Segment Bytes : %u\n"
                        "Valid Page Hdr: %s\n\n",
                        &opkt.ptype, opkt.pid,
                        gstmp.gstr(), &oph.version, &oph.ptype, 
                        &oph.segcount, &oph.segBytes,
                        (char*)(valid_hdr ? "true" : "false") ) ;
         ofs << gsOut.ustr() ;
         #endif   // DEBUG_EMDATA

         if ( verbose )
         {
            gsOut.compose( "Ogg/Vorbis ID Header\n"
                           "--------------------\n"
                           "version   : %u\n"
                           "channels  : %hhu\n"
                           "samprate  : %d\n"
                           "bitratemax: %d\n"
                           "bitratenom: %d\n"
                           "bitratemin: %d\n"
                           "blksize0  : %u\n"
                           "blksize1  : %u\n"
                           "frameflag : %hhx\n\n",
                           &oid.version, &oid.channels, &oid.samprate,
                           &oid.bitratemax, &oid.bitratenom, &oid.bitratemin,
                           &oid.blksize0, &oid.blksize1, &oid.frameflag ) ;
            ofs << gsOut.ustr() ;
         }

         //* If ID header is valid, continue to Comment Header *
         //* Note that an error in the Comment header is       *
         //* defined as a non-fatal error.                     *
         if ( valid_hdr )
         {
            //* Read header of second page *
            char segTable[gsMAXCHARS] ;
            oph.reset() ;
            ifs.read ( segTable, OGG_PAGEHDR_BYTES ) ;
            gstmp = segTable ;   // (version number should also be the null terminator)
            gstmp.copy( oph.tag, OGG_TAG_BYTES ) ;
            oph.version = segTable[4] ;
            oph.ptype   = segTable[5] ;
                                       // granpos
                                       // streamser
                                       // pageseq
            oph.segcount = segTable[26] ;
            //* Read the segment table (0-255 bytes) *
            //* and count the total segment bytes.   *
            ifs.read ( segTable, oph.segcount ) ;
            for ( UCHAR i = ZERO ; i < oph.segcount ; ++i )
               oph.segBytes += (UCHAR)(segTable[i]) ;

            //* Read packet header *
            UINT remBytes = oph.segBytes ; // unread bytes in Comment header
            opkt.reset() ;
            if ( remBytes >= OGG_PKTHDR_BYTES )
            {
               ifs.read ( ibuff, OGG_PKTHDR_BYTES ) ;
               remBytes -= OGG_PKTHDR_BYTES ;
               opkt.ptype  = ibuff[0] ;
               opkt.pid[0] = ibuff[1] ;
               opkt.pid[1] = ibuff[2] ;
               opkt.pid[2] = ibuff[3] ;
               opkt.pid[3] = ibuff[4] ;
               opkt.pid[4] = ibuff[5] ;
               opkt.pid[5] = ibuff[6] ;
            }

            #if DEBUG_EMDATA != 0
            gsOut.compose( "Header Packet : %02hhX '%s'\n"
                           "Ogg/Vorbis tag:'%S'\n"
                           "Version       : %02hhX\n"
                           "Page Type     : %02hhXh\n"
                           "Segment Count : %hhu\n"
                           "Segment Bytes : %u\n"
                           "Valid Page Hdr: %s\n\n", 
                           &opkt.ptype, opkt.pid,
                           gstmp.gstr(), &oph.version, &oph.ptype, 
                           &oph.segcount, &oph.segBytes,
                           (char*)(valid_hdr ? "true" : "false") ) ;
            ofs << gsOut.ustr() ;
            #endif   // DEBUG_EMDATA

            //* Read the Comment Header *
            if ( remBytes > ZERO )
            {
               Ogg_Comment ogc ;

               //* Read the watermark message - reported only in verbose mode *
               if ( remBytes >= 4 )
               {
                  ogc.reset() ;           // clear the buffer
                  ifs.read ( ibuff, 4 ) ; // read length of comment vector
                  remBytes -= 4 ;
                  ogc.clen = ogc.intConv( (UCHAR*)ibuff ) ;
                  // Programmer's Note: There is a logical error here. 
                  // If ogc.clen > gsMAXBYTES, we will get an out-of-bounds error.
                  // The likelihood of this happening is remote.
                  ifs.read ( ibuff, ogc.clen ) ;   // read comment text
                  remBytes -= ogc.clen ;
                  if ( verbose )
                  {
                     ibuff[ogc.clen] = NULLCHAR ;
                     gsOut.compose( "%s\n", ibuff ) ;
                     ofs << gsOut.ustr() ;
                  }
               }
               //* Get the number of comment vectors *
               UINT cVectors = ZERO ;
               if ( remBytes >= 4 )
               {
                  ifs.read ( ibuff, 4 ) ; // read length of comment vector
                  remBytes -= 4 ;
                  cVectors = ogc.intConv( (UCHAR*)ibuff ) ;
                  if ( verbose )
                  {
                     gsOut.compose( "\nTotal Comments: %u\n\n", &cVectors ) ;
                     ofs << gsOut.ustr() ;
                  }
               }

               //* Read and report each vector *
               if ( cVectors > ZERO )
               {
                  short cIndex ;
                  for ( UINT v = ZERO ; v < cVectors ; ++v )
                  {
                     ogc.reset() ;           // clear the buffer
                     ifs.read ( ibuff, 4 ) ; // read length of comment vector
                     ogc.clen = ogc.intConv( (UCHAR*)ibuff ) ;
                     if ( ogc.clen < (gsMAXBYTES - 2) )
                     {
                        ifs.read ( ibuff, ogc.clen ) ;   // read comment text

                        //* If an Image Vector, decode the header *
                        //* and discard the image itself.         *
                        if ( !(oggDecodeImageVector ( ibuff, ogc )) )
                        {
                           ibuff[ogc.clen] = NULLCHAR ;
                           gstmp = ibuff ;
                           gstmp.copy( ogc.ctxt, gsMAXBYTES ) ;
                        }
                     }
                     else     // output limited to four(4) kbyte blocks
                     {
                        // Programmer's Note: We may split a multi-byte UTF-8
                        // character here, but the gString object will automagically
                        // discard the partial character.
                        ifs.read ( ibuff, (gsMAXBYTES - 2) ) ;

                        //* If an Image Vector, decode the header *
                        //* and discard the image itself.         *
                        if ( !(oggDecodeImageVector ( ibuff, ogc )) )
                        {
                           ibuff[gsMAXBYTES - 2] = NULLCHAR ;
                           gstmp = ibuff ;
                           gstmp.copy( ogc.ctxt, gsMAXBYTES ) ;
                        }

                        remBytes = ogc.clen - (gsMAXBYTES - 2) ;
                        while ( remBytes > gsMAXBYTES )  // discard the remaining data
                        {
                           ifs.read ( ibuff, gsMAXBYTES ) ;
                           remBytes -= gsMAXBYTES ;
                        }
                        if ( remBytes > ZERO )
                           ifs.read ( ibuff, remBytes ) ;
                     }
                     // Programmer's Note: The specification states that the
                     // comment name is ASCII, so 1 byte == 1 character.
                     // The comment text however, is UTF-8.
                     gstmp = ogc.ctxt ;
                     if ( (cIndex = gstmp.find( L'=' )) >= ZERO ) // if valid comment format
                     {
                        if ( (gstmp.find( oggPicComment )) == ZERO )
                        {
                           gstmp.shiftChars( -(cIndex + 1) ) ;
                           ofs << gstmp.ustr() << '\n' ;
                        }
                        else
                        {
                           gstmp.limitChars( cIndex ) ;
                           gstmp.append( L' ' ) ;
                           while ( gstmp.gschars() < CNAME_COLS )
                              gstmp.append( L'-' ) ;
                           gstmp.append( L' ' ) ;
                           ofs << gstmp.ustr() << &ogc.ctxt[cIndex + 1] << '\n' ;
                        }
                     }
                     else if ( verbose )
                        ofs << "Non-standard encoding: '" << ibuff << "'\n" ;
                  }
               }
               else
                  ofs << noComment ;
            }
            else
               ofs << noComment ;
         }
         else
            ofs << BadOgg ;
      }
      else
         ofs << BadOgg ;

      ifs.close() ;        // close the source file
   }
   else
      ofs << "Sorry, unable to open file for reading.\n" ;
   ofs << endl ;

   #undef DEBUG_EMDATA

}  //* End ExtractMetadata_OGG() *

//***************************
//*  oggDecodeImageVector   *
//***************************
//******************************************************************************
//* Read and decode an OGG image vector                                        *
//*   1) Read the setup and description data and save it for later.            *
//*   2) Read the binary image data and,                                       *
//*      a) save it to the specified file, or                                  *
//*      b) discard it
//*                                                                            *
//* Input  : ibuff   : contains the first gsMAXBYTES of the raw Image Vector   *
//*          ogc     : (by reference) receives a summary report of image header*
//*                                                                            *
//* Returns: number of bytes read from the input stream                        *
//******************************************************************************

static bool oggDecodeImageVector ( const char* ibuff, Ogg_Comment& ogc )
{
   const char* ImgTemplate = "%s=  Image: %S bytes, MIME:%s, Type:%s, Desc:%s" ;

   bool isImage = false ;     // return value

   //* Do the easy test first *
   if ( (ibuff[0] == 'M') || (ibuff[0] == 'C') || (ibuff[0] == 'm') || (ibuff[0] == 'c') )
   {
      gString gstmp( ibuff, opcBYTES + 1 ) ;    // read the comment name + '='
      if ( ((gstmp.find( oggPicComment )) == ZERO) ||
           ((gstmp.find( oggPicCommentOld )) == ZERO) )
      {
         char mimType[gsMAXBYTES],  // MIME-type string
              txtDesc[gsMAXBYTES] ; // description string
         popMeter popm ;            // borrowed for conversion of big-endian integers
         int32_t hdrIndex = ZERO ;  // index of image-vector header
         UINT32 mimLen = ZERO,      // bytes in MIME-type string
                descLen = ZERO,     // bytes in description string
                imgBytes = ZERO,    // bytes of ENCODED image data
                picType = ZERO ;    // picture type code
         //* 'true' if standard image-vector format, 'false' if obsolete format*
         bool   stdFmt = bool((ibuff[0] == 'M') || (ibuff[0] == 'm')) ;

         //* Extract header data *
         if ( (hdrIndex = gstmp.find( L'=' )) >= ZERO )
         {
            ++hdrIndex ;               // index first integer value
            isImage = true ;           // valid image vector name format
            if ( stdFmt )              // standard image format
            {
               picType = popm.intConv( (UCHAR*)&ibuff[hdrIndex] ) ;
               hdrIndex += 4 ;         // step over integer value
               mimLen = popm.intConv( (UCHAR*)&ibuff[hdrIndex] ) ;
               hdrIndex += 4 ;         // step over integer value
               gstmp.loadChars( &ibuff[hdrIndex], mimLen ) ;
               gstmp.copy( mimType, gsMAXBYTES ) ;
               hdrIndex += mimLen ;    // step over MIME type
               descLen = popm.intConv( (UCHAR*)&ibuff[hdrIndex] ) ;
               hdrIndex += 4 ;         // step over integer value
               gstmp.loadChars( &ibuff[hdrIndex], descLen ) ;
               gstmp.copy( txtDesc, gsMAXBYTES ) ;
               hdrIndex += descLen ;   // step over description
               hdrIndex += 4 * 4 ;     // step over image dimension data (4 integers)
               imgBytes = popm.intConv ( (UCHAR*)&ibuff[hdrIndex] ) ;
            }
            else                 // obsolete image format (i.e. no header info)
            {
               imgBytes = ogc.clen - hdrIndex ;
               gstmp = "none" ;
               gstmp.copy( mimType, gsMAXBYTES ) ;
               gstmp = "(obsolete OGG image format)" ;
               gstmp.copy( txtDesc, gsMAXBYTES ) ;
            }

            //* Create display text *
            gstmp.formatInt( imgBytes, 13, true ) ;
            gString gsOut( ImgTemplate, oggPicComment, gstmp.gstr(), 
                           mimType, pType[picType], txtDesc ) ;
            gsOut.copy( ogc.ctxt, gsMAXBYTES ) ;
         }
         else
         { /* invalid comment-header format */ }
      }
   }
   return isImage ;

}  //* End mptDecodeImageVector() *

//******************************************************************************
//*********************   MPEG-4 Audio Files (M4A)   ***************************
//******************************************************************************

#define DEBUG_M4A (0)      // Set to non-zero for debugging only
#if DEBUG_M4A != 0
gString gsdbg1, gsdbg2 ;
#endif   // DEBUG_M4A

const uint32_t ubuffLEN = 0x001E8480 ; // size of input buffer (2.0 megabytes)
const uint32_t intLEN = 4 ;            // bytes in a 32-bit integer
const uint32_t idLEN = (intLEN * 2) ;  // size for initial header read


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  *
//* List of box types defined by ISO/IEC 14496-12:2005(E)   *
//* Not all types will be found in M4A (audio-only) files.  *
const uint32_t metaINDEX = ZERO, 
               metaGROUP = 17,
               freeINDEX = metaINDEX + metaGROUP,
               freeGROUP = 1,
               skipINDEX = freeINDEX + freeGROUP,
               skipGROUP = 1,
               udtaINDEX = skipINDEX + skipGROUP,
               udtaGROUP = 2,
               uuidINDEX = udtaINDEX + udtaGROUP,
               uuidGROUP = 1,
               ftypINDEX = uuidINDEX + uuidGROUP,
               ftypGROUP = 1,
               pdinINDEX = ftypINDEX + ftypGROUP,
               pdinGROUP = 1,
               moovINDEX = pdinINDEX + pdinGROUP,
               moovGROUP = 38,
               moofINDEX = moovINDEX + moovGROUP,
               moofGROUP = 8,
               mfraINDEX = moofINDEX + moofGROUP,
               mfraGROUP = 3,
               mdatINDEX = mfraINDEX + mfraGROUP,
               mdatGROUP = 1,
               MAX_BOXTYPES = mdatINDEX + mdatGROUP ;

//* Additional indices *
const uint32_t schiINDEX = metaINDEX + 11,
               xmlINDEX  = schiINDEX + 1,
               bxmlINDEX = xmlINDEX + 1 ;

const char* mpeg4_boxType[MAX_BOXTYPES] = 
{
   "meta",              // "meta" group
   "hdlr",                 // declared metadata handler typpe
   "dinf",                 // data-information-box container
   "dref",                 // declares source of metadata items
   "ipmc",                 // IPMP Control Box
   "iloc",                 // item location
   "ipro",                 // item protection (DRM)
   "sinf",                 // protection scheme information
   "frma",                 // original format
   "imif",                 // IMPM information
   "schm",                 // scheme type
   "schi",                 // scheme information
   "xml ",                 // XML container
   "bxml",                 // binary XML container
   "iinf",                 // item information
   "pitm",                 // primary item reference
   "ilst",                 // contains metadata (not part of MPEG-4 specification)
                           // This is an iTunes construct.

   "free",              // "free" space group

   "skip",              // "skip" group
   "udta",              // user data group
   "cprt",                 // copyright info

   "uuid",               // "uuid group" (user-defined type)

   "ftyp",              // "ftyp" group (file type and compatibility)

   "pdin",              // "pdin" group (progressive download information)

   "moov",              // "moov" group (container for all metadata)
   "mvhd",                 // (this box may contain data for user view)
   "trak",                 // 
   "tkhd",                 // (this box may contain data for user view 'track_ID')
   "tref",                 // 
   "edts",                 // 
   "elst",                 // 
   "mdia",                 // 
   "mdhd",                 // 
   "hdlr",                 // 
   "minf",                 // 
   "vmhd",                 // 
   "smhd",                 // 
   "hmhd",                 // 
   "mnhd",                 // 
   "dinf",                 // 
   "dref",                 // 
   "stbl",                 // 
   "stsd",                 // 
   "stts",                 // 
   "ctts",                 // 
   "stsc",                 // 
   "stsz",                 // 
   "stz2",                 // 
   "stco",                 // 
   "co64",                 // 
   "stss",                 // 
   "stsh",                 // 
   "padb",                 // 
   "stdp",                 // 
   "sdtp",                 // 
   "sbgp",                 // 
   "sgpd",                 // 
   "subs",                 // 
   "mvex",                 // 
   "mehd",                 // 
   "trex",                 // 
   "ipmc",                 // 

   "moof",              // "moof" group (movie fragment)
   "mfhd",                 // 
   "traf",                 // 
   "tfhd",                 // 
   "trun",                 // 
   "sdtp",                 // 
   "sbgp",                 // 
   "subs",                 // 

   "mfra",              // "mfra" group (movie fragment random access)
   "tfra",                 // 
   "mfro",                 //
   
   "mdat",              // "mdat" group (media data container)
} ;
// Programmer's Note: Certain box types were used in "pre-standard" files.
// clip, crgn, matt, kmat, pnot, ctab, load, imap;
// Track reference types: tmcd, chap, sync, scpt, ssrc.

//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  *
//* Atom identifiers defined in the Apple(tm) iTunes database of tag    *
//* identifiers, along with a human-readable version of the identifier. *
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  *
const uint32_t looneyITEMS = 44 ;   // items in liTunes[] array
//* Item index for first integer-value record. *
//* (First two items in this group ('trkn' and 'disk' are two-value records)
const uint32_t looneyINTVALUE = 24 ;
//* Nominal column width for tag name display string.*
const short tagDescWIDTH = 23 ;
//* Special non-ASCII character used by the iTunes dictionary *
const uint8_t BYTE_A9 = 0x0A9 ;
//* Number of bytes in luniTunesDB::hName member. *
const uint32_t looneyLEN = 32 ;

class luniTunesDB
{
   public:
   luniTunesDB ( void )    // default constructor
   {
      this->reset () ;
   }
   //* Initialization constructor *
   luniTunesDB ( const uint8_t* a, const char* h )
   {
      uint32_t indx ;
      for ( indx = ZERO ; indx <= intLEN ; ++indx ) // copy atom name
      {
         if ( (this->aName[indx] = a[indx]) == '\0' )
            break ;
      }
      for ( indx = ZERO ; indx < looneyLEN ; ++indx ) // copy human-readable name
      {
         if ( (this->hName[indx] = h[indx]) == '\0' )
            break ;
      }
   }
   void reset ( void )
   {
      this->aName[0] = '\0' ;
      this->hName[0] = '\0' ;
   }
   uint8_t aName[intLEN + 1] ;         // atom name defined by iTunes
   char    hName[looneyLEN + 1] ;      // human-readable atom description
} ;

luniTunesDB liTunes[looneyITEMS] = 
{  //* : dataConst1 S/B 0x01
   { (const uint8_t*)"\xa9""nam", "Title------------------", },
   { (const uint8_t*)"\xa9""alb", "Album------------------", },
   { (const uint8_t*)"\xa9""ART", "Artist-----------------", },
   { (const uint8_t*)"aART",      "Album Artist-----------", },
   { (const uint8_t*)"\xa9""wrt", "Composer---------------", },
   { (const uint8_t*)"\xa9""day", "Year-------------------", },
   { (const uint8_t*)"\xa9""cmt", "Comment----------------", },
   { (const uint8_t*)"desc",      "Description------------", },
   { (const uint8_t*)"purd",      "Purchase Date----------", },
   { (const uint8_t*)"\xa9""grp", "Grouping---------------", },
   { (const uint8_t*)"\xa9""gen", "Genre------------------", },
   { (const uint8_t*)"\xa9""lyr", "Lyrics-----------------", },
   { (const uint8_t*)"purl",      "Podcast URL------------", },
   { (const uint8_t*)"egid",      "Podcast Episode GUID---", },
   { (const uint8_t*)"catg",      "Podcast Category-------", },
   { (const uint8_t*)"keyw",      "Podcast Keywords-------", },
   { (const uint8_t*)"\xa9""too", "Encoded By-------------", },
   { (const uint8_t*)"cprt",      "Copyright--------------", },
   { (const uint8_t*)"tvsh",      "Show Name--------------", },
   { (const uint8_t*)"\xa9""wrk", "Work-------------------", },
   { (const uint8_t*)"\xa9""mvn", "Movement---------------", },
   { (const uint8_t*)"cpil",      "Part of Compilation----", },
   { (const uint8_t*)"pgap",      "Part of Gapless Album--", },
   { (const uint8_t*)"apID",      "User ID----------------", },

   //* Obsolete Tags *
   { (const uint8_t*)"gnre",      "Genre------------------", }, // (see "\a9gen")

   //* Integer Group: dataConst1 S/B 0x00 *
   { (const uint8_t*)"trkn",      "Track Number-----------", }, // two-value (xx/yy)
   { (const uint8_t*)"disk",      "Disc Number------------", }, // two-value (xx/yy)
   { (const uint8_t*)"tmpo",      "Tempo/BPM--------------", },
   { (const uint8_t*)"\xa9""mvc", "Movement Count---------", },
   { (const uint8_t*)"\xa9""mvi", "Movement Index---------", },
   { (const uint8_t*)"shwm",      "Work/Movemenmt---------", },
   { (const uint8_t*)"stik",      "Media Kind-------------", },
   { (const uint8_t*)"hdvd",      "HD Video---------------", },
   { (const uint8_t*)"rtng",      "Content Rating---------", },
   { (const uint8_t*)"tves",      "TV Episode-------------", },
   { (const uint8_t*)"tvsn",      "TV Season--------------", },

   //* iTunes Internal Codes: dataConst1 S/B 0x15 *
   { (const uint8_t*)"plID",      "(iTunes internal)------", },
   { (const uint8_t*)"cnID",      "(iTunes internal)------", },
   { (const uint8_t*)"geID",      "(iTunes internal)------", },
   { (const uint8_t*)"atID",      "(iTunes internal)------", },
   { (const uint8_t*)"sfID",      "(iTunes internal)------", },
   { (const uint8_t*)"cmID",      "(iTunes internal)------", },
   { (const uint8_t*)"akID",      "(iTunes internal)------", },

   //* Cover Art Codes: dataConst1 S/B 0x0D *
   { (const uint8_t*)"covr",      "Cover Art--------------", },

   #if 0  //* Non-public tags, ignored *
   { (const uint8_t*)"soal",      "Album Sort Order-------", }, // non-public
   { (const uint8_t*)"soaa",      "Album Artist Sort Order", }, // non-public
   { (const uint8_t*)"soar",      "Artist Sort Order------", }, // non-public
   { (const uint8_t*)"sonm",      "Title Sort Order-------", }, // non-public
   { (const uint8_t*)"soco",      "Composer Sort Order----", }, // non-public
   { (const uint8_t*)"sosn",      "Show Sort Order--------", }, // non-public
   { (const uint8_t*)"pcst",      "Podcast (import only)--", }, // non-public
   #endif //* Non-public tags, ignored *
} ;

luniTunesDB genericTag = {(const uint8_t*)"data", "Generic Tag------------"} ;


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//*   Decode and store an iTunes atom dictionary entry.                        *
//    Programmer's Note: For integer values, our only clue about the size of 
//    the integer is the bytes remaining in the stream. Unfortunately, this is 
//    not always enough.
//    -- For for a single-value tags, we must rely on the 1, 2, or 4 bytes 
//       remaining and decode whatever remains as a single value.
//       Example: 'gnre' apparently takes a 16-bit value.
//    -- For the two-value tags 'trkn' and 'disk', it seems that two 32-bit 
//       integers are provided; HOWEVER, the second value should be 
//       interpreted as 16-bit and the final 16 bits ignored. 
//       This analysis is based on files encoded by iTunes, which for better 
//       or worse must be considered as the gold standard. Wouldn't it be 
//       nice if the specification actually specified the format?
//    Rant: Your author wrote the original Award Software BIOS specification,
//          and you may be assured that _everything_ was fully specified.
//          Unfortunately, our work at Apple was related to the II-GS and 
//          early Macs, and not the iTunes specification. (their loss :-)
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *

//* The 'dataConst1' member indicates the type of data stored in that atom.*
const uint32_t itaMETA_NUMERIC = (0x000) ;   // numeric (integer) data
const uint32_t itaMETA_TEXT    = (0x001) ;   // text data
const uint32_t itaMETA_ITUNES  = (0x015) ;   // internal iTunes codes
const uint32_t itaMETA_JFIF    = (0x00D) ;   // image data in JFIF format

const uint32_t itaMETA_MAX = gsMAXBYTES ; // size of data-capture buffer (4.0KB)
class itAtom
{
   public:

   //******************
   //* Public Methods *
   //******************
   ~itAtom ( void )              // destructor
   {
      if ( this->metaStr != NULL )
         delete [] this->metaStr ;
   }

   itAtom ( void )               // default constructor
   {
      this->reset () ;

      //* Allocate a data-capture buffer.*
      this->metaStr = new char[itaMETA_MAX] ;
   }

   //***********************************************************
   //* Decode the binary data and initialize all data members. *
   //***********************************************************
   // Programmer's Note: 'binData' SHOULD BE const; however, 
   // see kludgy fix below.
   bool initialize ( uint32_t s,  uint32_t a, const char* n, uint8_t* binData )
   {
      this->reset () ;
      this->atomSize = s ;                            // copy atom size
      this->aIndex   = a ;                            // copy liTunes-database index
      for ( uint8_t i = ZERO ; i <= intLEN ; ++i )    // copy atom name
         this->atomName[i] = n[i] ;

      this->dataSize = this->decodeInt32 ( binData ) ;
      uint32_t indx = intLEN ;                        // set index after data-size field
      bool dnOk = false ;
      for ( uint8_t i = ZERO ; i < intLEN ; ++i )     // copy data name
         this->dataName[i] = binData[indx++] ;
      this->dataName[intLEN] = '\0' ;  // terminate the string
      if ( (this->dataName[0] == genericTag.aName[0]) && 
           (this->dataName[1] == genericTag.aName[1]) &&
           (this->dataName[2] == genericTag.aName[2]) && 
           (this->dataName[1] == genericTag.aName[3]) )
         dnOk = true ;

      //* Decode the "data" constants *
      this->dataConst1 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;
      this->dataConst2 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;
      uint32_t eod = this->atomSize - indx + idLEN ; // bytes of actual metadata

      //* Validate the internal data structure.*
      if ( dnOk &&                                          // valid "data" atom name
           (this->dataSize == (this->atomSize - idLEN)) &&  // valid data-atom size
           (this->aIndex < looneyITEMS ) &&                 // valid database index
           (this->dataConst2 == ZERO)                       // valid constant value
         )
      {
         gString gstmp ;         // formatting

         //* If atom contains text data *
         if ( this->dataConst1 == itaMETA_TEXT )
         {
            #if 1    // KLUDGY FIX
            // Programmer's Note: At least one of our sample file places a 0xA9 
            // byte as the first byte character of the metadata in a "\xa9com" atom.
            // This is unbelievably stupid because it is not UTF-8.
            if ( binData[indx] == BYTE_A9 )
               binData[indx] = ' ' ;
            #endif   // KLUDGEY FIX

            //* Copy the text metadata *
            uint32_t trg = ZERO ;
            for ( ; (indx < eod) && (trg < (itaMETA_MAX - 1)) ; ++trg )
            {
               this->metaStr[trg] = binData[indx++] ;
               if ( this->metaStr[trg] == '\0' ) break ;
            }
            this->metaStr[trg] = '\0' ;   // ensure termination

            #if 1    // Enable Beautification
            //* Strip leading and trailing whitespace from *
            //* the data, for a more professional display. *
            gstmp = this->metaStr ;
            gstmp.strip() ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            #endif   // Beautification

            this->valid = true ;
         }

         else if ( this->dataConst1 == itaMETA_NUMERIC )
         {
            //* Capture the first value               *
            //* (See note above about integer width.) *
            if ( (eod - indx) >= intLEN )
            {
               this->metaVal1 = this->decodeInt32 ( &binData[indx] ) ;
               indx += intLEN ;
            }
            else if ( (eod - indx) >= (intLEN / 2) )
            {
               this->metaVal1 = this->decodeInt16 ( &binData[indx] ) ;
               indx += (intLEN / 2) ;
            }
            else if ( (eod - indx) >= (intLEN / 4) )
            {
               this->metaVal1 = binData[indx] ;
               ++indx ;
            }
            else
               this->metaVal1 = ZERO ;
            gstmp.compose( "%u", &this->metaVal1 ) ;

            //* If this atom type defines a second value, it should *
            //* be only for the 'trkn' and 'disk' tags. For both of *
            //* these, it is assumed that the second value is       *
            //* encoded in 16 bits or less.                         *
            if ( indx < eod )
            {
               #if 0    // disable 32-bit capture
               if ( (eod - indx) >= intLEN )
                  this->metaVal2 = this->decodeInt32 ( &binData[indx] ) ;
               else
               #endif   // disable 32-bit capture
               if ( (eod - indx) >= (intLEN / 2) )
                  this->metaVal2 = this->decodeInt16 ( &binData[indx] ) ;
               else if ( (eod - indx) >= (intLEN / 4) )
                  this->metaVal2 = binData[indx] ;
               else
                  this->metaVal2 = ZERO ;
               indx += intLEN ;
               gstmp.append( "/%02u", &this->metaVal2 ) ;
            }
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;

            this->valid = true ;
         }

         else if ( this->dataConst1 == itaMETA_JFIF )
         {
            //* A "JFIF" container holds a JPEG image. The format for the  *
            //* container is:
            //* 0xFFD8 0xFFE0 JFIF-APPOsegment [JFXX-APPOsegment][other markers]
            //* 0xFFDA .... 0xFFD9
            //* Legend:
            //*   0xFFD8      Start Of Image
            //*   0xFFE0      JFIF APPO marker
            //*      s1s2     16-bit big-endian: length of segment (excl. APPO marker)
            //*      JFIF\0   literal "JFIF" (null terminated)
            //*      MMmm     two bytes representing Major and minor version
            //*      du       8-bits Density Units: 00h, 01h, 02h
            //*      hpd      16-bit horizontal pixel density
            //*      vpd      16-bit vertical pixel density
            //*      hthumb   8-bit horizontal thumbnail pixel count (0 if none)
            //*      vthumb   8-bit vertical thumbnail pixel count (0 if none)
            //*      thdata   thumbnail data (24-bit RGB format)
            //*   [Optional JFXX APPO extension block, describes thumbnail]
            //*   0xFFDA      Start Of Scan
            //*   ....        JPEG compressed image data. Per spec. image is 
            //*                square and <= 3000x3000 pixels
            //*   0xFFD9      End Of Image

            //* Calculate and report image dimensions *
            // Programmer's Note: There is a potential compiler bug near here
            // regarding variable alignment. (g++ 10.1.1 20200507), Defining 
            // the 'char' array first masks the problem, but keep an eye on it.
            char     segName[64] ;
            char     imgSize[64] ;
            char     pixels[64] ;
            uint32_t soi,        // start-of-image (0xFFD8)
                     appo,       // JFIF APPO (0xFFE0)
                     appolen,    // segment size (excl. 'appo')
                     imgsize ;   // image size in bytes (approximate)
            uint32_t vmaj,       // major version
                     vmin,       // minor version
                     denunits,   // density units: 00h==no units
                                 //                01h==pixels-per-inch (2.54cm)
                                 //                02==pixels-per-centimeter
                     xden,       // horizontal pixel density
                     yden ;      // vertical pixel density
            soi = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            appo = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            appolen = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            for ( uint32_t trg = ZERO ; trg <= (intLEN + 1) ; ++trg )
               segName[trg] = binData[indx++] ;
            vmaj = binData[indx++] ;
            vmin = binData[indx++] ;
            denunits = binData[indx++] ;
            xden = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            yden = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            #if 0    // CAPTURE AND REPORTING OF OPTIONAL THUMBNAIL IMAGE NOT IMPLEMENTED
            //* Thumbnails supported only by iTunes specification v:1.2 and higher *
            if ( (vmax >= 1) && (vmin >= 2 )
            {
               uint32_t hthumb = binData[indx++],
                        vthumb = binData[indx++] ;
               //* If embedded thumbnail image *
               if ( (hthumb > ZERO) && (vthumb > ZERO) )
               {
               }
            }
            #endif   // THUMBNAIL IMAGE INFO
            imgsize = this->dataSize - appolen - 2 ;
            gstmp.formatInt( imgsize, 5, true, false, false, fiKb ) ;
            gstmp.copy( imgSize, 64 ) ;
            
            if ( denunits == 0x00 )
               gstmp.compose( "%ux%u pixels", &xden, &yden ) ;
            else if ( denunits == 0x01 )
               gstmp.compose( "%u pixels-per-inch", &xden ) ;
            else  // (denunits==0x02)
               gstmp.compose( "%u pixels-per-cm", &xden ) ;
            gstmp.copy( pixels, 64 ) ;

            #if DEBUG_M4A == 0   // PRODUCTION
            gstmp.compose( "%s, %s, Encoding:%s", imgSize, pixels, segName ) ;

            // Programer's Note: This is only to silence the compiler warning 
            // about unused variables when not in debugging mode.
            if ( (soi == 0xFFD8) || (appo == 0xFFE0) || (vmaj >= 1) || (vmin >= 1) )
            { /* do nothing */ }
            #else    // DEBUG ONLY
            gstmp.compose( "%s, %s, Encoding:%s\n"
                           "soi:%04X appo:%04X (%u x%04X bytes) "
                           "v:%02u.%02u dunits:%02X xden:%u yden:%u",
                           imgSize, pixels, segName,
                           &soi, &appo, &appolen, &appolen, &vmaj, 
                           &vmin, &denunits, &xden, &yden ) ;
            #endif   // DEBUG_M4A

            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            this->valid = true ;
         }

         #if DEBUG_M4A != 0
         //* Code for iTunes internal coded data. *
         //* Reported only in debugging mode.     *
         else if ( this->dataConst1 == itaMETA_ITUNES )
         {
            //* Capture the value *
            if ( (eod - indx) >= intLEN )
               this->metaVal1 = this->decodeInt32 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 2) )
               this->decodeInt16 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 4) )
               this->metaVal1 = binData[indx] ;
            else
               this->metaVal1 = ZERO ;
            gstmp.compose( "%04Xh", &this->metaVal1 ) ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            
            this->valid = true ;
         }
         #endif   // DEBUG_M4A

         //* Unknown data format code. Currently ignored.*
         //else
         //{
         //}
      }

      return this->valid ;
   }

   //* Scan a 32-bit integer (4 bytes) from the binary data *
   uint32_t decodeInt32 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(binData[3])
              + (uint32_t(binData[2]) << 8)
              + (uint32_t(binData[1]) << 16)
              + (uint32_t(binData[0]) << 24) ;

      return uiVal ;
   }

   //* Scan a 16-bit integer (2 bytes) from the binary data *
   uint32_t decodeInt16 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;
      uiVal =    uint32_t(binData[1])
              + (uint32_t(binData[0]) << 8) ;

      return uiVal ;
   }

   void reset ( void )           // set all data members to default values
   {
      this->atomSize = this->dataSize = 
      this->dataConst1 = this->dataConst2 = ZERO ;
      this->atomName[0] = this->dataName[0] = '\0' ;
      this->valid = false ;
   }


   //***********************
   //* Public Data Members *
   //***********************
   uint32_t atomSize ;                 // atom size in bytes
   uint32_t dataSize ;                 // size of "data" atom in bytes
   uint32_t dataConst1 ;               // const value #1: itaMETA_ NUMERIC, 
                                       // or itaMETA_TEXT, or itaMETA_ITUNES
   uint32_t dataConst2 ;               // const value #2: always ZERO
   uint32_t aIndex ;                   // index of matching entry in liTunes[] database
   uint32_t metaVal1 ;                 // for numeric metadata, first value
   uint32_t metaVal2 ;                 // for numeric metadata, second value
   char*    metaStr ;                  // for text metadata, pointer to capture buffer
   char     atomName[intLEN + 1] ;     // atom name
   uint8_t  dataName[intLEN + 1] ;     // embedded name (usually "data")
   bool     valid ;                    // 'true' if valid data record captured

} ;   // End itAtom


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//*   Decode and store an (unencapsulated) "data" Atom.                        *
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//* Data records take the format:                                              *
//* "nnnndataval1val2meta"                                                     *
//* 'nnnn' size of the "data" atom                                             *
//* 'data' literal token                                                       *
//* 'val1' const value for "data" == 0x01 || 0x00 || 0x15                      *
//* 'val2' const value for "data" == 0x00                                      *
//* 'meta' the text or numeric metadata                                        *
//*        Text data are probably underminated UTF-8 data, but see hdlr.format.*
//*        Numeric data are assumed to be a single 32-bit, 16-bit or 8-bit     *
//*        integer value.                                                      *
//*                                                                            *
//* Note that "data" atoms are usually embedded within other atoms, but are    *
//* occasionally encountered wandering free. There may also be a "data" atom   *
//* within an unrecognized container atom, in which case we would skip over    *
//* the unrecognized container, but still identify the "data" atom it contains.*
//* In such circumstances, this class may be used to decode the atom.          *

class dataAtom
{
   public:

   //******************
   //* Public Methods *
   //******************
   ~dataAtom ( void )               // destructor
   {
      if ( this->metaStr != NULL )
         delete [] this->metaStr ;
   }

   dataAtom ( void )                // default constructor
   {
      this->reset () ;

      //* Allocate a data-capture buffer.*
      this->metaStr = new char[itaMETA_MAX] ;
   }

   //***********************************************************
   //* Decode the binary data and initialize all data members. *
   //***********************************************************
   bool initialize ( uint32_t s,  const char* n, const uint8_t* binData )
   {
      this->reset () ;
      this->atomSize = s ;                            // copy atom size
      for ( uint8_t i = ZERO ; i <= intLEN ; ++i )    // copy atom name
         this->atomName[i] = n[i] ;
      bool dkeyOk = (bool)((this->atomName[0] == genericTag.aName[0]) && 
                           (this->atomName[1] == genericTag.aName[1]) && 
                           (this->atomName[2] == genericTag.aName[2]) && 
                           (this->atomName[3] == genericTag.aName[3])) ;

      uint32_t indx = ZERO,               // scan index
               eod = atomSize - idLEN ;   // end-of-data index

      //* Decode the "data" constants *
      this->dataConst1 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;
      this->dataConst2 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Validate the record format *
      if ( dkeyOk && (this->dataConst2 == 0x00) &&
           ((this->dataConst1 == itaMETA_TEXT) ||
            (this->dataConst1 == itaMETA_NUMERIC) || 
            (this->dataConst1 == itaMETA_ITUNES) ||
            (this->dataConst1 == itaMETA_JFIF)) )
      {
         gString gstmp ;         // formatting

         //* If atom contains text data *
         if ( this->dataConst1 == itaMETA_TEXT )
         {
            #if 0    // KLUDGY FIX - iTunes ONLY
            // Programmer's Note: At least one of our sample file places a 0xA9 
            // byte as the first byte character of the metadata in a "\xa9com" atom.
            // This is unbelievably stupid because it is not UTF-8.
            if ( binData[indx] == BYTE_A9 )
               binData[indx] = ' ' ;
            #endif   // KLUDGEY FIX

            //* Copy the text metadata *
            uint32_t trg = ZERO ;
            for ( ; (indx < eod) && (trg < (itaMETA_MAX - 1)) ; ++trg )
            {
               this->metaStr[trg] = binData[indx++] ;
               if ( this->metaStr[trg] == '\0' ) break ;
            }
            this->metaStr[trg] = '\0' ;   // ensure termination

            #if 1    // Enable Beautification
            //* Strip leading and trailing whitespace from *
            //* the data, for a more professional display. *
            gstmp = this->metaStr ;
            gstmp.strip() ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            #endif   // Beautification

            this->valid = true ;
         }

         else if ( this->dataConst1 == itaMETA_NUMERIC )
         {
            //* Capture the first value               *
            //* (See note above about integer width.) *
            if ( (eod - indx) >= intLEN )
            {
               this->metaVal1 = this->decodeInt32 ( &binData[indx] ) ;
               indx += intLEN ;
            }
            else if ( (eod - indx) >= (intLEN / 2) )
            {
               this->metaVal1 = this->decodeInt16 ( &binData[indx] ) ;
               indx += (intLEN / 2) ;
            }
            else if ( (eod - indx) >= (intLEN / 4) )
            {
               this->metaVal1 = binData[indx] ;
               ++indx ;
            }
            else
               this->metaVal1 = ZERO ;
            gstmp.compose( "%u", &this->dataConst1 ) ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;

            this->valid = true ;
         }

         else if ( this->dataConst1 == itaMETA_JFIF )
         {
            //* Calculate and report image dimensions *
            // Programmer's Note: There is a potential compiler bug near here
            // regarding variable alignment. (g++ 10.1.1 20200507), Defining 
            // the 'char' array first masks the problem, but keep an eye on it.
            char     segName[64] ;
            char     imgSize[64] ;
            char     pixels[64] ;
            uint32_t soi,        // start-of-image (0xFFD8)
                     appo,       // JFIF APPO (0xFFE0)
                     appolen,    // segment size (excl. 'appo')
                     imgsize,    // image size in bytes (approximate)
                     vmaj,       // major version
                     vmin,       // minor version
                     denunits,   // density units: 00h==no units
                                 //                01h==pixels-per-inch (2.54cm)
                                 //                02==pixels-per-centimeter
                     xden,       // horizontal pixel density
                     yden ;      // vertical pixel density

            soi = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            appo = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            appolen = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            for ( uint32_t trg = ZERO ; trg <= (intLEN + 1) ; ++trg )
               segName[trg] = binData[indx++] ;
            vmaj = binData[indx++] ;
            vmin = binData[indx++] ;
            denunits = binData[indx++] ;
            xden = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            yden = this->decodeInt16 ( &binData[indx] ) ;
            indx += (intLEN / 2) ;
            imgsize = this->atomSize - appolen - 2 ;
            gstmp.formatInt( imgsize, 5, true, false, false, fiKb ) ;
            gstmp.copy( imgSize, 64 ) ;
            
            if ( denunits == 0x00 )
               gstmp.compose( "%ux%u pixels", &xden, &yden ) ;
            else if ( denunits == 0x01 )
               gstmp.compose( "%u pixels-per-inch", &xden ) ;
            else  // (denunits==0x02)
               gstmp.compose( "%u pixels-per-cm", &xden ) ;
            gstmp.copy( pixels, 64 ) ;

            #if DEBUG_M4A == 0   // PRODUCTION
            gstmp.compose( "%s, %s, Encoding:%s", imgSize, pixels, segName ) ;

            // Programer's Note: This is only to silence the compiler warning 
            // about unused variables when not in debugging mode.
            if ( (soi == 0xFFD8) || (appo == 0xFFE0) || (vmaj >= 1) || (vmin >= 1) )
            { /* do nothing */ }
            #else    // DEBUG ONLY
            gstmp.compose( "%s, %s, Encoding:%s\n"
                           "soi:%04X appo:%04X (%u x%04X bytes) "
                           "v:%02u.%02u dunits:%02X xden:%u yden:%u",
                           imgSize, pixels, segName,
                           &soi, &appo, &appolen, &appolen, &vmaj, 
                           &vmin, &denunits, &xden, &yden ) ;
            #endif   // DEBUG_M4A

            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            this->valid = true ;
         }

         #if DEBUG_M4A != 0
         //* Code for iTunes internal coded data. *
         //* Reported only in debugging mode.     *
         else if ( this->dataConst1 == itaMETA_ITUNES )
         {
            //* Capture the value *
            if ( (eod - indx) >= intLEN )
               this->metaVal1 = this->decodeInt32 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 2) )
               this->decodeInt16 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 4) )
               this->metaVal1 = binData[indx] ;
            else
               this->metaVal1 = ZERO ;
            gstmp.compose( "%04Xh", &this->metaVal1 ) ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            
            this->valid = true ;
         }
         #endif   // DEBUG_M4A

         //* Unknown data format code. Currently ignored.*
         //else
         //{
         //}
      }

      return this->valid ;
   }

   //* Scan a 32-bit integer (4 bytes) from the binary data *
   uint32_t decodeInt32 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(binData[3])
              + (uint32_t(binData[2]) << 8)
              + (uint32_t(binData[1]) << 16)
              + (uint32_t(binData[0]) << 24) ;

      return uiVal ;
   }

   //* Scan a 16-bit integer (2 bytes) from the binary data *
   uint32_t decodeInt16 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(binData[1])
              + (uint32_t(binData[0]) << 8) ;
              

      return uiVal ;
   }

   void reset ( void )              // set all data members to default values
   {
      this->atomSize = 
      this->dataConst1 = this->dataConst2 = this->metaVal1 = ZERO ;
      this->atomName[0] = '\0' ;
      this->valid = false ;
   }

   //***********************
   //* Public Data Members *
   //***********************
   uint32_t atomSize ;                 // atom size in bytes
   char     atomName[intLEN + 1] ;     // atom name
   uint32_t dataConst1 ;               // const value #1: always 0x01 || 0x00 || 0x15
   uint32_t dataConst2 ;               // const value #2: always 0x00
   uint32_t metaVal1 ;                 // for numeric metadata, first value
   char*    metaStr ;                  // pointer to metadata-text capture buffer
   bool     valid ;                    // 'true' if valid data record captured

} ;   // End dataAtom


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//*   User-defined tag atom.                                                   *
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//* Freeform records take the format:                                          *
//* "nnnn----meanzzzzaaaannnnnamezzzzNNNNnnnndatameta"                         *
//* 'nnnn' is the integer size of the record                                   *
//* '----' is the literal freeform type                                        *
//* 'mean' literal token                                                       *
//* 'zzzz' is a 32-bit integer value == 0                                      *
//* 'aaaa' this is usually "com.apple.iTunes"                                  *
//* 'nnnn' size of the "name" sub-atom                                         *
//* 'name' is the unique identifier for the record                             *
//* 'zzzz' is a 32-bit integer value == 0                                      *
//* 'NNNN' a unique tag description                                            *
//* 'nnnn' size of the "data" sub-atom                                         *
//* 'data' literal token                                                       *
//* 'val1' const value for "data" == 0x01 || 0x00 || 0x15                      *
//* 'val2' const value for "data" == 0x00                                      *
//* 'meta' the value associated with the custom name above                     *
//*        (probably underminated UTF-8 data, but see hdlr.format)             *
//*                                                                            *

luniTunesDB userdefTag = {(const uint8_t*)"----", "User-defined Tag-------"} ;
const uint8_t* userdefTag_mean = (uint8_t*)"mean" ;

class udefAtom
{
   public:

   //******************
   //* Public Methods *
   //******************
   ~udefAtom ( void )               // destructor
   {
      if ( this->metaStr != NULL )
         delete [] this->metaStr ;
   }

   udefAtom ( void )                // default constructor
   {
      this->reset () ;

      //* Allocate a data-capture buffer.*
      this->metaStr = new char[itaMETA_MAX] ;
   }

   //* Decode the binary data and initialize all data members.*
   bool initialize ( uint32_t s,  const char* n, uint8_t* binData )
   {
      this->reset () ;
      this->atomSize = s ;                            // copy atom size
      for ( uint8_t i = ZERO ; i <= intLEN ; ++i )    // copy atom name
         this->atomName[i] = n[i] ;

      uint32_t indx = ZERO,               // scan index
               trg,                       // target index
               eod = atomSize - idLEN ;   // end-of-data index
      bool     mkeyOk = false,            // true if "mean" key valid
               nkeyOk = false,            // true if "name" key valid
               dkeyOk = false ;           // true if "data" key valid

      //* Decode length of "mean" data *
      this->meanSize = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Decode the "mean" key *
      for ( trg = ZERO ; trg < intLEN ; ++trg )
         this->meanKey[trg] = binData[indx++] ;
      this->meanKey[trg] = '\0' ;
      mkeyOk = (bool)((this->meanKey[0] == 'm') && (this->meanKey[1] == 'e') && 
                      (this->meanKey[2] == 'a') && (this->meanKey[3] == 'n')) ;

      //* Decode the "mean" constant value (0x00) *
      this->meanConst = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Copy the text data of the "mean" atom *
      for ( trg = ZERO ; (trg < (this->meanSize - (intLEN * 3))) && (trg < 63) ; ++trg )
         this->meanDesc[trg] = binData[indx++] ;
      this->meanDesc[trg] = '\0' ;

      //* Decode length of "name" atom *
      this->nameSize = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Decode the "name" key *
      for ( trg = ZERO ; trg < intLEN ; ++trg )
         this->nameKey[trg] = binData[indx++] ;
      this->nameKey[trg] = '\0' ;
      nkeyOk = (bool)((this->nameKey[0] == 'n') && (this->nameKey[1] == 'a') && 
                      (this->nameKey[2] == 'm') && (this->nameKey[3] == 'e')) ;

      //* Decode the "name" constant value (0x00) *
      this->nameConst = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Copy the text data of the "name" atom *
      for ( trg = ZERO ; (trg < (this->nameSize - (intLEN * 3))) && (trg < 63) ; ++trg )
         this->nameDesc[trg] = binData[indx++] ;
      this->nameDesc[trg] = '\0' ;

      //* Decode length of "data" atom *
      this->dataSize = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Decode the "data" key *
      for ( trg = ZERO ; trg < intLEN ; ++trg )
         this->dataKey[trg] = binData[indx++] ;
      this->dataKey[trg] = '\0' ;
      dkeyOk = (bool)((this->dataKey[0] == 'd') && (this->dataKey[1] == 'a') && 
                      (this->dataKey[2] == 't') && (this->dataKey[3] == 'a')) ;

      //* Decode the "data" constants *
      this->dataConst1 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;
      this->dataConst2 = this->decodeInt32 ( &binData[indx] ) ;
      indx += intLEN ;

      //* Validate the record format *
      if ( mkeyOk && nkeyOk && dkeyOk && 
           (this->meanConst == 0x00) && (this->nameConst == 0x00) &&
           (this->dataConst2 == 0x00) &&
           #if 0     // Allow text metadata only
           (this->dataConst1 == itaMETA_TEXT) )
           #else     // Allow both text and numeric metadata
           ((this->dataConst1 == itaMETA_TEXT) ||
            (this->dataConst1 == itaMETA_NUMERIC) || 
            (this->dataConst1 == itaMETA_ITUNES)) )
           // Note: Numeric metadata is not supported at this time.
           #endif   // 
      {
         gString gstmp ;         // formatting

         //* Copy the user-defined text metadata to our data member *
         if ( this->dataConst1 == itaMETA_TEXT )
         {
            for ( trg = ZERO ; trg < (itaMETA_MAX - 1) ; ++trg )
            {
               this->metaStr[trg] = binData[indx++] ;
               if ( this->metaStr[trg] == '\0' ) break ;
               if ( indx >= eod ) break ;
               if ( trg >= (this->dataSize - (intLEN * 4)) ) break ;
            }
            this->metaStr[trg] = '\0' ;   // ensure termination

            #if 1    // Enable Beautification
            //* Strip leading and trailing whitespace from *
            //* the data, for a more professional display. *
            gstmp = this->metaStr ;
            gstmp.strip() ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;
            #endif   // Beautification

            valid = true ;
         }

         //* Decode and store user-defined integer metadata.*
         else if ( this->dataConst1 == itaMETA_NUMERIC )
         {
            //* Capture the value *
            if ( (eod - indx) >= intLEN )
               this->metaVal1 = this->decodeInt32 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 2) )
               this->metaVal1 = this->decodeInt16 ( &binData[indx] ) ;
            else if ( (eod - indx) >= (intLEN / 4) )
               this->metaVal1 = ZERO ;
            gstmp.compose( "%u", &this->metaVal1 ) ;
            gstmp.copy( this->metaStr, itaMETA_MAX ) ;

            this->valid = true ;
         }
      }

      return this->valid ;
   }

   //* Scan a 32-bit integer (4 bytes) from the binary data *
   uint32_t decodeInt32 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(binData[3])
              + (uint32_t(binData[2]) << 8)
              + (uint32_t(binData[1]) << 16)
              + (uint32_t(binData[0]) << 24) ;

      return uiVal ;
   }

   //* Scan a 16-bit integer (2 bytes) from the binary data *
   uint32_t decodeInt16 ( const uint8_t* binData ) const
   {
      uint32_t uiVal = ZERO ;
      uiVal =    uint32_t(binData[1])
              + (uint32_t(binData[0]) << 8) ;

      return uiVal ;
   }

   void reset ( void )              // set all data members to default values
   {
      this->atomSize = 
      this->meanSize = this->meanConst = 
      this->nameSize = this->nameConst = 
      this->dataSize = this->dataConst1 = this->dataConst2 = this->metaVal1 = ZERO ;
      this->atomName[0] = 
      this->meanKey[0] = this->meanDesc[0] = 
      this->nameKey[0] = this->nameDesc[0] = 
      this->dataKey[0] = '\0' ;
      this->valid = false ;
   }

   //***********************
   //* Public Data Members *
   //***********************
   uint32_t atomSize ;                 // atom size in bytes
   char     atomName[intLEN + 1] ;     // atom name

   uint32_t meanSize ;                 // size of "mean" atom
   char     meanKey[intLEN + 1] ;      // "mean" keyword
   char     meanDesc[64] ;             // description (usually "com.apple.iTunes")
   uint32_t meanConst ;                // constant following "mean" (s/b zero)
   
   uint32_t nameSize ;                 // size of "name" atom
   char     nameKey[intLEN + 1] ;      // "name" keyword
   uint32_t nameConst ;                // constant following "name" (s/b zero)
   char     nameDesc[64] ;             // custom tag description

   uint32_t dataSize ;                 // size of "data" atom
   char     dataKey[intLEN + 1] ;      // "data" keyword
   uint32_t dataConst1 ;               // const value #1: always 0x01 || 0x00 || 0x15
   uint32_t dataConst2 ;               // const value #2: always 0x00
   uint32_t metaVal1 ;                 // for numeric metadata
   char*    metaStr ;                  // pointer to metadata-text capture buffer

   bool     valid ;                    // 'true' if valid data record captured

} ;   //* End udefAtom *


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//*   Metadata Box structure to contain M4A metadata.                          *
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
class mdBox
{
   public:

   //******************
   //* Public Methods *
   //******************
   virtual ~mdBox ( void )             // Destructor
   {
      //* Release dynamic memory allocation.*
      if ( this->metadata != NULL )
         delete [] this->metadata ;
   }

   mdBox ( void )                      // Default constructor
   {
      this->reset() ;
   }

   //* Initialize data members from the specified mdBox object.*
   //* The 'next' and 'prev' members are not modified.         *
   void operator= ( const mdBox* srcPtr )
   {
      this->size      = srcPtr->size ;
      this->size64    = srcPtr->size64 ;
      this->metasize  = srcPtr->metasize ;
      this->tindex    = srcPtr->tindex ;
      this->metadata  = srcPtr->metadata ;
      this->textmeta  = srcPtr->textmeta ;
      this->imagemeta = srcPtr->imagemeta ;
      this->valid     = srcPtr->valid ;
      for ( uint32_t indx = ZERO ; indx <= intLEN ; ++indx )
         this->type[indx] = srcPtr->type[indx] ;
   }

   //* Decode the object type (name) and length.                    *
   //* Returns: 'true' if successful. Initialized data members:     *
   //*                 'size', 'type', 'metasize', 'tindex'         *
   //*          'false' if name is not ASCII or if                  *
   //*                  name is not on the list of registered names.*
   //*                  or if length >= MAX_BOXLEN                  *
   bool decodeHeader ( const uint8_t* ubuff )
   {
      // Programmer's Note: We arbitrarily limit the size field on the 
      // assumption that no music file will be larger that this limit, and 
      // that if the size is larger, then it is not a valid size field.
      // ALSO note the special size values:
      //    1 == actual size is an embedded 64-bit value
      //    0 == box size not given (lazy) because box continues to end-of-file
      const uint32_t MAX_BOXLEN = 0x01312D00 ;     // 20 megabytes

      this->size = ZERO ;     // reinitialize target data members
      *this->type = '\0' ;
      this->valid = false ;   // (return value)

      //* If the type field contains only ASCII characters, *
      //* it MAY BE a box type identifier.                  *
      if ( (this->isAsciiPrint ( &ubuff[intLEN] )) )
      {
         //* Isolate the (potential) box length and *
         //* the (potential) box name fields.       *
         uint32_t sz ;
         if ( (sz = this->decodeInt32 ( ubuff )) < MAX_BOXLEN )
         {
            //* Copy the (potential) name to temp buffer.*
            char tmpType[intLEN + 1] ;
            for ( uint32_t trg = ZERO, src = intLEN ; trg < intLEN ; )
               tmpType[trg++] = ubuff[src++] ;
            tmpType[intLEN] = '\0' ;
            uint32_t tableIndex ;
            if ( (tableIndex = this->validateType ( tmpType )) < MAX_BOXTYPES )
            {
               //* Null-terminated 'type' (box name) *
               for ( uint32_t i = ZERO ; i <= intLEN ; ++i )
                  this->type[i] = tmpType[i] ;
               this->tindex = tableIndex ;            // index into 'type' table
               this->size = sz ;                      // object data bytes
               this->metasize = this->size - idLEN ;  // data bytes excl. header

               #if 0    // currently unused
               //* Decode 64-bit box size.                   *
               //* (This should not happen in a music file.) *
               if ( this->size == 1 )
                  this->size64 = this->decodeInt64 ( &ubuff[intLEN * 2] ) ;
               #endif   // U/C

               this->valid = true ;
            }
         }
      }
      return this->valid ;
   }

   //* Scan a 32-bit integer (4 bytes) from the raw data *
   uint32_t decodeInt32 ( const uint8_t* ubuff ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(ubuff[3])
              + (uint32_t(ubuff[2]) << 8)
              + (uint32_t(ubuff[1]) << 16)
              + (uint32_t(ubuff[0]) << 24) ;

      return uiVal ;
   }

   //* Compare raw data to the list of MPEG-4 box types. *
   //* If a valid box type, initialize 'type' member.    *
   //*           (case-insentitive comparison)           *
   uint32_t validateType ( const char* ubuff )
   {
      const uint8_t UPPERCASE = 0x30 ;
      uint32_t src, tab, upp ;
      uint32_t indx = ZERO, offset ;
      for ( ; indx < MAX_BOXTYPES ; ++indx )
      {
         offset = ZERO ;
         for ( offset = ZERO ; offset < intLEN ; ++offset )
         {
            src = (uint32_t)ubuff[offset] ;
            tab = (uint32_t)mpeg4_boxType[indx][offset] ;   // lowercase
            upp = tab + UPPERCASE ;                         // uppercase
            if ( (src != tab) && (src != upp) )
               break ;
         }
         if ( offset == intLEN ) // if matching substring found
            break ;
      }
      return indx ;
   }

   //* Analyze four(4) bytes of raw data. If each of the four bytes *
   //* is in the range of ASCII printing characters, it MAY be the  *
   //* name of an MPEG-4 box-object name.                           *
   bool isAsciiPrint ( const uint8_t* ubuff )
   {
      bool status = false ;
   
      if ( (ubuff[0] >= ' ') && (ubuff[0] <= '~') &&
           (ubuff[1] >= ' ') && (ubuff[1] <= '~') &&
           (ubuff[2] >= ' ') && (ubuff[2] <= '~') &&
           (ubuff[3] >= ' ') && (ubuff[3] <= '~') )
         status = true ;
      return status ;
   }  //* End isAsciiPrint() *

   //* Reset all data members to default values.*
   void reset ( void )
   {
      this->size = this->metasize = this->handler = ZERO ;
      this->tindex = MAX_BOXTYPES ;
      this->size64 = ZERO ;
      this->type[0] = '\0' ;
      this->metadata = NULL ;
      this->textmeta = this->imagemeta = this->valid = false ;
   }


   //***********************
   //* Public Data Members *
   //***********************
   uint32_t size ;               // record size
   uint64_t size64 ;             // record size (if 64-bit)
   uint32_t metasize ;           // for metadata boxes only: size of metadata
   uint32_t tindex ;             // index of box type name in mpeg4_boxType[] array
   uint32_t handler ;            // data handler type code
   uint8_t  type[intLEN + 1] ;   // record type (name)
   uint8_t *metadata ;           // if 'textmeta', then contains UTF-8 text
                                 // if 'imagemeta' contains filespec of image file
   bool     textmeta ;           // 'true' if text metadata captured
   bool     imagemeta ;          // 'true' if image metadata captured
   bool     valid ;              // 'true' if both 'size' and 'type' fields initialized

} ;   //* End mdBox *


//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
//* A valid M4A File Header Box will be of the form:                           *
//* nnnnnnnn tttttttt MMMMMMMM mmmmmmmm (compatible brand info)                *
//* where nnnnnnnn == record length (32-bit BigEndian)                         *
//*       tttttttt == "ftyp"                                                   *
//*       MMMMMMMM == major brand   (32-bit)                                   *
//*       mmmmmmmm == minor version (32-bit)                                   *
//* A TYPICAL File Header Box will contain something similar to:               *
//*     00000018   == record length, in this example 18 hex (24 bytes decimal) *
//*       "ftyp"   == constant substring                                       *
//*       "M4A "   == major file-type identifier (brand)                       *
//*     00000200   == minor version (informational only)                       *
//*     "isomiso2" == where "isom" is optional and indicates the first version *
//*                   of the MP-4 specification, and "iso2" indicates          *
//*                   compliance with the ammended version of the specification*
//*                   Additional compatibility tokens which may be seen are    *
//*                   "M4A ", "mp42"                                           *
//*                                                                            *
//* NOTE: The spec does not instist that this be the first record, but should  *
//*       be "placed as early as possible". For instance, a Quicktime container*
//*       tag might be seen first; however, we look only at files with an      *
//*       ".m4a" extension, so the M4A header should be seen first.            *
//*                                                                            *
//* Programmer's Note: It would be a minor effort to include extraction of     *
//* metadata from .MP4 (video) files. Some valid 'ftype' records would be      *
//* "M4V ", "MP42", "isom". 'Compatibility' records have similar differences   *
//* from M4A audio files.                                                      *
//*                                                                            *
//*   ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----  ----   *
const uint32_t MAX_COMPAT = (intLEN * 8) ;   // size of 'compat' buffer
class m4aFileHdr
{
   public:

   //******************
   //* Public Methods *
   //******************
   // Default constructor *
   m4aFileHdr ( void )
   {
      this->reset() ;
   }

   //* Initialization constructor *
   m4aFileHdr ( uint32_t boxSize, const char* binaryRecord )
   {
      this->reset() ;
      this->size = boxSize ;
      uint32_t src = ZERO,
               eod  = this->size - intLEN,
               trg ;
      for ( trg = ZERO ; trg < intLEN && src < eod ; ++trg )
         this->type[trg] = binaryRecord[src++] ;
      this->type[trg] = '\0' ;
   }

   //* initialize the data members with the provided raw information.*
   //* Input  : boxSize: total number of bytes of input data
   //*          boxType: box type string (s/b "ftyp")
   //*          binData: raw data to be formatted
   //* Returns: 'true' if valid M4A file header
   //*          'false' if validation failed
   bool initialize ( uint32_t boxSize, const uint8_t* boxType, 
                     const uint8_t* binData )
   {
      this->reset() ;                     // reset to default values
      this->size = boxSize ;              // record size

      uint32_t eod = this->size - idLEN,  // loop control
               src = ZERO, trg = ZERO ;   // buffer indices

      //* Record Type *
      for ( ; trg <= intLEN ; ++trg )
         this->type[trg] = boxType[trg] ;
      this->type[intLEN] = '\0' ;

      //* "major" version (s/b "MP4 ") *
      for ( trg = ZERO ; trg < intLEN && src < eod ; ++trg )
         this->major[trg] = binData[src++] ;
      this->major[trg] = '\0' ;

      //* "minor" version (s/b  ) *
      this->minor = this->decodeInt32 ( &binData[src] ) ;
      src += intLEN ;

      //* "Compatibility Information: MPEG-4/ISO specification version *
      for ( trg = ZERO ; (trg < MAX_COMPAT) && (src < eod) ; ++trg )
         this->compat[trg] = binData[src++] ;
      this->compat[trg] = '\0' ;

      return ( this->validate() ) ;
   }  // initialize() *

   //* Scan a 32-bit integer (4 bytes) from the binary data *
   uint32_t decodeInt32 ( const uint8_t* binaryRecord ) const
   {
      uint32_t uiVal = ZERO ;

      uiVal =    uint32_t(binaryRecord[3])
              + (uint32_t(binaryRecord[2]) << 8)
              + (uint32_t(binaryRecord[1]) << 16)
              + (uint32_t(binaryRecord[0]) << 24) ;

      return uiVal ;
   }

   //* Test for valid file-header data *
   bool validate ( void )
   {
      const char* recType = "ftyp" ;
      const char* recMaj  = "M4A " ;
      const char* verOne  = "isom" ;
      const char* verTwo  = "iso2" ;

      this->valid = false ;         // assume an invalid record
      gString gs( this->type ) ;    // comparisons NOT case-sensitive
      if ( (gs.compare( recType, false )) == ZERO )
      {
         gs = this->major ;
         if ( (gs.compare( recMaj, false )) == ZERO )
         {
            gs = this->compat ;
            if ( ((gs.find( verTwo )) >= ZERO) ||
                 ((gs.find( verOne )) >= ZERO) )
            {
               this->valid = true ;
            }
         }
      }
      return this->valid ;
   }

   //* Initialize all data members to default values *
   void reset ( void )
   {
      this->size = this->minor = ZERO ;
      this->type[0] = this->major[0] = this->compat[0] = '\0' ;
      this->valid = false ;
   }

   //***********************
   //* Public Data Members *
   //***********************
   uint32_t size ;               // record size
   char     type[intLEN + 1] ;   // record type ( "ftyp" )
   char     major[intLEN + 1] ;  // major version ( "M4A " )
   uint32_t minor ;              // minor version
   char     compat[MAX_COMPAT] ; // additional compatible types
   bool     valid ;              // 'true' if object contains a valid 'ftyp' record
} ;   // m4aFileHdr


//* Define a class for passing data among the M4A methods.*
class m4aParms
{
   public:
   virtual ~m4aParms ( void )          // destructor
   {
      if ( this->ofs.is_open() )       // if target file is open, close it
         this->ofs.close() ;
      if ( this->ubuff != NULL )       // release dynamic allocation
         delete [] this->ubuff ;
   }

   m4aParms ( void )                   // default constructor
   {
      this->reset () ;
   }

   //* Initialization constructor.                                            *
   //*                                                                        *
   //* Input:                                                                 *
   //*   bufflen  : number of bytes to allocate for input buffer              *
   //*   srcsize  : total bytes in source file                                *
   //*   trgspec  : filespec of existing temporary file to contain captured   *
   //*              and formatted metadata                                    *
   //*   verb     : if 'true' produce verbose output (currently unused)       *
   //*              if 'false' produce standard output                        *
   //*                                                                        *
   m4aParms ( uint32_t bufflen, uint32_t srcsize, const char* trgspec, bool verb )
   {
      this->reset () ;
      //* Allocate an input buffer *
      this->ubuff = new uint8_t[bufflen] ;
      //* Save size of source file *
      this->srcBytes = srcsize ;
      //* Open the target file (append) *
      this->ofs.open( trgspec, (ofstream::out | ofstream::app) ) ;
      //* Flag signals standard vs. verbose metadata reporting *
      this->verbose = verb ;
   }

   void reset ( void )                 // set data members to default values
   {
      this->ubuff = NULL ;
      this->mdbList = NULL ;
      this->mdbCount = this->fileBytes = this->srcBytes = ZERO ;
      this->eof = this->verbose = false ;
      #if DEBUG_M4A != 0
      this->srcoffset = ZERO ;
      #endif   // DEBUG_M4A
   }

   //** Data Members **
   std::ifstream ifs ;              // source file (input) stream
   std::ofstream ofs ;              // target file (output) stream
   uint8_t *ubuff ;                 // input stream buffer
   m4aFileHdr fileHdr ;             // file header info
   mdBox   *mdbList ;               // anchor for linked list of records
   mdBox    mdbMeta ;               // info for a "meta" box object
   mdBox    mdbHdlr ;               // info for a "meta.hdlr" box object
   uint32_t mdbCount ;              // number of metadata records captured to 'mdbList'
   uint32_t fileBytes ;             // index into source file (bytes read)
   uint32_t srcBytes ;              // source file total number of bytes
   bool     eof ;                   // 'true' if end-of-file reached
   bool     verbose ;               // verbose output (currently unused)
   #if DEBUG_M4A != 0
   uint32_t srcoffset ;             // for debugging only: report offset in source file
   #endif   // DEBUG_M4A
} ;   // m4aParms

//* Prototypes for non-member methods.*
static void m4aScan4Metadata ( m4aParms& p ) ;
static void m4aExtractMetadata ( m4aParms& p, mdBox& metabox ) ;
static void m4aExtractMD_ilst ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_xml ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_bxml ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_schi ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_free ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_skip ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_uuid ( m4aParms& p, const mdBox& mdbCon ) ;
static bool m4aReadFileHeader ( m4aParms& p ) ;
static uint32_t m4aReadBytes ( m4aParms& p, uint32_t byteCnt, uint32_t offset = ZERO ) ;
static uint32_t m4aReadHeader ( m4aParms& p ) ;
static uint32_t m4aShiftHeader ( m4aParms& p ) ;
static uint32_t m4aDiscardBytes ( m4aParms& p, uint32_t byteCnt ) ;
static uint32_t m4aIsLooneyTunes ( const uint8_t* atomName ) ;


//*************************
//*  ExtractMetadata_M4A  *
//*************************
//******************************************************************************
//* Read the media file and write the metadata to the temp file.               *
//*                                                                            *
//* Input  : ofs    : open output stream to temporary file                     *
//*          srcPath: filespec of media file to be read                        *
//*          trgPath: filespec of open temp file (see 'ofs')                   *
//*          vebose : (optional, 'false' by default)                           *
//*                   if 'false', display all text-frame records               *
//*                   if 'true', ALSO display extended technical data and      *
//*                              non-text frames (used primarily for debugging)*
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Notes:                                                                     *
//* ------                                                                     *
//* Tree Structure:                                                            *
//* Box structures may be, and often are nested to an arbitrary level.         *
//* This nested design provides flexibility, but mandates an incredibly slow   *
//* slow processing time when scanning for tag data.                           *
//*                                                                            *
//* Alignment:                                                                 *
//* The specification states that all box objects are to be aligned on an      *
//* 8-byte boundary. Unsurprisingly, this part of the specification is         *
//* completely ignored by coders.                                              *
//* For this reason, the most straight-forward scanning method is to read the  *
//* file in groups of eight(8) bytes (two 32-bit integers). When synchronized  *
//* with the data, this yields the basic information needed to identify the    *
//* object.                                                                    *
//*    Integer 0: is the size of the object in bytes. (32-bits big-endian)     *
//*               Two exceptions:                                              *
//*               a) A size value of zero (0x00000000) indicates that the      *
//*                  object continues to the end-of-file.                      *
//*               b) A size value of one (0x00000001) indicates that the       *
//*                  object contains more data that can be represented in      *
//*                  32 bits. In this case, the actual object size is contained*
//*                  in a 64-bit big-endian value located immediately after the*
//*                  object name. Because music files are not that large, we   *
//*                  make little or no effort to support objects which are     *
//*                  of 64-bit size.                                           *
//*    Integer 1: is four(4) ASCII-alpha bytes which constitute the name of    *
//*               the object. Two exceptions:                                  *
//*               a) The Apple(tm) iTunes dictionary of tag names uses the     *
//*                  idiotic value of 0xA9 as the first byte for several of    *
//*                  the defined tag names.                                    *
//*               b) The free-form tag name is the literal "----" i.e. four    *
//*                  consecutive hyphen characters.                            *
//* In order to remain synchronized with the data, it is necessary to carefully*
//* count the number of bytes read for each object, and if the length of the   *
//* object is not an even multiple of eight bytes, the raw data must be parsed *
//* to locate the start of the next object.                                    *
//*                                                                            *
//* File Header:                                                               *
//* ------------                                                               *
//* Every file must have an "ftyp" box object as close to the top of the file  *
//* as possible. If a file type or container type requires a signature at the  *
//* top of the file, then the "ftyp" object will follow it. The structure and  *
//* content of this object is discussed elsewhere in this module.              *
//*                                                                            *
//* Container Box Objects:                                                     *
//* ----------------------                                                     *
//* Metadata must be contained within another box object. The most common is   *
//* the "meta" box object. Not all metadata is human readable. This algorithm  *
//* identifies each object, but decodes only those objects which contain       *
//* text or image data.                                                        *
//*                                                                            *
//*                                                                            *
//*                                                                            *
//******************************************************************************

void FileDlg::ExtractMetadata_M4A ( ofstream& ofs, const gString& srcPath, 
                                    const gString& trgPath, bool verbose )
{
   gString gsOut, gstmp ;        // text formatting
   tnFName fn ;                  // file stats
   this->fmPtr->GetFileStats ( fn, srcPath, true ) ;
   gstmp.formatInt( fn.fBytes, 11, true ) ;

   //* Close the target file so we can open it with our own stream.*
   //* Will be re-opened with caller's object before return.       *
   ofs.close() ;

   //* Initialize parameter block *
   m4aParms p ( ubuffLEN, fn.fBytes, trgPath.ustr(), verbose ) ;

   //* Open the source file *
   p.ifs.open( srcPath.ustr(), ifstream::in ) ;
   if ( (p.ifs.is_open()) && (p.ofs.is_open()) )
   {
      short fni = srcPath.findlast( fSLASH ) + 1 ;
      p.ofs << "FILE NAME  : \"" << &srcPath.ustr()[fni] 
          << "\"  (" << gstmp.ustr() << " bytes)\n\n" ;

      //************************************************
      //* Read the file header and verify that we have *
      //* a valid M4A (MPEG-4 audio) file, then scan   *
      //* the remainder of the file. Extract, format   *
      //* and display text and image metadata.         *
      //************************************************
      if ( (m4aReadFileHeader ( p )) )
         m4aScan4Metadata ( p ) ;
      else
         p.ofs << "Error: not a valid MPEG-4 audio (.m4a) file.\n" ;

      p.ifs.close() ;           // close the source file

   }     // ifs.is_open()
   else
      p.ofs << "Sorry, unable to open file for reading.\n" ;
   p.ofs << endl ;

   //* Close our local output-stream handler.*
   p.ofs.close() ;
   //* Re-open caller's target file (append) *
   ofs.open( trgPath.ustr(), (ofstream::out | ofstream::app) ) ;

}  //* End ExtractMetadata_M4A() *

//*************************
//*   m4aScan4Metadata    *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Caller has read the file header and has verified that the file contains    *
//* MPEG-4 audio data (M4A).                                                   *
//*                                                                            *
//* 1) Scan the source data for recognized box-object names.                   *
//* 2) For box objects which may contain human readable (text or image) data,  *
//* 3) Extract the human-readable data and format it for display.              *
//* 4) Write the formatted data to the temp file to be displayed to the user.  *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*                                                                            *
//* Returns: pointer to a linked list of mdBox objects                         *
//*          If no text or image metadata found, then a NULL pointer will      *
//*          be returned.                                                      *
//******************************************************************************
//* 1) Data format:                                                            *
//*    A valid M4A "Box" header will be of the form:                           *
//*    nnnnnnnn tttttttt (exsize) vv ffffff                                    *
//*    where nnnnnnnn == record length (32-bit Big Endian)                     *
//*          tttttttt == record type (32-bit, usually 4 ASCII bytes)           *
//*          exsize   == expanded size (optional 64-bit field)                 *
//*                      If record length == 1, then the true record length is *
//*                      is this 64-bit value following the record type        *
//*          vv       == 8-bit version                                         *
//*          ffffff   == 24-bit flags field                                    *
//*                                                                            *
//*    Like most specifications in the audio/video world, this structure       *
//*    cannot be relied upon. Fortunately, the version and flags are currently *
//*    fixed at zero (0x00000000), so whether or not they are actually present,*
//*    it is unnecessary to decode them.                                       *
//*                                                                            *
//* 2) Logic:                                                                  *
//*    a) According to the specification, all box objects must be aligned on   *
//*       an 8-byte boundary. Unfortunately, the real-world sample files show  *
//*       that this is not always the case.                                    *
//*       When a misaligned box object is encountered, boxSize mod 8 != ZERO.  *
//*       It is necessary to re-synch the stream so we do not miss any box     *
//*       objects.                                                             *
//*    b) A box object has a minimum of eight(8) bytes, a 32-but integer size  *
//*       member and a 32-bit name (usually) consisting of four ASCII          *
//*       characters.                                                          *
//*    c) A non-empty object will have a minimum of sixteen(16) bytes.         *
//*    d) A box object may contain other box objects.                          *
//*       To contain another object the container box size must be at least    *
//*       sixteen(16) bytes. (at least 24 bytes to contain a non-empty object) *
//*                                                                            *
//* 3) Implementation:                                                         *
//*    a) On entry, the input stream is on an 8-byte boundry.                  *
//*    b) The input stream is read in increments of eight(8) bytes.            *
//*       Each increment is tested for a format that is potentially a box      *
//*       header (see item 2b).                                                *
//*    c) When a potential box header is identified, the name field is         *
//*       compared with the list of box-object names to verify that a box      *
//*       object has been located.                                             *
//*    d) If the object name is a recognized metadata box name, the object is  *
//*       decoded and the contents formatted for display.                      *
//*       Certain local classes are defined for object decoding, storage and   *
//*       formatting. See definition of the mdBox class.                       *
//*                                                                            *
//******************************************************************************

static void m4aScan4Metadata ( m4aParms& p )
{
   mdBox    mdbRef ;                // temp object for data analysis
   uint32_t idLen ;                 // bytes returned by source read

   #if DEBUG_M4A != 0
   #define DEBUG_M4A_VERBOSE (0)
   uint32_t modulo ;                // modulus of eight(8) bytes
   #endif   // DEBUG_M4A

   p.mdbCount = ZERO ;        // initialize caller's counter

   //* Load buffer with new test data.*
   idLen = m4aReadHeader ( p ) ;

   //* Read until end-of-file reached *
   while ( ! p.eof )
   {
      #if DEBUG_M4A != 0 && DEBUG_M4A_VERBOSE != 0
      gsdbg1.compose( "RAW: %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX",
                      &p.ubuff[0], &p.ubuff[1], &p.ubuff[2], &p.ubuff[3], 
                      &p.ubuff[4], &p.ubuff[5], &p.ubuff[6], &p.ubuff[7] ) ;
      p.ofs << gsdbg1.ustr() << endl ;
      #endif   // DEBUG_M4A && DEBUG_M4A_VERBOSE

      //* If all four characters of the name field are ASCII printing *
      //* characters AND if name is a registered box type.            *
      if ( (mdbRef.decodeHeader ( p.ubuff )) )
      {
         #if DEBUG_M4A != 0
         modulo = mdbRef.size % mdbRef.size ;
         gsdbg2.formatInt( mdbRef.size, 11, true ) ;
         gsdbg1.compose( "BOX: %s (%s bytes at x%08X)", 
                        mdbRef.type, gsdbg2.ustr(), &p.srcoffset ) ;
         p.ofs << gsdbg1.ustr() ;
         if ( modulo != ZERO )
         {
            gsdbg1.compose( " mod:%u", &modulo ) ;
            p.ofs << gsdbg1.ustr() ;
         }
         p.ofs << endl ;
         // Programmer's Note: The reported position of the box object in the 
         // stream (p.srcoffset) is approximate only due to misalignment of 
         // preceeding objects. According to the specification, all box objects 
         // must be aligned on an 8-byte boundary, but in practice, they are not.
         #endif   // DEBUG_M4A

         //* If the current box object is a "meta" box, *
         //* it may contain text or image metadata.     *
         //* Extract and display the metadata.          *
         if ( mdbRef.tindex == metaINDEX )
         {
            m4aExtractMetadata ( p, mdbRef ) ;
         }
         //* If the current box object is a "free" box, *
         //* it may contain interesting data, but no    *
         //* embedded boxes.                            *
         else if ( mdbRef.tindex == freeINDEX )
         {
            m4aExtractMD_free ( p, mdbRef ) ;
         }
         //* If the current box object is a "skip" box, *
         //* it may contain interesting data, either    *
         //* directly or in an embedded "udta" or       *
         //* "cprt" box.                                *
         else if ( mdbRef.tindex == skipINDEX )
         {
            m4aExtractMD_skip ( p, mdbRef ) ;
         }
         //* If the current box object is a "udta" box, *
         //* it may contain interesting data, either a  *
         //* "cprt" box or another user-oriented box    *
         //* such as a "meta" box.                      *
         else if ( mdbRef.tindex == udtaINDEX )
         {
            /* Do nothing. Allow the main loop to identify  *
             * boxes/atoms contained within the 'udata' box.*/
         }
         //* If the current box object is a "uuid" box, *
         //* it may contain interesting user-defined    *
         //* data, but no embedded boxes.               *
         else if ( mdbRef.tindex == uuidINDEX )
         {
            m4aExtractMD_uuid ( p, mdbRef ) ;
         }
         //* If the current box object is a "schi" box, *
         //* it may contain the name of the person who  *
         //* downloaded the file from iTunes.           *
         if ( mdbRef.tindex == schiINDEX )
         {
            m4aExtractMD_schi ( p, mdbRef ) ;
         }
         //* Box objects which are "known" to contain   *
         //* neither interesting data nor embedded boxes*
         else if ( (mdbRef.tindex == pdinINDEX) ||
                   (mdbRef.tindex == mdatINDEX) )
         {
            m4aDiscardBytes ( p, mdbRef.metasize ) ;
         }
         //* Else, unhandled box type. Ignore it and    *
         //* scan its contents for embedded box objects.*
         else
         {
         }

         //* Load buffer with new test data.*
         if ( (idLen = m4aReadHeader ( p )) < idLEN )
         {  //* Fewer than idLEN bytes read. *
            //* Read to end-of-file.         *
            m4aReadBytes ( p, idLEN ) ;
         }
      }              // decodeHeader()
      else
      {  //* Invalid header. Shift, read and retest.*
         m4aShiftHeader ( p ) ;
      }
   }                    // while()

   #undef DEBUG_M4A_VERBOSE
}  //* End m4aScan4Metadata() *

//*************************
//*  m4aExtractMetadata   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Caller has read the header information for the "meta" box object           *
//* ('metabox' parameter). This is a container for other box objects, some of  *
//* which may contain text or image metadata.                                  *
//*                                                                            *
//* Read and decode each object in the container, capturing any human-readable *
//* data.                                                                      *
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          metabox: (by reference) contains basic information for "meta"     *
//*                    container: 'size', 'type', 'metasize', 'tindex', 'valid'*
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Notes:                                                                     *
//* ------                                                                     *
//* The "meta" container box may contain any of several different box types;   *
//* however, most are data-stream information.                                 *
//* We are interested primarily in the 'ilst' object which contains any text   *
//* metadata based on the Apple iTunes(tm) dictionary.                         *
//* Additional interesting types within a "meta" container are: 'xml ' and     *
//* 'bxml'.                                                                    *
//* Other contained objects such as the additional boxes listed under "meta"   *
//* in the specification (see below) as well as 'free', 'skip', 'udta', 'cprt',*
//* 'uuid' may contain some user-display data, but currently we have no samples*
//* of those for testing, so they are ignored.                                 *
//*                                                                            *
//* A) The 'meta' object contains various types of metadata.                   *
//*    uint32_t box size                                                       *
//*    uint32_t "meta"                                                         *
//*    uint32_t version == 0                                                   *
//*                                                                            *
//*    1) The 'hdlr' type (handler) object must be present. All other types    *
//*       are optional.                                                        *
//*       a) uint32_t version == 0 (box-type version)                          *
//*       b) uint32_t byte 3 == version == 0, bytes 2-0 predefined == 0x000000 *
//*       c) uint32_t handler type (format of meta box contents)               *
//*          The MPEG-4 specification only says that this is: "an appropriate  *
//*          value to indicate the format of the meta box contents".           *
//*          Not very bloody helpful, but it is assumed that 0x00 == UTF-8.    *
//*          For Quicktime, this is "mdta".                                    *
//*          For non-metadata, this should be 'vide', 'soun', 'hint            *
//*       d) 3 x uint32_t reserved, all set to 0                               *
//*          [NOTE: This may not be true for non-Quicktime files.]             *
//*          In the real world, this is two(2) integers set to zero, with the  *
//*          third set to "mdir" and the name "appl", OR two zero integers     *
//*          followed by the name "mdirappl". Dealer's choice....              *
//*       e) A null-terminated UTF-8 string: name of handler                   *
//*                                                                            *
//*    2) The 'ilst' sub-type (item list) contains the formatted metadata.     *
//*       Note that 'ilst' is NOT part of the MPEG-4 specification.            *
//*       The 'ilst' header is of the form: "nnnnilstxxxxyyyy", where          *
//*       'nnnn' is the size of the ilst box                                   *
//*       'ilst' is the box name                                               *
//*       'xxxx' is an integer value indicating ??                             *
//*       'yyyy' is an integer value indicating ??                             *
//*                                                                            *
//*       a) Metadata chunks are defined by the iTunes database of tag names.  *
//*          See liTunes[] array.                                              *
//*          Format: "nnnnAAAAdatainfo"                                        *
//*          'nnnn' is the integer size of the record                          *
//*          'AAAA' is a 4-byte identifier, either four ASCII alpha bytes _OR_ *
//*                 the hex value 0xa9 followed by three ASCII alpha bytes     *
//*          'data' is a "data" atom record: four 32-bit integers + string data*
//*                 (see "dataAtom" class and "genericTag" const above)        *
//*                                                                            *
//*                                                                            *
//*    3) The 'xml' type object contains XML string data, usually UTF-8 format,*
//*       but potentially another format specified by the handler object.      *
//*                                                                            *
//*    4) The 'bxml' type object contains unsigned-byte (uint8_t) binary data, *
//*       which fills the box.                                                 *
//*                                                                            *
//*    5) The 'iloc' type (location) object specifies the location of metadata *
//*       either in the current file or an external file.                      *
//*       a) This object begins with three 4-bit integer values:               *
//*          -- offset_size: bytes of the offset field                         *
//*          -- length_size: bytes of the length field                         *
//*          -- base_offset_size: bytes of the base_offset field               *
//*       b) The values must be of the set {0,4,8}                             *
//*          -- An offset of 0 indicates beginning of file.                    *
//*          -- A length of 0 indicates the entire file.                       *
//*       c) The 4th integer (4-bit) is reserved.                              *
//*       d) The 5th integer (16-bit) is the item count                        *
//*       e) The items follow:                                                 *
//*          -- item_ID (16-bit)                                               *
//*          -- data_reference_index (16-bit)                                  *
//*          -- base_offset (base_offset_size * 8)                             *
//*          -- extent_count (16-bit)                                          *
//*             -- extent_offset (offset_size * 8)                             *
//*             -- extent_length (length_size * 8)                             *
//*                                                                            *
//*    6) The 'pitm' type (primary item) object                                *
//*                                                                            *
//*    7) The 'ipro' type (item protection) object                             *
//*                                                                            *
//*    8) The 'iinf' type (item information) object                            *
//*                                                                            *
//*    9) URLs may refer to other files OR to the current file.                *
//*                                                                            *
//*   10) Static Metadata: ISO/IEC 14496-12, section 8.44.8                    *
//*       a) simple textual                                                    *
//*          -- The only defined user-data box is the copyright 'cprt'.        *
//*          -- Other registered box types                                     *
//*          -- The UUID escape type                                           *
//*          -- Registered tags from MPEG-7 ISO/IEC 15938                      *
//*             -- Text: handler == 'mp7t' Unicode                             *
//*                An 'xml' box contains the text data, OR a primary item box  *
//*                which identifies the item containing the MPEG-7 XML data.   *
//*             -- Binary: handler == 'mp7b' compressed binary in BIM format   *
//*                A 'bxml' box contains the configuration info, followed by   *
//*                the binarized 'xml' data.                                   *
//*                                                                            *
//******************************************************************************

static void m4aExtractMetadata ( m4aParms& p, mdBox& metabox )
{
   mdBox mdbRef, 
         mdbHdlr ;                        // "hdlr" box contents
   int64_t  metaSize = metabox.metasize ; // unread bytes in the "meta" container
   uint32_t gCnt ;                        // bytes read for each source read

   //* Read and decode the "meta.version" field *
   metaSize -= gCnt = m4aReadBytes ( p, intLEN ) ;

   #if DEBUG_M4A != 0
   uint32_t modulo ;                      // box size multiple of eight(8) bytes?
   gsdbg2.formatInt( metabox.size, 11, true ) ;
   gsdbg1.compose( "\n"
                   "** m4aExtractMetadata **\n"
                   " type    : %s\n"
                   " size    : %s (%Xh)\n",
                   metabox.type, gsdbg2.ustr(), &metabox.size ) ;
   p.ofs << gsdbg1.ustr() ;
   gsdbg2.formatInt( metabox.metasize, 11, true ) ;
   gsdbg1.compose( " metasize: %s (%Xh)\n"
                   " valid   : %s\n",
                   gsdbg2.ustr(), &metabox.metasize, 
                   (metabox.valid ? "yes" : "no") ) ;
   p.ofs << gsdbg1.ustr() << endl ;
   #endif   // DEBUG_M4A

   if ( gCnt == intLEN )
   {
      if ( (mdbRef.decodeInt32 ( p.ubuff )) == ZERO )
      {
         //* Read and decode the "hdlr" (handler) box.*
         gCnt = m4aReadHeader ( p ) ;
         metaSize -= gCnt ;
         if ( gCnt == idLEN )
         {
            if ( (mdbHdlr.decodeHeader ( p.ubuff )) )
            {
               #if DEBUG_M4A != 0
               modulo = mdbHdlr.size % idLEN ;
               gsdbg2.formatInt( mdbHdlr.size, 11, true ) ;
               gsdbg1.compose( " type    : %s\n"
                               " size    : %s (%X)",
                               mdbHdlr.type, gsdbg2.ustr(), &mdbHdlr.size ) ;
               if ( modulo != ZERO )
                  gsdbg1.append( " mod:%u", &modulo ) ;
               gsdbg2.formatInt( mdbHdlr.metasize, 11, true ) ;
               gsdbg1.append( "\n metasize: %s (%Xh)\n",
                              gsdbg2.ustr(), &mdbHdlr.metasize ) ; 
               p.ofs << gsdbg1.ustr() ;
               p.ofs.flush() ;
               #endif   // DEBUG_M4A

               //* Read to the end of the 'hdlr' box object. *
               //* It is assumed that it will fit the buffer.*
               gCnt = m4aReadBytes ( p, mdbHdlr.metasize ) ;
               metaSize -= gCnt ;
               if ( gCnt == mdbHdlr.metasize )
               {
                  //* Validate 'version' and 'handler-type' integers.*
                  //* Get the null-terminated handler-type string.   *
                  uint32_t hdlrVersion = mdbHdlr.decodeInt32( p.ubuff ) ;
                  mdbHdlr.handler = mdbHdlr.decodeInt32( &p.ubuff[intLEN] ) ;
                  mdbHdlr.metadata = new uint8_t[mdbHdlr.metasize - (intLEN * 2) + 1] ;
                  uint32_t srcindx = (intLEN * 2),  // index handler name
                           trgindx = ZERO ;         
                  while ( srcindx < mdbHdlr.metasize )
                     mdbHdlr.metadata[trgindx++] = p.ubuff[srcindx++] ;
                  mdbHdlr.metadata[trgindx] = NULLCHAR ; // be sure string is terminated

                  //* If handler data verified,     *
                  //* scan remainder of 'meta' box. *
                  if ( hdlrVersion == ZERO )
                     mdbHdlr.valid = true ;
                  else
                     mdbHdlr.valid = false ;

                  #if DEBUG_M4A != 0
                  gsdbg1.compose( " version : %02Xh\n"
                                  " hdlrtype: %02Xh\n"
                                  " hdlrname: %s\n"
                                  " valid   : %s\n", 
                                  &hdlrVersion, &mdbHdlr.handler, mdbHdlr.metadata, 
                                  (mdbHdlr.valid ? "yes" : "no") ) ;
                  p.ofs << gsdbg1.ustr() << endl ;
                  #endif   // DEBUG_M4A
               }
               #if DEBUG_M4A != 0
               else
                  p.ofs << "Error! \"meta.hdlr\" truncated." << endl ;
               #endif   // DEBUG_M4A
            }
            #if DEBUG_M4A != 0
            else
            {
               p.ofs << "Error! \"meta.hdlr\" out-of-synch." << endl ;
            }
            #endif   // DEBUG_M4A
         }
         #if DEBUG_M4A != 0
         else
            p.ofs << "Error! \"meta.hdlr\" incorrectly formated." << endl ;
         #endif   // DEBUG_M4A
      }
      #if DEBUG_M4A != 0
      else
         p.ofs << "Error! \"meta\" version != ZERO" << endl ;
      #endif   // DEBUG_M4A
   }
   #if DEBUG_M4A != 0
   else
      p.ofs << "Error! \"meta\" incorrectly formatted." << endl ;
   #endif   // DEBUG_M4A

#if 0    // EXPERIMENTAL - DUMP LiTunes[] ARRAY
ofs << "\n** LiTunes[] **" << endl ;
for ( uint32_t i = ZERO ; i < looneyITEMS ; ++i )
{
   p.ofs << liTunes[i].hName << liTunes[i].aName << endl ;
}
p.ofs << endl ;
#endif   // EXPERIMENTAL

   //************************************************************
   //* If 'hdlr' box verified, scan the remainder of 'meta' box *
   //* and extract any text or image data.                      *
   //************************************************************
   if ( mdbHdlr.valid )
   {
      gString gsDisplay, gstmp ; // formatting
      uint32_t ltIndex ;         // index into liTunes[] array

      //* Read and decode the remainder of the "meta" box object.*
      gCnt = m4aReadHeader ( p ) ;
      metaSize -= gCnt ;

      if ( gCnt == idLEN )
      {
         do
         {
            mdbRef.reset() ;
            if ( (mdbRef.decodeHeader ( p.ubuff )) )
            {
               metaSize -= mdbRef.metasize ;    // bytes of "meta" box remaining

               #if DEBUG_M4A !=  0
               modulo = mdbRef.size % idLEN ;
               gsdbg2.formatInt( mdbRef.size, 11, true ) ;
               gsdbg1.compose( "BOX: %s (%s bytes at x%08X)", 
                              mdbRef.type, gsdbg2.ustr(), &p.srcoffset ) ;
               p.ofs << gsdbg1.ustr() ;
               if ( modulo != ZERO )
               {
                  gsdbg1.compose( " mod:%u", &modulo ) ;
                  p.ofs << gsdbg1.ustr() ;
               }
               p.ofs << endl ;
               #endif   // DEBUG_M4A

// DO NOT DISPLAY RECORDS WITH EMPTY STRINGS
// IMPLEMENT SUPPORT FOR 'XML' RECORDS.
// IMPLEMENT SUPPORT FOR 'BXML' RECORDS.
// IMPLEMENT SUPPORT FOR IMAGE FIELDS.
// IMPLEMENT SUPPORT FOR METADATA WHICH IS NOT IN THE 'META' BOX.
               //* Select metadata processing method.*
               gstmp = (char*)mdbRef.type ;
               if ( (gstmp.compare( "ilst" )) == ZERO )
               {
                  m4aExtractMD_ilst ( p, mdbRef ) ;
                  metaSize -= mdbRef.metasize ;
               }
               else if ( (gstmp.compare( "xml " )) == ZERO )
               {
                  m4aExtractMD_xml ( p, mdbRef ) ;
                  metaSize -= mdbRef.metasize ;
               }
               else if ( (gstmp.compare( "bxml" )) == ZERO )
               {
                  m4aExtractMD_bxml ( p, mdbRef ) ;
                  metaSize -= mdbRef.metasize ;
               }
               else if ( (gstmp.compare( "free" )) == ZERO )
               {
                  m4aExtractMD_free ( p, mdbRef ) ;
                  metaSize -= mdbRef.metasize ;
               }
               else if ( (gstmp.compare( "skip" )) == ZERO )
               {
                  m4aExtractMD_skip ( p, mdbRef ) ;
                  metaSize -= mdbRef.metasize ;
               }
               else if ( (ltIndex = m4aIsLooneyTunes ( mdbRef.type )) < looneyITEMS )
               {
                  itAtom itaRef ;

                  //* Read the remainder of the atom.*
                  metaSize -= m4aReadBytes ( p, (mdbRef.size - idLEN ) ) ;

                  if ( (itaRef.initialize ( mdbRef.size, ltIndex, 
                                            (char*)mdbRef.type, p.ubuff )) )
                  {
                     //* Strip leading and trailing whitespace from *
                     //* the data, for a more professional display. *
                     gstmp = itaRef.metaStr ;
                     gstmp.strip() ;
                     gstmp.copy( itaRef.metaStr, itaMETA_MAX ) ;

                     #if DEBUG_M4A == 0   // PRODUCTION
                     p.ofs << liTunes[ltIndex].hName << itaRef.metaStr << endl ;
                     #else                // DEBUG ONLY
                     //* Replace BYTE_A9 (0xA9) with '_' (0x5F).*
                     //*     (0xA9 is not valid UTF-8 data.)    *
                     if ( (uint8_t)itaRef.atomName[0] == BYTE_A9 )
                        itaRef.atomName[0] = '_' ;

                     uint32_t foffset = p.srcoffset - idLEN ;
                     gsdbg1.compose( "ATOM: %s (bytes: %u x%02X @ x%08X)"
                                     " constA:%08X || %s%s",
                                     itaRef.atomName, 
                                     &itaRef.atomSize, &itaRef.atomSize,
                                     &foffset, 
                                     &itaRef.dataConst1, 
                                     liTunes[itaRef.aIndex].hName, itaRef.metaStr ) ;
                     p.ofs << gsdbg1.ustr() << endl ;
                     #endif   // DEBUG_M4A
                  }
               }
               else  // unparsed/unreported metadata
               {
                  metaSize -= m4aDiscardBytes ( p, mdbRef.metasize ) ;
               }
            }

            //* Not a valid box header, shift left, append *
            //* a byte from the input stream and continue. *
            else
            {
               metaSize -= m4aShiftHeader ( p ) ;
            }
         }
         while ( metaSize > ZERO ) ;
         #if DEBUG_M4A != 0
         p.ofs << "** End m4aExtractMetadata **" << endl ;
         #endif   // DEBUG_M4A
      }
      else
      {
         #if DEBUG_M4A != 0
         p.ofs << "Read error before end of \"meta\" reached." << endl ;
         #endif   // DEBUG_M4A
      }
   }        // (mdbHdlr.valid)

}  //* End m4aExtractMetadata() *

//*************************
//*   m4aExtractMD_ilst   *
//*************************
//******************************************************************************
//* Non-member Method: called only by m4aExtractMetadata()                     *
//* ------------------                                                         *
//* Process the contents of an "ilst" box object.                              *
//*                                                                            *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: number of bytes read from input stream                            *
//******************************************************************************

static void m4aExtractMD_ilst ( m4aParms& p, const mdBox& mdbCon )
{
   gString gsDisplay, gstmp ; // formatting and display
   mdBox    mdbRef ;          // temp for decoding and reference (MPEG-4 boxes)
   dataAtom dtaRef ;          // temp for decoding and reference ("data" atom)
   itAtom   itaRef ;          // temp for decoding and reference (iTunes atoms)
   udefAtom udfRef ;          // temp for decoding and reference (user-defined atom)

   uint32_t atomSize,         // decoded byte count for an atom
            trg = ZERO,       // index into 'atomName'
            ltIndex ;         // index into liTunes[] array
   int64_t  ilst_remain = mdbCon.metasize ; // unread bytes in the 'ilst' object
   char     atomName[intLEN + 1] ;

   while ( ilst_remain > ZERO )
   {
      //* Read 8 bytes from the input stream *
      if ( (m4aReadHeader ( p )) == idLEN )
      {
         ilst_remain -= idLEN ;

         //* Scan for ASCII-alpha atom name,         *
         //* plus "----" (freeform) atom name.       *
         //* plus 0xA9 used by the iTunes dictionary.*
         trg = ZERO ;
         for ( uint32_t indx = intLEN ; indx < idLEN ; ++ indx )
         {
            if ( ((p.ubuff[indx] >= 'a') && (p.ubuff[indx] <= 'z')) ||
                 ((p.ubuff[indx] >= 'A') && (p.ubuff[indx] <= 'Z')) ||
                 (p.ubuff[indx] == '-') || (p.ubuff[indx] == BYTE_A9) )
            {
               atomName[trg++] = p.ubuff[indx] ;
            }
         }

         //* If four consecutive potential atom-name characters captured *
         if ( trg == intLEN )
         {
            //* Compare atom identifier against the list of *
            //* Apple(tm) iTunes(tm) dictionary entries.    *
            //* These have a 4-byte fixed-length key.       *
            if ( (ltIndex = m4aIsLooneyTunes ( (uint8_t*)atomName )) < looneyITEMS )
            {
               //* Decode the number of bytes in the atom.*
               atomSize = mdbCon.decodeInt32( p.ubuff ) ;

               //* Read the remainder of the atom.*
               ilst_remain -= m4aReadBytes ( p, (atomSize - idLEN ) ) ;

               //* Decode and validate the atom.*
               if ( (itaRef.initialize ( atomSize, ltIndex, atomName, p.ubuff )) )
               {
                  //* Strip leading and trailing whitespace from *
                  //* the data, for a more professional display. *
                  gstmp = itaRef.metaStr ;
                  gstmp.strip() ;
                  gstmp.copy( itaRef.metaStr, itaMETA_MAX ) ;

                  #if DEBUG_M4A == 0   // PRODUCTION
                  p.ofs << liTunes[ltIndex].hName << itaRef.metaStr << endl ;
                  #else                // DEBUG ONLY
                  //* Replace BYTE_A9 (0xA9) with '_' (0x5F).*
                  //*     (0xA9 is not valid UTF-8 data.)    *
                  if ( (uint8_t)itaRef.atomName[0] == BYTE_A9 )
                     itaRef.atomName[0] = '_' ;

                  uint32_t foffset = p.srcoffset - idLEN ;
                  gsdbg1.compose( "ATOM: %s (bytes: %u x%02X @ x%08X)"
                                  " constA:%08X || %s%s",
                                  itaRef.atomName, 
                                  &itaRef.atomSize, &itaRef.atomSize,
                                  &foffset, 
                                  &itaRef.dataConst1, 
                                  liTunes[itaRef.aIndex].hName, itaRef.metaStr ) ;
                  p.ofs << gsdbg1.ustr() << endl ;
                  #endif   // DEBUG_M4A
               }
            }

            //*****************************************************
            //* If not looney-toons, test for MPEG-4 type name    *
            //*****************************************************
            else if ( (mdbRef.decodeHeader ( (uint8_t*)atomName )) )
            {
               //* If all four characters of the name field are ASCII printing *
               //* characters AND if name is a registered box type.            *
               #if DEBUG_M4A != 0
               uint32_t modulo = mdbRef.size % mdbRef.size ;
               gsdbg2.formatInt( mdbRef.size, 11, true ) ;
               gsdbg1.compose( "BOX : %s (%s bytes at x%08X)", 
                               mdbRef.type, gsdbg2.ustr(), &p.srcoffset ) ;
               p.ofs << gsdbg1.ustr() ;
               if ( modulo != ZERO )
               {
                  gsdbg1.compose( " mod:%u", &modulo ) ;
                  p.ofs << gsdbg1.ustr() ;
               }
               p.ofs << endl ;
               #endif   // DEBUG_M4A

               // There may never be an MPEG-4 box inside an 'islt' container. 
               // If such a box object is identified during testing, we will 
               // implement processing here.
            }

            //*****************************************************
            //* '----' freeform atom                              *
            //* This is a variable-length atom identifier.        *
            //*****************************************************
            else if ( (atomName[0] == '-') && (atomName[1] == '-') &&
                      (atomName[2] == '-') && (atomName[3] == '-') )
            {
               //* Decode the number of bytes in the atom.*
               atomSize = mdbCon.decodeInt32( p.ubuff ) ;

               //* Read the remainder of the atom.*
               ilst_remain -= m4aReadBytes ( p, (atomSize - idLEN ) ) ;

               //* Decode and validate the atom.*
               if ( (udfRef.initialize ( atomSize, atomName, p.ubuff )) )
               {
                  //* Strip leading and trailing whitespace from *
                  //* the data, for a more professional display. *
                  gstmp = udfRef.metaStr ;
                  gstmp.strip() ;
                  gstmp.copy( udfRef.metaStr, itaMETA_MAX ) ;

                  gstmp = udfRef.nameDesc ;
                  gstmp.padCols( tagDescWIDTH, L'-' ) ;
                  
                  #if DEBUG_M4A == 0   // PRODUCTION
                  p.ofs << gstmp.ustr() << udfRef.metaStr << endl ;
                  #else                // DEBUG ONLY
                  uint32_t foffset = p.srcoffset - idLEN ;
                  gsdbg1.compose( "ATOM: %s (bytes: %u x%02X @ x%08X)\n"
                                  "      %s (bytes: %u x%02X, const:%02X) '%s'\n"
                                  "      %s (bytes: %u x%02X const:%02X) '%s'\n"
                                  "      %s (bytes: %u x%02X const1:%02X const2:%02X)\n"
                                  "      %s%s",
                                  udfRef.atomName, 
                                  &udfRef.atomSize, &udfRef.atomSize, &foffset,
                                  udfRef.meanKey, &udfRef.meanSize, &udfRef.meanSize,
                                  &udfRef.meanConst, udfRef.meanDesc,
                                  udfRef.nameKey, &udfRef.nameSize, &udfRef.nameSize,
                                  &udfRef.nameConst, udfRef.nameDesc,
                                  udfRef.dataKey, &udfRef.dataSize, &udfRef.dataSize,
                                  &udfRef.dataConst1, &udfRef.dataConst2,
                                  gstmp.ustr(), udfRef.metaStr ) ;
                  p.ofs << gsdbg1.ustr() << endl ;
                  #endif   // DEBUG_M4A
               }
            }        // freeform atom

            //*****************************************************
            //* Else, test for an un-encapsulated "data" atom.    *
            //*****************************************************
            else if ( (atomName[0] == genericTag.aName[0]) && 
                      (atomName[1] == genericTag.aName[1]) &&
                      (atomName[2] == genericTag.aName[2]) && 
                      (atomName[3] == genericTag.aName[3]) )
            {
               //* Decode the number of bytes in the atom.*
               atomSize = mdbCon.decodeInt32( p.ubuff ) ;

               //* Read the remainder of the atom.*
               ilst_remain -= m4aReadBytes ( p, (atomSize - idLEN ) ) ;

               if ( (dtaRef.initialize( atomSize, atomName, p.ubuff )) )
               {

                  #if DEBUG_M4A == 0   // PRODUCTION
                  p.ofs << genericTag.hName << dtaRef.metaStr << endl ;
                  #else                // DEBUG ONLY
                  uint32_t foffset = p.srcoffset - idLEN ;
                  gsdbg1.compose( "ATOM: %s (bytes: %u x%02X @ x%08X)"
                                  " constA:%08X || %s%s",
                                  dtaRef.atomName, 
                                  &dtaRef.atomSize, &dtaRef.atomSize,
                                  &foffset, 
                                  &dtaRef.dataConst1, 
                                  genericTag.hName, dtaRef.metaStr ) ;
                  p.ofs << gsdbg1.ustr() << endl ;
                  #endif   // DEBUG_M4A
               }
            }        // "data" atom

            //* Unrecognized box/atom name *
            else
            {
               if ( (m4aShiftHeader ( p )) != 1 )
                  break ;  // end-of-file (unlikely)
               --ilst_remain ;
            }
         }  // (trg==intLEN)

         //* Else out-of-synch. Shift in another byte and continue.*
         else
         {
            if ( (m4aShiftHeader ( p )) != 1 )
               break ;  // end-of-file (unlikely)
            --ilst_remain ;
         }
      }  // m4aReadHeader()

      else  // end-of-file (unlikely)
         break ;

   }  // while()

}  //* End m4aExtractMD_ilst() *

//*************************
//*   m4aExtractMD_xml    *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_xml ( m4aParms& p, const mdBox& mdbCon )
{

#if 1    // TEMP - m4aExtractMD_xml NOT IMPLEMENTED
gString gstmp( "m4aExtractMD_xml( %s %u bytes)", mdbCon.type, &mdbCon.size ) ;
p.ofs << gstmp.ustr() << endl ;
m4aDiscardBytes ( p, mdbCon.metasize ) ;
#endif   // U/C

}  //* End m4aExtractMD_xml() *

//*************************
//*   m4aExtractMD_bxml   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_bxml ( m4aParms& p, const mdBox& mdbCon )
{

#if 1    // TEMP - m4aExtractMD_bxml NOT IMPLEMENTED
gString gstmp( "m4aExtractMD_bxml( %s %u bytes)", mdbCon.type, &mdbCon.size ) ;
p.ofs << gstmp.ustr() << endl ;
m4aDiscardBytes ( p, mdbCon.metasize ) ;
#endif   // U/C

}  //* End m4aExtractMD_bxml() *

//*************************
//*   m4aExtractMD_schi   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* The current box object is a "schi" box.                                    *
//* It may contain a wide variety information, but we are interested ONLY in   *
//* the name (or account name) of the person who downloaded the file from      *
//* iTunes if it was recorded here.                                            *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_schi ( m4aParms& p, const mdBox& mdbCon )
{
   luniTunesDB schiTag = {(const uint8_t*)"name", "Downloaded By----------"} ;

   //* Read in the remainder of the box object.*
   if ( (m4aReadBytes ( p, mdbCon.metasize )) == mdbCon.metasize )
   {
      gString  gstmp ;
      uint8_t  nameBuff[intLEN + 1] ;     // receives "name" const string
      uint32_t indx = ZERO,               // p.ubuff index
               nbi  = ZERO ;              // nameBuff index

      //* Scan for the const "name" *
      while ( indx < mdbCon.metasize )
      {
         if ( (p.ubuff[indx] == schiTag.aName[0]) && (nbi == 0) )
            nameBuff[nbi++] = p.ubuff[indx] ;
         else if ( (p.ubuff[indx] == schiTag.aName[1]) && (nbi == 1) )
            nameBuff[nbi++] = p.ubuff[indx] ;
         else if ( (p.ubuff[indx] == schiTag.aName[2]) && (nbi == 2) )
            nameBuff[nbi++] = p.ubuff[indx] ;
         else if ( (p.ubuff[indx] == schiTag.aName[3]) && (nbi == 3) )
         {
            nameBuff[nbi++] = p.ubuff[indx] ;
            nameBuff[nbi] = NULLCHAR ;
            gstmp = (char*)nameBuff ;
            break ;
         }
         else
            nbi = ZERO ;
         ++indx ;
      }

      if ( (gstmp.compare ( (char*)schiTag.aName )) == ZERO )
      {
         //* Capture and display the name *
         gstmp.compose( "%s%s", schiTag.hName, &p.ubuff[++indx] ) ;
         p.ofs << gstmp.ustr() << endl ;
      }
   }

}  //* End m4aExtractMD_schi() *

//*************************
//*   m4aExtractMD_free   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* The current box object is a "free" box.                                    *
//* Scan the data for any human-readable information, and if found report it.  *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_free ( m4aParms& p, const mdBox& mdbCon )
{
   if ( (m4aReadBytes ( p, mdbCon.metasize )) == mdbCon.metasize )
   {
      #if DEBUG_M4A != 0
      gString gstmp( "Contents of \"free\" atom:" ) ;
      uint32_t baseCnt = gstmp.gschars(),
               indx = ZERO ;

      while ( indx < mdbCon.metasize )
      {
         if (   (((p.ubuff[indx] >= ' ') && (p.ubuff[indx] >= '~')) 
             || (p.ubuff[indx] == '\n')) && (gstmp.gschars() < gsMAXCHARS) )
         {
            gstmp.append( (wchar_t)p.ubuff[indx] ) ;
         }
         ++indx ;
      }
      if ( (uint32_t)(gstmp.gschars()) > baseCnt )
      {
         gstmp.strip() ;
         p.ofs << gstmp.ustr() << endl ;
      }
      #endif   // DEBUG_M4A
   }

}  //* End m4aExtractMD_free() *

//*************************
//*   m4aExtractMD_skip   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* The current box object is a "skip" box.                                    *
//* Scan the data for any human-readable information, and if found report it.  *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_skip ( m4aParms& p, const mdBox& mdbCon )
{
   if ( (m4aReadBytes ( p, mdbCon.metasize )) == mdbCon.metasize )
   {
      #if DEBUG_M4A != 0
      gString gstmp( "Contents of \"skip\" atom:" ) ;
      uint32_t baseCnt = gstmp.gschars(),
               indx = ZERO ;

      while ( indx < mdbCon.metasize )
      {
         if (   (((p.ubuff[indx] >= ' ') && (p.ubuff[indx] >= '~')) 
             || (p.ubuff[indx] == '\n')) && (gstmp.gschars() < gsMAXCHARS) )
         {
            gstmp.append( (wchar_t)p.ubuff[indx] ) ;
         }
         ++indx ;
      }
      if ( (uint32_t)(gstmp.gschars()) > baseCnt )
      {
         gstmp.strip() ;
         p.ofs << gstmp.ustr() << endl ;
      }
      #endif   // DEBUG_M4A
   }

}  //* End m4aExtractMD_skip() *

#if 0    // CURRENTLY UNUSED
//*************************
//*   m4aExtractMD_udta   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* The current box object is a "udta" box.                                    *
//* Scan the data for any human-readable information, and if found report it.  *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_udta ( m4aParms& p, const mdBox& mdbCon ) ;
static void m4aExtractMD_udta ( m4aParms& p, const mdBox& mdbCon )
{
   if ( (m4aReadBytes ( p, mdbCon.metasize )) == mdbCon.metasize )
   {
      #if DEBUG_M4A != 0
      #endif   // DEBUG_M4A
   }

}  //* End m4aExtractMD_udta() *
#endif   // CURRENTLY UNUSED

//*************************
//*   m4aExtractMD_uuid   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* The current box object is a "uuid" box.                                    *
//* Scan the data for any human-readable information, and if found report it.  *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          mdbCon : (by reference) header information for containing object  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

static void m4aExtractMD_uuid ( m4aParms& p, const mdBox& mdbCon )
{
   if ( (m4aReadBytes ( p, mdbCon.metasize )) == mdbCon.metasize )
   {
      #if DEBUG_M4A != 0
#if 1    // UNDER CONSTRUCTION - REPORT "uuid" ATOM
#endif   // U/C
      #endif   // DEBUG_M4A
   }

}  //* End m4aExtractMD_uuid() *

//*************************
//*   m4aReadFileHeader   *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Caller has opened the source file but has not yet read any data from it.   *
//* Scan the input stream until the file header is found.                      *
//*                                                                            *
//* The "ftyp" box is usually the first box defined; however, the specification*
//* states that it MUST be positioned as close to the beginning of the file as *
//* possible, and before any variable-size boxes.                              *
//*                                                                            *
//* When the file header has been located, parse the data and verify           *
//* formatting.                                                                *
//*                                                                            *
//* If the file header has not been identified within the first 2.0KB of the   *
//* file, it is assumed that the file is not an M4A file.                      *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*                   p.fileHdr receives the file header data                  *
//*                                                                            *
//* Returns: 'true' if file header successfully read and format verified       *
//*          'false' if file header not found OR if invalid format             *
//******************************************************************************

static bool m4aReadFileHeader ( m4aParms& p )
{
   mdBox mdbRef ;          // temp storage
   gString gstmp ;         // data analysis
   uint32_t idLen ;        // bytes returned from m4aReadHeader()

   p.fileHdr.reset() ;     // re-initialize the target structure

   while ( ! p.fileHdr.valid && ! p.eof && (p.fileBytes < 0x0800) )
   {
      //* Read the 8-byte header record.*
      if ( (idLen = m4aReadHeader ( p )) == idLEN )
      {
         if ( (mdbRef.decodeHeader( p.ubuff )) )
         {
            //* If "ftyp" object identified *
            if ( mdbRef.tindex == ftypINDEX )
            {
               //* Read the remainder of the object.*
               if ( (m4aReadBytes ( p, mdbRef.metasize )) == mdbRef.metasize )
               {
                  //* Decode and format the raw data.         *
                  //* If successful, p.fileHdr is initialized.*
                  p.fileHdr.initialize( mdbRef.size, mdbRef.type, p.ubuff ) ;
                  
                  #if DEBUG_M4A != 0
                  gsdbg2.formatInt( p.fileHdr.size, 11, true ) ;
                  gsdbg1.compose( "M4A File Header:\n"
                                  " size  : %s (%Xh)\n"
                                  " type  : '%s'\n"
                                  " major : '%s'\n"
                                  " minor : 0x%08X\n"
                                  " compat: '%s'\n"
                                  " valid : %s\n",
                                  gsdbg2.ustr(), &p.fileHdr.size, p.fileHdr.type, 
                                  p.fileHdr.major, &p.fileHdr.minor, p.fileHdr.compat,
                                  (p.fileHdr.valid ? "yes" : "no") ) ;
                  p.ofs << gsdbg1.ustr() << endl ;
                  #endif   // DEBUG_M4A
               }
            }

            //* Else, not the file header. Discard the object.*
            else
            {
               m4aReadBytes ( p, mdbRef.metasize ) ;
            }
         }
      }
      else           // file read error, unable to continue
         break ;
   }
   return p.fileHdr.valid ;

}  //* End m4aReadFileHeader() *

//*************************
//*     m4aReadBytes      *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Read 'byteCount' bytes from the input stream into the buffer.              *
//* This method consolidates all reads from the "xxx.m4a" source file.         *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          byteCnt: number of bytes to read                                  *
//*          offset : (optional, ZERO by default)                              *
//*                   if non-zero, write new data at specified p.ubuff offset  *
//*                   if zero, write new data at beginning of p.ubuff          *
//*                                                                            *
//* Returns: number of bytes read from input stream                            *
//*          p.ubuff      receives the data                                    *
//*          p.fileBytes  accumulator updated                                  *
//*          p.eof        set if bytes read < bytes requested                  *
//*          p.srcoffset  previous file offset (for debugging only)            *
//******************************************************************************

static uint32_t m4aReadBytes ( m4aParms& p, uint32_t byteCnt, uint32_t offset )
{
   if ( byteCnt > ubuffLEN )     // prevent buffer overflow
      byteCnt = ubuffLEN ;
   if ( (offset > ubuffLEN) || (byteCnt > (ubuffLEN - offset)) )
      offset = ZERO ;
   uint32_t gCnt = ZERO ;

   if ( (byteCnt > ZERO) && ! p.eof ) // if not yet at end-of-file
   {
      p.ifs.read( (char*)&p.ubuff[offset], byteCnt ) ;
      gCnt = p.ifs.gcount() ;
      if ( gCnt < byteCnt )
         p.eof = true ;

      #if DEBUG_M4A != 0
      p.srcoffset = p.fileBytes ;   // save previous file offset
      #endif   // DEBUG_M4A

      p.fileBytes += gCnt ;         // update current file offset
   }

   return ( gCnt ) ;

}  //* End m4aReadBytes() *

//*************************
//*     m4aReadHeader     *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Read idLEN bytes from the input stream into the buffer.                    *
//*                                                                            *
//* For convenience, a null character '\0' is appended after bytes read.       *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*                                                                            *
//* Returns: number of bytes read from input stream (idLEN if successful)      *
//******************************************************************************

static uint32_t m4aReadHeader ( m4aParms& p )
{

   uint32_t gCnt = m4aReadBytes ( p, idLEN ) ;
   p.ubuff[gCnt] = NULLCHAR ;

   return ( gCnt ) ;

}  //* End m4aReadHeader() *

//*************************
//*    m4aShiftHeader     *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Shift the first idLEN bytes in the buffer left by one byte and append      *
//* one byte from the input stream.                                            *
//*                                                                            *
//* Programmer's Note: We use the 'offset' parameter of m4aReadBytes() to      *
//* bring in the new byte at buffer offset (idLEN - 1).                        *
//*                                                                            *
//* For convenience, a null character '\0' is appended to the shifted data.    *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*                                                                            *
//* Returns: number of bytes read from input stream (1 if success, else 0)     *
//******************************************************************************

static uint32_t m4aShiftHeader ( m4aParms& p )
{
   #if DEBUG_M4A != 0
   #define DEBUG_M4A_VERBOSE (0)
   #if DEBUG_M4A_VERBOSE != 0
   gsdbg1.compose( "B: %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX",
                   &p.ubuff[0], &p.ubuff[1], &p.ubuff[2], &p.ubuff[3], 
                   &p.ubuff[4], &p.ubuff[5], &p.ubuff[6], &p.ubuff[7] ) ;
   p.ofs << gsdbg1.ustr() << endl ;
   #endif   // DEBUG_M4A_VERBOSE
   #endif   // DEBUG_M4A

   #if DEBUG_M4A != 0
   p.srcoffset = p.fileBytes ;
   #endif   // DEBUG_M4A

   uint32_t src = 1 ;
   for ( uint32_t trg = ZERO ; src < idLEN ; ++trg, ++src )
      p.ubuff[trg] = p.ubuff[src] ;

   --src ;                          // position the index
   uint32_t gCnt = m4aReadBytes ( p, 1, src ) ;
   if ( gCnt == 1 )
      p.ubuff[idLEN] = NULLCHAR ;   // terminate the (potential) box name string
   else
      p.ubuff[idLEN - 1] = NULLCHAR ;

   #if DEBUG_M4A != 0 && DEBUG_M4A_VERBOSE != 0
   gsdbg1.compose( "A: %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX\n",
                   &p.ubuff[0], &p.ubuff[1], &p.ubuff[2], &p.ubuff[3], 
                   &p.ubuff[4], &p.ubuff[5], &p.ubuff[6], &p.ubuff[7] ) ;
   p.ofs << gsdbg1.ustr() << endl ;
   #endif   // DEBUG_M4A && DEBUG_M4A_VERBOSE

   return ( p.ifs.gcount() ) ;

   #undef DEBUG_M4A_VERBOSE
}  //* End m4aShiftHeader() *

//*************************
//*    m4aDiscardBytes    *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Read and discard the specified number of bytes from the input stream.      *
//* This method may be called to scan past the data of a box object which      *
//* contains neither human-readable metadata (text/images) nor other           *
//* interesting box objects.                                                   *
//*                                                                            *
//* Input  : p      : (by reference) M4A-capture parameters                    *
//*          byteCnt: number of bytes to read                                  *
//*                                                                            *
//* Returns: number of bytes read from input stream                            *
//*          p.ubuff      receives the data                                    *
//*          p.fileBytes  accumulator updated                                  *
//*          p.eof        set if bytes read < bytes requested                  *
//*          p.srcoffset  previous file offset (for debugging only)            *
//******************************************************************************

static uint32_t m4aDiscardBytes ( m4aParms& p, uint32_t byteCnt )
{
   uint32_t sub = ubuffLEN,      // stream divisor (efficiency w/o buffer overflow)
            bytesRead = ZERO,    // total bytes read (return value)
            gCnt ;               // bytes read each pass

   for ( int64_t remain = byteCnt ; remain > ZERO ; )
   {
      if ( remain < sub )
         sub = remain ;

      gCnt = m4aReadBytes ( p, sub ) ;
      remain -= gCnt ;
      bytesRead -= gCnt ;
   }
   return bytesRead ;

}  //* End m4aDiscardBytes() *

//*************************
//*   m4aIsLooneyTunes    *
//*************************
//******************************************************************************
//* Non-member Method:                                                         *
//* ------------------                                                         *
//* Compare the specified string to the elements of the liTunes[] array which  *
//* implements the Apple iTunes dictionary of tag identifiers.                 *
//*                                                                            *
//* Caller's data compared to the 'aName' element of the record.               *
//* If a match, then the 'hName' (human-readable name) can be used to format   *
//* and display the metadata record.                                           *
//*                                                                            *
//* Input  : atomName: name of atom to be tested                               *
//*                                                                            *
//* Returns: index of matching record in the dictionary                        *
//*          or looneyITEMS if no match found                                  *
//******************************************************************************

static uint32_t m4aIsLooneyTunes ( const uint8_t* atomName )
{
   uint32_t ltIndex ;

   for ( ltIndex = ZERO ; ltIndex < looneyITEMS ; ++ltIndex )
   {
      if ( (atomName[0] == liTunes[ltIndex].aName[0]) &&
           (atomName[1] == liTunes[ltIndex].aName[1]) &&
           (atomName[2] == liTunes[ltIndex].aName[2]) &&
           (atomName[3] == liTunes[ltIndex].aName[3]) &&
           (atomName[4] == NULLCHAR) )
         break ;
   }

   return ltIndex ;

}  //* End m4aIsLoneyTunes() *

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

